//
// Copyright (c) 1993 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// AssociationMgr.h - This class handles all the informations related
//                    to the associations.

// By Luca Deri, July 1993

#include "AssocMgr.h"

#include "Top.h"

#if defined(ACCESS)
#include "LabelInit.h"
#include "ACRule.h"
#include "GlobalRule.h"
#endif

#include "CmisAgent.h"

#if defined(SECURITY)
#include "mintegrity.h"
#endif

AssociationMgr *theAssocMgr = (AssociationMgr *) 0;

static int
monotonic(const int, const int prev) { return (prev + 1); }

AssociationDescriptor::AssociationDescriptor()
{
  /* Default values */
  theFileDescriptor = -1;
  theInitiatorName = NULL;
  userRights.theChoice = UNKNOWN_USER;
  userRights.theRights.unknownUser.defaultAccessPermission = 0; /* None */
  userRights.theRights.unknownUser.defaultDenialGranularity = 
    DENIAL_ON_REQUEST;

  /* Added by Saleem, Feb 1994 */
  _agentId = 0;
  _prevAgentId = 0;
  _managerId = 0;
  _prevManagerId = 0;
  _invokeIdFunc = monotonic;
};

/* Modified by Saleem */
AssociationDescriptor::~AssociationDescriptor()
{
#if defined(SECURITY)
    clearMd5Key(theFileDescriptor);
    clearId(theFileDescriptor);
#endif

  if(theInitiatorName != NULL)
    delete theInitiatorName;
}

/* Modified by Saleem, Feb 1994 */
char *
AssociationDescriptor::GetSessionKey()
{
#if defined(SECURITY)
    return getMd5Key(theFileDescriptor);
#else
    return (char *) 0;
#endif
}

/* Modified by Saleem, Feb 1994 */
void
AssociationDescriptor::SetSessionKey(char *theKey)
{
#if defined(SECURITY)
     setMd5Key(theFileDescriptor, theKey);
#if 0
     printf("Session key: '");
     for(int u=0; u<8; u++) 
         printf("%02x", (unsigned int) ((unsigned char) theKey[u]));
     printf("'\n");
#endif
#endif
}

/* Added by Saleem, Feb 1994 */
int
AssociationDescriptor::makeAgentId()
{
    if (theFileDescriptor == -1) {
#if defined(DEBUG)
        advise((char *) 0, "AssociationDescriptor::makeAgentId(): bad fd");
#endif
        return 0;
    }

#if defined(SECURITY)
    register
    char *key = getMd5Key(theFileDescriptor);

    if (key != (char *) 0)
        _invokeIdFunc = ::makeResponderId;
#if defined(DEBUG)
     else
         advise((char *) 0, "AssociationDescriptor::makeAgentId(): no key fd=%d", theFileDescriptor);
#endif
#else
    _invokeIdFunc = monotonic;
#endif

    _prevAgentId = _agentId;
    _agentId = (*_invokeIdFunc)(theFileDescriptor, _prevAgentId);

    return _agentId;
}

/* Added by Saleem, Feb 1994 */
int
AssociationDescriptor::makeManagerId()
{
    if (theFileDescriptor == -1) {
#if defined(DEBUG)
        advise((char *) 0, "AssociationDescriptor::makeManagerId(): bad fd");
#endif
        return 0;
    }

#if defined(SECURITY)
    register
    char *key = getMd5Key(theFileDescriptor);

    if (key != (char *) 0)
        _invokeIdFunc = ::makeInitiatorId;
#if defined(DEBUG)
    else
        advise((char *) 0, "AssociationDescriptor::makeManagerId(): no key fd=%d", theFileDescriptor);
#endif
#else
    _invokeIdFunc = monotonic;
#endif

    _prevManagerId = _managerId;
    _managerId = (*_invokeIdFunc)(theFileDescriptor, _prevManagerId);

    return _managerId;
}

/* Modified by Saleem, Feb 1994 */
AssociationMgr::AssociationMgr()
{
  numDescriptors = 0;
  theDescriptor = NULL;

#if defined(SECURITY)
  static
  int flag = 0;

  if (!flag) {
      initialiseMd5KeySpace();
      flag = 1;
  }
#endif
};

AssociationMgr::~AssociationMgr()
{
  for(register i=0; i<numDescriptors; i++)
    delete theDescriptor[i];

  // free((char*) theDescriptor);
};


/**********************************************************************
  
  Name:	                InsertNewAssocDescriptor
  
  Description:	        This function inserts a new association
                        descriptor inside the internal descriptor list.
                        
  Input parameters:     int
                        The descriptor id.

  Output parameters:    OsimisErr
                        OK: no error has occurred
		        DESCRIPTOR_ALREADY_EXISTING: descriptor already used
			

  Modification history:                         

  Coded by:             Luca Deri
  
  Date:                 July 9th, 1993
  
***********************************************************************/
OsimisErr AssociationMgr::InsertNewAssocDescriptor(int theFileDescriptor)
{
  int i=0; 

  while(i<numDescriptors)
    {
      if(theDescriptor[i]->GetFileDescriptor() == theFileDescriptor)
	{
#ifdef DEBUG
	  advise(NULLCP, "AssociationMgr::InsertNewAssocDescriptor() failed");
#endif
	  return(DESCRIPTOR_ALREADY_EXISTING);
	}
     else if(theDescriptor[i]->GetFileDescriptor() <= theFileDescriptor)
	i++;
      else
	break;
    }

  /* I found the place for the new descriptor */
  numDescriptors++;
  
  if(numDescriptors == 1)
    theDescriptor =
	      (AssociationDescriptor**)malloc(
		(unsigned)(sizeof(AssociationDescriptor*)));
  else
    theDescriptor = 
	      (AssociationDescriptor**)realloc((char*)theDescriptor, 
	       (unsigned)(sizeof(AssociationDescriptor*)*numDescriptors));
  
  theDescriptor[numDescriptors-1] = theDescriptor[i];
  theDescriptor[i] = new AssociationDescriptor;
  theDescriptor[i]->SetFileDescriptor(theFileDescriptor);

  return(OK);
};


/**********************************************************************
  
  Name:	                RemoveAssocDescriptor
  
  Description:	        This function deletes the block related to this
                        file descriptor.
                        
  Input parameters:     int
                        The descriptor id.

  Output parameters:    OsimisErr
                        OK: no error has occurred
		        UNKNOWN_DESCRIPTOR: the descriptor was not found.
			
  Modification history:                         

  Coded by:             Luca Deri
  
  Date:                 July 9th, 1993
  
***********************************************************************/
OsimisErr AssociationMgr::RemoveAssocDescriptor(int theFileDescriptor)
{
  for(register int i=0; i<numDescriptors; i++)
    if(theDescriptor[i]->GetFileDescriptor() == theFileDescriptor)
      {
	if((--numDescriptors) == 0)
	  {
	    delete theDescriptor[0];
	    delete theDescriptor;
	  }
	else
	  {
	    /* Exchange the element with le last one */
	    theDescriptor[i] = theDescriptor[numDescriptors];
	    theDescriptor = 
	      (AssociationDescriptor**)realloc((char*)theDescriptor, 
		 (unsigned)(sizeof(AssociationDescriptor*)*(numDescriptors)));
	  }

	return(OK);
      }
    else if(theDescriptor[i]->GetFileDescriptor() > theFileDescriptor)
      break; /* Not found */
  
#ifdef DEBUG
  advise(NULLCP, "AssociationMgr::RemoveAssocDescriptor() failed");
#endif
  return(UNKNOWN_DESCRIPTOR);
}

/**********************************************************************
  
  Name:	                GetSessionKey
  
  Description:	        This function returns the session key related to the
                        association identified by the file descriptor.
                        
  Input parameters:     int
                        The descriptor id.

  Output parameters:    char*
                        The session key, NULL a problem has occurred.

  Modification history:                         

  Coded by:             Luca Deri
  
  Date:                 July 9th, 1993
  
***********************************************************************/
char* AssociationMgr::GetSessionKey(int theFileDescriptor)
{
  for(register int i=0; i<numDescriptors; i++)
  {
    if(theDescriptor[i]->GetFileDescriptor() == theFileDescriptor)
      return(theDescriptor[i]->GetSessionKey());
    else if(theDescriptor[i]->GetFileDescriptor() > theFileDescriptor)  
      break; /* Not found */
  }
  
#ifdef DEBUG
  advise(NULLCP, "AssociationMgr::GetSessionKey() failed");
#endif

  return(NULL);
}



/**********************************************************************
  
  Name:	                StoreSessionInfo
  
  Description:	        This function store the session key for the
                        association identified by the file descriptor.
                        
  Input parameters:     int
                        The descriptor id.

			char*
			The session key.

			char*
			The initiator name.

  Output parameters:    OsimisErr
                        OK: no error has occurred
		        UNKNOWN_DESCRIPTOR: the descriptor was not found.

  Modification history:                         

  Coded by:             Luca Deri
  
  Date:                 July 9th, 1993
  
***********************************************************************/
OsimisErr AssociationMgr::StoreSessionInfo(int theFileDescriptor,
					   char* theSessionKey,
					   char* theInitiatorName)
{
  for(register int i=0; i<numDescriptors; i++)
  {
    if(theDescriptor[i]->GetFileDescriptor() == theFileDescriptor)
      {
	theDescriptor[i]->SetSessionKey(theSessionKey);
	theDescriptor[i]->SetInitiatorName(theInitiatorName);
	theDescriptor[i]->UpdateSecurityDatas(); /* Update of the datas */ 
	return(OK);
      }
    else if(theDescriptor[i]->GetFileDescriptor() > theFileDescriptor)  
      break; /* Not found */
  }

#ifdef DEBUG
  advise(NULLCP, "AssociationMgr::StoreSessionInfo() failed");
#endif
  
  return(UNKNOWN_DESCRIPTOR);
};



/**********************************************************************
  
  Name:	                UpdateSecurityDatas
  
  Description:	        This function update the security datas for
                        the association descriptor.
                        
  Input parameters:     - none -

  Output parameters:    - none -

  Modification history:                         

  Coded by:             Luca Deri
  
  Date:                 July 14th, 1993
  
***********************************************************************/
void AssociationDescriptor::UpdateSecurityDatas()
{
#if defined(ACCESS)
  MO *top, **mobj;
  LabelInitiator *tmpObj;
  char *tmpName, *theSubordinateName, theInstName[128];
  struct type_ACtrl_SecurityLabel* theLabel;
  int len;
  Bool foundUser=False, firstLoop=True;
  struct type_ACtrl_DenialResponse* theDenialResponse;
  CMISScope theScope;
  CMISFilter* theFilter;
  MNListVal* theMemberValue;
  struct type_ACtrl_AccessPermission *theAccPermission;
  AccessControlRule *theDefaultRule;
  struct type_ACtrl_DefaultAccess *theDefaultAccess;
  struct type_ACtrl_DenialGranularity *theDenialGranularity;
  
  theScope.sc_type = Sc_BaseToNthLevel;
  theScope.sc_level = 3; /* The Global rule is at 3rd level */
  
  top = MO::getRoot(); /* This is the root of the MIB */
  mobj = top->Scope(&theScope, True, DISCARDAVA);
  
  /* Perform filtering: find all the objects that are member of the
    GlobalRule MOC */
  for (int i = 0; mobj[i];)
    if (mobj[i]->Filter((theFilter = 
			 str2mfilter("(objectClass=globalRule)"))) != True,
			 DISCARDAVA)
      for (int j = i; mobj[j]; j++)
	mobj[j] = mobj[j+1];
    else
      i++;
  
  mfilter_free(theFilter); /* Free the filter */
  
  /* For all the global rule instances... */
  if(theInitiatorName != NULL) /* Not a guest */
    for (int nobjs = 0; ((mobj[nobjs] != NULL) && (!foundUser)); nobjs++)
      {
	theMemberValue = ((GlobalRule *)mobj[nobjs])->GetMembers();
	
	while((theMemberValue != NULL) && (!foundUser))
	  {
	    tmpObj = 
	      (LabelInitiator*)top->getMO(theMemberValue->mnl_mn.mn_dn);
	    
	    if(tmpObj == NULL)
	      {
		/* Can't find the instance */
#ifdef DEBUG
		advise(NULLCP,
	       "AssociationDescriptor::UpdateSecurityDatas() failed (1)");
#endif
		return;      
	      }
	    
	    /* The security label associated with the object */
	    theLabel = tmpObj->GetLabel(); 
	    
	    while(theLabel != NULL) /* Scan the label */
	      {
		if(theLabel->member_ACtrl_0->offset == 
		   choice_ACtrl_0_initiatorLabel)
		  {
	       if((theLabel->member_ACtrl_0->un.initiatorLabel != NULL)
    && (theLabel->member_ACtrl_0->un.initiatorLabel->category != NULL))
		      {
			/* Get the string inside the security label */
			tmpName = bitstr2strb(prim2bit(theLabel->member_ACtrl_0->un.initiatorLabel->category), &len);
			
			if(len > 0) /* String not empty */
			  {
			    if(strcmp(tmpName, theInitiatorName) == 0)
			      {
				foundUser = True;
				break; /* Found the initiator: 
				end of the while cycle */
			      }
			  }
		      }
		    
		  } /* End if */
		
		/* Go to next element of the label */
		theLabel = theLabel->next; 
	      } /* End while */
	    theMemberValue = theMemberValue->mnl_next; /* Go to next element */
	  } /* End while */
	
      } /* End for */
  
      
  if(foundUser) /* The user has been found */
    {
      userRights.theChoice = KNOWN_USER;
      theAccPermission = ((Rule *)mobj[nobjs-1])->GetAccessPermission();
      if(theAccPermission != NULL)
	{
	  if(theAccPermission->parm == 
	     int_ACtrl_AccessPermission_deny)
	    userRights.theRights.knownUser.accessType = ACCESS_DENIED;
	  else
	    userRights.theRights.knownUser.accessType = 
	      ACCESS_PERMITTED;
	  
	  if((theDenialResponse = 
	      ((Rule *)mobj[nobjs-1])->GetDenialResponse()) != NULL)
	    switch(theDenialResponse->parm)
	      {
	      case int_ACtrl_DenialResponse_responseGiven:
		userRights.theRights.knownUser.denialResponse = 
		  DENIAL_RESPONSE_GIVEN;
		break;
	      case int_ACtrl_DenialResponse_noResponse:
		userRights.theRights.knownUser.denialResponse = 
		  DENIAL_NO_RESPONSE;
		break;
	      case int_ACtrl_DenialResponse_abortAssociation:
		userRights.theRights.knownUser.denialResponse = 
		  DENIAL_ABORT_ASSOCIATION;
		break;
	      case int_ACtrl_DenialResponse_falseResponse:
		userRights.theRights.knownUser.denialResponse = 
		  DENIAL_FALSE_RESPONSE;
		break;
	      default:	/* Value out of range */
		userRights.theRights.knownUser.denialResponse = 
		  DENIAL_RESPONSE_GIVEN;
	      }
	  else
	    userRights.theRights.knownUser.denialResponse = 
	      DENIAL_RESPONSE_GIVEN;
	}
      else
	userRights.theRights.knownUser.accessType = ACCESS_DENIED;
      
#ifdef DEBUG
      advise(NULLCP, "Updated security infos for known user:\n\t%s",
	     theInitiatorName);
#endif
    }
  else
    {
      userRights.theChoice = UNKNOWN_USER;
      /* Get the default rule: user not found in global rule */
      theSubordinateName = 
	(MO::getRoot()->getSubordinate())->getRDNString();
      sprintf(theInstName, "%s@accessControlObjectName=DomainRule", 
	      theSubordinateName);
      
      theDefaultRule = (AccessControlRule*)(top->getMO(theInstName));
      if(theDefaultRule != NULL) /* Found */
	{
	  theDefaultAccess = theDefaultRule->GetDefaultAccess();
	  /* Reset value */
	  userRights.theRights.unknownUser.defaultAccessPermission = 0; 
	  if(theDefaultAccess != NULL)
	    {
	      if((theDefaultAccess->actionOp != NULL) 
		 && (theDefaultAccess->actionOp->parm ==
		     int_ACtrl_AccessPermission_allow))
	SET_PRESENT(userRights.theRights.unknownUser.defaultAccessPermission, 
		    ACTION_OPERATION);
	      if((theDefaultAccess->createOp != NULL)    
		 && (theDefaultAccess->createOp->parm ==
		     int_ACtrl_AccessPermission_allow))
	SET_PRESENT(userRights.theRights.unknownUser.defaultAccessPermission, 
		    CREATE_OPERATION);
	      if((theDefaultAccess->deleteOp != NULL)    
		 && (theDefaultAccess->deleteOp->parm ==
		     int_ACtrl_AccessPermission_allow))
	SET_PRESENT(userRights.theRights.unknownUser.defaultAccessPermission, 
		    DELETE_OPERATION);
	      if((theDefaultAccess->getOp != NULL)    
		 && (theDefaultAccess->getOp->parm ==
		     int_ACtrl_AccessPermission_allow))
	SET_PRESENT(userRights.theRights.unknownUser.defaultAccessPermission, 
		    GET_OPERATION);
	      if((theDefaultAccess->replaceOp != NULL)    
		 && (theDefaultAccess->replaceOp->parm ==
		     int_ACtrl_AccessPermission_allow))
	 SET_PRESENT(userRights.theRights.unknownUser.defaultAccessPermission,
		     REPLACE_OPERATION);
	      if((theDefaultAccess->addValueOp != NULL)    
		 && (theDefaultAccess->addValueOp->parm ==
		     int_ACtrl_AccessPermission_allow))
	 SET_PRESENT(userRights.theRights.unknownUser.defaultAccessPermission, 
		     ADD_OPERATION);
	      if((theDefaultAccess->removeValueOp != NULL)    
		 && (theDefaultAccess->removeValueOp->parm ==
		     int_ACtrl_AccessPermission_allow))
         SET_PRESENT(userRights.theRights.unknownUser.defaultAccessPermission,
		     REMOVE_OPERATION);
	      if((theDefaultAccess->replaceWithdefaultOp != NULL)    
		 && (theDefaultAccess->replaceWithdefaultOp->parm ==
		     int_ACtrl_AccessPermission_allow))
	 SET_PRESENT(userRights.theRights.unknownUser.defaultAccessPermission,
		     REPLACEWITHDEFAULT_OPERATION);
	      if((theDefaultAccess->multipleObjectSelectionOp != NULL)  
		 && (theDefaultAccess->multipleObjectSelectionOp->parm ==
		     int_ACtrl_AccessPermission_allow))
	 SET_PRESENT(userRights.theRights.unknownUser.defaultAccessPermission,
		     MULTIPLEOBJECTSELECTION_OPERATION);  
	      if((theDefaultAccess->filterOp != NULL)    
		 && (theDefaultAccess->filterOp->parm ==
		     int_ACtrl_AccessPermission_allow))
	 SET_PRESENT(userRights.theRights.unknownUser.defaultAccessPermission,
		     FILTER_OPERATION);
	    }
	  
	  theDenialGranularity = 
	    theDefaultRule->GetDenialGranularity();
	  if(theDenialGranularity != NULL)
	    {
	      switch(theDenialGranularity->parm)
		{
		case int_ACtrl_DenialGranularity_request:
		  userRights.theRights.unknownUser.defaultDenialGranularity = 
		    DENIAL_ON_REQUEST;
		  break;
		case int_ACtrl_DenialGranularity_object:
		  userRights.theRights.unknownUser.defaultDenialGranularity = 
		    DENIAL_ON_OBJECT;
		  break;
		case int_ACtrl_DenialGranularity_attribute:
		  userRights.theRights.unknownUser.defaultDenialGranularity = 
		    DENIAL_ON_ATTRIBUTE;
		  break;
		default:
		  userRights.theRights.unknownUser.defaultDenialGranularity =
		    DENIAL_ON_REQUEST;
		  break;
		}
	    }
	  else
	    userRights.theRights.unknownUser.defaultDenialGranularity = 
	      DENIAL_ON_REQUEST;
#ifdef DEBUG
	  advise(NULLCP,"Updated security infos for unknown user:\n\t%s",
		 theInitiatorName);
#endif
	}
      else
	{
	  /* The default instance can't be found: I allow every operation */
	  userRights.theChoice = UNKNOWN_USER;
	  userRights.theRights.unknownUser.defaultAccessPermission = 
	    ALL_OPERATION_ALLOWED; 
	  userRights.theRights.unknownUser.defaultDenialGranularity = 
	    DENIAL_ON_REQUEST;
#ifdef DEBUG
	  advise(NULLCP,
	  "AssociationDescriptor::UpdateSecurityDatas() using default values");
#endif
	}
    } /* End default rule */
#endif /* ACCESS */
};

/**********************************************************************
  
  Name:	                AssociationMgr::VerifySessionPermission
  
  Description:	        This method verify for the single request
                        if this operation may be executed according to
			the security requirements. If the operation may be 
			executed, the method returns True, otherwise it returns
			False and sends (if applicable) to the manager
			the correct response according to the security
			policy.
                        
  Input parameters:     int
                        The association descriptor (file descriptor)

                        int
			The invokeId of the request operation

			int
			An enumerated that specify the operation type

			int
			A bit valued integer that specifies the access
			permissions required for the operation. This field
			is exploited only if the initiator in unknown.

		        CMISAgent*
			The agent that is executing there operations.

  Output parameters:    Bool
                        True if the operation may be excuted
			False otherwise and a response (if applicable) is
			sent to the remote initiator.

  Modification history:                         

  Coded by:             Luca Deri
  
  Date:                 July 15th, 1993
  
***********************************************************************/
Bool AssociationMgr::VerifySessionPermission(int theAssocDescr,
			       int theInvokeId, int operationType, 
	                       int AccPermissionRequired, CMISAgent *theAgent)
{
  UserRights *userRights;
  int operationId;
  
  for(register int i=0; i<numDescriptors; i++)
    {
      if(theDescriptor[i]->GetFileDescriptor() == theAssocDescr)
	{
	  /* Before I check if the datas have to be updated */
	  if(theDescriptor[i]->IsToBeUpdated() == True)
	    theDescriptor[i]->UpdateSecurityDatas(); /* I do the update... */
	  
	  userRights = theDescriptor[i]->GetUserRights();
	  switch(userRights->theChoice)
	    {
	    case KNOWN_USER:
	      if(userRights->theRights.knownUser.accessType == ACCESS_PERMITTED)
		return(True); /* Access permitted */
	      else
		switch(userRights->theRights.knownUser.denialResponse)
		  {
		  case DENIAL_RESPONSE_GIVEN:
		    operationId = -1;
		    switch(operationType)
		      {
		      case MI_GET:
			operationId = M_GET;
			break;
		      case MI_SETC:
			operationId = M_SETC;
			break;
		      case MI_ACTIONC:
			operationId = M_ACTIONC;
			break;
		      case MI_CREATE:
			operationId = M_CREATE;
			break;
		      case MI_DELETE:
			operationId = M_DELETE;
			break;
		      default:
			/* Other operations that don't require
			  a response to be sent */
			break;
		      }
		    if(operationId != -1)

		      (void) theAgent->MError(theAssocDescr, theInvokeId, 0, 
					operationId, m_accessDenied);

		    break;
		  case DENIAL_NO_RESPONSE:
		    /* Nothing to do */
		    break;
		  case DENIAL_ABORT_ASSOCIATION:
		    MSAPIndication mis;
		    M_AbortReq(theAssocDescr, NULLEXTERN, &mis);
		    advise(NULLCP, "AssocMgr: assoc. %d aborted.",
			   theAssocDescr);

		    if(RemoveAssocDescriptor(theAssocDescr) != OK)
		      advise(NULLCP, 
			     "AssocMgr: error deleting assoc. descriptor");
		    break;
		  case DENIAL_FALSE_RESPONSE:
		    /* I always send the m_resourceLimitation error response */
		    operationId = -1;
		    switch(operationType)
		      {
		      case MI_GET:
			operationId = M_GET;
			break;
		      case MI_SETC:
			operationId = M_SETC;
			break;
		      case MI_ACTIONC:
			operationId = M_ACTIONC;
			break;
		      case MI_CREATE:
			operationId = M_CREATE;
			break;
		      case MI_DELETE:
			operationId = M_DELETE;
			break;
		      default:
			/* Other operations that don't require
			  a response to be sent */
			break;
		      }
		    if(operationId != -1)
		      (void) theAgent->MError(theAssocDescr, theInvokeId, 0, 
					operationId, m_resourceLimitation);
		  }
#ifdef DEBUG
	      advise(NULLCP, 
		     "AssociationMgr::VerifySessionPermission() Acc. denied");
#endif
	      return(False); /* Access denied */
	    case UNKNOWN_USER:
	      if((AccPermissionRequired &
		  userRights->theRights.unknownUser.defaultAccessPermission)
		 == AccPermissionRequired)
		return(True); /* Operation allowed */
	      else
	     switch(userRights->theRights.unknownUser.defaultDenialGranularity)
		  {
		  case DENIAL_ON_ATTRIBUTE:
		    /* This feature is not supported up to now and 
		      will be supported with the introduction of the item
			rules. The behaviour is the same as the denial
			  on request */
		  case DENIAL_ON_OBJECT:
		 /* This in implemented like a denial on request, according to 
		    the note contained in 8.1.2.2.3 of ISO/IEC 10164-9 */
		  case DENIAL_ON_REQUEST:
		    operationId = -1;
		    switch(operationType)
		      {
		      case MI_GET:
			operationId = M_GET;
			break;
		      case MI_SETC:
			operationId = M_SETC;
			break;
		      case MI_ACTIONC:
			operationId = M_ACTIONC;
			break;
		      case MI_CREATE:
			operationId = M_CREATE;
			break;
		      case MI_DELETE:
			operationId = M_DELETE;
			break;
		      default:
			/* Other operations that don't require
			  a response to be sent */
			break;
		      }
		    if(operationId != -1)
		      (void) theAgent->MError(theAssocDescr, theInvokeId, 
					0, operationId,  m_accessDenied);
		  }
	    } /* End user type: known/unknown user */
	  /* NOTREACHED */
	} /* Association descriptor not found */
      else if(theDescriptor[i]->GetFileDescriptor() > theAssocDescr)  
	break; 
    }
  
#ifdef DEBUG
  advise(NULLCP, "AssociationMgr::VerifySessionPermission() failed");
#endif
  
  return(False); /* Unknown Descriptor */
};


/**********************************************************************
  
  Name:	                AssociationMgr::VerifySessionPermission
  
  Description:	        This method verify for the single request
                        if this operation may be executed according to
			the security requirements. If the operation may be 
			executed, the method returns True, otherwise it returns
			False and sends (if applicable) to the manager
			the correct response according to the security
			policy.
                        
  Input parameters:     Bool
                        True is an ACRule instance has been updated
			False otherwise

  Output parameters:    - none -

  Modification history:                         

  Coded by:             Luca Deri
  
  Date:                 July 16th, 1993
  
***********************************************************************/
void AssociationMgr::MarkAssocDescriptor(Bool isACRule)
{
  UserRights *assocRights;

  for(register int i=0; i<numDescriptors; i++)
    {
      assocRights = theDescriptor[i]->GetUserRights();

      if(isACRule) 
	{
	  /* It is an update on ACRule so I have to update 
	    only unknown users */
	  if(assocRights->theChoice == UNKNOWN_USER)
	    theDescriptor[i]->SetToBeUpdated();
	}
      else
	{
	  /* It is an update to an item or a global rule so I have
	     to update known an unknown users. In fact, an unknown user may 
	     become 'known' because has been added to to an initiator
	     instance */
	  theDescriptor[i]->SetToBeUpdated();
	}
    }
};

/* Added by Saleem, Feb 1994 */
AssociationDescriptor *
AssociationMgr::find(const int fd)
{
    register
    int i;

    for(i = 0; i < numDescriptors; ++i) {
        if (theDescriptor[i]->GetFileDescriptor() == fd)
            return theDescriptor[i];
        else if (theDescriptor[i]->GetFileDescriptor() > fd)
            break; /* Not found */
    }

#ifdef DEBUG
    advise(NULLCP, "AssociationMgr::find(): failed fd=%d", fd);
#endif

    return (AssociationDescriptor *) 0;
}


/* Added by Saleem Feb 1994 */
int
AssociationMgr::getAgentId(const int fd)
{
    register
    AssociationDescriptor *ad = find(fd);

    if (ad == (AssociationDescriptor *) 0)
        return 0;

    return ad->getAgentId();
}

/* Added by Saleem Feb 1994 */
int
AssociationMgr::getPrevAgentId(const int fd)
{
    register
    AssociationDescriptor *ad = find(fd);

    if (ad == (AssociationDescriptor *) 0)
        return 0;

    return ad->getPrevAgentId();
}

/* Added by Saleem Feb 1994 */
int
AssociationMgr::makeAgentId(const int fd)
{
    register
    AssociationDescriptor *ad = find(fd);

    if (ad == (AssociationDescriptor *) 0)
        return 0;

    return ad->makeAgentId();
}


/* Added by Saleem Feb 1994 */
int
AssociationMgr::getManagerId(const int fd)
{
    register
    AssociationDescriptor *ad = find(fd);

    if (ad == (AssociationDescriptor *) 0)
        return 0;

    return ad->getManagerId();
}

/* Added by Saleem Feb 1994 */
int
AssociationMgr::getPrevManagerId(const int fd)
{
    register
    AssociationDescriptor *ad = find(fd);

    if (ad == (AssociationDescriptor *) 0)
        return 0;

    return ad->getPrevManagerId();
}

/* Added by Saleem Feb 1994 */
int
AssociationMgr::makeManagerId(const int fd)
{
    register
    AssociationDescriptor *ad = find(fd);

    if (ad == (AssociationDescriptor *) 0)
        return 0;

    return ad->makeManagerId();
}
