//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// CmisAgent.cc - CMIS Agent object realisation

// By George Pavlou, February 1991


#include <stdlib.h>
#include "mtime.h"
#include "sys/time.h"
#if !defined(SOLARIS)
extern "C" { int gettimeofday(struct timeval*, struct timezone*); }
#endif
#if	defined(__SUNPRO_CC)
extern "C" { void bzero(char*, int); }
#endif
#include "msap.h"
#include "Error.h"
#include "Create.h"
#include "CmisAgent.h"
#include "AssocMgr.h"
#include "RecordInfo.h"
#include "eventForwardingDiscriminator.h"

#ifdef	SECURITY
#include "mauth.h"
#include "mintegrity.h"
#endif


/*
 * the following should be always defined as it keeps memory requirements low
 * low without any noticeable performance degradation
 */
#define	MINIMISE_MEMORY

const int AssocInfo::MsdCmp = 1;
const int AssocInfo::AetCmp = 2;
int  AssocInfo::cmpmode;

DN   agentNameDomainDN = NULLDN;
int  agentNameDomainRdnNo = 0;
Bool globalNames = False;
extern "C" { int  dn_mgmt_order_cmp(DN, DN); DN dn_dec(PE); }

extern "C" { int TGetReadMask(fd_set*, int*); }


class Request {
public:

    MSAPIndication _mi;
    PE             _pdu;
    int            _check;
    Bool           _done;

    Request();
};

Request::Request()
{
    memset((void *) &_mi, 0, sizeof(MSAPIndication));
    _pdu = (PE) 0;
    _check = 0;
    _done = False;
}

#ifdef	SECURITY
extern AssociationMgr *theAssocMgr;
#endif

CMISAgent* CMISAgent::_instance;



CMISAgent::CMISAgent (int argc, char** argv, char* service,
						Bool singleInstanceAgent)
{
    init();

    TSAPdisconnect tds;
    if (initialise(argc, argv, service, &tds, singleInstanceAgent) == NOTOK)
	error(NULLCP, "isode error: [%s]", TErrString(tds.td_reason));
}


CMISAgent::~CMISAgent ()
{
    if (_emptyDN) dn_free(_emptyDN);
    if (_aei) delete _aei;
    oid_free(_processingFailureOID);
#ifdef MINIMISE_MEMORY
    cancelWakeUps(NULLCP);
#endif

#ifdef ACCESS_TEST
    dn_free(_initr);
#endif
}


void  CMISAgent::doTime (char* msg)
{
    if (!msg) {
	gettimeofday(&_prevTime, NULL);
	return;
    }

    timeval now;
    gettimeofday(&now, NULL);

    long diffsec = now.tv_sec - _prevTime.tv_sec, diffusec;
    if ((diffusec = now.tv_usec - _prevTime.tv_usec) < 0)
        --diffsec, diffusec += 1000000;

    _prevTime = now;	// struct copy

    printf("%s in %d.%06d seconds\n", msg, diffsec, diffusec);
}


void  CMISAgent::init ()
{
    _emptyDN = NULLDN;
    _processingFailureOID = name2oid("processingFailure");
    _maxAssocs = DFLT_MAXASSOCS;
    _curAssocs = _thisAssoc = 0;
    _psapEndpoint = 0;
    _aei = NULLCP;
    _iterator.setList(&_requests);
    _instance = this;
    _createOperation = False;
    _invokeId = 0;
#ifdef USE_X500
    _agentService = NULLCP;
    _singleInstanceAgent = False;
    _ltFailed = False;
#endif

    notifyShutdown();		// to release the MIB
#ifdef MINIMISE_MEMORY
    scheduleWakeUps(GARBAGE_COLLECTION_INTVL, NULLCP);
#endif

#ifdef ACCESS_TEST
    char* cp;
    _initr = str2dn(cp = strdup(TEST_INITIATOR));
    free(cp);
#endif
}


int  CMISAgent::initialise (int argc, char** argv, char* service,
				TSAPdisconnect* td, Bool singleInstanceAgent)
{
    if (argc > 1)
	error(NULLCP, "static server: should be started without args");
    AEI aei;
    char* myhost;

    fd_set mask; int nfds = 0;
    mask.fds_bits[0] = 0;
    TGetReadMask(&mask, &nfds);
    fd_mask prevMask = mask.fds_bits[0];

#ifdef USE_X500
    _singleInstanceAgent = singleInstanceAgent;
    myhost = singleInstanceAgent ? NULLCP : getlocalhost();

    struct PSAPaddr pa[FD_SETSIZE]; /* list of possible psap addresses */
    extern  struct PSAPaddr listen_pa[FD_SETSIZE]; /* see $TOP/lt/ltserver.c */
    char * agentDN;
    char * agent_service;

    bzero ((char *) pa, sizeof pa);
    if ( lt_server_psaps(service, pa) == 0 )
      advise(NULLCP, "Supported isode stacks were found");

    if ( lt_server_init(argc, argv, pa, td) == NOTOK )
      return NOTOK;

    agent_service = strdup(service);

    _agentService = strdup (agent_service);

    agentDN = lt_str2dn(myhost, agent_service);

    if ( lt_update(agentDN, listen_pa, UP) == 0 )
        advise(NULLCP, "Directory succesfully updated on startup");
    else {
	_ltFailed = True;
        advise(NULLCP, "Reverting to static AE information through \
isoentities");
    }

    if (_ltFailed)
#endif /* USE_X500 */
    {
#ifndef	LOCALHOST
    myhost = getlocalhost();
#else
    myhost = "localhost";
#endif
    if ((aei = str2aei(myhost, service)) == NULLAEI)
	error(NULLCP, "unknown application entity information %s-%s",
			myhost, service);
    if (iserver_init(argc, argv, aei, NULL, td) == NOTOK)
	return NOTOK;
    AEIFREE(aei);
    }

    TGetReadMask(&mask, &nfds);
    fd_mask maskFdVal = mask.fds_bits[0] ^ prevMask; // XOR to find the new fd
    _psapMask.fds_bits[0] = maskFdVal;
    _psapEndpoint = 1;
    while ((maskFdVal /= 2) != 1) ++_psapEndpoint;
    startListen(_psapEndpoint);

#ifdef SECURITY
    initialiseOsisec(argc, argv);
    initialiseIdSpace();
#endif

    char buf[128];
    sprintf(buf, "%s-%s", service, myhost);
    advise(NULLCP, "application entity title: %s", _aei = strdup(buf));
    return OK;
}


int  CMISAgent::wakeUp (char*)
{
    /*
     * we cannot release the DN after emitted event reports as we may
     * be in the middle of a CMIS operation and this can cause side effects -
     * it can be fixed with some tricky code but this is the easy way out...
     */
    MO** mo = MO::getRoot()->getWholeSubtree();
    for (register int i = 0; mo[i]; i++)
	mo[i]->releaseDN();
    return OK;
}


int  CMISAgent::associationRequest (int vecp, char** vec)
{
    int msd, status = ACS_ACCEPT;
    MSAPStart mss, *ms = &mss;
    MSAPIndication mis, *mi = &mis;
    MSAPAbort *ma = &mi -> mi_abort;
    External *userInfo = (External *) 0;

    if (M_Init(vecp, vec, ms, mi) == NOTOK) {
        advise(NULLCP, "CMISAgent - initialisation fails: %s", ma -> ma_data);
        return NOTOK;
    }

    if ((ms->ms_info.mai_funits | AGENT_FUNUNITS) != AGENT_FUNUNITS) {
	status = ACS_PERMANENT;
#ifdef DEBUG
	advise(NULLCP, "\nCMIS association refused - requested \
non-supported functional unit, requested 0x%x, supported 0x%x",
			ms->ms_info.mai_funits, AGENT_FUNUNITS);
#endif
    }
    else
    if (_curAssocs >= _maxAssocs) {
	status = ACS_TRANSIENT;		// too many associations at present
#ifdef DEBUG
	advise(NULLCP, "\nCMIS association refused - max number %d established",
			_maxAssocs);
#endif
    }

#ifdef SECURITY
    register
    int e;
    AuthAssocIntegrityInfo *aaii = (AuthAssocIntegrityInfo *) 0;

    if (status == ACS_ACCEPT) {
        /*
        ** Check the credentials in the ACSE info
        */

        e = getAcseUserInfo(ms->ms_info.mai_info, &aaii);
#if defined(DEBUG)
        if (e == 0) {
            printf("CMISAgent: secure association accepted %s\n",
                   sprintaei(&ms->ms_calling));

            register
            char *nameStr = (char *) 0, *cnStr;

            cnStr = "cn=";
            if ((nameStr = strstr(aaii->peer, cnStr)) == (char *) 0) {
                cnStr = "commonName=";
                nameStr = strstr(aaii->peer, cnStr);
            }
            if (nameStr != (char *) 0)
                nameStr += strlen(cnStr);
            fprintf(stderr, "CMISAgent: I believe I am talking to %s\n\n",
                    nameStr != (char *) 0 ? nameStr : aaii->peer);
        }
#endif /* DEBUG */

        if (e != 0) { /* Association refused: security error */
            advise((char *) 0, "CMISAgent: assoc refused (security error) with %s - %s", sprintaei(&ms->ms_calling), authErrorMessage(e));
            status = ACS_PERMANENT;
        }
        else if ((e = makeAcseUserInfo(aaii, &userInfo)) != 0) {
            /*
            ** Agent sends credentials to manager for authentication
            */
            advise((char *) 0,"CMISAgent: error constructing authentication info - %s", authErrorMessage(e));
            status = ACS_TRANSIENT; /* Manager re-try? Attack possible! */
        }
    }

#endif /* SECURITY */

    msd = ms -> ms_sd;

    if (M_InitialiseResp(ms, status, NULLAEI, &ms->ms_start.ps_called, NULLCP,
		CMIP_VERSION2, ms->ms_info.mai_funits & AGENT_FUNUNITS,
		NULLEXTERN, userInfo, mi) == NOTOK) {
        advise(NULLCP, "CMISAgent A-ASSOCIATE.RESPONSE: %s", ma -> ma_data);
#ifdef	SECURITY
        AuthAssocIntegrityInfo_free(aaii);
#endif
        if (userInfo != (External *) 0)
            external_free(userInfo);
        ms_free(ms);
        return NOTOK;
    }

    if (userInfo != (External *) 0)
        external_free(userInfo);

    if (status != ACS_ACCEPT) {
	ms_free(ms);
#ifdef	SECURITY
        AuthAssocIntegrityInfo_free(aaii);
#endif
	return NOTOK;
    }

    _curAssocs++;

#ifdef	SECURITY
    if (theAssocMgr->InsertNewAssocDescriptor(msd) != OK)
        advise(NULLCP, "CMISAgent: error creating new assoc. descriptor");

    /* Store session key in the slot identified by the file descriptor */
    if (theAssocMgr->StoreSessionInfo(msd, aaii->sessionKey, aaii->peer) != OK)
        advise(NULLCP, "CMISAgent: error storing the session info.");
    aaii->peer = (char *) 0;
    AuthAssocIntegrityInfo_free(aaii);
#endif

    startListen(msd);
    AssocInfo* assocInfo = new AssocInfo;
    assocInfo->msd = msd;
    assocInfo->aet = NULLDN;
    if (ms->ms_calling.aei_ap_title)
	assocInfo->aet = dn_dec(ms->ms_calling.aei_ap_title);
    _assocInfoList.append(assocInfo);
#ifdef DEBUG
    advise(NULLCP, "\nCMIS association established - msd %d", msd);
#endif

    ms_free(ms);
    return msd;
}


int  CMISAgent::shutdown (int msd)
{

#ifdef USE_X500

    extern  struct PSAPaddr listen_pa[FD_SETSIZE]; // in $TOP/lt/ltserver.c

    char* myhost = _singleInstanceAgent ? NULLCP : getlocalhost();
    char* agentDN = lt_str2dn(myhost, _agentService);

    if ( !_ltFailed && lt_update(agentDN, listen_pa, DOWN) == 0 )
      advise(NULLCP, "\nDirectory succesfully updated on shutdown");

#endif

    if (msd == -1) {
	MO::terminateMIB();

#ifdef	SECURITY
	delete theAssocMgr;
#endif
	if (notificationMapper)
	    delete notificationMapper;
	free_oid_table();
    }
    else
    if (msd != _psapEndpoint) {
        MSAPIndication mis;
        M_AbortReq(msd, NULLEXTERN, &mis);
    }

    return OK;
}


int  CMISAgent::sendEventReport (MO* mo, MID obj_class, MN obj_inst, MID event,
				 PE info, char* event_time, AET* aet, int msd)
{
    int invokeId;
    AEI aei;
    DN  aetdn;
    char title[128], qualifier[128];
    AssocInfo assocInfo, *ai;
    PSAPaddr *pa = NULL;
    isobject* isobj; OID oid;
    MSAPIndication mis;
    MSAPConnect mcs;
    External *userInfo = NULLEXTERN;
#ifdef	SECURITY
    register char *key;
    AuthAssocIntegrityInfo *aaii;
    register int e;
#endif
#ifdef ACCESS
    int operMask = 0;
    operMask |= AC_GET; operMask |= AC_SCOPE;
#endif


    // check if a "thisAssociation" ie. empty destination was used
    if (msd > 0) {
	aet = NULL;
#ifdef ACCESS
	if (mo && _adf.checkMOAccess(getInitiator(msd), mo, operMask, True) !=
					int_Access_EnforcementAction_allow)
	    return OK;
#endif
	goto thisAssoc;
    }

    // check if we have an association open to the destination AET and use it
    assocInfo.aet = aet->aet;
    if (ai = _assocInfoList.find(&assocInfo)) {
	msd = ai->msd;
	aet = NULL;
#ifdef ACCESS
	if (mo && _adf.checkMOAccess(getInitiator(msd), mo, operMask, True) !=
					int_Access_EnforcementAction_allow)
	    return OK;
#endif
	goto thisAssoc;
    }

    if ((aei = str2aei(aet->qualifier, aet->title)) == NULLAEI) {
        advise(NULLCP, "CMISAgent::sendEventReport: %s-%s: \
unknown application entity title",
			aet->qualifier, aet->title);
	return NOTOK;
    }
    if ((pa = aei2addr(aei)) == NULLPA) {
	advise(NULLCP, "AEI address translation for event report failed - \
%s-%s",
			aet->qualifier, aet->title);
	return NOTOK;
    }

#ifdef	ACCESS
    char buf[256];
    sprintf(buf, "%s:%s", aet->title, aet->qualifier);
    aetdn = str2aet(buf);
    if (mo && _adf.checkMOAccess(aetdn, mo, operMask, True) !=
					int_Access_EnforcementAction_allow) {
	dn_free(aetdn);
	return OK;
    }
    dn_free(aetdn);
#endif

#ifdef	SECURITY
    aaii = AuthAssocIntegrityInfo_make();
    key = makeMd5Key();
    memcpy(aaii->sessionKey, key, 8);

    if ((e = makeAcseUserInfo(aaii, &userInfo)) != 0) {
        advise((char *) 0, "Could not make authentication info");
        AuthAssocIntegrityInfo_free(aaii);
        return NOTOK;
    }
#endif

    if (M_InitialiseReq(NULLAEI, NULLPA, NULLAEI, pa, CONTEXT, CMIP_VERSION2,
			0x0, NULLEXTERN, userInfo, &mcs, &mis) == NOTOK) {
#ifdef DEBUG
	advise(NULLCP, "Cannot establish association for event report to \
%s-%s",
			aet->qualifier, aet->title);
#endif
        if (userInfo != (External *) 0)
            external_free(userInfo);
#ifdef	SECURITY
        AuthAssocIntegrityInfo_free(aaii);
#endif
	return NOTOK;
    }
    msd = mcs.mc_sd;
    mc_free(&mcs);
    if (userInfo != (External *) 0)
        external_free(userInfo);

#ifdef SECURITY
    if(theAssocMgr->InsertNewAssocDescriptor(msd) != OK)
        advise(NULLCP, "CMISAgent: error creating new assoc. descriptor");

    /* Store session key in the slot identified by the file descriptor */
    if (theAssocMgr->StoreSessionInfo(msd, aaii->sessionKey, aaii->peer) != OK)
        advise(NULLCP, "CMISAgent: error storing the session info.");
    else
        aaii->peer = (char *) 0;

    AuthAssocIntegrityInfo_free(aaii);

    /*
    ** Authenticate manager
     */
    e = getAcseUserInfo(mcs.mc_info.mai_info, &aaii);

#ifdef	DEBUG
    if (e == 0) {
        register
        char *nameStr = (char *) 0, *cnStr;

        cnStr = "cn=";
        if ((nameStr = strstr(aaii->peer, cnStr)) == (char *) 0) {
            cnStr = "commonName=";
            nameStr = strstr(aaii->peer, cnStr);
        }
        if (nameStr != (char *) 0)
            nameStr += strlen(cnStr);
        fprintf(stderr, "I believe I am talking to %s\n\n",
                nameStr != (char *) 0 ? nameStr : aaii->peer);

        /* Sanity check ... not fatal */
        if (memcmp(aaii->sessionKey, key, 8) != 0)
            fprintf(stderr, "Warning: different sessionKey from peer!\n");
    }
#endif
    AuthAssocIntegrityInfo_free(aaii);

    /*
    ** If authentication fails, for now, drop the assoc. Actually,
    ** behaviour should be determined by policy statements.
    ** Dream on ...
    */
    if (e != 0) {
        advise((char *) 0, "CMISAgent: Event Report");
        advise((char *) 0, "         : %s", authErrorMessage(e));
        advise((char *) 0, "         : Aborting assoc %d", dest->da_msd);
        if (M_AbortReq(msd, (External *) 0, &mis) != OK) {
#ifdef	DEBUG
            advise((char *) 0, "CMISAgent: M_AbortReq() %s\n",
                    mis.mi_abort.ma_data);
#endif
            mi_free(&mis);
        }
        return NOTOK;
    }
#endif /* SECURITY */

    advise(NULLCP, "CMISAgent: sending event report to %s-%s",
     		    title, qualifier);

thisAssoc:
#ifndef	SECURITY
    invokeId = ++_invokeId;
#else
    /* We must use the initiator sequence, i.e. the 'manager' sequence */
    invokeId = theAssocMgr->makeManagerId(msd);
#endif

    if (!aet)
        advise(NULLCP, "CMISAgent: sending event report on msd %d", msd);
    if (M_EventRep(msd, invokeId, obj_class, obj_inst, event,
                       event_time, info, &mis) == NOTOK) {
#ifdef DEBUG
	advise(NULLCP, "Event Report failed - %s", mis.mi_preject.mpr_data);
#endif
	mi_free(&mis);
	return NOTOK;
    }

    if (aet) {		// we have established an association
	if (M_TerminateReq(msd, ACF_NORMAL, NULLPE, &mis) != OK) {
	    advise(NULLCP, "CMISAgent::sendEventReport: terminate request \
failed: reason = %s\n",
			mis.mi_abort.ma_data);
	    return NOTOK;
        }
        if (userInfo)
	    external_free(userInfo);
	mc_free(&mcs);
    }
    return OK;
}


int  CMISAgent::associationInit (int vecp, char** vec)
{
    return CMISAgent::getInstance()->associationRequest(vecp, vec);
}


int  CMISAgent::associationWork (int msd)
{
    return CMISAgent::getInstance()->readCommEndpointAux(msd);
}


int  CMISAgent::readCommEndpoint (int msd)
{
    TSAPdisconnect tds;
    if (msd == _psapEndpoint)		// association request
        iserver_wait(CMISAgent::associationInit, NULL, NULL,
			0, NULLFD, NULLFD, NULLFD, NOTOK, &tds);
    else {
	int nfds = msd+1;
	fd_set ifds;
	ifds.fds_bits[0] = 1 << msd;
        iserver_wait(NULL, CMISAgent::associationWork, NULL,
			nfds, &ifds, NULLFD, NULLFD, NOTOK, &tds);
    }
    return OK;
}


int  CMISAgent::readCommEndpointAux (int msd)
{
    MSAPIndication mis;
    PE pdu;

    _thisAssoc = msd;

#ifdef TIMING
    doTime(NULLCP);
#endif

    M_WaitReqAux(msd, OK, &mis, &pdu);

#ifdef TIMING
    if (mis.mi_type != MI_TERM)
	doTime("did read CMIS request");
#endif

    if (serviceCmisRequest(msd, &mis, pdu) == NOTOK)
	return NOTOK;

#ifdef TIMING
    doTime("finished with CMIS request");
#endif

    if (mis.mi_type != MI_GET || !_requests.first())
	return OK;

    // requests queued while looking for a Cancel-Get
    register Request *r;
    while (r = (Request *) _requests.get()) {
        if (r->_done) {
            mi_free(&(r->_mi));
#ifdef	SECURITY
            if (r->_pdu != (PE) 0)
                pe_free(r->_pdu);
            /* advance the id sequence */
            if (r->_mi.mi_type != MI_CANCEL_GET)
                theAssocMgr->makeManagerId(msd);
#endif
        }
        else
	if (serviceCmisRequest(msd, &(r->_mi), r->_pdu) == NOTOK)
            return NOTOK;
        delete r;
    }

    return OK;
}


int  CMISAgent::checkForCancelGet (int msd, int currentGetInvoke)
{
    MSAPIndication mis;
    Request *r;
    PE pdu;

    while (M_WaitReqAux(msd, OK, &mis, &pdu) != NOTOK) {
        r = new Request;
        r->_mi = mis;           // struct copy
#ifdef	SECURITY
        r->_pdu = pdu;
#endif
        _requests.append(r);
        if (mis.mi_type == MI_CANCEL_GET) {
#ifdef	SECURITY
            /*
            ** Integrity check for CancelGet
            */
            register
            int i, id;
            char *state;

            /* get the correct id value */
            id = theAssocMgr->getManagerId(msd);
            saveInitiatorIdState(msd, &state);
            for(i = 0; i < _requests.getCount(); ++i)
                id = makeInitiatorId(msd, id);
            restoreInitiatorIdState(msd, state);

            MD5Value md5;

            if (makeMd5Value(msd, r->_pdu, &md5) != 0) {
#if defined(DEBUG)
                advise((char *) 0, "CMISAgent: makeMd5Value() failed");
#endif /* DEBUG */
                return NOTOK;
            }

            register
            int invoke = ENCODE_ID(id, md5.md5_words[0]);

            if (mis.mi_invoke != invoke) {
#if defined(DEBUG)
                advise((char *) 0, "CMISAgent: CancelGet integrity check failed");
                advise((char *) 0, "         : id=%08x mi=%08x invoke=%08x",
                       (u_int) id, (u_int) mis.mi_invoke, (u_int) invoke);
#endif /* DEBUG */
                return NOTOK;
            }
#endif /* SECURITY */
            r->_done = True;
            return CancelGet(msd, mis.mi_cancelgetinv.id,
                             mis.mi_cancelgetinv.args.cga_getinvoke,
                             currentGetInvoke);
            /* mi_free(&mis) is not needed for MI_CANCEL_GET */
        }
    }

    return OK;
}


int  CMISAgent::serviceCmisRequest (int msd, MSAPIndication* mi, PE pdu)
{
    int operationType = 0, id;
    AssocInfo assocInfo, *ai;
    assocInfo.aet = NULL;

#ifdef	SECURITY
    /* Save state of sequence generator in case check fails */
    char *state;
    saveInitiatorIdState(msd, &state);

    id = theAssocMgr->makeManagerId(msd); /* The expected invokeId value */

    /*
    ** Perform integrity check for PDU
    */
    if (pdu != (PE) 0) {

        if (mi->mi_invoke != 0) {

            MD5Value md5;

            if (makeMd5Value(msd, pdu, &md5) != 0) {
#if defined(DEBUG)
                advise((char *) 0, "CMISAgent: could not make integrity check");
#endif
                pe_free(pdu);
                return NOTOK;
            }
            pe_free(pdu);

            register
            int v = ENCODE_ID(id, md5.md5_words[0]);
            if (mi->mi_invoke != v) {
                advise((char *) 0, "CMISAgent: integrity check failed");
                advise((char *) 0, "         : fd=%d  mi=%08x",
                       msd, (unsigned long) mi->mi_invoke);
                advise((char *) 0, "         : expectd id=%08x chk=%08x %08x",
                       (u_int) id, md5.md5_words[0], (u_int) v);
                /*
                ** Return a CMIS error if integrity check fails? ###
                */

                /* restore state of sequence generaator */
                restoreInitiatorIdState(msd, state);

                mi_free(mi);
                return NOTOK;
            }
        }
        else {
            advise((char *) 0, "CMISAgent: zero invokeId!");
            mi_free(mi);
            return NOTOK;
        }
    }
#endif /* SECURITY */

    switch (mi->mi_type) {

    case MI_PREJECT:
	switch (mi->mi_preject.mpr_reason) {
	case ROS_TIMER:
	    return OK;

	case ROS_ACS:		// ACS disconnect
	default:
#ifdef	SECURITY
            if(theAssocMgr->RemoveAssocDescriptor(msd) != OK)
              advise(NULLCP, "CMISAgent: error deleting assoc. descriptor");
#endif
	    _curAssocs--;
	    eventForwardingDiscriminator::checkOnDisconnect(msd);
	    stopListen(msd);
	    assocInfo.msd = msd;
	    ai = _assocInfoList.remove(&assocInfo);
	    _assocInfoList.deleteElem(ai);
#ifdef DEBUG
	    advise(NULLCP, "\nProvider Reject - id %d, reason %s",
			mi->mi_preject.mpr_id, mi->mi_preject.mpr_data);
	    advise(NULLCP, "CMIS association broken - msd %d\n", msd);
#endif
	    return NOTOK;
	}
	break;

    case MI_ABORT:
	_curAssocs--;
	eventForwardingDiscriminator::checkOnDisconnect(msd);
	stopListen(msd);
	assocInfo.msd = msd;
	ai = _assocInfoList.remove(&assocInfo);
	_assocInfoList.deleteElem(ai);
#ifdef DEBUG
	advise(NULLCP, "CMIS association aborted - msd %d\n", msd);
#endif
	return NOTOK;

    case MI_GET:
    {
	CMISGetArg* g = &mi->mi_getinv.args;
#if defined(SECURITY) && defined(DEBUG)
        advise(NULLCP, "CMIS Get indication accepted");
#endif

	Get(msd, mi->mi_getinv.id, &g -> ga_class, &g -> ga_inst,
		&g -> ga_scope, g -> ga_filter, g -> ga_access, g -> ga_sync,
		g -> ga_nattrs, g -> ga_attrs);
    }
	break;

    case MI_CANCEL_GET:
	CancelGet(msd, mi->mi_cancelgetinv.id,
			mi->mi_cancelgetinv.args.cga_getinvoke, NOTOK);
	break;

    case MI_SET:
    case MI_SETC:
    {
	CMISSetArg* s = &mi->mi_setinv.args;
#if defined(SECURITY) && defined(DEBUG)
        advise(NULLCP, "CMIS Set indication accepted");
#endif

	Set(msd, mi->mi_type == MI_SETC ? True : False,
		mi->mi_setinv.id, &s -> sa_class, &s -> sa_inst,
		&s -> sa_scope, s -> sa_filter, s -> sa_access, s -> sa_sync,
		s -> sa_nattrs, s -> sa_attrs);
    }
	break;

    case MI_ACTION:
    case MI_ACTIONC:
    {
	CMISActionArg* a = &mi->mi_actioninv.args;
#if defined(SECURITY) && defined(DEBUG)
        advise(NULLCP, "CMIS Action indication accepted");
#endif

	Action(msd, mi->mi_type == MI_ACTIONC ? True : False,
		mi->mi_actioninv.id, &a -> aa_class, &a -> aa_inst,
		&a -> aa_scope, a -> aa_filter, a -> aa_access, a -> aa_sync,
		&a -> aa_info);
    }
	break;

    case MI_CREATE:
    {
	CMISCreateArg* c = &mi->mi_createinv.args;
	_createOperation = True;
#if defined(SECURITY) && defined(DEBUG)
        advise(NULLCP, "CMIS Create indication accepted");
#endif

	Create(msd, mi->mi_createinv.id, &c -> ca_class, &c -> ca_inst,
		c -> ca_inst_type, &c -> ca_reference_inst, c -> ca_access,
		c -> ca_nattrs, c -> ca_attrs);
	_createOperation = False;
    }
	break;

    case MI_DELETE:
    {
	CMISDeleteArg* d = &mi->mi_deleteinv.args;
#if defined(SECURITY) && defined(DEBUG)
        advise(NULLCP, "CMIS Delete indication accepted");
#endif

	Delete(msd, mi->mi_deleteinv.id, &d -> da_class, &d -> da_inst,
		&d -> da_scope, d -> da_filter, d -> da_access, d -> da_sync);
    }
	break;

    case MI_TERM:
#ifdef	SECURITY
        if(theAssocMgr->RemoveAssocDescriptor(msd) != OK)
          advise(NULLCP, "CMISAgent: error deleting assoc. descriptor");
#endif
	M_TerminateResp(msd, ACS_ACCEPT, ACR_NORMAL, NULLPE, mi);
	_curAssocs--;
	eventForwardingDiscriminator::checkOnDisconnect(msd);
	stopListen(msd);
	assocInfo.msd = msd;
	ai = _assocInfoList.remove(&assocInfo);
	_assocInfoList.deleteElem(ai);
#ifdef DEBUG
	advise(NULLCP, "CMIS association closed - msd %d\n", msd);
#endif
	return NOTOK;

    case MI_UREJECT:
    case MI_ERROR:
    case MI_EVENT_REP_RES:
    case MI_GET_RES:
    case MI_SET_RES:
    case MI_ACTION_RES:
    case MI_CREATE_RES:
    case MI_DELETE_RES:
    case MI_EMPTY_RES:
    case MI_LINKED_REPLY:
    case MI_EVENT_REP:
    case MI_EVENT_REPC:
	advise(NULLCP, "CMISAgent: received result/error, linked reply or \
event report - ignored!");
	break;

    case MI_UNKNOWN:
	break;

    default:
	advise(NULLCP, "CMISAgent - readAssoc: Unknown indication type %d",
			mi->mi_type);
	break;
    }

    mi_free(mi);
    return OK;
}


DN  CMISAgent::mn2localdn (MN mn)
{
    if (mn->mn_type == MN_LOCALDN)
	return mn->mn_dn;

    int cmp = dn_mgmt_order_cmp(agentNameDomainDN, mn->mn_dn);
    if (! (cmp == -1 || cmp == 0))	// less than or equal
	return NULLDN;
    register DN cur = mn->mn_dn;
    for (register int i = 0; i <= agentNameDomainRdnNo; i++)
	cur = cur->dn_parent;

    if (cur == NULLDN) {
	if (!_emptyDN)
	    _emptyDN = dn_comp_new(NULLRDN);
	cur = _emptyDN;
    }
    return cur;
}


#ifdef ACCESS

int  CMISAgent::checkMOAccess(DN init, MO* mo,
				CMISScope* scope, CMISFilter* filter, int oper)
{
    int operMask = 0, operCount = 0;
    if (scope->sc_type != Sc_BaseObject) {
	operMask |= AC_SCOPE; ++operCount;
    }
    if (filter) {
	operMask |= AC_FILTER; ++operCount;
    }
    operMask |= oper; ++operCount;
    return _adf.checkMOAccess(init, mo,
			      operMask, operCount > 1 ? True : False);
}

#endif


static void  free_mo_array (MO** mo, int nmo)
{
    for (register int i = 0; i < nmo; i++)
	mo[i] = NULLMO;
    delete[] mo;
}


int  CMISAgent::Get (int msd, int invoke, MID obj_class, MN obj_inst,
	CMISScope* scope, CMISFilter* filter, External*, CMISSync sync,
	int nattrs, MID attrs)
{
    MO *base, *top, **mobj_tmp;
    CMISGetAttr result[MO_MAXATTRS];
    CMISErrors error = m_noError;
    struct MIDentifier my_class; my_class.mid_type = MID_GLOBAL;
    MName my_inst; my_inst.mn_type = globalNames ? MN_DN : MN_LOCALDN;
    MSAPIndication mis;
    int nres, linked, nobjs, nobjs1, nrepl = 0;
    register int i, j;
    int invokeId;
    Bool allomorphic = False;
    OID  allomorphicClass;
    AVA* err = NULLAVA;

    // check parameters, find base object, check class and do scoping

    if (!(error = checkParms(obj_class, obj_inst, NULL, scope, filter, sync))) {
	top = MO::getRoot() -> getSubordinate();

	DN localdn;
	if ((localdn = mn2localdn(obj_inst)) == NULLDN)
	    error = m_noSuchObjectInstance;
	else
	if ((base = top -> getMO(localdn, True, err)) == NULLMO)
	    error = err ? err -> getError() : m_noSuchObjectInstance;
	else
        if (obj_class -> mid_global &&
		base -> checkClass(obj_class -> mid_global, &allomorphic)
			== NOTOK)
	    error = m_classInstanceConflict;
	else {
	    mobj_tmp = base -> Scope(scope, True, err);
	    if (err)
		error = err -> getError();
	}
    }

    if (error) {
#ifdef	SECURITY
        invoke = theAssocMgr->getManagerId(msd);
#endif
	return MError(msd, invoke, 0, M_GET, error, obj_class, obj_inst,
				scope, filter, sync, NULL, NULL, err);
    }

#ifdef DEBUG
    char* cp;
    advise(NULLCP, "Get request: id %08x class %s inst {%s}\nscope %s",
                (u_int) invoke,
                obj_class->mid_global ? _operationClass : "",
		cp = dn2str(obj_inst->mn_dn), sprintscope(scope));
    free(cp);
    if (filter)
	advise(NULLCP, "filter %s", sprintmfilter(filter));
#endif


    // copy the MO pointers in order to avoid side effects
    // (Scope returns the actual array, not a copy)
    for (nobjs1 = 0; mobj_tmp[nobjs1]; nobjs1++);
    MO** mobj = new MO* [nobjs1+1];
    for (i = 0; i <= nobjs1; i++)
	mobj[i] = mobj_tmp[i];


    // perform filtering
    if (filter) {
	for (i = 0; mobj[i];) {
	    Bool flt = mobj[i] -> Filter(filter, err);
	    if (err) {
	        error = m_processingFailure;	// the only error allowed
#ifdef	SECURITY
		invoke = theAssocMgr->getManagerId(msd);
#endif
		free_mo_array(mobj, nobjs1);
		return MError(msd, invoke, 0, M_GET, error, obj_class, obj_inst,
				scope, filter, sync, NULL, NULL, err);
	    }
	    if (flt != True)
		for (j = i; mobj[j]; j++)
		    mobj[j] = mobj[j+1];
	    else
		i++;
	}
    }
    for (nobjs = 0; mobj[nobjs]; nobjs++);

#ifdef DEBUG
    if (scope->sc_type || filter)
	advise(NULLCP, "%d objects selected after %s",
		nobjs, scope->sc_type && filter ? "scoping and filtering" :
		(scope->sc_type ? "scoping" : "filtering"));
#endif

    // perform get operations and return results/errors
    linked = nobjs > 1 ? invoke : 0;

    for (i = 0; i < nobjs; i++) {
	if (mobj[i] == base) {
	    my_class.mid_global = obj_class -> mid_global ?
			obj_class -> mid_global : mobj[i] -> getClass();
	    allomorphicClass	= allomorphic ? my_class.mid_global : NULLOID;
	}
	else {
	    my_class.mid_global = mobj[i] -> getClass();
	    allomorphicClass	= NULLOID;
	}
	my_inst.mn_dn = mobj[i] -> getDN(globalNames ? True : False);

	error = m_noError;
#ifdef ACCESS
	int enforceAction =
		checkMOAccess(getInitiator(), mobj[i], scope, filter, AC_GET);

	if (enforceAction == int_Access_EnforcementAction_denyWithoutResponse)
	    continue;

	if (enforceAction == int_Access_EnforcementAction_denyWithResponse)
	    if (!linked)
		error = m_accessDenied;
	    else {
		error = m_getListError;	// get/set linked replies
					// cannot have accessDenied
		nres = 0;
	    }
#endif
	if (!error) {
	    nres = 0;
	    error = mobj[i] -> Get(allomorphicClass,
				   nattrs, attrs, &nres, result, err);
	}

#ifdef TIMING
	doTime("performed CMIS Get");
#endif

	if (nobjs > 1 && i % CANCELGETBLOCK == 0 &&
				checkForCancelGet(msd, invoke) == DONE) {
	    advise(NULLCP, "Get operation cancelled after %d \
linked replies\n", i);
	    free_mo_array(mobj, nobjs);
	    return OK;		// operation cancelled
	}

        /* Advance the invokeId value */
        invokeId = nobjs == 1 ?
#ifndef	SECURITY
		invoke : ++_invokeId;
#else
		theAssocMgr->getManagerId(msd) :
		theAssocMgr->makeAgentId(msd);
#endif

	if (!error || error == m_getListError) {
#ifdef DEBUG
	    if (error) {
		char *clss = mobj[i] -> getClassName();
		char *inst = mobj[i] -> getDNString();
		advise(NULLCP, "CMIS GetListError: class %s, instance %s",
			clss, inst);
		free(inst);

		for (j = 0; j < nres; j++)
		    if (result[j].ga_error &&
				result[j].ga_id.mid_type == MID_GLOBAL)
			advise(NULLCP, "Attribute %s: %s",
				oid2name(result[j].ga_id.mid_global, OIDPART),
				MErrString(result[j].ga_error));
	    }
#endif
	    if (M_GetRes(msd, invokeId, linked,
				&my_class, &my_inst,
				mtime(), nres, result, error, NULL, &mis)
			== NOTOK)
		return NOTOK;
	    if (err)
		delete err;
#ifdef TIMING
	    doTime("sent CMIS result");
#endif
	}
	else
	    if (MError(msd, invokeId, linked,
				M_GET, error, &my_class, &my_inst,
				NULL, NULL, NULLMSYNC, NULL, NULL, err)
			== NOTOK)
		return NOTOK;
	++nrepl;

#ifdef MINIMISE_MEMORY
	mobj[i] -> releaseEncodedData();
#endif
    }

    if (nrepl == 0 || linked) {	// final empty result
#ifdef	SECURITY
        invoke = theAssocMgr->getManagerId(msd);
#endif
	M_GetRes(msd, invoke, 0, NULL, NULL, NULL, 0, NULL,
				m_noError, NULL, &mis);
#ifdef	DEBUG
	advise(NULLCP, "Get result: dispatched %d linked replies", nrepl);
#endif
    }
    free_mo_array(mobj, nobjs);
    return OK;
}


int  CMISAgent::CancelGet (int msd, int invoke,
				int cancelGetInvoke, int currentGetInvoke)
{
    MSAPIndication mis, *mi;

    // check first if it relates to the current Get operation

    if (cancelGetInvoke == currentGetInvoke) {
#ifdef	SECURITY
        currentGetInvoke = theAssocMgr->makeAgentId(msd);
#endif
	M_GetRes(msd, currentGetInvoke, 0, NULL, NULL,
                 mtime(), 0, NULL, m_operationCancelled, NULL, &mis);
#ifdef	SECURITY
        invoke = theAssocMgr->getManagerId(msd);
#endif
	M_CancelGetRes(msd, invoke, m_noError, NULL, &mis);
	return DONE;
    }

    // check then if it relates to any of the queued Get operations
    Request *r;
    int id;

    _iterator.reset();
    while ((r = (Request *) _iterator.getNext()) && (r->_pdu != (PE) 0)) {
        mi = &(r->_mi);

	if (mi->mi_type == MI_GET && mi->mi_invoke == cancelGetInvoke) {
#ifdef	SECURITY
            /*
            ** Need the id to send back in the EmptyResult
            */
            if (r->_check == 0) {
                MD5Value md5;

                /* if checksum fails, leave request in list for now ... */
                if (makeMd5Value(msd, r->_pdu, &md5) != 0)
                    fprintf(stderr, "makeMd5Value() failed\n");

                else r->_check = (int) md5.md5_words[0];
            }

            id = DECODE_ID(cancelGetInvoke, r->_check);
#endif
	    advise(NULLCP, "Queued Get operation with id %08x cancelled\n",
				cancelGetInvoke);
#ifndef	SECURITY
            id = mi->mi_invoke;
#else
            id = theAssocMgr->makeAgentId(msd);
#endif
	    M_GetRes(msd, id, 0, NULL, NULL,
                     mtime(), 0, NULL, m_operationCancelled, NULL, &mis);
#ifndef	SECURITY
            id = cancelGetInvoke;
#endif
	    M_CancelGetRes(msd, id, m_noError, NULL, &mis);
            r->_done = True;
	    return OK;
	}
    }

    // a noSuchInvokeId error

    CMISErrorInfo errorInfo;
    errorInfo.ei_noSuchInvokeId = cancelGetInvoke;
#ifdef	SECURITY
    invoke = theAssocMgr->makeAgentId(msd);
#endif
    M_CancelGetRes(msd, invoke, m_noSuchInvokeId, &errorInfo, &mis);
    advise(NULLCP, "CancelGet error: noSuchInvokeId %08x\n",
           (u_int) cancelGetInvoke);
    return OK;
}


int  CMISAgent::Set (int msd, Bool confirmed,
	int invoke, MID obj_class, MN obj_inst,
	CMISScope* scope, CMISFilter* filter, External*, CMISSync sync,
	int nattrs, CMISSetAttr* attrs)
{
    MO *base, *top, **mobj_tmp;
    CMISSetAttr result[MO_MAXATTRS];
    CMISErrors error = m_noError;
    struct MIDentifier my_class; my_class.mid_type = MID_GLOBAL;
    MName my_inst; my_inst.mn_type = globalNames ? MN_DN : MN_LOCALDN;
    MSAPIndication mis;
    int nres, linked, nobjs, nobjs1, nrepl = 0;
    register int i, j;
    int invokeId;
    Bool allomorphic = False;
    OID  allomorphicClass;
    AVA* err = NULLAVA;

    // check parameters, find base object, check class and do scoping

    if (!(error = checkParms(obj_class, obj_inst, NULL, scope, filter, sync))) {
	top = MO::getRoot() -> getSubordinate();

	DN localdn;
	if ((localdn = mn2localdn(obj_inst)) == NULLDN)
	    error = m_noSuchObjectInstance;
	else
	if ((base = top -> getMO(localdn, True, err)) == NULLMO)
	    error = err ? err -> getError() : m_noSuchObjectInstance;
	else
        if (obj_class -> mid_global &&
		base -> checkClass(obj_class -> mid_global, &allomorphic)
			== NOTOK)
	    error = m_classInstanceConflict;
	else {
	    mobj_tmp = base -> Scope(scope, True, err);
	    if (err)
		error = err -> getError();
	}
    }

    if (error)
	if (confirmed) {
#ifdef	SECURITY
            invoke = theAssocMgr->getManagerId(msd);
#endif
	    return MError(msd, invoke, 0, M_SETC, error, obj_class, obj_inst,
				scope, filter, sync, NULL, NULL, err);
        }
	else
	    return OK;

#ifdef DEBUG
    char* cp;
    advise(NULLCP, "%sConfirmed Set request: id %08x class %s, inst {%s}\nscope %s",
		confirmed ? "" : "Non", (u_int) invoke,
		obj_class->mid_global ? _operationClass : "",
		cp = dn2str(obj_inst->mn_dn), sprintscope(scope));
    free(cp);
    if (filter)
	advise(NULLCP, "filter %s", sprintmfilter(filter));
#endif


    // copy the MO pointers in order to avoid side effects
    // (Scope returns the actual array, not a copy)
    for (nobjs1 = 0; mobj_tmp[nobjs1]; nobjs1++);
    MO** mobj = new MO* [nobjs1+1];
    for (i = 0; i <= nobjs1; i++)
	mobj[i] = mobj_tmp[i];


    // perform filtering
    if (filter) {
	for (i = 0; mobj[i];) {
	    Bool flt = mobj[i] -> Filter(filter, err);
	    if (err) {
	        error = m_processingFailure;	// the only error allowed
#ifdef	SECURITY
		invoke = theAssocMgr->getManagerId(msd);
#endif
		free_mo_array(mobj, nobjs1);
		if (confirmed)
		    return MError(msd, invoke, 0, M_SETC, error,
				obj_class, obj_inst,
				scope, filter, sync, NULL, NULL, err);
		delete err; err = NULLAVA;
		flt = False;
	    }
	    if (flt != True)
		for (j = i; mobj[j]; j++)
		    mobj[j] = mobj[j+1];
	    else
		i++;
	}
    }
    for (nobjs = 0; mobj[nobjs]; nobjs++);

#ifdef DEBUG
    if (scope->sc_type || filter)
	advise(NULLCP, "%d objects selected after %s",
		nobjs, scope->sc_type && filter ? "scoping and filtering" :
		(scope->sc_type  ? "scoping" : "filtering"));
#endif

    // perform set operations and return results/errors
    linked = nobjs > 1 ? invoke : 0;

    for (i = 0; i < nobjs; i++) {
	if (mobj[i] == base) {
	    if (confirmed)
		my_class.mid_global = obj_class -> mid_global ?
			obj_class -> mid_global : mobj[i] -> getClass();
	    allomorphicClass = allomorphic ? my_class.mid_global : NULLOID;
	}
	else {
	    if (confirmed)
		my_class.mid_global = mobj[i] -> getClass();
	    allomorphicClass = NULLOID;
	}
	my_inst.mn_dn = mobj[i] -> getDN(globalNames ? True : False);

	error = m_noError;
#ifdef ACCESS
	int enforceAction =
		checkMOAccess(getInitiator(), mobj[i], scope, filter, AC_SET);

	if (enforceAction == int_Access_EnforcementAction_denyWithoutResponse)
	    continue;

	if (enforceAction == int_Access_EnforcementAction_denyWithResponse)
	    if (!linked)
		error = m_accessDenied;
	    else {
		error = m_setListError;	// get/set linked replies
					// cannot have accessDenied
		nres = 0;
	    }
#endif
	if (!error) {
	    nres = 0;
	    error = mobj[i] -> Set(confirmed, allomorphicClass,
				nattrs, attrs, &nres, result, err);
	}
	if (!confirmed) {
	    if (err) {
		delete err; err = NULLAVA;
	    }
	    continue;
	}

        /* Advance the invokeId value */
        invokeId = nobjs == 1 ?
#ifndef	SECURITY
		invoke : ++_invokeId;
#else
		theAssocMgr->getManagerId(msd) :
		theAssocMgr->makeAgentId(msd);
#endif

	if (!error || error == m_setListError) {
#ifdef DEBUG
	    if (error) {
		char *clss = mobj[i] -> getClassName();
		char *inst = mobj[i] -> getDNString();
		advise(NULLCP, "CMIS SetListError: class %s, instance %s",
			clss, inst);
		free(inst);

		for (j = 0; j < nres; j++)
		    if (result[j].sa_error &&
				result[j].sa_id.mid_type == MID_GLOBAL)
			advise(NULLCP, "Attribute %s: %s",
				oid2name(result[j].sa_id.mid_global, OIDPART),
				MErrString(result[j].sa_error));
	    }
#endif
	    if (M_SetRes(msd, invokeId, linked,
				&my_class, &my_inst,
				mtime(), nres, result, error, NULL, &mis)
			== NOTOK)
		return NOTOK;
	    if (err)
		delete err;
	}
	else
	    if (MError(msd, invokeId, linked,
				M_SETC, error, &my_class, &my_inst,
				NULL, NULL, NULLMSYNC, NULL, NULL, err)
			== NOTOK)
		return NOTOK;
	++nrepl;

#ifdef MINIMISE_MEMORY
	mobj[i] -> releaseEncodedData();
#endif
    }

    if ((nrepl == 0 || linked) && confirmed) {	// final empty result
#ifdef	SECURITY
        invoke = theAssocMgr->getManagerId(msd);
#endif
	M_SetRes(msd, invoke, 0, NULL, NULL, NULL, 0, NULL,
				m_noError, NULL, &mis);
#ifdef DEBUG
	advise(NULLCP, "Set result: dispatched %d linked replies", nrepl);
#endif
    }
    free_mo_array(mobj, nobjs);
    return OK;
}


int  CMISAgent::Action (int msd, Bool confirmed,
	int invoke, MID obj_class, MN obj_inst,
	CMISScope* scope, CMISFilter* filter, External*, CMISSync sync,
	CMISParam* info)
{
    MO *base, *top, **mobj_tmp;
    CMISParam result;
    bzero((char*) &result, sizeof(result)); result.mp_id.mid_type = MID_GLOBAL;
    struct MIDentifier my_class; my_class.mid_type = MID_GLOBAL;
    MName my_inst; my_inst.mn_type = globalNames ? MN_DN : MN_LOCALDN;
    CMISErrors error = m_noError;
    MSAPIndication mis;
    int linked, nobjs, nobjs1, nrepl = 0;
    register int i, j;
    int invokeId;
    Bool allomorphic = False;
    OID  allomorphicClass;
    AVA* err = NULLAVA;

    // check parameters, find base object, check class and do scoping

    if (!(error = checkParms(obj_class, obj_inst, NULL, scope, filter, sync))) {
	top = MO::getRoot() -> getSubordinate();

	DN localdn;
	if ((localdn = mn2localdn(obj_inst)) == NULLDN)
	    error = m_noSuchObjectInstance;
	else
	if ((base = top -> getMO(localdn, True, err)) == NULLMO)
	    error = err ? err -> getError() : m_noSuchObjectInstance;
	else
        if (obj_class -> mid_global &&
		base -> checkClass(obj_class -> mid_global, &allomorphic)
			== NOTOK)
	    error = m_classInstanceConflict;
	else {
	    mobj_tmp = base -> Scope(scope, True, err);
	    if (err)
		error = err -> getError();
	}
    }

    if (error)
	if (confirmed) {
#ifdef	SECURITY
            invoke = theAssocMgr->getManagerId(msd);
#endif
	    return MError(msd, invoke, 0, M_ACTIONC, error, obj_class, obj_inst,
				scope, filter, sync, NULL, NULL, err);
        }
	else
	    return OK;

#ifdef DEBUG
    char* cp;
    advise(NULLCP, "%sConfirmed Action request: id %08x class %s, inst {%s}\nscope %s, action \"%s\"",
		confirmed ? "" : "Non", (u_int) invoke,
		obj_class->mid_global ? _operationClass : "",
		cp = dn2str(obj_inst->mn_dn), sprintscope(scope),
		oid2name(info->mp_id.mid_global, OIDPART));
    free(cp);
    if (filter)
	advise(NULLCP, "filter %s", sprintmfilter(filter));
#endif


    // copy the MO pointers in order to avoid side effects
    // (Scope returns the actual array, not a copy)
    for (nobjs1 = 0; mobj_tmp[nobjs1]; nobjs1++);
    MO** mobj = new MO* [nobjs1+1];
    for (i = 0; i <= nobjs1; i++)
	mobj[i] = mobj_tmp[i];


    // perform filtering
    if (filter) {
	for (i = 0; mobj[i];) {
	    Bool flt = mobj[i] -> Filter(filter, err);
	    if (err) {
	        error = m_processingFailure;	// the only error allowed
#ifdef	SECURITY
		invoke = theAssocMgr->getManagerId(msd);
#endif
		free_mo_array(mobj, nobjs1);
		if (confirmed)
		    return MError(msd, invoke, 0, M_ACTIONC, error,
				obj_class, obj_inst,
				scope, filter, sync, NULL, NULL, err);
		delete err; err = NULLAVA;
		flt = False;
	    }
	    if (flt != True)
		for (j = i; mobj[j]; j++)
		    mobj[j] = mobj[j+1];
	    else
		i++;
	}
    }
    for (nobjs = 0; mobj[nobjs]; nobjs++);

#ifdef DEBUG
    if (scope->sc_type || filter)
	advise(NULLCP, "%d objects selected after %s",
		nobjs, scope->sc_type && filter ? "scoping and filtering" :
		(scope->sc_type ? "scoping" : "filtering"));
#endif

    // perform action operations and return results/errors
    linked = nobjs > 1 ? invoke : 0;

    for (i = 0; i < nobjs; i++) {
	if (mobj[i] == base) {
	    if (confirmed)
		my_class.mid_global = obj_class -> mid_global ?
			obj_class -> mid_global : mobj[i] -> getClass();
	    allomorphicClass = allomorphic ? my_class.mid_global : NULLOID;
	}
	else {
	    if (confirmed)
		my_class.mid_global = mobj[i] -> getClass();
	    allomorphicClass = NULLOID;
	}
	my_inst.mn_dn = mobj[i] -> getDN(globalNames ? True : False);

	error = m_noError;
#ifdef ACCESS
	int enforceAction =
	    checkMOAccess(getInitiator(), mobj[i], scope, filter, AC_ACTION);

	if (enforceAction == int_Access_EnforcementAction_denyWithoutResponse)
	    continue;

	if (enforceAction == int_Access_EnforcementAction_denyWithResponse)
		error = m_accessDenied;
#endif
	if (!error)
	    error = mobj[i] -> Action(confirmed, allomorphicClass,
					info, &result, err);
	if (!confirmed) {
	    if (err) {
		delete err; err = NULLAVA;
	    }
	    continue;
	}

        /* Advance the invokeId value */
        invokeId = nobjs == 1 ?
#ifndef	SECURITY
		invoke : ++_invokeId;
#else
		theAssocMgr->getManagerId(msd) :
		theAssocMgr->makeAgentId(msd);
#endif

	if (!error) {
	    if (M_ActionRes(msd, invokeId, linked,
				&my_class, &my_inst, mtime(),
				result.mp_val ? &result : NULLMPARM,
				m_noError, NULL, &mis)
			== NOTOK)
		return NOTOK;
	    mparm_free(&result);
	}
	else {
	    MIDListVal id_list;
	    id_list.midl_mid = info->mp_id;	// struct cpy
	    id_list.midl_next = NULL;
	    if (MError(msd, invokeId, linked,
				M_ACTIONC, error, &my_class, &my_inst,
				NULL, NULL, NULLMSYNC,
				&id_list, info->mp_val, err)
			== NOTOK)
		return NOTOK;
	}
	++nrepl;

    }

    if ((nrepl == 0 || linked) && confirmed) {	// final empty result
#ifdef	SECURITY
        invoke = theAssocMgr->getManagerId(msd);
#endif
	M_ActionRes(msd, invoke, 0, NULL, NULL, NULL, NULL,
				m_noError, NULL, &mis);
#ifdef	DEBUG
	advise(NULLCP, "Action result: dispatched %d linked replies", nrepl);
#endif
    }
    free_mo_array(mobj, nobjs);
    return OK;
}


int  CMISAgent::Create (int msd, int invoke, MID obj_class, MN obj_inst,
	int inst_type, MN ref_inst, External* access,
	int nattrs, CMISParam* attrs)
{
    MO *superior = NULLMO, *ref = NULLMO, *top, *mo = NULLMO;
    CMISParam result[MO_MAXATTRS]; int nres = 0;
    CMISErrorInfo err_info; CMISErrors error = m_noError;
    bzero((char*) &err_info, sizeof(CMISErrorInfo));
    MName my_inst;
    my_inst.mn_type = globalNames ? MN_DN : MN_LOCALDN;
    my_inst.mn_dn = NULLDN;
    MSAPIndication mis;
    RDN rdn = NULLRDN;
    char *cp;
    DN objinst = NULLDN, refinst = NULLDN;
    MIDListVal id_list, *id_listp = &id_list;
    id_list.midl_mid.mid_type = MID_GLOBAL;
    id_list.midl_mid.mid_global = NULLOID;
    id_list.midl_next = NULL;
    PE value = NULLPE;
    AVA* err = NULLAVA;

    // check parameters
    if (error = checkParms(obj_class, obj_inst, ref_inst,
			NULLMSCOPE, NULLMFILTER, NULLMSYNC)) {
#ifdef	SECURITY
        invoke = theAssocMgr->getManagerId(msd);
#endif
	return MError(msd, invoke, 0, M_CREATE, error, obj_class,
			error == m_noSuchReferenceObject ? ref_inst : obj_inst);
    }

    // check object instance
    if (obj_inst -> mn_dn != NULL) {
	if ((objinst = mn2localdn(obj_inst)) == NULLDN) {
#ifdef	SECURITY
            invoke = theAssocMgr->getManagerId(msd);
#endif
	    return MError(msd, invoke, 0, M_CREATE, m_invalidObjectInstance,
			obj_class, obj_inst);
        }

	top = MO::getRoot() -> getSubordinate();

	if (inst_type == CA_OBJECT_INST) {
	    // check for duplicate instance

	    MO* ref = top -> getMO(objinst, True, err);
	    if (err || ref != NULLMO)
		return MError(msd, invoke, 0, M_CREATE,
			err ? m_processingFailure :
			      m_duplicateManagedObjectInstance,
			obj_class,
			err ? NULLMN : obj_inst,
			NULL, NULL, NULLMSYNC, NULL, NULL, err);

	    // break object instance into superior instance and rdn
	    for (register DN* cur = &obj_inst -> mn_dn;
			(*cur) -> dn_parent; cur = &(*cur) -> dn_parent);
	    rdn = (*cur) -> dn_rdn;
	    (*cur) -> dn_rdn = NULLRDN;
	    dn_free(*cur);
	    *cur = NULLDN;
	}

	// check if superior object exists
	objinst = mn2localdn(obj_inst);
	AVA* err = NULLAVA;
	if ((superior = top -> getMO(objinst, False, err)) == NULLMO) {
	    if (err)
		return MError(msd, invoke, 0, M_CREATE, m_processingFailure,
			obj_class, NULL, NULL, NULL, NULLMSYNC, NULL, NULL,err);
	    // reconstruct distinguished name
	    if (inst_type == CA_OBJECT_INST) {
		if (obj_inst -> mn_dn)
		    dn_append(obj_inst -> mn_dn, dn_comp_new(rdn));
		else
		    obj_inst -> mn_dn = dn_comp_new(rdn);
		rdn = NULLRDN;
	    }
#ifdef	SECURITY
            invoke = theAssocMgr->getManagerId(msd);
#endif
	    return MError(msd, invoke, 0, M_CREATE, m_invalidObjectInstance,
				obj_class, obj_inst);
	}
    }

    // check reference instance
    if (ref_inst -> mn_dn != NULL) {
	if ((refinst = mn2localdn(obj_inst)) == NULLDN) {
#ifdef	SECURITY
            invoke = theAssocMgr->getManagerId(msd);
#endif
	    return MError(msd, invoke, 0, M_CREATE, m_noSuchReferenceObject,
			obj_class, ref_inst);
        }

	top = MO::getRoot() -> getSubordinate();

	if ((ref = top -> getMO(refinst, True, err)) == NULLMO) {
	    if (rdn)
		rdn_free(rdn);
	    return MError(msd, invoke, 0, M_CREATE,
				err ? m_processingFailure :
				      m_noSuchReferenceObject,
				obj_class, err ? NULLMN : ref_inst);
	}
    }

#ifdef DEBUG
    advise(NULLCP, "Create request: id %08x class %s", (u_int) invoke,
			_operationClass);
    cp = NULLCP;
    if (obj_inst->mn_dn)
	advise(NULLCP, "superior instance {%s}", cp = dn2str(obj_inst->mn_dn));
    if (cp) {
	free(cp); cp = NULLCP;
    }
    if (rdn)
	advise(NULLCP, "rdn {%s}", cp = rdn2str(rdn));
    if (cp) {
        free(cp); cp = NULLCP;
    }
    if (ref_inst->mn_dn)
	advise(NULLCP, "reference instance {%s}", cp = dn2str(ref_inst->mn_dn));
    if (cp)
	free(cp);
#endif

#ifdef ACCESS
    if (_adf.checkCreateAccess(getInitiator(), obj_class->mid_global) !=
					int_Access_EnforcementAction_allow)
	error = m_accessDenied;
#endif
    // perform create operation and return result/error
    if (!error)
        mo = MO::Create(obj_class->mid_global, &my_inst.mn_dn,
			rdn, superior, ref, access,
			nattrs, attrs, &error, &err_info);

    if (!mo) {
	if (!error)	// class code did not set it
	    error = m_processingFailure;
    }
    else
	error = mo -> Get(obj_class->mid_global, &nres, result);

    if (!error) {
#ifdef	SECURITY
        invoke = theAssocMgr->getManagerId(msd);
#endif
	M_CreateRes(msd, invoke, obj_class, &my_inst, mtime(), nres, result,
				m_noError, NULL, &mis);
    }
    else {
	switch (error) {
	case m_noSuchAttribute:
	    id_list.midl_mid = err_info.ei_noSuchAttribute;	// struct cpy
	    break;
	    
	case m_missingAttributeValue:
	    id_listp = err_info.ei_missingAttributeValue;
	    break;

	case m_invalidAttributeValue:
	    id_list.midl_mid = err_info.ei_invalidAttributeValue.mp_id;
	    value = err_info.ei_invalidAttributeValue.mp_val;
	    break;

	case m_processingFailure:
	    if (err_info.ei_processingFailure.pf_error.mp_id.mid_global)
	        id_list.midl_mid.mid_global =
			err_info.ei_processingFailure.pf_error.mp_id.mid_global;
	    value = err_info.ei_processingFailure.pf_error.mp_val;

	case m_invalidObjectInstance:
	    // reconstruct distinguished name
	    if (inst_type == CA_OBJECT_INST) {
		if (obj_inst -> mn_dn)
		    dn_append(obj_inst -> mn_dn, dn_comp_new(rdn));
		else
		    obj_inst -> mn_dn = dn_comp_new(rdn);
		rdn = NULLRDN;
	    }
	    if (my_inst.mn_dn) {
		dn_free(obj_inst -> mn_dn);
		obj_inst -> mn_type = my_inst.mn_type;
		obj_inst -> mn_dn = my_inst.mn_dn;
	    }
	    break;

	case m_accessDenied:
	    break;

	default:
	    error = m_processingFailure;
	    id_list.midl_mid.mid_global = NULLOID; value = NULLPE;
	    advise(NULLCP, "Invalid create error %s!", MErrString(error));
	    break;
	}

	if (rdn)
	    rdn_free(rdn);
#ifdef	SECURITY
        invoke = theAssocMgr->getManagerId(msd);
#endif
	return MError(msd, invoke, 0, M_CREATE, error, obj_class, obj_inst,
			NULL, NULL, NULLMSYNC, id_listp, value);
    }

#ifdef DEBUG
    char* inst = dn2str(my_inst.mn_dn);
    advise(NULLCP, "Create result: created MO with inst {%s}", inst);
    advise(NULLCP, "%s", cp = mo->print());
    free(inst); free(cp);
#endif

#ifdef MINIMISE_MEMORY
    if (mo)
	mo -> releaseEncodedData();
#endif
    return OK;
}


int  CMISAgent::Delete (int msd, int invoke, MID obj_class, MN obj_inst,
	CMISScope* scope, CMISFilter* filter, External*, CMISSync sync)
{
    MO *base, *top, **mobj_tmp;
    CMISErrors error = m_noError;
    struct MIDentifier my_class; my_class.mid_type = MID_GLOBAL;
    MName my_inst; my_inst.mn_type = globalNames ? MN_DN : MN_LOCALDN;
    MSAPIndication mis;
    int linked, nobjs, nobjs1, nrepl = 0;
    register int i, j;
    int invokeId;
    AVA* err = NULLAVA;

    // check parameters, find base object, check class and do scoping

    if (!(error = checkParms(obj_class, obj_inst, NULL, scope, filter, sync))) {
	top = MO::getRoot() -> getSubordinate();

	DN localdn;
	if ((localdn = mn2localdn(obj_inst)) == NULLDN)
	    error = m_noSuchObjectInstance;
	else
	if ((base = top -> getMO(localdn, True, err)) == NULLMO)
	    error = err ? err -> getError() : m_noSuchObjectInstance;
	else
        if (obj_class -> mid_global &&
		base -> checkClass(obj_class -> mid_global) == NOTOK)
	    error = m_classInstanceConflict;
	else {
	    mobj_tmp = base -> Scope(scope, True, err);
	    if (err)
		error = err -> getError();
	}
    }

    if (error) {
#ifdef	SECURITY
        invoke = theAssocMgr->getManagerId(msd);
#endif
	return MError(msd, invoke, 0, M_DELETE, error, obj_class, obj_inst,
				scope, filter, sync, NULL, NULL, err);
    }

#ifdef DEBUG
    char* cp;
    advise(NULLCP, "Delete request: id %08x class %s, inst {%s}\nscope %s",
           (u_int) invoke,
		obj_class->mid_global ? _operationClass : "",
		cp = dn2str(obj_inst->mn_dn), sprintscope(scope));
    free(cp);
    if (filter)
	advise(NULLCP, "filter %s", sprintmfilter(filter));
#endif


    // copy the MO pointers in order to avoid side effects
    // (Scope returns the actual array, not a copy)
    for (nobjs1 = 0; mobj_tmp[nobjs1]; nobjs1++);
    MO** mobj = new MO* [nobjs1+1];
    for (i = 0; i <= nobjs1; i++)
	mobj[i] = mobj_tmp[i];


    // perform filtering
    if (filter) {
	for (i = 0; mobj[i];) {
	    Bool flt = mobj[i] -> Filter(filter, err);
	    if (err) {
	        error = m_processingFailure;	// the only error allowed
#ifdef	SECURITY
		invoke = theAssocMgr->getManagerId(msd);
#endif
		free_mo_array(mobj, nobjs1);
		return MError(msd, invoke, 0, M_DELETE, error,
				obj_class, obj_inst,
				scope, filter, sync, NULL, NULL, err);
	    }
	    if (flt != True)
		for (j = i; mobj[j]; j++)
		    mobj[j] = mobj[j+1];
	    else
		i++;
	}
    }
    for (nobjs = 0; mobj[nobjs]; nobjs++);

#ifdef DEBUG
    if (scope->sc_type || filter)
	advise(NULLCP, "%d objects selected after %s",
		nobjs, scope->sc_type && filter ? "scoping and filtering" :
		(scope->sc_type ? "scoping" : "filtering"));
#endif

    // perform delete operations and return results/errors
    linked = nobjs > 1 ? invoke : 0;

    for (i = nobjs-1; i >= 0; i--) {
	if (mobj[i] == base)
	    my_class.mid_global = obj_class->mid_global ?
		oid_cpy(obj_class->mid_global) : oid_cpy(mobj[i]->getClass());
	else
	    my_class.mid_global = oid_cpy(mobj[i] -> getClass());
	my_inst.mn_dn = dn_cpy(mobj[i] -> getDN(globalNames ? True : False));

	error = m_noError;
#ifdef ACCESS
	int enforceAction =
	    checkMOAccess(getInitiator(), mobj[i], scope, filter, AC_DELETE);

	if (enforceAction == int_Access_EnforcementAction_denyWithoutResponse)
	    continue;

	if (enforceAction == int_Access_EnforcementAction_denyWithResponse)
		error = m_accessDenied;
#endif
	if (!error)
	    error = mobj[i] -> Delete(err);

        /* Advance the invokeId value */
        invokeId = nobjs == 1 ?
#ifndef	SECURITY
		invoke : ++_invokeId;
#else
		theAssocMgr->getManagerId(msd) :
		theAssocMgr->makeAgentId(msd);
#endif

	if (!error) {
	    if (M_DeleteRes(msd, invokeId, linked,
				&my_class, &my_inst,
				mtime(), m_noError, NULL, &mis)
			== NOTOK)
		return NOTOK;
	}
	else
	    if (MError(msd, invokeId, linked,
				M_DELETE, error, &my_class, &my_inst,
				NULL, NULL, NULLMSYNC, NULL, NULL, err)
			== NOTOK)
		return NOTOK;
	++nrepl;

	oid_free(my_class.mid_global);
	dn_free(my_inst.mn_dn);
    }

    if (nrepl == 0 || linked) {	// final empty result
#ifdef	SECURITY
        invoke = theAssocMgr->getManagerId(msd);
#endif
	M_DeleteRes(msd, invoke, 0, NULL, NULL, NULL, m_noError, NULL, &mis);
#ifdef	DEBUG
	advise(NULLCP, "Delete result: dispatched %d linked replies", nrepl);
#endif
    }
    free_mo_array(mobj, nobjs);
    return OK;
}


int  CMISAgent::MError (int msd, int invoke, int linked,
			int operation, CMISErrors error,
			MID obj_class, MN obj_inst,
			CMISScope* scope, CMISFilter* filter, CMISSync sync,
			MIDListVal* id_list, PE value, AVA*& errInfo)
{
    CMISErrorInfo err_info;
    MSAPIndication mis;
    int res;

    // fill in the error information

    switch (error) {
    case m_invalidScope:
	err_info.ei_invalidScope = *scope;		// struct copy
	break;

    case m_invalidFilter:
	err_info.ei_invalidFilter = filter;
	break;

    case m_syncNotSupported:
	err_info.ei_syncNotSupported = sync;
	break;

    case m_resourceLimitation:
	break;

    case m_noSuchObjectClass:
	err_info.ei_noSuchObjectClass = *obj_class;	// struct copy
	break;

    case m_noSuchObjectInstance:
	err_info.ei_noSuchObjectInstance = *obj_inst;	// struct copy
	break;

    case m_classInstanceConflict:
	err_info.ei_classInstanceConflict.bmo_class = *obj_class; // struct copy
	err_info.ei_classInstanceConflict.bmo_inst  = *obj_inst;  //  ..
	break;

    case m_accessDenied:
	if (operation == M_ACTIONC && linked)
	    err_info.ei_linkedActionAccessDenied = id_list->midl_mid; // ..
	break;

    case m_processingFailure:
	err_info.ei_processingFailure.pf_class = *obj_class;	// struct copy
	if (!err_info.ei_processingFailure.pf_class.mid_global)
	    err_info.ei_processingFailure.pf_class.mid_global =
						ACTUALCLASS_ASN_OID;
	if (obj_inst)
	    err_info.ei_processingFailure.pf_inst = *obj_inst;	// struct copy
	else
	    bzero((char*)&err_info.ei_processingFailure.pf_inst, sizeof(MName));

	err_info.ei_processingFailure.pf_error.mp_id.mid_type = MID_GLOBAL;

	if (errInfo != NULLAVAREF) {
	    if (errInfo->getOid())
	        err_info.ei_processingFailure.pf_error.mp_id.mid_global =
					errInfo->getOid();
	    else
	        err_info.ei_processingFailure.pf_error.mp_id.mid_global =
					oid_cpy(_processingFailureOID);
	    if (errInfo->getValue())
	        err_info.ei_processingFailure.pf_error.mp_val =
					errInfo->getValue()->encode();
	    else
	        // this is mandatory, so if the user has not supplied it
	        // we assume a GraphicString syntax

	        err_info.ei_processingFailure.pf_error.mp_val =
					gfxs2prim("reason unknown", 14);
	}
	else {	// id_list and value may be set
	    if (id_list && id_list->midl_mid.mid_global)
	        err_info.ei_processingFailure.pf_error.mp_id.mid_global =
					id_list->midl_mid.mid_global;
	    else
	        err_info.ei_processingFailure.pf_error.mp_id.mid_global =
					oid_cpy(_processingFailureOID);
	    if (value)
	        err_info.ei_processingFailure.pf_error.mp_val = value;
	    else
	        // this is mandatory, so if the user has not supplied it
	        // we assume a GraphicString syntax

	        err_info.ei_processingFailure.pf_error.mp_val =
					gfxs2prim("reason unknown", 14);
	}
	break;

    case m_noSuchAction:
	err_info.ei_noSuchAction.nsa_class = *obj_class;	// struct copy
	err_info.ei_noSuchAction.nsa_type  = id_list->midl_mid;	//  ..
	break;

    case m_noSuchArgument:
	err_info.ei_noSuchArgument.nsa_class = *obj_class;	// struct copy
	err_info.ei_noSuchArgument.nsa_type  = id_list->midl_mid; //  ..
	break;

    case m_invalidArgumentValue:
	err_info.ei_invalidArgumentValue.iav_id = id_list->midl_mid;
								// struct copy
	err_info.ei_invalidArgumentValue.iav_val = value;
	break;

    case m_invalidObjectInstance:
	err_info.ei_invalidObjectInstance = *obj_inst;		// struct copy
	break;

    case m_duplicateManagedObjectInstance:
	err_info.ei_duplicateObjectInstance = *obj_inst;	// struct copy
	break;

    case m_noSuchReferenceObject:
	err_info.ei_noSuchReferenceObject = *obj_inst;		// struct copy
	break;

    case m_noSuchAttribute:
	err_info.ei_noSuchAttribute = id_list->midl_mid;	// struct copy
	break;

    case m_invalidAttributeValue:
	err_info.ei_invalidAttributeValue.mp_id = id_list->midl_mid; //  ..
	err_info.ei_invalidAttributeValue.mp_val = value;
	break;

    case m_missingAttributeValue:
	err_info.ei_missingAttributeValue = id_list;
	break;

    default:
	advise(NULLCP, "CMISAgent::MError: operation %s, invalid error code %s",
			MOperString(operation), MErrString(error));
	return NOTOK;
    }

    // send the error reply

    switch (operation) {
    case M_GET:
	res = M_GetRes(msd, invoke, linked, NULL, NULL, mtime(), 0, NULL,
			error, &err_info, &mis);
	break;

    case M_SETC:
	res = M_SetRes(msd, invoke, linked, NULL, NULL, mtime(), 0, NULL,
			error, &err_info, &mis);
	break;

    case M_ACTIONC:
	res = M_ActionRes(msd, invoke, linked, obj_class, obj_inst, mtime(),
                          NULL, error, &err_info, &mis);
	break;

    case M_CREATE:
	res = M_CreateRes(msd, invoke, NULL, NULL, mtime(), 0, NULL,
                          error, &err_info, &mis);
	break;

    case M_DELETE:
	res = M_DeleteRes(msd, invoke, linked, obj_class, obj_inst, mtime(),
			error, &err_info, &mis);
	break;

    default:
	advise(NULLCP, "CMISAgent::MError: invalid operation/error %s/%s!",
			MOperString(operation), MErrString(error));
	return NOTOK;
    }

    // release memory if needed

    switch (error) {
    case m_processingFailure:
	if (errInfo != NULLAVAREF) {
	    delete errInfo;
	    errInfo = NULLAVAREF;
	}
	else
	    mparm_free(&err_info.ei_processingFailure.pf_error);
	break;

    case m_invalidAttributeValue:
	mparm_free(&err_info.ei_invalidAttributeValue);
	break;

    case m_missingAttributeValue:
	free_CMIP_AttributeIdList(err_info.ei_missingAttributeValue);
	break;

    default:
	break;
    }

#ifdef DEBUG
    char *clss = NULLCP, *inst = NULLCP;

    if (obj_class && obj_class->mid_type == MID_GLOBAL && obj_class->mid_global)
	clss = oid2name(obj_class->mid_global, OIDPART);
    if (obj_inst && obj_inst->mn_type != MN_NONSPEC)
	inst = dn2str(obj_inst->mn_dn);

    advise(NULLCP, "CMIS %s error \"%s\": class %s, instance %s",
		    MOperString(operation), MErrString(error), clss, inst);
    if (inst) free(inst);
#endif

    return res == OK ? OK : NOTOK;
}


CMISErrors  CMISAgent::checkParms (MID obj_class, MN obj_inst, MN ref_inst,
	    CMISScope* scope, CMISFilter* filter, CMISSync sync)
{
    MOClass* moClass = NULL;

    if (obj_class &&
	    (obj_class -> mid_type != MID_GLOBAL ||
		(obj_class -> mid_global &&
		(moClass = MOClassInfo::getClassArray()
			-> findClass(
				MO::setOperationClass(obj_class -> mid_global)))
		== NULL)))
	return m_noSuchObjectClass;

    if (moClass)
	_operationClass = moClass -> className;

    if (obj_inst && obj_inst -> mn_type == MN_NONSPEC)
	return m_noSuchObjectInstance;

    if (ref_inst && ref_inst -> mn_type == MN_NONSPEC)
	return m_noSuchReferenceObject;

    if (scope && check_scope(scope) == NOTOK)
	return m_invalidScope;

    if (filter && check_filter(filter) == NOTOK)
	return m_invalidFilter;

    if (sync == s_atomic && scope &&
		!(scope->sc_type == Sc_BaseObject) &&
		!(scope->sc_type != Sc_WholeSubtree &&
		  scope->sc_type != Sc_FirstLevel && scope->sc_level == 0))
	return m_syncNotSupported;

    return m_noError;
}

