//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// EventLog.cc - Logging for Event Reports

// Saleem N. Bhatti
// April 1992

#if !defined(_EventLog_cc_)
#define _EventLog_cc_

#if defined(__GNUG__)
#include <std.h>
#elif defined(__cplusplus)
#include <stdlib.h>
#ifdef AIXV3
#include <sys/stat.h>
#endif
#endif

#include "msap.h"
#if defined(LOGGING)
#include "Log.h"
#endif
#include "EventLog.h"
#include "Error.h"

#include "ObjectClass.h"
#include "ManagedObjectInstance.h"
#include "EventType.h"
#include "Time.h"
#include "String.h"
#include "Integer.h"


extern Bool globalNames;	/* in CmisAgent.cc */

/*
** LogRecord methods
*/
const int
LogRecord::_level = 1;

MOClassInfo LogRecord::_info;

LogRecord::LogRecord(RDN rdn, MO *superior)
: Top (rdn, superior)
{
    /* Initialise Class */
    if (initialiseClass() == NOTOK)
        error(NULL, "LogRecord::LogRecord() : initialiseClass() failed\n");

    _attrs[I_logRecordId] = new Integer(*((int *) rdn->rdn_av.av_struct));
    _attrs[I_loggingTime] = new Time;

    /* Register class */
    registerClass(I_logRecordNAttrs, _attrs, &_info, False);
}

int
LogRecord::initialiseClass()
{
    if (_info.initialised()) return(OK);

    /* set parent class (meta-class tree) */
    _info.setParent(Top::getClassInfo());

    if (
        /* MO class name */
        (_info.setClass("logRecord", 1, I_logRecordNAttrs, 0, 0, 0) == NOTOK) ||

#define BAD_ATTR(s, i) (_info.setAttr(s, i) == NOTOK)

        BAD_ATTR("logRecordId", I_logRecordId) ||
        BAD_ATTR("loggingTime", I_loggingTime) ||

#undef BAD_ATTR

        /* Name binding - not really needed as it is not checked */
        (_info.setNameBinding("logRecord-log", "log", "logRecordId") == NOTOK)

     ) {
#if defined(DEBUG)
        printf("LogRecord::initialiseClass() failed\n");
#endif
        return(NOTOK);
    }

    return(OK);
}

CMISErrors
LogRecord::cmisDelete()
{
#if defined(LOGGING)
    register
    Log *log = (Log *) getSuperior();

    register
    int i = OK, j;

#if defined(PERSISTENCY)
    i = log->remove(logRecordId()->getint());
#if defined(DEBUG)
    if (i == NOTOK) {
        register
        char *cp = getDNString();

        printf("LogRecord::cmisDelete() : log->remove() failed for %s record %d [%s]\n", log->getLogName(), j, cp);
        free(cp);
    }
#endif

    delete this;
    return(i == NOTOK ? m_processingFailure : m_noError);

#else /* PERSISTENCY */

    delete this;
    return m_noError;
#endif /* PERSISTENCY */

#endif /* LOGGING */
}

/*
** EventLogRecord methods
*/
MOClassInfo EventLogRecord::_info;
const int EventLogRecord::_level = 2;

EventLogRecord::EventLogRecord(RDN rdn, MO *superior, Bool leaf)
: LogRecord(rdn, superior)
{
    /* Initialise Class */
    if (initialiseClass() == NOTOK)
        error(NULL, "EventLogRecord::EventLogRecord() : initialiseClass() failed\n");

     /* Attributes */
    _attrs[I_managedObjectClass]    = new ObjectClass;


    // the following is needed for correct filtering on global names
    _attrs[I_managedObjectInstance] = new ManagedObjectInstance;
    ((ManagedObjectInstance*) _attrs[I_managedObjectInstance])
						->setContainingMO(this);

    _attrs[I_eventType]             = new EventType;
    _attrs[I_eventTime]             = new Time;
    _attrs[I_additionalText]        = new String;

    /* Register Class */
    registerClass(I_eventLogRecordNAttrs, _attrs, &_info, False);

    /* Check name binding */
    if (leaf && (checkNameBinding() == NOTOK)) {
#if defined(DEBUG)
        printf("EventLogRecord::EventLogRecord() : checkNameBinding() failed\n");
#endif
        delete this;
        return;
    }
}

EventLogRecord::~EventLogRecord()
{
    /*
    ** This destructor is only for the test object used by the GMS to process
    ** a notification for generating Event Reports or for logging.
    **
    ** The values in the test object could be in use elsewhere.
    ** To prevent them being deleted when the test object is deleted,
    ** the _attrs[] pointers for the test object are set to NULL.
    */
    if (logRecordId()->getint() < 0) {
        _attrs[I_managedObjectClass]->replace(NULLVD);
        _attrs[I_managedObjectInstance]->replace(NULLVD);
        _attrs[I_eventType]->replace(NULLVD);
    }
}

int
EventLogRecord::initialiseClass()
{
    if (_info.initialised()) return(OK);

    /* set parent class (meta-class tree) */
    _info.setParent(LogRecord::getClassInfo());

    if (
        /* MO class name */
        (_info.setClass("eventLogRecord", 1, I_eventLogRecordNAttrs, 0, 0, 0) == NOTOK) ||

#define BAD_ATTR(s, i) (_info.setAttr(s, i) == NOTOK)

        BAD_ATTR("managedObjectClass", I_managedObjectClass) ||
        BAD_ATTR("managedObjectInstance", I_managedObjectInstance) ||
        BAD_ATTR("eventType", I_eventType) ||
        BAD_ATTR("eventTime", I_eventTime) ||
        BAD_ATTR("additionalText", I_additionalText) ||

#undef BAD_ATTR

        /* Name binding */
        (_info.setNameBinding("logRecord-log", "log", "logRecordId") == NOTOK)

    ) {
#if defined(DEBUG)
        printf("EventLogRecord::initialiseClass() failed\n");
#endif
        return(NOTOK);
    }

    _info.setCreate(EventLogRecord::create);

    return(OK);
}

void
EventLogRecord::setInfo(OID object, DN name, OID event)
{
    /*
    ** This method is for the use of the test object!
    */
    if (logRecordId()->getint() > 0)
        return;

    static Bool first = True;
    static struct MIDentifier objClass;
    static struct MIDentifier eventType;
    static MName       objInst;


    if (first) {
	_attrs[I_managedObjectClass]->setval(NULL);
	_attrs[I_managedObjectInstance]->setval(NULL);
	_attrs[I_eventType]->setval(NULL);
	first = False;
    }

    objClass.mid_type = eventType.mid_type = MID_GLOBAL;

    objClass.mid_global = object;
    _attrs[I_managedObjectClass]->replval(&objClass);

    eventType.mid_global = event;
    _attrs[I_eventType]->replval(&eventType);

    objInst.mn_type = globalNames ? MN_DN : MN_LOCALDN;
    objInst.mn_dn = name;

    _attrs[I_managedObjectInstance]->replval(&objInst);

    ((Time *) _attrs[I_eventTime])->set();
}

LogRecordInfo
EventLogRecord::setInfo(const void *)
{
    /*
    ** If this is a derived class, then this method should have been
    ** redefined to map the notification to the particular record type!
    */
    if (strcmp("eventLogRecord", getClassName()) != 0) {
        advise((char *) 0, "EventLogRecord::setInfo(): This should be redefined for class %s", getClassName());
        ((String *) _attrs[I_additionalText])->set((void *) "<Unknown Value>");
        return LogRecordInfo_wrongRecordType;
    }

    /* This notification has no associated record type */
    return LogRecordInfo_noRecordType;
}

void
EventLogRecord::setInfo(const PE info)
{
    if (info == (PE) 0) {
        advise((char *) 0, "EventLogRecord::setInfo(): NULL info!");
        ((String *) _attrs[I_additionalText])->set((void *) "<NULL>");
        return;
    }
    /* Make a hexdump copy of the BER and store it */
    /* Non-standard behaviour */

    char *str = (char *) 0;
    int length;

    if (pe2mem(info, &str, &length) == NOTOK)
        ((String *) _attrs[I_additionalText])->set((void *) "<Bad value>");
    else {
        register
        char *bp, *sp, *buffer = new char[2*length + 1];
        register
        int i;

        for(i = 0, bp = buffer, sp = str; i < length; ++i, bp += 2, ++sp)
            sprintf(bp, "%02x", (u_int) *((u_char *) sp));
        *bp = '\0';

        ((String *) _attrs[I_additionalText])->set((void *) buffer);
	delete buffer;
    }

    if (str != (char *) 0)
        free(str);
}

#endif /* _EventLog_cc_ */
