//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// GenericMO.cc - generic managed object realisation

// By George Pavlou, February 1991


#include <stdlib.h>
#include <signal.h>
#include <string.h>
#if defined(__GNUG__)
#include <std.h>
#include <unistd.h>
#include <sys/stat.h>
extern "C" { int  strncasecmp(const char*, const char*, size_t); }
#elif defined(__cplusplus)
#include <sysent.h>
#endif

#include "msap.h"
#include "Error.h"
#include "GenericMO.h"
#include "Create.h"
#include "Top.h"
#include "CmisAgent.h"
#include "Log.h"
#include "RecordInfo.h"
#ifdef  SECURITY
#include "AssocMgr.h"
#endif

#include "ObjectClass.h"
#include "ObjectClassList.h"


char *logDir = NULLCP;
int  logRecordAccess = 0;

extern char *etcdir; // In msap/syntax.c
extern char *tailor; // In msap/syntax.c

MO*		MO::_root;
MO**		MO::_scopedMOs;
Bool		MO::_wrongnb;
#ifdef PERSISTENCY
DbmFile*	MO::_db;
#endif
char*		MO::_agentMibFile;
Bool		MO::_notificationFunction;

OID		MO::_operationClass;
PE		MO::_attrSetPE;
External*	MO::_accessControl;

static int	__nattrs  [MO_MAXDERIVED];
static Attr**	__attrs   [MO_MAXDERIVED];
static MOClassInfo*	__classes [MO_MAXDERIVED];
static int	__nclasses;


// a version of dn_free that does not free the contained rdns

void  dn_free1 (DN dn)
{
    register DN cur, next;

    for (cur = dn; cur; cur = next) {
	next = cur->dn_parent;
	free((char*) cur);
    }
}


int  MO::checkClass (OID thisclass, Bool* allomorphic)
{
    // the allomorphic flag is not supplied, all the classes
    // of the inheritance branch are checked

    if (allomorphic == (Bool *) 0) {
	for (register int i = _nclasses-1; i >= 0; i--)
	    if (oid_equal(thisclass, _classes[i]->getClass()))
		return OK;
	return NOTOK;
    }


    // ...else only the actual and allomorphic classes

    *allomorphic = False;

    // check the actual class first
    if (oid_equal(thisclass,
		((ObjectClass*) _attrs[TOP_LEVEL][OBJECTCLASS])->getOID()))
	return OK;

    // then the list of allomorphic classes (if any)
    ObjectClassList* allomorphs = (ObjectClassList*)
					_attrs[TOP_LEVEL][ALLOMORPHS];
    if (!allomorphs)
	return NOTOK;

    MIDListVal* val = (MIDListVal*) allomorphs -> get();
    while (val) {
	if (oid_equal(thisclass, val->midl_mid.mid_global)) {
	    if (allomorphic)
		*allomorphic = True;
	    return OK;
	}
	val = val -> midl_next;
    }
    return NOTOK;
}


int  MO::checkPackage (OID package)
{
    for (register int i = 0; i < _nclasses; i++)
	for (register int j = 0; j < _classes[i]->_npackages; j++)
	    if (oid_equal(package, _classes[i]->_packageOids[j]))
		return OK;
    return NOTOK;
}


NameBinding*  MO::getNameBinding ()
{
    NameBinding* nb;
    if (nb = getClassInfo() -> checkNameBinding(_superior, _rdn))
	return nb;
    advise(NULLCP, "No or wrong name binding for class %s!", getClassName());
    return (NameBinding *) NULL;	// not found
}


int  MO::checkNameBinding ()
{
    Attr* namingAttr;

    // copy the class / attr information
    _nclasses = __nclasses;
    _classes  = new MOClassInfo*[__nclasses];
    _nattrs   = new int[__nclasses];
    _attrs    = new Attr**[__nclasses];

    for (register int i = 0; i < __nclasses; i++) {
	_classes[i] = __classes[i];
	_nattrs[i]  = __nattrs[i];
	_attrs[i]   = __attrs[i];
    }
    __nclasses = 0;


    // check the name binding for MOs other than:
    //  - the root, top, and dummies (i.e. not in the MIT)

    _wrongnb = False;
    if (_superior && _superior->_superior &&
				(_activeNB = getNameBinding()) == NULL)
	goto wrongnb;


    // check for duplicate instance

    if (_superior) {
	register MO* mo;
        for (mo = _superior->_subordinate, i=0; mo; mo = mo->_peer)
	    if (rdn_cmp(_rdn, mo->getRDN()) == 0)
	        i++;

        if (i > 1) {
	    char* cp;
	    advise(NULLCP, "MO of class %s with rdn %s already exists!",
			_classes[_nclasses-1]->getClassName(),cp=rdn2str(_rdn));
	    free(cp);
	    goto wrongnb;
        }
    }


    // initialise special attributes

    initialiseSpecialAttrs();

    // set the naming attribute
    // (check if the naming attribute makes a copy, else we make it)

    if (_superior) {	// not for the root
	namingAttr = getAttr(getNameBinding()->getNamingAttr());
	namingAttr -> set(_rdn->rdn_av.av_struct);
	if (namingAttr -> get() == _rdn->rdn_av.av_struct)
	    _rdn->rdn_av.av_struct = namingAttr -> copy();
    }

    return OK;

wrongnb:
    __nclasses = 0;
    _wrongnb = True;
    return NOTOK;
}


int  MO::registerClass(int nattrs, Attr** attrs,
					MOClassInfo* classinfo, Bool link)
{
#ifdef DEBUG
    for (register int i = 0; i < nattrs; ) {
	if (!attrs[i])
	    continue;
	if (!classinfo->getAttrOid(i++)) {
	    advise(NULLCP, "MO::registerClass %s - attrs badly registered - \
check constructor and classInitialise methods!",
			classinfo->getClassName());
	    return NOTOK;
	}
    }
#endif

    __nattrs[__nclasses] = nattrs;
    __attrs[__nclasses] = attrs;
    __classes[__nclasses++] = classinfo;

    // temporary, to enable addressing of already instantiated attributes
    if (_nclasses == 0) {
	_classes  = __classes;
	_nattrs   = __nattrs;
	_attrs    = __attrs;
    }
    _nclasses = __nclasses;

    // maintain linked list of class objects (if asked)
    _classnext = NULLMO;
    if (!link)
	return OK;

    MO* first = classinfo->getFirst();
    if (first) {
	for (register MO* cur = first; cur->_classnext; cur = cur->_classnext);
	cur->_classnext = this;
    }
    else
	classinfo->setFirst(this);	// first instance

    return OK;
}


Attr*  MO::findAttr (OID attrOid, char** attrName)
{
    int attrid;
    for (register int i = 0; i < _nclasses; i++)
	if ((attrid = _classes[i] -> getAttrId(attrOid)) != NOTOK) {
	    if (attrName)
		*attrName = _classes[i] -> _attrNames[attrid];
	    return _attrs[i][attrid];
	}

    return NULLATTR;
}


Attr*  MO::findAttr (char* attrName, int* attrId, int* classLevel)
{
    int attrid;
    for (register int i = 0; i < _nclasses; i++)
	if ((attrid = _classes[i] -> getAttrId(attrName)) != NOTOK) {
	    if (attrId)
		*attrId = attrid;
	    if (classLevel)
		*classLevel = i;
	    return _attrs[i][attrid];
	}
    return NULLATTR;
}


int  MO::findAttr (OID attrOid, int* attrId, int* classLevel,
		OID allomClass, Bool* settable, Attr** dflt)
{
    for (register int i = 0; i < _nclasses; i++)
	if ((*attrId = _classes[i] -> getAttrId(attrOid, settable, dflt))
		!= NOTOK) {
	    if (_attrs[*classLevel = i][*attrId] == NULLATTR)
		return NOTOK;	// possibly allomorphs, packages
	    return OK;
	}
	else
	if (allomClass && oid_equal(allomClass, _classes[i] -> getClass()))
	    return *classLevel = NOTOK;	// don't look any further

    return *classLevel = NOTOK;
}


Bool  MO::isState (int attrId, int classLevel)
{
    char* syntax = _attrs[classLevel][attrId] -> getSyntax();
    int   len    = strlen(syntax);

    if (len && strcasecmp(syntax+len-5, "State") == 0 ||
                        strcasecmp(syntax+len-6, "Status") == 0)
        return True;
    return False;
}


int  MO::findAction (OID actionOid, int* actionId, int* classLevel,
			OID allomClass, Attr** infoTmpl, Attr** replyTmpl)
{
    for (register int i = 1; i < _nclasses; i++)
	if ((*actionId = _classes[i]
		-> getActionId(actionOid, infoTmpl, replyTmpl)) != NOTOK) {
	    *classLevel = i;
	    return OK;
	}
	else
	if (allomClass && oid_equal(allomClass, _classes[i] -> getClass()))
	    return *classLevel = NOTOK;	// don't look any further

    return *classLevel = NOTOK;
}


OID  MO::findEvent (char* eventName)
{
    OID eventOid;
    for (register int i = 1; i < _nclasses; i++)
	if ((eventOid = _classes[i] -> getEventOid(eventName)) != NULLOID)
	    return eventOid;
    return NULLOID;
}


int  MO::findEvent (OID eventOid, int* eventId, int* classLevel,
							char** eventName)
{
    for (register int i = 1; i < _nclasses; i++)
	if ((*eventId = _classes[i] -> getEventId(eventOid))
		!= NOTOK) {
	    *classLevel = i;
	    if (eventName)
		*eventName = _classes[i] -> _eventNames[*eventId];
	    return OK;
	}
    if (eventName)
	*eventName = NULLCP;
    return *eventId = *classLevel = NOTOK;
}


void  MO::buildDN ()
{
    register MO* cur = _superior;
    if (cur -> _superior == NULLMO) {
	_dn = dn_comp_new(NULLRDN);	// the system MO
	return;
    }

    register DN mydn = dn_comp_new(getRDN()), dncomp;
    while (cur -> _superior -> _superior) {
	dncomp = dn_comp_new(cur -> getRDN());
	dn_append(dncomp, mydn);
	mydn = dncomp;
	cur = cur -> _superior;
    }
    _dn = mydn;
}


void  MO::nullifyDN ()
{
    if (_dn)
	dn_free1(_dn);
    _dn = NULLDN;

    for (register MO* mo = _subordinate; mo; mo = mo -> _peer)
	mo -> nullifyDN();
}


int  MO::setRDN (RDN rdn)
{
    // set the object's RDN
    if (_rdn)
	rdn_free(_rdn);
    _rdn = rdn;

    // allign the naming attribute
    NameBinding* nb = getNameBinding();
    if (!nb)
	return NOTOK;

    // set the naming attribute
    // check if the naming attribute makes a copy, else make it
    Attr* namingAttr = getAttr(nb->getNamingAttr());
    namingAttr -> set(rdn->rdn_av.av_struct);
    if (namingAttr -> get() == rdn->rdn_av.av_struct)
	rdn->rdn_av.av_struct = namingAttr -> copy();

    // nullify the DNs of the whole subtree
    nullifyDN();

    return OK;
}


OID*  MO::getGroup (OID groupid, int* ngroupattrs, OID allomClass)
{
    OID* group;
    for (register int i = 0; i < _nclasses; i++)
	if (group = _classes[i] -> getGroupAttrs(groupid, ngroupattrs))
	    return group;
	else
	if (allomClass && oid_equal(allomClass, _classes[i] -> getClass()))
	    return (OID *) 0;	// don't look any further

    return (OID *) 0;
}


Attr*  MO::getUpdatedAttr (OID attrOid, AVA*& err)
{
    int attrId, classLevel;
    AVA* err_aux = NULLAVA;

    if (findAttr(attrOid, &attrId, &classLevel) == NOTOK)
	return NULLATTR;
    if (get(attrId, classLevel, err == DISCARDAVA ? err_aux : err ) == NOTOK ||
	    get(MO_NOMOREATTRS, MO_ALLEVELS, err == DISCARDAVA ? err_aux : err) == NOTOK) {
	if (err_aux)
	    delete err_aux;
	return NULLATTR;
    }
    return _attrs[classLevel][attrId];
}


CMISErrors  MO::Get (OID allomClass, int* nattrs, CMISParam* attrs)
{
    // this Get does not have a AVA*& errInfo argument as the only
    // possible error here is due to a MIB bug (attribute encoding)

    register int i, j, k = 0;
    int classNAttrs;
    Attr** classAttrs;

    for (i = 0; i < _nclasses; i++) {
	classAttrs = _attrs[i];

	for (j = classNAttrs = 0; classNAttrs < _nattrs[i]; j++) {
	    if (!classAttrs[j])
		continue;
	    classNAttrs++;
	    attrs[k].mp_id.mid_type = MID_GLOBAL;
	    attrs[k].mp_id.mid_global = _classes[i]->getAttrOid(j);
	    if (classAttrs[j] -> get(&attrs[k++].mp_val) != m_noError) {
		advise(NULLCP, "Error in encoding attribute %s of class %s - \
processingFailure!\n",
		    _classes[i]->getAttrName(j), _classes[i]->getClassName());
		
		return m_processingFailure;
	    }
	}

	if (allomClass && oid_equal(allomClass, _classes[i] -> getClass())) {
	    *nattrs = k;
	    return m_noError;
	}
    }

    *nattrs = k;
    return m_noError;
}


#define MemoryValueSize 2048
#define HexValueSize 4096 /* 2 * MemoryValueSize */

#ifdef PERSISTENCY

/*
** The dbm file has entries of the format :
**
**   <MO ClassName>:<MO DN>:<nAttrs>:<AVA>:<AVA>: ... :
**
** with nAttr AVAs. <MO ClassName> will be used to recreate the MIB, when
** such a mechanism is introduced into the GMS. AVA has the format :
**
**   <AttributeName>=<hex-byte representation of PE>
**
** AttributeName is as in oidtable.at
*/
datum *
MO::getKey(int k)
{ if (_db == (DbmFile *) 0) return (datum *) 0; return dbmGetKey(_db, k); }

int
MO::store(datum *key, datum *entry)
{
    if (_db == (DbmFile *) 0) return(NOTOK);

    register
    int i;

    i = dbmStore(_db, key, entry);
#if defined(DEBUG)
    if (i < 0) { printf("MO::store() : dbmStore() failed\n"); perror(""); }
#endif
    return(i < 0 ? NOTOK : OK);
}

int
MO::fetch(datum *key, datum *entry)
{
    if (_db == (DbmFile *) 0) return(NOTOK);

    datum *e = dbmFetch(_db, key);

    if (e != (datum *) 0) *entry = *e;
#if defined(DEBUG)
    if (e == (datum *) 0) printf("MO::fetch() : dbmFetch() failed\n");
#endif
    return(e == (datum *) 0 ? NOTOK : OK);
}

int
MO::remove(datum *key)
{
    if (_db == (DbmFile *) 0) return(NOTOK);

    datum entry;

    if (fetch(key, &entry) == NOTOK) return(OK); // never saved

    if (dbmDelete(_db, key) < 0) {
#if defined(DEBUG)
        printf("MO::remove(%s) : dbmDelete() failed\n", key->dptr);
#endif
        return(NOTOK);
    }
    return(OK);
}

int
MO::save()
{
    if (_db == (DbmFile *) 0) return(NOTOK);

    char buffer[DBM_MaxRecordSize + 1];
    int length = DBM_MaxRecordSize + 1;

    if (save(buffer, &length) == NOTOK) {
#if defined(DEBUG)
        printf("MO::save() : save(buffer, length) failed\n");
#endif
        return(NOTOK);
    }

    register
    char *dnStr = getDNString();
    datum key, entry;

    key.dptr = dnStr;
    key.dsize = strlen(dnStr);
    entry.dptr = buffer;
    entry.dsize = strlen(buffer);

    register
    int i = store(&key, &entry);
#if defined(DEBUG)
    if (i == NOTOK) printf("MO::save() : MO::store() failed\n");
#endif
    free(dnStr);
    return(i);
}

int
MO::remove()
{
    if (_db == (DbmFile *) 0) return(NOTOK);

    datum key;

    key.dptr = getDNString();
    key.dsize = strlen(key.dptr);

    register
    int i = remove(&key);
#if defined(DEBUG)
    if (i == NOTOK)
        printf("MO::remove() : dbmDelete() failed for %s\n", key.dptr);
#endif
    free(key.dptr);
    return(i);
}

int
MO::restore()
{
    if (_db == (DbmFile *) 0) return(NOTOK);

    datum key, entry;
    register
    char *dnStr = getDNString();

    key.dptr = dnStr;
    key.dsize = strlen(dnStr);

    register
    int i = fetch(&key, &entry);
#if defined(DEBUG)
    if (i == NOTOK) printf("MO::restore() : MO::fetch() failed for key %s\n", dnStr);
#endif
    if (i == NOTOK) { free(dnStr); return(NOTOK); }

    char data[DBM_MaxRecordSize + 1];
    bcopy(entry.dptr, data, entry.dsize);
    data[entry.dsize] = '\0';

    i = restore(data);
#if defined(DEBUG)
    if (i == NOTOK) printf("MO::restore() : MO::restore(data) failed for %s\n", dnStr);
#endif
    free(dnStr);
    return(i);
}

#endif	/* PERSISTENCY */

int
MO::save(char *buffer, int *length)
{
    register
    char *dnStr;
    register
    int i, nAttrs;

    /*
    ** The buffer has to be big enough for at least storage of the following :
    **
    **   class name
    **   DN
    **   nAttrs
    **   objectClass
    **   nameBindings
    **
    ** (and usually a few more attributes!), so I think a buffer space of at
    ** least 256 bytes is called for. Any less than this and I don't bother
    ** trying. Ideally, use a size of DBM_MaxRecordSize + 1.
    */
    if (*length < 256) {
#if defined(DEBUG)
        printf("MO::save(char *, int *) : not enough buffer space\n");
#endif
        return(NOTOK);
    }

    for(i = 0, nAttrs = 0; i < _nclasses; ++i)
        nAttrs += _nattrs[i];
    /* Record "header" */
    sprintf(buffer, "%s:%s:%d:",
		getClassName(), dnStr = getDNString(), nAttrs); // _nattrs

    char value[HexValueSize + 1];
    register
    char *bp;

    /* AVAs */
    for(i = 0; i < _nclasses; ++i) {

        register
        MOClassInfo *info = _classes[i];
        register
        int j;

        for(j = 0, bp = buffer + strlen(buffer), nAttrs = _nattrs[i];
        (j < nAttrs ) && (bp < ((char *) (buffer + *length - 1)));
        ++j, bp = buffer + strlen(buffer)) {

            /* Get attribute name and value */
            register
            char *name = info->getAttrName(j);
            char memory[MemoryValueSize];
            int length1 = 0;
            register
            Attr *ap = findAttr(name);
            PE pe = ap->encode();

            if ((pe != (PE) 0) && (ps_get_abs(pe) <= MemoryValueSize)) {
                char *mp = memory;
                if (pe2mem(pe, &mp, &length1) == NOTOK) {
#if defined(DEBUG)
                    printf("MO::save(char *, int *) : pe2mem() failed for %s\n", name);
#endif
                    free(dnStr);
                    length1 = 0;
		    return(NOTOK);
                }
            }
#if defined(DEBUG)
            if ((pe != (PE) 0) && (ps_get_abs(pe) > MemoryValueSize))
                printf("MO::save(char *, int *) : out of buffer space for %s\n", name);
#endif
            /* Convert binary to hex-byte */
            register
            char *cp0, *cp1;
            register
            int k;

            if (length1 == 0) {
                value[0] = '-';
                value[1] = '\0';
            }
            else {
                for(k = 0, cp0 = value, cp1 = memory; k < length1; ++k, cp0 += 2, ++cp1)
                    sprintf(cp0, "%02x", (unsigned int) *((unsigned char *) cp1));
            }

            /* write into buffer */
            if (strlen(name) + strlen(value) + 2 + strlen(bp) > *length) {
#if defined(DEBUG)
                printf("MO::save() : out of buffer space\n");
#endif
                free(dnStr);
                return(NOTOK);
            }
            sprintf(bp, "%s=%s:", name, value);
        }
    }
    *length = strlen(buffer);

    free(dnStr);
    return(OK);
}

int
MO::restore(char *data)
{
    char name[80];
    char value[HexValueSize + 1];
    register
    char *cp0 = data, *cp1;
    register
    int i, nAttrs;

#if defined(DEBUG)
    sscanf(cp0, "%[^:]", value); // MO class
    printf("MO::restore() : %s ", value);
#endif
    cp1 = strchr(cp0, ':'); cp0 = cp1 + 1;
#if defined(DEBUG)
    sscanf(cp0, "%[^:]", value); // DN
    printf("{%s}\n", value);
#endif
    cp1 = strchr(cp0, ':'); cp0 = cp1 + 1;
    sscanf(cp0, "%[^:]", value); // nAttrs
    nAttrs = atoi(value);

    for(i = 0; i < nAttrs; ++i) {

        cp1 = strchr(cp0, ':'); cp0 = cp1 + 1;
        sscanf(cp0, "%[^=]=%[^:]", name, value);

        if (value[0] == '-') continue; // no PE value was written

        register
        Attr *ap = findAttr(name);
        if (ap == (Attr *) 0) {
#if defined(DEBUG)
            printf("MO::restart() : no such attribute - %s\n", name);
#endif
            continue;
        }

        register
        char *vcp0, *vcp1;
        char memory[MemoryValueSize];
        register
        int l;
        int v;

        /* Covert hex-byte to binary */
        for(l = 0, vcp0 = value, vcp1 = memory; *vcp0 != '\0';
            ++l, vcp0 += 2, ++vcp1) {
            sscanf(vcp0, "%02x", &v);
            *vcp1 = (char) v;
        }

        PE pe;
        if (mem2pe(&pe, memory, l) == NOTOK) {
#if defined(DEBUG)
            printf("MO::restore() : mem2pe() failed for %s\n", name);
            continue;
#else
            return(NOTOK);
#endif
        }
        ap->set(pe);
    }

    return(OK);
}

#undef MemoryValueSize
#undef HexValueSize


MO::MO ()
{
    if (_root == (MO *) 0)
	_root = this;
    _rdn = NULLRDN;
    _dn = NULLDN;
    _peer = _subordinate = _superior = NULLMO;
    _nclasses = 0;
    _updated = _persistent = False;
    _activeNB = NULL;

#ifdef PERSISTENCY

    /*
    ** create/open the database file ...
    */
    char fileName[512];

    /* Make directory */
    static logDirInit = 0;

    if (logDirInit == 0) {
        FILE *fp;
        char buffer[256]; /* I hope this is enough! */
        char pathName[256]; /* I hope this is enough */
        char access[5];
        int pathFlag = 0, accessFlag = 0;

        if ((fp = fopen(tailor, "r")) == (FILE *) 0)
            error((char *) 0, "MO:MO() : could not fopen() tailor %s", tailor);

        while(fgets(buffer, (sizeof(buffer) - 1), fp) && (*buffer != EOF)) {

            if (!pathFlag && (sscanf(buffer, "gmsLogDir : %255s", pathName) == 1))
                pathFlag = 1;

            if (!accessFlag && (sscanf(buffer, "gmsLogRecordAccess : %4s", access) == 1))
                accessFlag = 1;

            if (pathFlag && accessFlag)
                break;
        }
        fclose(fp);

        if (pathFlag == 0)
            error((char *) 0, "MO:MO() : Could not read gmsLogDir from %s", tailor);

        if (accessFlag == 0)
            error((char *) 0, "MO:MO() : Could not read gmsLogRecordAccess from %s", tailor);
        if (strcmp(access, "on") == 0)
            logRecordAccess = 1;
        else if (strcmp(access, "off") == 0)
            logRecordAccess = 0;
        else
            error((char *) 0, "MO::MO() : Bad value for gmsLogRecordAccess - %s", access);

#if defined(LOGGING) && !defined(PERSISTENCY)
        /*
        ** If LOGGING is to be used without PERSISTENCY, then insist that
        ** logRecords acn always be accessed via CMIS.
        */
        if (!logRecordAccess)
            error((char *) 0, "MO::MO() : Set gmsLogRecordAccess to 'on'\n");
#endif

        logDir = strdup(pathName);
        logDirInit = 1;
    }

    mode_t mode = umask(0);

    sprintf(fileName, "%s/%s", logDir, MO_AgentMibDirectory);
    if ((mkdir(fileName, MO_AgentDirectoryMode) == -1) && (errno != EEXIST))
        error((char *) 0, "MO::MO() : mkdir() failed for %s", fileName);

    /* Open file */
    sprintf(fileName, "%s/%s/%s-%s", logDir, MO_AgentMibDirectory,
        CMISAgent::getInstance()->getAEI(), MO_AgentMibFile);
    _agentMibFile = strdup(fileName);
    if ((_db = dbmOpen(_agentMibFile, DBM_ReadWrite, MO_AgentMibFileMode))
        == (DbmFile *) 0)
        error((char *) 0, "MO::MO() : dbmOpen() failed for %s", _agentMibFile);

    umask(mode);

    /*
    ** If file already exists, then the MIB it contains should be restarted.
    ** The mechanism for this "restart" is to be added later.
    */

#else

    logDir = NULLCP;

#endif	/* PERSISTENCY */

}


MO::MO (RDN myrdn, MO* mysuperior)
{
    // initialise
    _rdn = myrdn;
    _dn = NULLDN;
    _nclasses = 0;
    _updated = _persistent = False;
    _activeNB = NULL;

    // put object in the containment hierarchy
    if (mysuperior) {
	for (register MO** cur = &mysuperior -> _subordinate;
				*cur; cur = &(*cur) -> _peer);
	*cur = this;
    }

    _superior = mysuperior;
    _peer = _subordinate = NULLMO;
}


MO::~MO ()
{
    // delete subordinates
    register MO *cur, *next;
    for (cur = _subordinate; cur; cur = next) {
	next = cur -> _peer;
	delete cur;	// recursive virtual destructor
    }

    if (_superior) {		// for objects in the MIT
				// unlink from the containment hierarchy
	for (register MO** prev = &_superior -> _subordinate;
				*prev; prev = &(*prev) -> _peer)
	    if (*prev == this) {
		*prev = _peer;
		break;
	    }
    }
    else
    if (this == _root) {	// before the root dies release static info
	MOClassInfo::_classArray -> clear();
	delete MOClassInfo::_classArray;
	delete _scopedMOs;
        delete _agentMibFile;
        if (logDir != NULLCP)
            free(logDir);
    }

    // unlink from linked list of same class objects
    MO* first = _classes[_nclasses-1]->getFirst();
    if (first)
	if (first == this)
	    _classes[_nclasses-1]->setFirst(_classnext);
	else {
	    for (register MO* cur = first; cur -> _classnext != this;
						cur = cur ->_classnext);
	    cur -> _classnext = _classnext;
	}

    // delete attributes
    register int i, j, classNAttrs;
    Attr** classAttrs;

    for (i = 0; i < _nclasses; i++) {
	classAttrs = _attrs[i];
	for (j = classNAttrs = 0; classNAttrs < _nattrs[i]; j++) {
	    if (!classAttrs[j])
                continue;
	    classNAttrs++;
	    classAttrs[j] -> clear();
	    delete classAttrs[j];
	}

	_classes[i] = (MOClassInfo *) 0; // to avoid calling the destructor
	_attrs[i]   = (Attr **) 0;
    }

    delete[] _classes;
    delete[] _nattrs;
    delete[] _attrs;

    // free the rdn / dn
    if (_rdn) rdn_free(_rdn);
    if (_dn) dn_free1(_dn);
}


/*
 **************************** ex-MibAccess.cc ********************************
 */


static int  nblocks;


// Methods to Find, Scope and Filter Managed Objects

MO*  MO::getSubordinate_aux (RDN rdn, Bool refresh, AVA*& err)
{
    if (refresh && refreshSubordinate(rdn, err) != OK)
	return NULLMO;

    for (register MO* cur = _subordinate; cur; cur = cur -> _peer)
	if (rdn_cmp(rdn, cur -> getRDN()) == 0)
	    return cur;

    return NULLMO;	// can't find it
}


MO*  MO::getMO_aux (DN dn, Bool refresh, AVA*& err)
{
    if (dn == NULLDN || dn -> dn_rdn == NULLRDN)
	return this;	// found it

    MO* part = getSubordinate_aux(dn -> dn_rdn, refresh, err);
    if (part == NULLMO)
	return NULLMO;	// can't find it or error

    return part -> getMO_aux(dn -> dn_parent, refresh, err);
}


MO**  MO::Scope (CMISScope* scope, Bool refresh, AVA*& err)
{
    if (!_scopedMOs) {
	_scopedMOs = (MO**) malloc(MO_SCOPEBLOCK * sizeof(MO*));// first time
	nblocks = 1;
    }
    else
    if (nblocks > 1)
	_scopedMOs = (MO**) realloc((char*) _scopedMOs,
				(nblocks=1) * MO_SCOPEBLOCK * sizeof(MO*));

    if (scope == NULL || scope -> sc_type == Sc_BaseObject) {
	_scopedMOs[0] = this;
	_scopedMOs[1] = NULLMO;
	return _scopedMOs;
    }

    int i = 0;
    AVA* err_aux = NULLAVA;

    switch (scope -> sc_type) {
    case Sc_FirstLevel:
	if (refresh && refreshSubordinates(err == DISCARDAVA ? err_aux : err) == NOTOK) {
	    if (err_aux)
		delete err_aux;
	    return (MO**) NULL;
	}
    {
	for (register MO* cur = _subordinate; cur; cur = cur -> _peer) {
	    _scopedMOs[i] = cur;
	    if (++i % MO_SCOPEBLOCK == 0)
		_scopedMOs = (MO**) realloc((char*) _scopedMOs,
				++nblocks*MO_SCOPEBLOCK*sizeof(MO*));
	}
    }
	break;

    case Sc_WholeSubtree:
    case Sc_IndividualLevel:
    case Sc_BaseToNthLevel:
	if (searchMIB(&i, scope, 0, refresh, err == DISCARDAVA ? err_aux : err) == NOTOK) {
	    if (err_aux)
		delete err_aux;
	    return (MO**) NULL;
	}
	break;

    default:
	return (MO**) NULL;	// never reached because of check_scope()
    }

    _scopedMOs[i] = NULLMO;
    return _scopedMOs;
}


int  MO::searchMIB (int* i, CMISScope* scope, int level, Bool refresh,
			AVA*& err)
{
    switch (scope -> sc_type) {
    case Sc_WholeSubtree:
	_scopedMOs[*i] = this;
	if (++(*i) % MO_SCOPEBLOCK == 0)
	    _scopedMOs = (MO**) realloc((char*) _scopedMOs,
				++nblocks * MO_SCOPEBLOCK * sizeof(MO*));
	if (refresh) {
	    refreshSubordinates(err);
	    if (err && err->getError())
		return NOTOK;
	}
	break;

    case Sc_IndividualLevel:
	if (level > scope -> sc_level)
	    return OK;
	if (level == scope -> sc_level) {
	    _scopedMOs[*i] = this;
	    if (++(*i) % MO_SCOPEBLOCK == 0)
		_scopedMOs = (MO**) realloc((char*) _scopedMOs,
				++nblocks*MO_SCOPEBLOCK*sizeof(MO*));
	}
	else				// level < scope -> sc_level
	if (refresh) {
	    refreshSubordinates(err);
	    if (err && err->getError())
		return NOTOK;
	}
	break;

    case Sc_BaseToNthLevel:
	if (level > scope -> sc_level)
	    return OK;
	_scopedMOs[*i] = this;
	if (++(*i) % MO_SCOPEBLOCK == 0)
	    _scopedMOs = (MO**) realloc((char*) _scopedMOs,
				++nblocks * MO_SCOPEBLOCK * sizeof(MO*));
	if (level < scope -> sc_level && refresh) {
	    refreshSubordinates(err);
	    if (err && err->getError())
		return NOTOK;
	}
	break;
    }

    if (_subordinate &&
		_subordinate -> searchMIB(i, scope, level+1, refresh, err)
	    == NOTOK)
	return NOTOK;

    if (_peer && level &&
		_peer -> searchMIB(i, scope, level, refresh, err) == NOTOK)
	return NOTOK;

    return OK;
}


Bool  MO::Filter (CMISFilter* flt, AVA*& err)
{
    AVA* err_aux = NULLAVA;

    if (get(MO_ALLATTRS, MO_ALLEVELS, err == DISCARDAVA ? err_aux : err) == NOTOK) {
	if (err_aux)
	    delete err_aux;
	return False;
    }

    _updated = True;
    Bool val = Filter_aux(flt);
    if (val == False)
	_updated = False;	// operation will not reset it
    return val;
}


Bool  MO::Filter_aux (CMISFilter* flt)
{
    switch (flt -> flt_type) {
    case Filter_Item:
    {
	Attr* attr;
        void* value;
	Bool  status;

	if (flt -> flt_item -> fi_type == FI_Substrings) {
	    register CMISFilterSubstrings* substr =
					flt->flt_item->fi_substrings;
	    if ((attr = findAttr(substr->fi_sub_attrid.mid_global)) == NULLATTR)
		return False;

	    for (; substr; substr = substr->fi_sub_next) {
		if ((value = attr -> decode(substr->fi_sub_string))
				== NULLVD &&
			substr->fi_sub_string->pe_errno
				!= PE_ERR_NONE)
		    return False;	// decoding failed

		Bool mode = (substr == flt->flt_item->fi_substrings) ?
					True : False;
		status = attr ->
			filterSubstring(substr->fi_sub_type, value, mode);

		attr -> ffree(value);	// free decoded value
		if (status == False)
		    return False;
	    }
	    return True;	// logical ANDing of substring list
	}

	if ((attr = findAttr(flt->flt_item->fi_attr.mp_id.mid_global))
		== NULLATTR)
	    return False;

	if (flt->flt_item->fi_type == FI_Present)
	    return True;

	if ((value = attr -> decode(flt->flt_item->fi_attr.mp_val)) == NULLVD
		&& flt->flt_item->fi_attr.mp_val->pe_errno != PE_ERR_NONE)
	    return False;	// decoding failed

	status = attr -> filter(flt->flt_item->fi_type, value);

	attr -> ffree(value);	// free decoded value
	return status;
    }

    case Filter_Not:
	return (Bool) !Filter_aux(flt -> flt_filter);

    case Filter_And:
    {
	for (register CMISFilter* cur = flt -> flt_filter;
					cur; cur = cur -> flt_next)
	    if (Filter_aux(cur) == False)
		return False;
    }
	return True;

    case Filter_Or:
    {
	register CMISFilter* cur = flt -> flt_filter;
	if (!cur)
	    return True;
	for ( ; cur; cur = cur -> flt_next)
	    if (Filter_aux(cur) == True)
		return True;
    }
	return False;

    default:
	return False;	// never reached because of check_filter()
    }
}


// Internal CMIS Interface to Managed Objects


CMISErrors  MO::Get (OID objClass, int nattrs, MID attrs,
			int* nres, CMISGetAttr* cmisRes, AVA*& err)
{
    CMISErrors error = m_noError;

    // check first if no or all attributes are requested and respond

    if (nattrs == GET_NOATTRS) {
	if (nres) *nres = 0;
	return m_noError;
    }
    else
    if (nattrs == GET_ALLATTRS) {
	if (!cmisRes) {
            if (nres) *nres = 0;
	    err = new AVA(m_processingFailure);	// only for a non-debugged MIB
            return m_processingFailure;		// the GET_ALLATTRS feature
        }					// is not supported locally

	CMISParam res[MO_MAXATTRS];
	if (!_updated && get(MO_ALLATTRS, MO_ALLEVELS, err) == NOTOK)
	    return m_processingFailure;		// the only allowed error

	error = Get(objClass, nres, res);

	if (!error)
	    for(register int i = 0; i < *nres; i++) {
		cmisRes[i].ga_error = m_noError;
		cmisRes[i].ga_ava = res[i];	// struct copy
	    }
	_updated = False;
	return error;
    }

    // ...else find the requested attributes first

    int attrid[MO_MAXATTRS], level[MO_MAXATTRS];

    for (register int i=0, cur=0; i < nattrs && cur < MO_MAXATTRS; i++, cur++) {
	if (cmisRes)
	    cmisRes[cur].ga_id = attrs[i];	// struct copy

	if (attrs[i].mid_type != MID_GLOBAL) {
	    level[cur] = -1;
	    continue;
	}

	if (findAttr(attrs[i].mid_global, &attrid[cur], &level[cur], objClass)
		== OK) {
	    if (level[cur] > 0 && !_updated &&
			get(attrid[cur], level[cur], err) == NOTOK)
		return m_processingFailure;	// only allowed error
	    continue;
	}

	// check in case it is a group attribute
	OID* group;
	int  ngroupattrs;

	if (!cmisRes ||		// group attributes are not supported locally
	    (group = getGroup(attrs[i].mid_global, &ngroupattrs, objClass))
			== NULL) {
	    level[cur] = -1;
	    continue;
	}

	// it IS a group attribute
	for (register int j = 0; j < ngroupattrs; j++, cur++)
	    if (findAttr(group[j], &attrid[cur], &level[cur], objClass) == OK) {
		if (!_updated && get(attrid[cur], level[cur], err) == NOTOK)
		    return m_processingFailure;	// only allowed error
		cmisRes[cur].ga_id.mid_type = MID_GLOBAL;
		cmisRes[cur].ga_id.mid_global = group[j];
	    }
	    else {
		// should not happen for a debugged MIB
		advise(NULLCP, "cannot find group attr %s!",
							sprintoid(group[j]));
		_updated = False;
		return m_processingFailure;
	    }
	cur--;
    }

    if (get(MO_NOMOREATTRS, MO_ALLEVELS, err) == NOTOK)
	return m_processingFailure;	// only allowed error

    if (nres) *nres = cur;
    _updated = False;

    // now do get them

    for (i=0; i < cur; i++) {
	if (level[i] == -1) {
	    error = m_getListError;
	    cmisRes[i].ga_error = m_noSuchAttribute;
	    continue;
	}

	switch(cmisRes[i].ga_error =
		_attrs[level[i]][attrid[i]] -> get(&cmisRes[i].ga_val)) {
	case m_noError:
	    break;

	case m_accessDenied:		// should not happen (yet)
	    error = m_getListError;
	    break;

	case m_processingFailure:
	    return m_processingFailure;

	default:
	    // should not happen for a debugged MIB
	    advise(NULLCP, "get returned invalid error code: %d !",
				(int) cmisRes[i].ga_error);
	    return m_processingFailure;
	}
    }

    return error;
}


CMISErrors  MO::Set (Bool confirmed, OID objClass,
		int nattrs, CMISSetAttr* attrs,
		int *nres, CMISSetAttr* result, AVA*& err)
{
    CMISErrors error = m_noError;

    _updated = False;

    // find and set the requested attributes

    int		attrid[MO_MAXATTRS],
		level[MO_MAXATTRS];
    Bool	settable[MO_MAXATTRS];
    void*	val[MO_MAXATTRS];
    CMISErrors	errors[MO_MAXATTRS];
    Attr*	dflt;

    for (register int i=0, cur=0; i < nattrs && cur < MO_MAXATTRS; i++, cur++) {
	errors[cur] = m_noError;
	val[cur] = NULLVD;

	result[cur].sa_ava = attrs[i].sa_ava;		// struct copy
	result[cur].sa_modify = attrs[i].sa_modify;

	if (attrs[i].sa_id.mid_type != MID_GLOBAL) {
	    level[cur] = -1;
	    errors[cur] = m_noSuchAttribute;
	    continue;
	}

	if (findAttr(attrs[i].sa_id.mid_global, &attrid[cur], &level[cur],
			objClass, &settable[cur], &dflt) == OK) {
	    Attr* attr = _attrs [level[cur]] [attrid[cur]];
	    if (!settable[i] ||
		    (attrs[i].sa_modify == m_setToDefault && !dflt))
		errors[cur] = m_invalidOperation;
	    else
	    if ((attrs[i].sa_modify == m_addValue ||
				attrs[i].sa_modify == m_removeValue)
			&& !attr->isMultiValued())
		errors[cur] = m_invalidOperator;
	    else {
		if (attrs[i].sa_modify == m_setToDefault)
		    val[cur] = dflt->copy();
		else {
		    if ((val[cur] = attr->decode(attrs[i].sa_val)) == NULLVD &&
				attrs[i].sa_val->pe_errno != PE_ERR_NONE) {
			errors[cur] = m_invalidAttributeValue;
			continue;
		    }
		}
		_attrSetPE = attrs[i].sa_val;
		if (set(attrs[i].sa_modify, attrid[cur], level[cur],
						val[cur], err) == NOTOK &&
			(errors[cur] = err->getError()) == m_processingFailure)
		    return m_processingFailure;
	    }
	    continue;
	}

	// check in case it is a group attribute
	OID* group;
	int  ngroupattrs;

	if ((group = getGroup(attrs[i].sa_id.mid_global,
		&ngroupattrs, objClass)) == NULL) {
	    level[cur] = -1;
	    errors[cur] = m_noSuchAttribute;
	    continue;
	}

	// it IS a group attribute
	if (attrs[i].sa_modify != m_setToDefault) {
	    level[cur] = -1;
	    errors[cur] = m_invalidOperation;
	    continue;
	}

	for (register int j = 0; j < ngroupattrs; j++, cur++)
	    if (findAttr(group[j], &attrid[cur], &level[cur],
			objClass, &settable[cur], &dflt) == OK) {
		if (!settable[cur] || !dflt)
		    errors[cur] = m_invalidOperation;
		else {
		    val[cur] = dflt->copy();
		    _attrSetPE = NULL;	// not needed for setToDefault

		    if (set(m_setToDefault, attrid[cur], level[cur],
						val[cur], err) == NOTOK &&
			(errors[cur] = err->getError())
						== m_processingFailure)
			return m_processingFailure;

		    result[cur].sa_id.mid_type = MID_GLOBAL;
		    result[cur].sa_id.mid_global = group[j];
		    result[cur].sa_modify = m_setToDefault;
		}
	    }
	    else {
		// should not happen for a debugged MIB
		advise(NULLCP, "cannot find group attr %s!",
							sprintoid(group[j]));
		return m_processingFailure;
	    }
	cur--;
    }

    if (set(m_noModifyOp, MO_NOMOREATTRS, MO_ALLEVELS, NULLVD, err) == NOTOK) {
	for (i=0; i < cur; i++)
	    _attrs[level[i]][attrid[i]] -> ffree(val[i]);
	return m_processingFailure;
    }

    *nres = cur;

    // now set the MO attribute values

    for (i=0; i < cur; i++) {
	result[i].sa_error = level[i] == -1 ? m_noSuchAttribute : errors[i];

	if (level[i] == -1 || result[i].sa_error) {
	    if (val[i])
		_attrs[level[i]][attrid[i]] -> ffree(val[i]);
	    error = m_setListError;
	    continue;
	}

	// keep the old value for the attrValChange notification
	void* oldAttrValue = _attrs[level[i]][attrid[i]] -> copy();

	switch(result[i].sa_modify) {
	case m_replace:
	    result[i].sa_error =
		_attrs[level[i]][attrid[i]] -> set(val[i]) == OK ?
			m_noError : m_invalidAttributeValue;
	    if (_attrs[level[i]][attrid[i]] -> getval() != val[i])
		_attrs[level[i]][attrid[i]] -> ffree(val[i]);
	    break;

	case m_addValue:
	    result[i].sa_error =
		_attrs[level[i]][attrid[i]] -> add(val[i]) == OK ?
			m_noError : m_invalidAttributeValue;
	    break;

	case m_removeValue:
	    result[i].sa_error =
		_attrs[level[i]][attrid[i]] -> remove(val[i]) == OK ?
			m_noError : m_invalidAttributeValue;
	    _attrs[level[i]][attrid[i]] -> ffree(val[i]);
	    break;

	case m_setToDefault:
	    result[i].sa_error =
		_attrs[level[i]][attrid[i]] -> setDefault(val[i]) == OK ?
			m_noError : m_invalidAttributeValue;
	    if (_attrs[level[i]][attrid[i]] -> getval() != val[i])
		_attrs[level[i]][attrid[i]] -> ffree(val[i]);
	    break;

	case m_noModifyOp:	
	default:
	    return m_processingFailure;		// will not happen
	}

	if (confirmed && !result[i].sa_error)
	    result[i].sa_error =
		(result[i].sa_val = _attrs[level[i]][attrid[i]] -> encode()) ?
		m_noError : m_processingFailure;

	switch (result[i].sa_error) {
	case m_noError:
	    if (isState(attrid[i], level[i]))
		((Top*) this) -> triggerStateChange(attrid[i], level[i],
				oldAttrValue,
				int_SMI_SourceIndicator_managementOperation);
	    else
		((Top*) this)->triggerAttributeValueChange(attrid[i], level[i],
				oldAttrValue,
				int_SMI_SourceIndicator_managementOperation);
	    break;

	case m_invalidAttributeValue:
	case m_invalidOperation:
	case m_invalidOperator:
	case m_accessDenied:		// this should not happen yet
	    _attrs[level[i]][attrid[i]] -> ffree(oldAttrValue);
	    error = m_setListError;
	    break;

	case m_processingFailure:
	    _attrs[level[i]][attrid[i]] -> ffree(oldAttrValue);
	    return m_processingFailure;

	default:
	    // should not happen for a debugged MIB
	    advise(NULLCP, "set returned invalid error code: %d !",
					(int) result[i].sa_error);
	    return m_processingFailure;
	}
    }

#ifdef PERSISTENCY
    if (_persistent)
	save();		// write object to secondary storage
#endif
    return error;
}


CMISErrors  MO::Action (Bool confirmed, OID objClass,
		CMISParam* info, CMISParam* result, AVA*& err)
{
    _updated = False;

    int actionid, level;
    Attr *infoTmpl, *replyTmpl;

    if (info->mp_id.mid_type != MID_GLOBAL ||
		findAction(info->mp_id.mid_global, &actionid, &level,
		objClass, &infoTmpl, &replyTmpl) != OK)
	return m_noSuchAction;

    void *value = NULLVD, *reply = NULLVD;
    Bool freeFlag = True;
    if (info->mp_val)
	if (!infoTmpl)
	    return m_noSuchArgument;
	else
	if ((value = infoTmpl -> decode(info->mp_val)) == NULLVD)
	    return m_invalidArgumentValue;

    if (action(actionid, level, value, confirmed ? reply:NULLVD, freeFlag, err)
					== OK) {
        if (confirmed) {
	    if (!reply)
	        result->mp_val = NULLPE;	// reply is optional
	    else {
	        // encode and free action reply

	        if (!replyTmpl)
		    return m_processingFailure;	// specific MO class bug
	        if ((result->mp_val = replyTmpl -> encode(reply)) == NULLPE) {
		    advise(NULLCP, "Cannot encode action reply!");
		    return m_processingFailure;
	        }
	        if (freeFlag)
	            replyTmpl -> ffree(reply);
	    }
	    result->mp_id.mid_global = info->mp_id.mid_global;
	    info->mp_id.mid_global   = NULLOID;
        }
    }

    // free action value
    if (value)
	infoTmpl -> ffree(value);

#ifdef PERSISTENCY
    if (_persistent)
	save();		// write object to secondary storage
#endif

    return err ? err->getError() : m_noError;
}


CMISErrors  MO::Delete (AVA*& err)
{
    MO** subords;
    int  nsubords;

    _updated = False;

    if (this->getClassInfo()->_gdmoCmplGenerated) {

	switch (_activeNB->getDeleteMode()) {

	case dm_noDelete:
	    err = new AVA(m_processingFailure, "nonDeletableObject");
	    return m_processingFailure;

	case dm_deleteOnlyIfNoContainedObjs:
	    if (getSubordinate() != NULL) {
	        err = new AVA(m_processingFailure, "containedObjects");
	        return m_processingFailure;
	    }
	    break;

	case dm_deleteContainedObjs:
	    // get all subordinates and count them
	    subords = getWholeSubtree(False, DISCARDAVA);
	    for (nsubords = 0; subords[nsubords]; nsubords++);

	    // delete them in reverse order apart from the base object
	    for (register int i = nsubords-1; i >= 1; i--) {
		subords[i] -> deleteRR(err);	// don't care for errors
		if (err) {
		    delete err; err = NULLAVA;
		}
		delete subords[i];
	    }
	    break;
	}
    }
    else {
        MO* superior = getSuperior();
        Bool deleted = True;

        // check delete behaviour according to name bindings first
        if (cmisDelete() != m_noError) {
	    err = new AVA(m_processingFailure) ;
	    return m_processingFailure;
	}

        // this is for compatibility reasons with the previous style
        // of cmisDelete where the MO was actually deleted
        MO** subords = superior -> getSubordinates(False, DISCARDAVA);
        for (register int i = 0; subords[i]; i++)
	    if (subords[i] == this) {
	        deleted = False;
	        break;
	    }
        if (deleted)
	    return m_noError;
    }


    // delete the base object itself

    if (deleteRR(err) != OK) {
	char* cp;
	advise(NULLCP, "%s - %s: deleteRR failed!",
			getClassName(), cp = getDNString());
	free(cp);
	return err ? err -> getError() : m_processingFailure;
    }

    // this will only trigger if it is supported
    ((Top*) this)->triggerObjectDeletion
			(int_SMI_SourceIndicator_managementOperation);
    delete this;
    return m_noError;
}


void  MO::releaseEncodedData ()
{
    register int i, j;
    int classNAttrs;
    Attr** classAttrs;

    for (i = 0; i < _nclasses; i++) {
	classAttrs = _attrs[i];

	for (j = classNAttrs = 0; classNAttrs < _nattrs[i]; j++) {
	    if (!classAttrs[j])
		continue;
	    classNAttrs++;
	    classAttrs[j] -> freePE();
	}
    }
}


/*
 ****************************** ex-Create.cc *********************************
 */


#define	SIGMASK	0x7fffffff	/* all signals */

extern char* etcdir;
extern NotificationMapper *notificationMapper;


// create initial managed objects reading etcdir/<file> and also
// logs from previous system incarnations

int  MO::initialiseMIB (char* file)
{
    char filename[512], line[512];
    FILE* fp;
    int status = OK;
    MOClass* moClass;
    MO *mo, *superior, *root, *top = NULLMO;
    RDN rdn;
    DN  dn = NULLDN;
    register int i;
    int input;
    AVA* crErr = NULLAVA;
    Bool first = True;
    CreationType crType = ct_initialisation;

    int oldmask = sigblock(SIGMASK);	// block signals

    // set up meta-class information first
#ifdef DEBUG
    advise(NULLCP, "\ninitialising meta-class information...");
#endif
    MOClassInfo::_classArray = new MOClassArray;
    for (i = 0; i < NCLASSES; i++) {
	if (moclasses[i].classInit() == NOTOK) {
	    advise(NULLCP, "%s:  class initialisation failed!",
						moclasses[i].className);
	    return NOTOK;
	}
	MOClassInfo* parent = moclasses[i].getClassInfo()->getParent();
	if (i && (!parent || !parent->_moClass)) {
	    advise(NULLCP, "%s:  parent class should be initialised first!",
						moclasses[i].className);
	    return NOTOK;
	}
	MOClassInfo::_classArray -> addClass(moclasses[i].className,
				moclasses[i].getClassInfo);
	advise(NULLCP, "%s", moclasses[i].className);
    }
    advise(NULLCP, "");

    // check if the fundamental classes are there

    if (MOClassInfo::_classArray -> findClass("top") == NULL) {
	advise(NULLCP, "\n'top' class not in meta-class tree!\n");
	return NOTOK;
    }
    char* cp;
    if (MOClassInfo::_classArray ->
			findClass(cp = "discriminator") == NULL ||
	MOClassInfo::_classArray ->
			findClass(cp = "eventForwardingDiscriminator") == NULL||
	MOClassInfo::_classArray ->
			findClass(cp = "logRecord") == NULL ||
	MOClassInfo::_classArray ->
			findClass(cp = "eventLogRecord") == NULL) {

	_notificationFunction = False;
	MOClassInfo::_classArray ->
			removeClass("discriminator");
	MOClassInfo::_classArray ->
			removeClass("eventForwardingDiscriminator");
	MOClassInfo::_classArray ->
			removeClass("logRecord");
	MOClassInfo::_classArray ->
			removeClass("eventLogRecord");
	advise(NULLCP, "'%s' class not in meta-class tree, \
no notification function", cp);
    }
    else
	_notificationFunction = True;
    
    // create the notification mapper

    if (_notificationFunction)
	notificationMapper = new NotificationMapper;
    else
	notificationMapper = NULL;

    // now create initial instances from mib.init.*
#ifdef DEBUG
    advise(NULLCP, "\ninitialising MIB...");
#endif

    if ((fp = fopen(strcat(strcpy(filename, etcdir), file), "r"))
		== NULL) {
	advise(NULLCP, "cannot open MIB initialisation file: %s !", filename);
	status = NOTOK;
	goto restore;
    }

    char moClassName[128], rdnstr[128], dnstr[512];

    root = new Top;	// the MIB root

    while ((line[0] = getc(fp)) != (char) EOF)
    {
	if (line[0] == '@')
	    break;
	if (line[0] == '\n')
	    continue;	// empty line
	i = 0;
	while ((line[++i] = getc(fp)) != '\n');	// get line
	if (line[0] == '#')
	    continue;	// comment line
	line[i] = '\0';

	input = sscanf(line, "%s %s %s", moClassName, rdnstr, dnstr);

	if (input != 2 && input != 3) {
	    advise(NULLCP, "mib_init: failed while reading MIB initialisation \
file %s", filename);
	    status = NOTOK;
	    goto restore;
	}

	if (input == 2)
	    dnstr[0] = '\0';

	if (first) {
	    if (moClass = MOClassInfo::_classArray->findClass(moClassName))
		if (moClass->getClassInfo()->getParent("system") == NULL)
		    advise(NULLCP, "mib_init: WARNING: top MIB instance not of \
class \"system\"!\n");
	        else {
		int offset = strlen(rdnstr) - strlen("hostname");

		if (strcmp(rdnstr+offset, "hostname") == 0) {
		    char localHost[128];

		    if (gethostname(localHost, 128) == NOTOK) {
			advise(NULLCP, "mib_init: cannot get local host name!");
			status = NOTOK;
			goto restore;
		    }

		    register char* cp = localHost;
		    while(*cp && *cp != '.') cp++;
		    *cp = '\0';
		    strcpy(rdnstr+offset, localHost);
		}
	        }
	}

	if ((moClass = MOClassInfo::_classArray->findClass(moClassName))
			== NULL ||
			moClass->getClassInfo()->_create == NULL) {
	    advise(NULLCP, "mib_init: class \"%s\" not in meta-class tree!",
				moClassName);
		status = NOTOK;
		goto restore;
	}

	if ((rdn = str2rdn(rdnstr)) == NULLRDN ||
		(input == 3 && (dn = str2dn(dnstr)) == NULLDN)) {
	    advise(NULLCP, "mib_init: class \"%s\": RDN or superior DN \
incorrect!", moClassName);
	    status = NOTOK;
	    goto restore;
	}

	superior = first ? root :
			   (input == 3 ? top->getMO(dn, False, DISCARDAVA) : top);
	if (superior == NULLMO) {
	    advise(NULLCP, "mib_init: class \"%s\", rdn %s: superior object \
{%s} not yet created!",
			moClassName, rdnstr, dnstr);
	    status = NOTOK;
	    goto restore;
	}

	if (dn) {
	    dn_free(dn);
	    dn = NULLDN;
	}

	mo = moClass -> getClassInfo() -> _create(rdn, superior);

	// check for wrong name binding
	if (_wrongnb) {
	    advise(NULLCP, "%s - %s: wrong name binding, create failed!",
				moClassName, rdnstr);
	    delete mo;
	    status = NOTOK;
	    goto restore;
	}

	if (first) {
	    // set the system title
	    if (strncasecmp(moClassName, "system", 6) == 0) {
		DN title = str2dn(rdnstr);
		mo->setAttrVal("systemTitle", title);
	    }
	    top = root->getSubordinate();
	    first = False;

#ifdef PERSISTENCY
            // recreate logs
#ifdef DEBUG
            advise(NULLCP, "restarting logs");
#endif
            Log::restartLogs();
#endif
	}

	// interact with the real resource
	if (mo->createRR(crErr, &crType) != OK) {
	    advise(NULLCP, "%s - %s: createRR failed!", moClassName, rdnstr);
	    if (crErr) {
		if (crErr -> getError())
		    advise(NULLCP, "%s", crErr->getErrorStr());
		delete crErr; crErr = NULLAVA;
	    }
	    delete mo;
	    status = NOTOK;
	    goto restore;
	}

	advise(NULLCP, "%s  {%s%s%s}", moClassName,
			dnstr, dnstr[0] != '\0' ? "@":"", rdnstr);
    }

    fclose(fp);

restore:
    if (status != OK) {
	terminateMIB();		// clean-up properly
    }
#ifdef	SECURITY
    else // OK
       theAssocMgr = new AssociationMgr; /* Initialise the ass. mgr */
#endif

    sigsetmask(oldmask);	// restore signals
    return status;
}


void  MO::terminateMIB ()
{
/*
 *  This is complex because an object may delete others which can be
 *  in different parts of the MIT; as such, scoping needs to be re-evaluated
 *  after every object been deleted. In order to alleviate the cost
 *  of wholeSubtree scoping, we delete separately every subtree below
 *  the first level MIT objects
 */
    AVA* err = NULLAVA;
    register MO **subords, *cur, *prev;
    register int nsubords;

    // check first if the top MIB object exists (may fail to initialise)
    if (! (cur = MO::getRoot() -> getSubordinate()) ) {
	delete MO::getRoot();
	return;
    }

    for (cur = cur -> getSubordinate(); cur; ) {
	// this is a firstLevel MIT MO
	if (!cur -> getSubordinate()) {
	    cur -> deleteRR(err);	// don't care for errors now
	    if (err) {
	 	delete err; err = NULLAVA;
	    }
	    prev = cur;
	    cur = cur -> getPeer();
	    delete prev;
	    continue;
	}

	// this is the subtree hanging from it
	while (cur -> getSubordinate()) {
	    subords = cur -> getWholeSubtree(False, DISCARDAVA);
	    for (nsubords = 0; subords[nsubords]; nsubords++);
	    subords[nsubords-1] -> deleteRR(err);	// don't care for errors
	    if (err) {
		delete err; err = NULLAVA;
	    }
	    delete subords[nsubords-1];
	}
    }

    // delete now the top and root MOs
    for (cur = MO::getRoot() -> getSubordinate(); cur; ) {
	cur -> deleteRR(err);	// will not do much for those MOs
	if (err) {
	    delete err; err = NULLAVA;
	}
	prev = cur;
	cur = cur -> getSuperior();
	delete prev;
    }
}


// create managed objects through the CMIS interface

MO*  MO::Create (OID objClass, DN* objInst,
		RDN rdn, MO* superior, MO* reference, External* access,
		int nattrs, CMISParam* attrs,
		CMISErrors* err, CMISErrorInfo* errInfo)
{
    MOClass* moClass;
    NameBinding* nb;
    AVA* crErr = NULLAVA;
    CreationType crType = ct_managementOperation;

    // check first if registered in Create.h (meta-class information)

    if ((moClass = MOClassInfo::_classArray
			-> findClass(_operationClass = objClass)) == NULL) {
	*err = m_noSuchObjectClass;
	return NULLMO;
    }

    // check is creation through CMIS is allowed according to name bindings

    if ((moClass -> getClassInfo() -> _gdmoCmplGenerated &&
	    (nb = moClass -> getClassInfo() -> checkCmisCreate()) == NULL) ||
	(!moClass -> getClassInfo() -> _gdmoCmplGenerated &&
	    moClass -> getClassInfo() -> _mcreate == NULLMCREATE)) {

	errInfo->ei_processingFailure.pf_error.mp_id.mid_global =
					name2oid("nonCreatableObject");
	errInfo->ei_processingFailure.pf_error.mp_val = NULLPE;
	*err = m_processingFailure;	// CMIS M_Create not allowed
	return NULLMO;
    }

    if (moClass -> getClassInfo() -> _gdmoCmplGenerated) {
        // check is rdn/superior have been supplied in createWithName case

	if (nb -> getCreateMode() == cm_createWithName && !rdn) {
	    errInfo->ei_processingFailure.pf_error.mp_id.mid_global =
					name2oid("createObjectWithName");
	    errInfo->ei_processingFailure.pf_error.mp_val = NULLPE;
	    *err = m_processingFailure;	// needs DN of MO to be created
	    return NULLMO;
        }

        // find the superior in createWithAutoInstanceName case

	if (nb -> getCreateMode() == cm_createWithAutoInstanceName &&
								!superior) {
	    // check the top object first and then the whole tree

	    MO* top = getRoot()->getSubordinate();
	    if (top->checkClass(nb->getSuperiorClass()) == OK)
		superior = top;
	    else {
		// check whole tree for the single superior
		int nsuperior = 0;
                MO** subords = top->getWholeSubtree(False, DISCARDAVA);

                for (register int i = 1; subords[i]; i++)
		    if (subords[i]->checkClass(nb->getSuperiorClass()) == OK) {
			superior = subords[i];
			++nsuperior;
			if (nsuperior > 1)
			    break;
		    }
		if (nsuperior != 1) {
		    if (nsuperior == 0)
			errInfo->ei_processingFailure.pf_error.mp_id.mid_global=
					name2oid("noSuperiorObject");
		    else // nsuperior > 1
			errInfo->ei_processingFailure.pf_error.mp_id.mid_global=
					name2oid("manySuperiorObjects");
		    errInfo->ei_processingFailure.pf_error.mp_val = NULLPE;
		    *err = m_processingFailure;
		    return NULLMO;
		}
	    }
	}
    }


    // check name binding before creating for compiler produced objects

    if (moClass -> getClassInfo() -> _gdmoCmplGenerated &&
	moClass -> getClassInfo() -> checkNameBinding(superior, rdn) == NULL) {
	advise(NULLCP, "%s: wrong name binding!", moClass->className);
	*err = m_invalidObjectInstance;
	return NULLMO;
    }


    // create the managed object...

    _accessControl = access;	// to be possibly retrieved by getAccessControl
				// in the case of a proxy
    OID eventId;		// for compatibility with the old API (not used)
    Top* mo;
    if (moClass -> getClassInfo() -> _gdmoCmplGenerated)
	mo = (Top*) moClass -> getClassInfo() -> _create(rdn, superior);
    else	// previous style with cmisCreate
	mo = (Top*) moClass -> getClassInfo() ->
			_mcreate(rdn, superior, reference,
				nattrs, attrs, err, errInfo, &eventId);

    // ... and check if it was created correctly

    if (!mo) {
	advise(NULLCP, "%s: cmisCreate failed!", moClass->className);
	*err = m_processingFailure;
	bzero((char*) &errInfo->ei_processingFailure,
		sizeof(CMISProcessingFailure));
	return NULLMO;
    }
    else
    if (_wrongnb || *err) {
	if (rdn)
	    mo -> _rdn = NULLRDN;	// not to be freed
        *objInst = dn_cpy(mo->getDN());
	delete mo;

	// for GDMO compiler genarated objects name bindings have already
	// been checked; the only failure possibility is for makeRdn
	// to make the same rdn as one of an existing object, in which
	// case we return processingFailure with duplicateObjectInstance

	if (moClass -> getClassInfo() -> _gdmoCmplGenerated) {
	    *err = m_processingFailure;
	    errInfo->ei_processingFailure.pf_error.mp_id.mid_global=
				name2oid("duplicateObjectInstance");
	    errInfo->ei_processingFailure.pf_error.mp_val = NULLPE;
	}
	else {
	    advise(NULLCP, "%s: cmisCreate failed!", moClass->className);
	    if (!*err)
	        *err = m_invalidObjectInstance;
	}

	return NULLMO;
    }


    // set initial attribute values (only for GDMO compiler generated MOs)

    if (moClass -> getClassInfo() -> _gdmoCmplGenerated) {
        for (register int j = 0; j < nattrs; j++) {
	    int attrId, classLevel; Bool settable;
	    if (attrs[j].mp_id.mid_type != MID_GLOBAL ||
		mo -> findAttr(attrs[j].mp_id.mid_global, &attrId, &classLevel,
				NULLOID, &settable) == NOTOK) {
	        *err = m_noSuchAttribute;
	        errInfo->ei_noSuchAttribute = attrs[j].mp_id;	// struct copy
		if (rdn)
		    mo -> _rdn = NULLRDN;	// not to be freed
		delete mo;
	        return NULLMO;
	    }
	    if (!settable ||
		(mo->MO::_attrs[classLevel][attrId]) -> set(attrs[j].mp_val)
			!= m_noError) {
	        *err = m_invalidAttributeValue;

	        errInfo->ei_invalidAttributeValue.mp_id = attrs[j].mp_id;
	        errInfo->ei_invalidAttributeValue.mp_val = attrs[j].mp_val;

		// the following two are freed by CMISAgent::MError() as
		// they have just had their references copied
		attrs[j].mp_id.mid_global = NULLOID;
		attrs[j].mp_val = NULLPE;

		if (rdn)
		    mo -> _rdn = NULLRDN;	// not to be freed
		delete mo;
	        return NULLMO;
	    }
	}
    }

    // interaction with the real resource

    if (mo->createRR(crErr, &crType) != OK) {
	advise(NULLCP, "%s: createRR failed!", moClass->className);

	if (crErr) {
	    switch (*err = crErr->getError()) {
	    case m_invalidAttributeValue:
		errInfo->ei_invalidAttributeValue.mp_id.mid_type =
					MID_GLOBAL;
		errInfo->ei_invalidAttributeValue.mp_id.mid_global =
					oid_cpy(crErr->getOid());
		errInfo->ei_invalidAttributeValue.mp_val =
			pe_cpy(mo->findAttr(crErr->getOid())->encode());
		break;

	    case m_missingAttributeValue:
		errInfo->ei_missingAttributeValue = new MIDListVal;
		errInfo->ei_missingAttributeValue->midl_mid.mid_type =
					MID_GLOBAL;
		errInfo->ei_missingAttributeValue->midl_mid.mid_global =
					oid_cpy(crErr->getOid());
		errInfo->ei_missingAttributeValue->midl_next = NULL;
		break;

	    case m_processingFailure:
		errInfo->ei_processingFailure.pf_error.mp_id.mid_global =
			crErr->getOid() ?
			oid_cpy(crErr->getOid()) : NULLOID;
		errInfo->ei_processingFailure.pf_error.mp_val =
			crErr->getValue() ?
			pe_cpy(crErr->getValue()->encode()) : NULLPE;
		break;

	    default:
	        *err = m_processingFailure;
		bzero((char*) &errInfo->ei_processingFailure,
			sizeof(CMISProcessingFailure));
		advise(NULLCP, "inconsistent return and error value by \
createRR");
		break;
	    }
	    delete crErr; crErr = NULLAVA;
	}
	else {
	    *err = m_processingFailure;
	    bzero((char*) &errInfo->ei_processingFailure,
			sizeof(CMISProcessingFailure));
	}

	mo->deleteRR(crErr);
	if (crErr)
	    delete crErr;

	if (rdn)
	    mo -> _rdn = NULLRDN;	// not to be freed
	delete mo;
	return NULLMO;
    }

    // all is ok, get the DN and triggerObjectCreation

    *objInst = mo->getDN();

    // this will only trigger it if the object supports it
    mo->triggerObjectCreation(int_SMI_SourceIndicator_managementOperation);

    return mo;
}

