#include <sys/stat.h>
#include "Coordinator.h"
#include "CmisAgent.h"
#include "eventLogRecord.h"
#include "Time.h"

extern "C" {		// for Sun AT&T CC
int unlink(const char *);
int rename(const char *, const char *);
}

int Log::_logId;


RDN Log::makeRdn (MO*)
{
    char buf[64];
    sprintf(buf, "logId=%d", getNewLogId());
    return str2rdn(buf);
}

int
Log::createRR (AVA*& err, void* info, int)
{
    Top::createRR(err);

    _mask = -1;

#if defined(PERSISTENCY)
    /*
    ** create/open log file
    */
    char pathName[512];
    mode_t mode = umask(0);

    sprintf(pathName, "%s/%s", logDir, LOG_Directory);
    if ((mkdir(pathName, LOG_DirectoryMode) == -1) && (errno != EEXIST))
        error(NULL, "Log:createRR() : mkdir() failed for %s\n", pathName);

    sprintf(pathName, "%s/%s/%s-log%d",
			logDir, LOG_Directory,
			CMISAgent::getInstance()->getAEI(),
			*(int *) logId()->get());
    _logName = strdup(pathName);

    if ((_db = dbmOpen(_logName, DBM_ReadWrite, LOG_FileMode)) == NULL)
        error(NULL, "Log::createRR() : could not dbmOpen() %s\n", _logName);

    umask(mode);

    /*
    ** If log already exists, update _recordId and _loRecordId
    */
    _recordId = 0;
    _loRecordId = 0;

    datum *key, *entry;

    for(key = dbmGetKey(_db, DBM_GetFirstKey); key != NULL;
	key = dbmGetKey(_db)) {

        entry = dbmFetch(_db, key);
        if (entry == NULL) continue;

        char name[64];
        int logId, recordId;

        memcpy(name, key->dptr, key->dsize);
        name[key->dsize] = (char) NULL;
        sscanf(name, "logId=%d@logRecordId=%d", &logId, &recordId);
        if (_loRecordId == 0) _loRecordId = recordId;
        if (recordId < _loRecordId) _loRecordId = recordId;
        if (recordId > _recordId) _recordId = recordId;
    }

    if (info && *(CreationType*) info == ct_managementOperation)
        save();
#endif /* PERSISTENCY */

    Coordinator::getInstance()->registerReportObj();
    _scheduledWakeUp = False;
    return OK;
}

int
Log::deleteRR (AVA*& err, void* info, Bool checkOnly, int)
{
    if (checkOnly)
        return OK;

    if (info && *(DeletionType*) info == dt_managementOperation &&
		mdelete() == NOTOK) {
	err = new AVA(m_processingFailure);
	return NOTOK;
    }

#if defined(PERSISTENCY)
    if (_db != NULL) {
        dbmClose(_db);
	_db = NULL;
    }
    free(_logName);
    if (_scheduledWakeUp) cancelWakeUps("");
#endif
    Coordinator::getInstance()->deregisterReportObj();

    return Top::deleteRR(err, info, checkOnly);
}

int
Log::mdelete ()
{
    /*
    ** When a log is deleted, it should delete the corresponding gdbm file.
    ** For now, instead of deleteing it rename it to ".old".
    ** Comment out the #define in log.inc.hcl to enable deleting.
    */

    register
    char *logName = getLogName();
    register int i;

#if defined(LOG_RenameOldLogFiles)

    /* Rename log files */
    char newName[517];

    sprintf(newName, "%s.old", logName);
    if (rename(logName, newName) == -1) {
        advise("", "Log::mdelete() : rename(\"%s\") failed\n", logName);
        return NOTOK;
    }

#else

    /* Delete log files */
    if (unlink(logName) == -1) {
        advise("", "Log::mdelete() : unlink(\"%s\") failed\n", logName);
        return NOTOK;
    }

#endif /* RenameOldLogFiles */

    /*
    ** Remove this object from the MIB dump file
    */
#if defined(PERSISTENCY)
    i = ((MO *) this)->remove();
#if defined(DEBUG)
    if (i == NOTOK) printf("Log::mdelete() : remove() failed\n");
#endif
#endif /* PERSISTENCY */
    return i;
}

int
Log::get(int attrId, int level, AVA*& err, Bool checkOnly, int)
{
    if ((level != _level) && (level != MO_ALLEVELS))
        return Top::get(attrId, level, err, checkOnly);

    if (checkOnly)
	return OK;	/* we can always do it */

#if defined(PERSISTENCY)
    static
    unsigned int attrs = 0;
    static
    Bool requestInProgress = False,
         needsRefresh      = False;

    switch (attrId) {

    case MO_ALLATTRS :
        attrs             = 0xffffffff;
        needsRefresh      = True;
        requestInProgress = False;
        break;

    case MO_NOMOREATTRS :
        requestInProgress = False;
        break;

    default :
        needsRefresh      = True;
        requestInProgress = True;
        attrs            |= (1 << attrId);
        break;
    }

    if (requestInProgress || !needsRefresh) return OK;
    needsRefresh = False;

    /*
    ** For updating currentLogSize and numberOfRecords, I currently
    ** inspect the disc file. This method is the only way to ensure
    ** I know what the disc file looks like, e.g. in case
    ** the file has been altered off-line or without the agent's
    ** knowledge.
    ** I do not known how much this will slow the performance of
    ** the agent when dealing with large log files.
    */
#define UPDATE(id) (attrs & (1 << (id)))

    datum *key, *entry;
    int i, size;

    size = 0; _loRecordId = _recordId;
    for(i = 0, key = dbmGetKey(_db, DBM_GetFirstKey);
        key != NULL; ++i, key = dbmGetKey(_db)) {

        char id[20];
        register
        int j;

        memcpy(id, key->dptr, key->dsize);
        id[key->dsize] = (char) NULL;
        if ((j = atoi(id)) < _loRecordId) _loRecordId = j;

        /*
        ** [Refering to ISO 10164-6 Sections 8.1.1.2.1 and 8.1.1.2.2]
        ** For log sizes, it is not totally clear to me if "data
        ** representation" should mean :
        **
        **  - the attribute values only
        **  - the C++ logRecord objects
        **  - the size of the record on disc
        **
        ** I have chosen the the last of these, as this implemenation will
        ** not hold all log records in memory, but on disc. The value will
        ** be the size of the record only, not including any overhead that
        ** is incurred when the record is written to disc.
        */
        if (UPDATE(I_currentLogSize)) {
            if ((entry = dbmFetch(_db, key)) != NULL)
                size += entry->dsize;
        }
    }

    if (UPDATE(I_currentLogSize))
        currentLogSize()->set((void *) &size);
    numberOfRecords()->set((void *) &i);

#undef UPDATE

    attrs = 0;

#endif /* PERSISTENCY */

    return OK;
}

int
Log::set(CMISModifyOp op, int attrId, int level, void* value, AVA*& err,
	 Bool checkOnly, int)
{
    if (level != _level)
        return Top::set(op, attrId, level, value, err, checkOnly);

    int ret = OK;
    type_SMI2_AvailabilityStatus elem; elem.next = NULL;

    switch ((int) op) {

    case (int) m_noModifyOp :
        break;

    case (int) m_replace :

        switch (attrId) {

        case I_maxLogSize :
            register
            int *n, oldVal, newVal, size;

            n = (int *) value;
            newVal = *n;
            if (newVal < 0) {
		if (checkOnly)
		    return NOTOK;
                err = new AVA(m_invalidAttributeValue);
		return NOTOK;
            }

            oldVal = *((int *) maxLogSize()->get());
            get(I_currentLogSize, _level, err); // update
            get(MO_NOMOREATTRS, _level, err);
            size = *((int *) currentLogSize()->get());
            if ((newVal != 0) && ((newVal < size) || (newVal < oldVal))) {
		if (checkOnly)
		    return NOTOK;
                err = new AVA(m_invalidAttributeValue);
		return NOTOK;
            }

	    if (checkOnly)
		return OK;

            /*
            ** maxLogSize has been increased.
            ** Update availabilityStatus.
            */
	    elem.member_SMI2_0 = int_SMI2_AvailabilityStatusElem_logFull;
            availabilityStatus()->remove(&elem);
            break;

       case I_logFullAction :
	    if (checkOnly)
		return OK;

            register
            type_SMI2_LogFullAction* l = (type_SMI2_LogFullAction *) value;

            /*
            ** if the new value is int_SMI2_LogFullAction_wrap, then a logFull
            ** condition need not apply
            */
            if (l->parm == int_SMI2_LogFullAction_wrap) {
		elem.member_SMI2_0 = int_SMI2_AvailabilityStatusElem_logFull;
                availabilityStatus()->remove(&elem);
	    }
            break;
        }

        break; /* case (int) m_replace */

    default :
        break;
    }

    return OK;
}

int
Log::store(logRecord *record)
{
    if (!_db)
	return OK;	/* it may be trying to store its own deletion! */

    /*
    ** Check status of log
    */
    AvailabilityStatus *av = availabilityStatus();
    AdminStateVal ad = *((AdminStateVal *) administrativeState()->get());
    OperStateVal op = *((OperStateVal *) operationalState()->get());

    /* ISO 10164-6, Section 8.1.1.5 */
    type_SMI2_AvailabilityStatus elem;
    elem.member_SMI2_0 = int_SMI2_AvailabilityStatusElem_logFull;
    elem.next = NULL;

    if (av->find(&elem) || (ad != as_unlocked) || (op == os_disabled)) {
#if defined(DEBUG)
        printf("Log::store() : %s - ", getLogName());
        type_SMI2_AvailabilityStatus *asv = (type_SMI2_AvailabilityStatus *)
					    av->get();
	char* cp = NULLCP;
        if (asv) printf(" %s", cp = av->print());
	if (cp) free(cp);
        if (ad == as_locked) printf(" locked");
        if (op == os_disabled) printf(" disabled");
        printf("\n");
#endif
        return NOTOK;
    }

    /* Set RDN */
    char name[128];
    sprintf(name, "logRecordId=%d", getNewRecordId());
    RDN newRdn = str2rdn(name);

#if defined(PERSISTENCY) /* tmp to compile ??? */
    /*
    ** record is a pointer to a static test object so we have to patch
    ** up values for the RDN. These patches should be "unpatched"
    ** after the record is save()d.
    */
    RDN rdn = rdn_cpy(record->getRDN());
    if (record->setRDN(newRdn) == NOTOK) {

        /* Even if setRDN() reports that it has failed, _rdn */
        /* is changed(!) so we must not rdn_free(newRDN) */
        rdn_free(rdn);
#if defined(DEBUG)
        printf("Log:store() : record->setRDN(%s) failed\n", name);
#endif
        return NOTOK;
    }
#else /* PERSISTENCY */

    /*
    ** Here we have logging without persistency ... hmm ... mine is
    ** not to question why but to do or go to the pub.
    ** Anyway, make a new eventLogRecord. This code will have to change
    ** when we have more record types.
    */
    register
    eventLogRecord *newRecord = eventLogRecord::create(newRdn, this);
    newRecord->createRR();

    /* copy the attribute values from record ... arrggh */
    ((ObjectClass *) newRecord->getAttr("managedObjectClass"))
        ->set(((ObjectClass *) record->getAttr("managedObjectClass"))->copy());
    ((ObjectInstance *) newRecord->getAttr("managedObjectInstance"))
        ->set(((ObjectInstance *) record->getAttr("managedObjectInstance"))->copy());
    ((EventType *) newRecord->getAttr("eventType"))
        ->set(((EventType *) record->getAttr("eventType"))->copy());
    /* Time does not need setting -- the default is current time */
    /* AdditionalText does not need copying as the String will do it */
    ((String *) newRecord->getAttr("additionalText"))
        ->set(((String *) record->getAttr("additionalText")));

    printf("Log::store() : record %d ok\n", _recordId);
#endif /* PERSISTENCY */

    /* Also, set the loggingTime */
    register
    Time *t = (Time *) record->getAttr("loggingTime");
    t->set(); /* now */

#if defined(PERSISTENCY)
    /*
    ** Store
    */
    char buffer[DBM_MaxRecordSize + 1];
    int length = DBM_MaxRecordSize + 1;

    /* Now save it */
    if (record->save(buffer, &length) == NOTOK) {
#if defined(DEBUG)
        printf("Log::store() : record->save(...) failed\n");
#endif
        return NOTOK;
    }

    /* "unpatch" */
    record->setRDN(rdn);

    /* Check log size */
    register
    int i, size, maxSize = *((int *) maxLogSize()->get());

    /*
    ** "Finite Log Size" behaviour
    */
    if (maxSize > 0) {

        /* Check size of record */
        if (length > maxSize) {
#if defined(DEBUG)
            printf("Log::store() : record %d exceeds maxLogSize\n", _recordId);
#endif
            return NOTOK;
        }

        /*
        ** Here, I check whether writing the next record will exceed
        ** maxLogSize. If this is the case and logFullAction=halt, I do
        ** not write the record, but I say that the log is "full". This
        ** may mean that the log is seen as full even though
        ** currentLogSize < maxLogSize. The first record that is too big to
        ** fit in the remaining space will render the log as "full", but
        ** I think this is nicer than rejecting furthter records for
        ** storage all the time, just because they do not fit in the
        ** space that is left. If the log is seen as "full" a manager
        ** can act accordingly.
        */
        get(I_currentLogSize, _level, NULLAVAREF, False); // update
        get(MO_NOMOREATTRS, _level, NULLAVAREF);
        size = *((int *) currentLogSize()->get());
        if (size + length >= maxSize) {

            type_SMI2_LogFullAction* lfa = (type_SMI2_LogFullAction *)
                                           logFullAction()->get();
            /*
            ** "Wrap" behaviour
            */
            if (lfa->parm == int_SMI2_LogFullAction_wrap) {

                /* remove older record */
                while(size + length > maxSize) {
                    remove(_loRecordId);
                    get(I_currentLogSize, _level, NULLAVAREF);
                    get(MO_NOMOREATTRS, _level, NULLAVAREF);
                    size = *((int *) currentLogSize()->get());
                }
            }

            /*
            ** "Halt" behaviour
            */
            if (lfa->parm == int_SMI2_LogFullAction_halt) {
                advise(NULL, "Log::store() : %s is full\n", getLogName());
                advise(NULL, "             : record %d not written\n", _recordId);
                --_recordId;
		type_SMI2_AvailabilityStatus* elem =
					new type_SMI2_AvailabilityStatus;
		elem->member_SMI2_0 = int_SMI2_AvailabilityStatusElem_logFull;
		elem->next = NULL;
                av->add(elem);
		save();
                return NOTOK;
            }
        }
    }

    /* Store */
    datum key, entry;

    /*
    ** Use local DN as key.
    ** Actually only need to use value of recordId, but use DN as that is
    ** the same behaviour as MO::save(). MO::save(), uses str2dn(), but I'll
    ** do it this way for now as its a bit quicker.
    */
    sprintf(name, "logId=%d@logRecordId=%d",
            *((int *) logId()->get()), _recordId);
    key.dptr = name;
    key.dsize = strlen(name);
    entry.dptr = buffer;
    entry.dsize = length;

    i = dbmStore(_db, &key, &entry);

    if (i < 0) {
        advise(NULL, "Log::store() : failed for %s", getLogName());
        advise(NULL, "             : record %d", _recordId);
        advise(NULL, "             : Can't guarantee integrity of log file");
        printf("System Error : %d - %s\n", errno
               /* , errno < sys_nerr ? sys_errlist[errno] : "??" */ );
    }
    else {
	/* save() in an attepmt to keep the mib-file up to date */
	save();
#if defined(DEBUG)
	printf("Log::store() : %s record %d ok\n", getLogName(), _recordId);
#endif
    }

    return (i < 0 ? NOTOK : OK);

#else /* PERSISTENCY */ /* ??? */

    return OK; /* ??? */

#endif /* PERSISTENCY */
}

int
Log::fetch(int recordId, char *buffer, int *length)
{
    *buffer = '\0';
    *length = 0;

#if defined(PERSISTENCY) /* tmp to compile ??? */
    datum *key, key1, *entry;
    char name[64];

    /* Get key */
    switch (recordId) {
    case LOG_FetchFirst : key = dbmGetKey(_db, DBM_GetFirstKey); break;
    case LOG_FetchNext : key = dbmGetKey(_db); break;

    default :
        sprintf(name, "logId=%d@logRecordId=%d",
                *((int *) logId()->get()), recordId);
        key1.dptr = name;
        key1.dsize = strlen(name);
        key = &key1;
        break;
    }

    if (key == (datum *) 0)
        return NOTOK;

    /* Fetch */
    entry = dbmFetch(_db, key);

    if (entry == (datum *) 0) {
#if defined(DEBUG)
        printf("Log::fetch() : dbmFetch() failed %s\n", name);
#endif
        return NOTOK;
    }

    memcpy(buffer, entry->dptr, entry->dsize);
    buffer[entry->dsize] = (char) NULL;
    *length = entry->dsize;

#endif /* PERSISTENCY */

    return OK;
}

int
Log::remove(int recordId)
{
#if defined(PERSISTENCY) /* tmp to compile ??? */
    char name[30];
    datum key;

    sprintf(name, "logId=%d@logRecordId=%d",
            *((int *) logId()->get()), recordId);
    key.dptr = name;
    key.dsize = strlen(name);

    register
    int i = dbmDelete(_db, &key);

#if defined(DEBUG)
    printf("Log::remove() : dbmDelete() %s [%s record %s]\n",
    (i < 0 ? "failed" : "ok"), getLogName(), name);
#endif

    /* The following get() is an attempt to keep _loRecordId up to date */
    if (recordId == _loRecordId) {
        get(I_numberOfRecords, _level, NULLAVAREF);
        get(MO_NOMOREATTRS, _level, NULLAVAREF);
    }

    return (i < 0 ? NOTOK : OK);

#else

    return OK;

#endif /* PERSISTENCY */
}

/*
** Should the following functionality should be part of the GMS!
** This hack will do for now.
*/
int
Log::restartLogs() // Dirty! Clean up later.
{
#if defined(PERSISTENCY)
    static
    int done = 0;

    if (done)
        return OK;

    register
    MO *root = getRoot();

    /* Check that system exists */
    if ((root == NULL) || (root->getSubordinate() == NULL))
        return NOTOK;

    datum *key, *entry;
    register
    int logIdHi = 0;
    register
    Log *mo = (Log *) 0;
    register
    DbmFile *db = root->getDbmFile();

    if (db == NULL)
        return NOTOK;

    /*
    ** Get any saved logs from disc
    */
    for(key = dbmGetKey(db, DBM_GetFirstKey); key != NULL;
        key = dbmGetKey(db)) {

        char buffer[80];

        /* Fetch saved entry */
        if ((entry = dbmFetch(db, key)) == NULL) continue;

        char data[DBM_MaxRecordSize + 1];

        /*
        ** Have to make a copy of data because dbmFetch() calls are
        ** made when a new Log object are created and they will
        ** overwrite the static space used by dbmFetch().
        */
        memcpy(data, entry->dptr, entry->dsize);
        data[entry->dsize] = (char) NULL;
        sscanf(data, "%[^:]", buffer);
        if (strcasecmp(buffer, "Log") != 0) continue;

        RDN rdn;
        register
        char *cp;
        char id[20];
        register
        int i;

        /*
        ** Find the logId. This will be part of the DN, the
        ** second field in the disc entry for this MO.
        */
        cp = strchr(data, ':'); // find RDN in DN
        ++cp;
        sscanf(cp, "logId=%[^:]", id); // dummy
        sprintf(buffer, "logId=%s", id);
        rdn = str2rdn(buffer);

        /* Restore the object */
        mo = (Log *) Log::create(rdn, root->getSubordinate());
	mo->createRR();

        if (mo->restore(data) == NOTOK) {
            delete mo;
            continue;
        }

        /* Keep track of logId */
        i = atoi(id);
        if (i > logIdHi) logIdHi = i;
    }

    /* update logId */
    if (mo != NULL)
        { while(logIdHi--) mo->getNewLogId(); }

    done = 1; // only do this once!

#endif

    return OK;
}

int
Log::refreshSubordinates(AVA*& err, int)
{
#if defined(PERSISTENCY)

    if (!logRecordAccess) {
	err = new AVA(m_resourceLimitation);
        return NOTOK;
    }

    /*
    ** Cancel any previous wakeUp
    */
    if (_scheduledWakeUp) {
        cancelWakeUps("");
        _scheduledWakeUp = False;
    }

    register
    eventLogRecord *mo;

    char buffer[DBM_MaxRecordSize + 1];
    int length = DBM_MaxRecordSize + 1;

    for(mo = (eventLogRecord *) getSubordinate(),
        fetch(LOG_FetchFirst, buffer, &length);
        length != 0;
        mo = (eventLogRecord *) mo->getPeer(),
        fetch(LOG_FetchNext, buffer, &length)) {

        /*
        ** Make a new eventLogRecord if required
        */
        char name[50];
        RDN  rdn;

        if (mo == NULL) {
            sprintf(name, "logRecordId=-12345"); // dummy
            rdn = str2rdn(name);
            mo = (eventLogRecord*) eventLogRecord::create(rdn, this);
	    mo->createRR();
        }

        /*
        ** update log record objects
        */
        if (mo->restore(buffer) == NOTOK) {
#if defined(DEBUG)
            printf("Log::refreshSubordintes() : mo->restore() failed\n");
#endif
            delete mo;
	    err = new AVA(m_processingFailure);
            return NOTOK;
        }

        /* Update RDN */
        sprintf(name, "logRecordId=%d",
                *((int *) mo->getAttrVal("logRecordId")));
        rdn = str2rdn(name);
        mo->setRDN(rdn);
    }


    /*
    ** Tidy up the list
    */
    if (mo != NULL)
        mo = (eventLogRecord *) mo->getPeer();
    while(mo != NULL) {
        eventLogRecord *peer = (eventLogRecord *) mo->getPeer();
        delete mo;
        mo = peer;
    }

    /*
    ** Schedule a wakeUp
    */
    scheduleWakeUp(LOG_TidyInterval, "");
    _scheduledWakeUp = True;

#if defined(DEBUG)
    printf("Log::refreshSubordinates() %s\n", getLogName());
#endif

#endif	/* PERSISTENCY */

    return OK;
}

int
Log::refreshSubordinate(RDN rdn, AVA*& err, int)
{
#if defined(PERSISTENCY)

    if (!logRecordAccess) {
	err = new AVA(m_resourceLimitation);
        return NOTOK;
    }

    /*
    ** Cancel any previous wakeUp
    */
    if (_scheduledWakeUp) {
       cancelWakeUps("");
       _scheduledWakeUp = False;
   }

    char buffer[DBM_MaxRecordSize + 1];
    int length = DBM_MaxRecordSize + 1;

    fetch(*((int *) rdn->rdn_av.av_struct), buffer, &length);
    if (length == 0) {
	err = new AVA(m_processingFailure);
        return NOTOK;
    }

    register
    eventLogRecord *mo = (eventLogRecord *) getSubordinate();

    char name[50];
    RDN  rdn2;

    if (mo == NULL) {
        sprintf(name, "logRecordId=-12345"); // dummy
        rdn2 = str2rdn(name);
        mo = (eventLogRecord*) eventLogRecord::create(rdn2, this);
	mo->createRR();
    }

    if (mo->restore(buffer) == NOTOK) {
#if defined(DEBUG)
        printf("Log::refreshSubordinate() : mo->restore() failed\n");
#endif
        delete mo;
	err = new AVA(m_processingFailure);
        return NOTOK;
    }

    /* Update RDN */
    sprintf(name, "logRecordId=%d", *((int *) mo->getAttrVal("logRecordId")));
    rdn2 = str2rdn(name);
    mo->setRDN(rdn2);

    /*
    ** Tidy up the list
    */
    mo = (eventLogRecord *) mo->getPeer();
    while(mo != NULL) {
        eventLogRecord *peer = (eventLogRecord *) mo->getPeer();
        delete mo;
        mo = peer;
    }

    /*
    ** Schedule a wakeUp
    */
    scheduleWakeUp(LOG_TidyInterval, "");
    _scheduledWakeUp = True;

#if defined(DEBUG)
    printf("Log::refreshSubordinate() %s\n", getLogName());
#endif

#endif /* PERSISTENCY */

    return OK;
}

int
Log::wakeUp(char *)
{
#if defined(PERSISTENCY)
    if (!_scheduledWakeUp)
        return NOTOK;

    /* For now, delete all logRecord objects */
    register
    eventLogRecord *mo, *peer;

    mo = (eventLogRecord *) getSubordinate();
    while(mo != NULL) {
        peer = (eventLogRecord *) mo->getPeer();
        delete mo;
        mo = peer;
    }

    _scheduledWakeUp = False;
#if defined(DEBUG)
        printf("Log::wakeUp() : %s\n", getLogName());
#endif

    /* do a save() in an attempt to keep the mib file up-to-date() */
    save();

#endif /* PERSISTENCY */

    return OK;
}

