//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// MOClassInfo.cc

// By George Pavlou, February 1991


#include <stdlib.h>
#ifdef	__GNUG__
extern "C" { int  strcasecmp (const char*, const char*); }
#endif
#include "ucbext.h"
#include "msap.h"
#include "Error.h"
#include "MOClassInfo.h"
#include "GenericMO.h"

MOClassArray*	MOClassInfo::_classArray;


// NameBinding methods

NameBinding::NameBinding (OID myNameBinding,
					OID mySuperiorClass, OID myNamingAttr)
{
    _nameBinding = myNameBinding;
    _superiorClass = mySuperiorClass;
    _namingAttr  = myNamingAttr;

    _createMode = cm_noCreate;
    _referenceObjMode = False;
    _deleteMode = dm_noDelete;
}


NameBinding::~NameBinding ()
{
    oid_free(_nameBinding);
    oid_free(_superiorClass);
    oid_free(_namingAttr);
}


// MOClassInfo methods

int  MOClassInfo::setClass (char* className, int nbindings, int nattrs,
			int ngroups, int nevents, int nactions, int npackages)
{
    if ((_moClass = name2oid(className)) == NULLOID)
	return NOTOK;
    _moClassName = className;

    register int i;

    if (nbindings)
	_nameBindings	= new NameBinding* [_nbindings = nbindings];
    for (i = 0; i < nbindings; i++)
	_nameBindings[i] = (NameBinding *) 0;

    if (nattrs) {
	_attrOids	= new OID [_nattrs = nattrs];
	_attrNames	= new char* [nattrs];
	_attrIds	= new int [nattrs];
	_settable	= new SetMode [nattrs];
	_defaults	= new Attr* [nattrs];
    }
    for (i = 0; i < nattrs; i++)
	_attrNames[i] = NULLCP;

    if (ngroups) {
	_ngroupattrs	= new int [_ngroups = ngroups];
	_attrGroups	= new OID* [ngroups];
	_groupNames	= new char* [ngroups];
	_groupIds	= new int [ngroups];
    }
    for (i = 0; i < ngroups; i++)
	_groupNames[i] = NULLCP;

    if (nevents) {
	_eventOids	= new OID [_nevents = nevents];
	_eventNames	= new char* [nevents];
	_eventIds	= new int [nevents];
    }
    for (i = 0; i < nevents; i++)
	_eventNames[i] = NULLCP;

    if (nactions) {
	_actionOids		= new OID [_nactions = nactions];
	_actionNames		= new char* [nactions];
	_actionIds		= new int [nactions];
	_actionInfoTmpls	= new Attr* [nactions];
	_actionReplyTmpls	= new Attr* [nactions];
    }
    for (i = 0; i < nactions; i++)
	_actionNames[i] = NULLCP;

    if (npackages) {
	_packageOids	= new OID [_npackages = npackages];
	_packageNames	= new char* [npackages];
	_packageIds	= new int [npackages];
    }
    for (i = 0; i < npackages; i++)
	_packageNames[i] = NULLCP;

    return OK;
}


MOClassInfo*  MOClassInfo::getParent (char* className)
{
    for (register MOClassInfo* cur = this; cur; cur = cur->_parent)
	if (strcasecmp(className, cur->_moClassName) == 0)
	    return cur;
    return NULL;	// not found
}


int  MOClassInfo::setNameBinding (char* myNameBinding,
				char* mySuperiorClass, char* myNamingAttr,
				CreateMode cm, Bool rm, DeleteMode dm)
{
    OID nameBinding, superiorClass, namingAttr;

    if ((nameBinding = name2oid(myNameBinding)) == NULLOID ||
		(superiorClass = name2oid(mySuperiorClass)) == NULLOID ||
		(namingAttr = name2oid(myNamingAttr)) == NULLOID) {
	advise(NULLCP, "Unknown name binding %s, superior class %s or naming \
attr %s!",
			myNameBinding, mySuperiorClass, myNamingAttr);
	return NOTOK;
    }

    register int i;
    for (i = 0; _nameBindings[i]; i++);
    _nameBindings[i] = new NameBinding(nameBinding, superiorClass, namingAttr);
    _nameBindings[i] -> setCreateMode(cm, rm);
    _nameBindings[i] -> setDeleteMode(dm);

    return OK;
}


NameBinding*  MOClassInfo::checkNameBinding (MO* superior, RDN rdn)
{
    for (register MOClassInfo* cur = this; cur->_parent; cur = cur->_parent)
        for (register int i = 0; i < cur->_nbindings; i++)
	    if (superior -> checkClass(cur->_nameBindings[i]->
					getSuperiorClass()) == OK &&
		(!rdn ||
		(rdn && oid_equal(cur->_nameBindings[i]->getNamingAttr(),
					rdn->rdn_at->oa_ot.ot_oid))))
	        return cur->_nameBindings[i];

    return (NameBinding *) NULL;
}


NameBinding*  MOClassInfo::checkCmisCreate (MO* superior)
{
    // here we look for the first name binding that allows creation
    // and satisfies the superior class,
    // starting from the bottom of the inheritance tree;
    // the assumption is that name bindings in children "overwrite"
    // those of parents

    for (register MOClassInfo* cur = this; cur->_parent; cur = cur->_parent)
	for (register int i = 0; i < cur->_nbindings; i++)
	    if (cur->_nameBindings[i]->getCreateMode() != cm_noCreate &&
		    superior->checkClass(
			cur->_nameBindings[i]->getSuperiorClass()) == OK)
		return cur->_nameBindings[i];
    return NULL;	// not found
}


NameBinding**  MOClassInfo::getNameBindings (int* nbindings)
{
    if (nbindings)
	*nbindings = _nbindings;
    return _nameBindings;
}


int  MOClassInfo::setAttr (char* attrName, int attrId,
					SetMode settable, Attr* dfltVal)
{
    register int i;
    for (i = 0; i < _nattrs && _attrNames[i]; i++)
	if (strcmp(attrName, _attrNames[i]) == 0 || _attrIds[i] == attrId) {
	    advise(NULLCP, "Class %s: attr %s with id %d badly registered!",
				_moClassName, attrName, attrId);
	    return NOTOK;
	}

    if (i == _nattrs) {
	advise(NULLCP, "Class %s: more attributes than declared number %d!", 
				_moClassName, _nattrs);
	return NOTOK;
    }

    if ((_attrOids[i] = name2oid(attrName)) == NULLOID)
	return NOTOK;

    _attrNames[i] = attrName;
    _attrIds[i] = attrId;
    _settable[i] = settable;
    _defaults[i] = dfltVal;

    return OK;
}


int  MOClassInfo::setDefault (int attrId, Attr* dfltVal)
{
    for (register int i = 0; i < _nattrs; i++)
	if (attrId == _attrIds[i])	// found it
	    if (_settable[i] == sm_set) {
		if (_defaults[i]) {	// free the previous one (if any)
		    _defaults[i] -> clear();
		    delete _defaults[i];
		}
		_defaults[i] = dfltVal;
		return OK;
	    }
	    else {
		advise(NULLCP, "Class %s: attr with id %d is not settable!",
				_moClassName, attrId);
		return NOTOK;
	    }

    advise(NULLCP, "setDefault: Class %s: attr with id %d is not registered!",
		    _moClassName, attrId);
    return NOTOK;
}


int  MOClassInfo::setGroupAttr (char* groupName, int groupId, int ngroupAttrs)
{
    register int i;
    for (i = 0; i < _ngroups && _groupNames[i]; i++)
	if (strcmp(groupName, _groupNames[i]) == 0 || _groupIds[i] == groupId) {
	    advise(NULLCP, "Class %s: group %s with id %d badly registered!",
				_moClassName, groupName, groupId);
	    return NOTOK;
	}

    if (i == _ngroups) {
	advise(NULLCP, "Class %s: more groups than declared number %d!", 
				_moClassName, _ngroups);
	return NOTOK;
    }

    OID groupOid;
    if ((groupOid = name2oid(groupName)) == NULLOID)
	return NOTOK;
    _attrGroups[i] = new OID[ngroupAttrs+1];
    _attrGroups[i][0] = groupOid;
    for (register int j = 1; j < ngroupAttrs+1; j++)
	_attrGroups[i][j] = NULLOID;

    _ngroupattrs[i] = ngroupAttrs;
    _groupNames[i] = groupName;
    _groupIds[i] = groupId;

    return OK;
}


int  MOClassInfo::addGroupAttr (int groupId, int attrId)
{
    register int i;
    for (i = 0; i < _ngroups && _groupNames[i]; i++)
	if (_groupIds[i] == groupId)
	    break;	// found the group

    if (i == _ngroups || !_groupNames[i]) {
	advise(NULLCP, "Class %s: group with id %d not registered!",
				_moClassName, groupId);
	return NOTOK;
    }

    OID attrOid;
    if ((attrOid = getAttrOid(attrId)) == NULLOID) {
	advise(NULLCP, "Class %s: attribute with id %d not registered!",
				_moClassName, attrId);
	return NOTOK;
    }
	
    register int j;
    for (j = 1; j < _ngroupattrs[i]+1 && _attrGroups[i][j]; j++)
	if (oid_equal(attrOid, _attrGroups[i][j])) {
	    advise(NULLCP, "Class %s: group with id %d has already attr with \
id %d!",
				_moClassName, groupId, attrId);
	    return NOTOK;
	}

    if (j == _ngroupattrs[i]+1) {
	advise(NULLCP, "Class %s: more group attrs than declared number %d \
for group %s!",
				_moClassName, _ngroupattrs[i], _groupNames[i]);
	return NOTOK;
    }

    _attrGroups[i][j] = attrOid;
    return OK;
}


int  MOClassInfo::setEvent (char* eventName, int eventId)
{
    register int i;
    for (i = 0; i < _nevents && _eventNames[i]; i++)
	if (strcmp(eventName, _eventNames[i]) == 0 || _eventIds[i] == eventId) {
	    advise(NULLCP, "Class %s: event %s with id %d doubly registered!",
				_moClassName, eventName, eventId);
	    return NOTOK;
	}

    if (i == _nevents) {
	advise(NULLCP, "Class %s: more events than declared number %d!", 
				_moClassName, _nevents);
	return NOTOK;
    }

    if ((_eventOids[i] = name2oid(eventName)) == NULLOID)
	return NOTOK;

    _eventNames[i] = eventName;
    _eventIds[i] = eventId;

    return OK;
}


int  MOClassInfo::setAction (char* actionName, int actionId,
				Attr* infoTmpl, Attr* replyTmpl)
{
    register int i;
    for (i = 0; i < _nactions && _actionNames[i]; i++)
	if (strcmp(actionName, _actionNames[i]) == 0 ||
					_actionIds[i] == actionId) {
	    advise(NULLCP, "Class %s: action %s with id %d double registered!",
				_moClassName, actionName, actionId);
	    return NOTOK;
	}

    if (i == _nactions) {
	advise(NULLCP, "Class %s: more actions than declared number %d!", 
				_moClassName, _nactions);
	return NOTOK;
    }

    if ((_actionOids[i] = name2oid(actionName)) == NULLOID)
	return NOTOK;

    _actionNames[i] = actionName;
    _actionIds[i] = actionId;
    _actionInfoTmpls[i] = infoTmpl;
    _actionReplyTmpls[i] = replyTmpl;

    return OK;
}


int  MOClassInfo::setAction (int actionId, Attr* infoTmpl, Attr* replyTmpl)
{
    for (register int i = 0; i < _nactions; i++)
	if (actionId == _actionIds[i]) {	// found it
	    if (_actionInfoTmpls[i] || _actionReplyTmpls[i]) {
		advise(NULLCP, "setAction: Class %s: action %s has templates!",
			_moClassName, _actionNames[i]);
		return NOTOK;
	    }
	    _actionInfoTmpls[i] = infoTmpl;
	    _actionReplyTmpls[i] = replyTmpl;
	    return OK;
	}

    advise(NULLCP, "setAction: Class %s: action with id %d is not registered!",
		    _moClassName, actionId);
    return NOTOK;
}


int  MOClassInfo::setPackage (char* packageName, int packageId)
{
    register int i;
    for (i = 0; i < _npackages && _packageNames[i]; i++)
	if (strcmp(packageName, _packageNames[i]) == 0 ||
					_packageIds[i] == packageId) {
	    advise(NULLCP, "Class %s: package %s with id %d double registered!",
				_moClassName, packageName, packageId);
	    return NOTOK;
	}

    if (i == _npackages) {
	advise(NULLCP, "Class %s: more packages than declared number %d!", 
				_moClassName, _npackages);
	return NOTOK;
    }

    if ((_packageOids[i] = name2oid(packageName)) == NULLOID)
	return NOTOK;

    _packageNames[i] = packageName;
    _packageIds[i] = packageId;

    return OK;
}


int  MOClassInfo::getAttrId (OID attrOid, SetMode* settable, Attr** dflt)
{
    for (register int i = 0; i < _nattrs; i++)
	if (oid_equal(attrOid, _attrOids[i])) {
	    if (settable)
		*settable = _settable[i];
	    if (dflt)
		*dflt = _defaults[i];
	    return _attrIds[i];	// found it
	}
    if (dflt)
	*dflt = NULLATTR;
    return NOTOK;
}


int  MOClassInfo::getAttrId (char* attrName)
{
    for (register int i = 0; i < _nattrs; i++)
	if (strcmp(attrName, _attrNames[i]) == 0)
	    return _attrIds[i];	// found it
    return NOTOK;
}


OID  MOClassInfo::getAttrOid (int attrId)
{
    for (register int i = 0; i < _nattrs; i++)
	if (attrId == _attrIds[i])
	    return _attrOids[i];	// found it
    return NULLOID;
}


int  MOClassInfo::getSetInfo (int attrId, SetMode* settable, Attr** dflt)
{
    for (register int i = 0; i < _nattrs; i++)
	if (attrId == _attrIds[i]) {
	    *settable = _settable[i];
	    *dflt = _defaults[i];
	    return OK;	// found it
	}
    return NOTOK;
}


int  MOClassInfo::getEventId (OID eventOid)
{
    for (register int i = 0; i < _nevents; i++)
	if (oid_equal(eventOid, _eventOids[i]))
	    return _eventIds[i];	// found it
    return NOTOK;
}


OID  MOClassInfo::getEventOid (int eventId)
{
    for (register int i = 0; i < _nevents; i++)
	if (eventId == _eventIds[i])
	    return _eventOids[i];	// found it
    return NULLOID;
}


OID  MOClassInfo::getEventOid (char* eventName)
{
    for (register int i = 0; i < _nevents; i++)
	if (strcmp(eventName, _eventNames[i]) == 0)
	    return _eventOids[i];	// found it
    return NULLOID;
}


int  MOClassInfo::getActionId (OID actionOid, Attr** infoTmpl, Attr** replyTmpl)
{
    for (register int i = 0; i < _nactions; i++)
	if (oid_equal(actionOid, _actionOids[i])) {
	    if (infoTmpl)
		*infoTmpl = _actionInfoTmpls[i];
	    if (replyTmpl)
		*replyTmpl = _actionReplyTmpls[i];
	    return _actionIds[i];	// found it
	}
    if (infoTmpl)
	*infoTmpl = NULLATTR;
    if (replyTmpl)
	*replyTmpl = NULLATTR;
    return NOTOK;
}


int  MOClassInfo::getPackageId (OID packageOid)
{
    for (register int i = 0; i < _npackages; i++)
	if (oid_equal(packageOid, _packageOids[i]))
	    return _packageIds[i];	// found it
    return NOTOK;
}


OID  MOClassInfo::getPackageOid (int packageId)
{
    for (register int i = 0; i < _npackages; i++)
	if (packageId == _packageIds[i])
	    return _packageOids[i];	// found it
    return NULLOID;
}


OID*  MOClassInfo::getGroupAttrs (OID attrOid, int* ngroupattrs)
{
    for (register int i = 0; i < _ngroups; i++)
	if (oid_equal(attrOid, _attrGroups[i][0])) {
	    *ngroupattrs = _ngroupattrs[i];
	    return &_attrGroups[i][1];	// found it
	}
    return (OID *) 0;
}


MO*  MOClassInfo::getMO (RDN rdn)
{
    for (register MO* cur = _first; cur; cur = cur -> getClassNext())
	if (rdn_cmp(rdn, cur -> getRDN()) == 0)
	    return cur;
    return NULLMO;	// can't find it
}


void  MOClassInfo::clear ()
{
    register int i;

    if (_moClass)
	oid_free(_moClass);

    if (_nbindings) {
	for (i = 0; i < _nbindings; i++) {
	    delete _nameBindings[i];
	    _nameBindings[i] = (NameBinding *) 0;
	}
	delete[] _nameBindings;
    }

    if (_nattrs) {
	for (i = 0; i < _nattrs; i++) {
	    oid_free(_attrOids[i]);
	    if (_defaults[i]) {
		_defaults[i] -> clear();
		delete _defaults[i];
		_defaults[i] = NULLATTR;
	    }
	}
	delete[] _attrOids;
	delete[] _attrNames;
	delete[] _attrIds;
	delete[] _settable;
	delete[] _defaults;
    }

    if (_ngroups) {
	for (i = 0; i < _ngroups; i++) {
	    oid_free(_attrGroups[i][0]);
	    delete[] _attrGroups[i];
	}
	delete[] _attrGroups;
    }

    if (_nevents) {
	for (i = 0; i < _nevents; i++)
	    oid_free(_eventOids[i]);
	delete[] _eventOids;
	delete[] _eventNames;
	delete[] _eventIds;
    }

    if (_nactions) {
	for (i = 0; i < _nactions; i++) {
	    oid_free(_actionOids[i]);
	    if (_actionInfoTmpls[i]) {
		_actionInfoTmpls[i] -> clear();
		delete _actionInfoTmpls[i];
		_actionInfoTmpls[i] = NULLATTR;
	    }
	    if (_actionReplyTmpls[i]) {
		_actionReplyTmpls[i] -> clear();
		delete _actionReplyTmpls[i];
		_actionReplyTmpls[i] = NULLATTR;
	    }
	}
	delete[] _actionOids;
	delete[] _actionNames;
	delete[] _actionIds;
	delete[] _actionInfoTmpls;
	delete[] _actionReplyTmpls;
    }

    if (_npackages) {
	for (i = 0; i < _npackages; i++)
	    oid_free(_packageOids[i]);
	delete[] _packageOids;
	delete[] _packageNames;
	delete[] _packageIds;
    }
}


MOClassInfo::MOClassInfo ()
{
    bzero((char*) this, sizeof(MOClassInfo));
}

