//
// NAME
//   monitorMetric.inc.cc
//
// DESCRIPTION
//   for monitorMetric MOC, via OSIMIS GDMO compiler
//   (based on ISO/IEC 10164-11 Monitor Objects MIB)
//
// AUTHORS
//   Jim Reilly
//   Technical Research Centre of Finland
//   icm.group@tel.vtt.fi
//
//   George Pavlou
//   University College London
//   gpavlou@cs.ucl.ac.uk
//


#if	defined(__GNUG__) || (defined(__cplusplus) && defined(AIXV3))
extern "C" { int  strncasecmp(const char*, const char*, size_t); }
#endif

#include "SmiSntx.h"	  /* for QosAlarmInfo + other notifications */
#ifdef ACCESS
#include "ADF.h"	  /* for accessControl */
#endif

#include "OperationalState.h"	/* for scanner::operationalState */
#include "AdministrativeState.h"/* for scanner::administrativeState */



static OID OID_ResourceRequestRate    = NULL;
static OID OID_ObservedObjectInstance = NULL;
static OID OID_ObservedAttributeId    = NULL;
	//
	// These are initialized (just once) in monitorMetric::createRR below.
	//
	// We can *NOT* use: name2oid("~~~") here, since it depends
	// on an OID hashtable being initialized first to work properly.
	// Hashtable is not initialized before above OID handles are.
	//

//
// force update of watched MO attribute
//
void* monitorMetric::forceGet (MO* pmo, OID oid) {

  if (pmo==NULL) return NULL;

  Attr* attr = pmo->getUpdatedAttr(oid, DISCARDAVA);
  if (!attr)
    return NULLVD;

  return attr->get();

} //monitorMetric::forceGet


//
// Managers should specify the attribute values for:
//	scannerId
//	observedObjectInstance
//	observedAttributeId
//	granularityPeriod
//	etc
// when they create a monitorMetric MO.
//

int monitorMetric::createRR(AVA*& err, void* pv, int) {

  //
  // initialize class instance variables first
  //
  _thldAlarmInfo		= NULL;
  _bNoPreviousCounterValue	= True;
  _pmoObserved			= NULL;
  _pvObservedMOattribute	= NULL;

  // correct OO style
  if (scanner::createRR(err, pv) == NOTOK)
    return NOTOK;

  static Bool bClassInitialised = False;


  MN mnObservedObjectInstance = (MN) observedObjectInstance()->get();
  if (!mnObservedObjectInstance ||
         (_dnObservedObjectInstance = mn2localdn(mnObservedObjectInstance))
		== NULLDN) {
    if (!mnObservedObjectInstance)
      err = new AVA(m_missingAttributeValue, "observedObjectInstance");
    else
      err = new AVA(m_invalidAttributeValue, "observedObjectInstance");
    advise (NULL, "NULL or wrong observedObjectInstance in monitorMetric::createRR()\n");
    return NOTOK;
  }
  _typeObservedObjectInstance = mnObservedObjectInstance->mn_type;
        //
	// Distinguished Name (local or global) of watched MO in this agent
        //

  _oidObservedAttribute = observedAttributeId()->getOID();
        //
	// OID of attribute inside MO to be watched
        //
  if (!_oidObservedAttribute)
  {
    err = new AVA(m_missingAttributeValue, "observedAttributeId");
    advise (NULL, "NULL observedAttributeId in monitorMetric::createRR()\n");
    return NOTOK;
  }

  MO* top = MO::getRoot() -> getSubordinate();

  _pmoObserved = top -> getMO(_dnObservedObjectInstance, True, DISCARDAVA);

  char* inst;

  if (_pmoObserved == NULLMO) {
    err = new AVA(m_invalidAttributeValue, "observedObjectInstance");
    advise (
      NULL,
      "Error: NULLMO for DN=\"%s\" in monitorMetric::createRR()\n",
      inst = dn2str(_dnObservedObjectInstance)
    );
    free(inst);
    _pvObservedMOattribute = NULL;
    return NOTOK;
  }
  _pvObservedMOattribute =
        _pmoObserved->getAttr(_oidObservedAttribute, &_strObservedAttribute);
  if (_pvObservedMOattribute == NULLVD) {
    err = new AVA(m_invalidAttributeValue, "observedAttributeId");
    // convert OID to Attribute pointer inside MO
    advise(
      NULL,
      "Error: no attribute %s in MO %s\n",
      _strObservedAttribute,
      inst = dn2str(_dnObservedObjectInstance)
    );
    free(inst);
    return NOTOK;
  }
#ifdef ACCESS
  if (ADF::getInstance()->checkMOAccess
	(MO::getInitiator(), _pmoObserved, AC_GET) !=
				int_Access_EnforcementAction_allow) {
    err = new AVA(m_processingFailure, "observedObjectAccessDenied");
    char* inst2;
    advise(
      NULL,
      "Error: no read access to MO %s by %s\n",
      inst = dn2str(_dnObservedObjectInstance),
      inst2 = dn2str(MO::getInitiator())
    );
    free(inst); free(inst2);
    return NOTOK;
  }
#endif

  //
  // Determine type for Monitor Gauges
  //
  char* sntx = _pvObservedMOattribute->getSyntax();
  if (strcasecmp(sntx, "ObservedValue") == 0) {
    _gaugeType = ((Gauge*)(_pvObservedMOattribute))->getType();
    _observedAttrType = OA_GAUGE;
  } else
  if (strcasecmp(sntx, "SnmpGauge") == 0) {
    _gaugeType = OV_INTEGER;
    _observedAttrType = OA_SNMPGAUGE;
  } else
  if (strncasecmp(sntx, "Count", 5) == 0 ||
		strcasecmp(sntx, "Integer") == 0 ||
		strcasecmp(sntx, "SnmpCounter") == 0) {
    _gaugeType = OV_INTEGER;
    _observedAttrType = OA_COUNTER;
  } else {
    err = new AVA(m_invalidAttributeValue, "observedAttributeId");
    advise(NULL, "attribute %s not a gauge or counter\n",
		_strObservedAttribute);
    return NOTOK;
  }

  if (_observedAttrType == OA_GAUGE || _observedAttrType == OA_SNMPGAUGE) {
    previousScanGaugeValue() -> setType(this->gaugeType());
  }
  else {
    previousScanGaugeValue() -> setType(OV_INTEGER);
  }


  //
  // the derivedGauge Attribute
  // set type for Gauge attributes because GDMO compiler can't know it:
  //   OV_INTEGER or OV_REAL
  //
  derivedGauge() -> setType(this->gaugeType());
  updateDerivedGauge();		// update it


  //
  // the severityIndicatingGaugeThreshold attribute
  // set type for GaugeThreshold attributes because GDMO compiler can't know it:
  //   OV_INTEGER or OV_REAL
  // the type should be same as derivedGauge
  //
  severityIndicatingGaugeThreshold() -> setType(this->gaugeType());
  //
  // set the initial value (GDMO compiler leaves it empty)
  // only if it has not already be set through CMIS
  // because these default values are "hacks" and monitorMetric
  // MOs can be applied to any situation the mode is "Off"
  //
  if (!severityIndicatingGaugeThreshold()->get()) {
     severityIndicatingGaugeThreshold()
       ->add(DERIVEDGAUGETHLDLOW, Off, DERIVEDGAUGETHLDHIGH, Off); 
  }

  //
  // - associate the threshold to the gauge
  //
  derivedGauge()->associateThreshold (severityIndicatingGaugeThreshold());

  //
  // - set the default value/mode only once for the class.
  //
  if (!bClassInitialised) {
    switch(this->gaugeType()) {

    case OV_INTEGER:
      _classInfo->setDefault (
          I_severityIndicatingGaugeThreshold,
          new GaugeThresholdInt (
            DERIVEDGAUGETHLDLOW, Off, DERIVEDGAUGETHLDHIGH, Off
          )
      );
      break;

    case OV_REAL:
      _classInfo->setDefault (
          I_severityIndicatingGaugeThreshold,
          new GaugeThresholdReal (
            DERIVEDGAUGETHLDLOW, Off, DERIVEDGAUGETHLDHIGH, Off
          )
      );
      break;

    }
  }

  // the severityIndicatingTideMarkMax attribute
  // need to set the type first to TM_MAX as the GDMO compiler does not "know"
  severityIndicatingTideMarkMax() -> setType(TM_MAX);

  // the severityIndicatingTideMarkMin attribute
  // need to set the type first to TM_MAX as the GDMO compiler does not "know"
  severityIndicatingTideMarkMin() -> setType(TM_MIN);


  // associate the tide-marks to the gauges
  // unless the observed attribute it is a non-difference counter

  if (_observedAttrType != OA_COUNTER ||
      counterOrGaugeDifference() -> getbool() != False) {

    derivedGauge() -> associateTideMarkMax(severityIndicatingTideMarkMax());
    derivedGauge() -> associateTideMarkMin(severityIndicatingTideMarkMin());
  }

  if (!bClassInitialised) {

    ::OID_ResourceRequestRate = name2oid("resourceRequestRate");
    ::OID_ObservedObjectInstance = name2oid("observedObjectInstance");
    ::OID_ObservedAttributeId = name2oid("observedAttributeId");

    bClassInitialised = True;
  }

  return OK;

} //monitorMetric::createRR 


int monitorMetric::deleteRR(AVA*& err, void* info, Bool checkOnly, int) {

  if (!checkOnly) {
    if (_thldAlarmInfo) {
      AlarmInfo_free_aux(_thldAlarmInfo);
      // frees parts of _thldAlarmInfo (see comment in gms/SmiSntx.c)
    }
  }

  return scanner::deleteRR(err, info, checkOnly);

} //monitorMetric::deleteRR
 

void monitorMetric::updateDerivedGauge () {

#if DEBUG
  advise (NULL, "monitorMetric::updateDerivedGauge()");
#endif

  // temporary holder for derivedGauge 
  double iGaugeValue  = 0;

  AdminStateVal admState = *(AdminStateVal*)
    scanner::administrativeState()->get();
  OperStateVal operState = *(OperStateVal*)
    scanner::operationalState()->get();

  if (admState == as_unlocked && operState == os_enabled) {
    //
    // If the administrative state has the value "unlocked",
    // the scanner is administratively permitted to perform scans.
    // If the administrative state has the value "locked",
    // the scanner is administratively prohibitted to perform scans.
    //
    // The operational state becomes "disabled" only if the observed MO dies.
    // A stateChange notification is sent so that the manager may
    // realise it and delete the monitorMetric MO.

    // it has to be checked every time if the watched MO is still there

    MO* top = MO::getRoot() -> getSubordinate();

    _pmoObserved = top -> getMO(_dnObservedObjectInstance, True, DISCARDAVA);

    //   Force a get on watched MO::attribute's value,
    //   so we have an up-to-date attribute value if other MO has
    //   not been e.g. refreshed
    //
    void* val;
    if (_pmoObserved == NULLMO ||
	(val = forceGet(_pmoObserved, _oidObservedAttribute)) == NULLVD) {
      void* oldStateVal = scanner::operationalState()->copy();
      scanner::operationalState() -> set(os_disabled);
      triggerStateChange(I_operationalState, scanner::_level, oldStateVal);

      advise (NULL, "Error: NULLMO in monitorMetric::updateDerivedGauge()\n");
      return;
    }


    if (_observedAttrType == OA_GAUGE || _observedAttrType == OA_SNMPGAUGE) {

      //
      // Gauge Attribute
      //

      double iCurrGaugeVal = 0;
      double iPrevGaugeVal = 0;

      //
      // get the current observed gauge value and save the previous one

      if (_observedAttrType == OA_GAUGE) {
        switch(this->gaugeType()) {
          case OV_INTEGER:
            iGaugeValue = iCurrGaugeVal = ((ObservedValue*) val)
	      -> ov_integer;
            iPrevGaugeVal = 
              ((ObservedValue*) previousScanGaugeValue() -> get())
              -> ov_integer;
            break;

          case OV_REAL:
            iGaugeValue = iCurrGaugeVal = ((ObservedValue*) val)
              -> ov_real;
            iPrevGaugeVal =
              ((ObservedValue*) previousScanGaugeValue() -> get())
              -> ov_real;
            break;
        }
      }
      else {	// OA_SNMPGAUGE
        iGaugeValue = iCurrGaugeVal = *(int*) val;
        iPrevGaugeVal = ((ObservedValue*) previousScanGaugeValue() -> get())
			-> ov_integer;
      }


      if (counterOrGaugeDifference() -> getbool() != False) {

        //
        // the derived gauge ("rate") is converted from the observed gauge:
        //     (G2 - G1) according to the text in 10164-11
        //

        int iPollInterval = 0;

        previousScanGaugeValue()->setreal(iCurrGaugeVal);

        if (_bNoPreviousCounterValue == True) {
          //
          // We need another sample, this is first time thru
          //

          iGaugeValue = 0;	// no meaningful difference
          _bNoPreviousCounterValue = False;

        } else {
  
          iPollInterval = TimePeriod2seconds(
            (type_Monitor_TimePeriod*) scanner::granularityPeriod()->get()
          );

          // calculate the rate

          iGaugeValue = iCurrGaugeVal - iPrevGaugeVal;
          // iGaugeValue /= iPollInterval;	// GP 16-2-94

#if DEBUG 
          advise(
            NULL,
            "monmet Gauge rate conversion: gcur %f gprev %f T %d G %f",
            iCurrGaugeVal, iPrevGaugeVal, iPollInterval, iGaugeValue
          );
#endif

        }
      }

    } else {	// OA_COUNTER

      int iCurrGaugeVal = 0;

      iGaugeValue = iCurrGaugeVal = *(int*) val;

      if (counterOrGaugeDifference() -> getbool() != False) {
        //
        // the derived gauge ("rate") is converted from the observed counter:
        //     (C2 - C1) according to the text in 10164-11
        //
  
        int iPrevGaugeVal = 0;
        int iPollInterval = 0;

        // get and reset previousScanCounterValue
        iPrevGaugeVal = previousScanCounterValue()->getint();
        previousScanCounterValue()->set(iCurrGaugeVal);
  
        if (_bNoPreviousCounterValue == True) {
          //
          // We need another sample, this is first time thru
          //

          iGaugeValue = 0;	// no meaningful difference
          _bNoPreviousCounterValue = False;
        
        } else {
  
          iPollInterval = TimePeriod2seconds(
            (type_Monitor_TimePeriod*) scanner::granularityPeriod()->get()
          );

          // calculate the rate
  
          iGaugeValue = iCurrGaugeVal - iPrevGaugeVal;
          // iGaugeValue /= iPollInterval;	// GP 16-02-94
  
#if DEBUG 
          advise(
            NULL,
            "monmet Cntr rate conversion: ccur %d cprev %d T %d G %f",
            iCurrGaugeVal, iPrevGaugeVal, iPollInterval, iGaugeValue
          );
#endif
  
        }
      }

    }

    //
    // In either case, Gauge or Counter, the correct value
    // to insert into the derivedGauge attribute is now
    // inside iGaugeValue.
    //
    // Note: Gauge::setreal(double) takes care of Gauge's type
    //       (OV_INTEGER/OV_REAL)
    //
    derivedGauge()->setreal(iGaugeValue);

#if DEBUG
    advise (NULL, "DBUG derivedGauge set to %f\n", iGaugeValue);
#endif
  

    if (derivedGauge()->triggeredThreshold() == On) {
      _qosAlarmGauge   = derivedGauge();
      _qosAlarmThldOid = _classInfo ->
				getAttrOid(I_severityIndicatingGaugeThreshold);
      triggerEvent(I_qualityofServiceAlarm, _level);
    }
  

  } // if unlocked (scan)
  
} //monitorMetric::updateDerivedGauge


int monitorMetric::buildReport (int eventId, int classLevel,
				void*& info, Bool& freeFlag) {
    if (classLevel != _level) {
        return scanner::buildReport(eventId, classLevel, info, freeFlag);
        // for correct OO style...
    }

    freeFlag = False;	// we take control for efficiency

    switch (eventId) {
 
    case I_qualityofServiceAlarm:
      if((_thldAlarmInfo == NULL) &&
         (_thldAlarmInfo=this->_thldAlarmInfo_alloc())==NULL
      ) {
        return NOTOK;	// not alloc'd and can't alloc
      }
      
      _thldAlarmInfo->thresholdInfo->triggeredThreshold->mid_global =
			_qosAlarmThldOid;	// NOT to be free'd
      
      _thldAlarmInfo->thresholdInfo->observedValue->offset = gaugeType();
      switch(gaugeType()) {
        case OV_INTEGER:
          _thldAlarmInfo->thresholdInfo->observedValue->un.integer =
            ((GaugeInt*) _qosAlarmGauge) -> getint();
          break;
        case OV_REAL:
          _thldAlarmInfo->thresholdInfo->observedValue->un.real =
            ((GaugeReal*) _qosAlarmGauge) -> getreal();
          break;
      }

      info = _thldAlarmInfo;
    break;

    default:
      return NOTOK;

    } /* end switch */
 
    return OK;

} //monitorMetric::buildReport


//
// used to initialize _thldAlarmInfo pointer during createRR()
// freed by deleteRR()
//
type_SMI_AlarmInfo*
monitorMetric::_thldAlarmInfo_alloc ()
{
    struct type_SMI_AlarmInfo* elem;
    if ((elem = ThresholdAlarmInfo_ProbsAttrs_alloc (NULL, 1, 2))== NULL) {
       return NULL;
    }

    /* specificProblems */

    ((MIDListVal*) elem->specificProblems)->midl_mid.mid_global =
        ::OID_ResourceRequestRate;
 
    /* monitoredAttrs */

    /* insert observedObjectInstance OID/value */
    ((CMISParamListVal*) elem->monitoredAttrs)->mpl_id.mid_global =
       ::OID_ObservedObjectInstance;
    ((CMISParamListVal*) elem->monitoredAttrs)->mpl_val =
        pe_cpy(observedObjectInstance()->encode());

    /* insert observedAttributeId OID/value */
    ((CMISParamListVal*) elem->monitoredAttrs)->mpl_next->mpl_id.mid_global =
       ::OID_ObservedAttributeId;
    ((CMISParamListVal*) elem->monitoredAttrs)->mpl_next->mpl_val =
        pe_cpy(observedAttributeId()->encode());

    return elem;
}


int  monitorMetric::set (
  CMISModifyOp op, int attrId, int classLevel, void* newval,
  AVA*& err, Bool checkOnly, int
) {

    if (classLevel != _level) {
        return scanner::set(op, attrId, classLevel, newval, err, checkOnly);
    }

#if DEBUG
    advise(NULL, "!!! monitorMetric::set");
#endif
 
    AdminStateVal admState = *(AdminStateVal*)
      scanner::administrativeState()->get();

    switch (attrId) {
    case I_severityIndicatingGaugeThreshold:
      if ((admState==as_unlocked)) {
        if (checkOnly)
	  return NOTOK;
        err = new AVA(m_processingFailure, "adminStateUnlocked");
        return NOTOK;
      }
      break;

    default:
      break;
    }
    return OK;

} // monitorMetric::set
