//
// NAME
//   monitoringSupport.inc.cc
//
// DESCRIPTION
//   for monitoringSupport MOC, via OSIMIS GDMO compiler
//
// IMPLEMENTORS
//   George Mykoniatis
//   NTUA
//   mykoniat@ektor.ntua.gr
//
// HISTORY
//   A first version was produced by
//   Kostas Kasapakis (Alpha), Panos Georgatsos (Alpha),
//   George Mykoniatis (NTUA), Jorge Sanchez (NTUA)
//

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "sumstat.h"    
#include "observed_attr.h"    
#include "AdministrativeState.h" /* for MSscanner::operationalState */
#include "OperationalState.h"    /* for MSscanner::operationalState */
#include "SmiSntx.h"      /* for QosAlarmInfo + other notifications */

// --------------------------------------------------------------
// deallocates memory 
// --------------------------------------------------------------
int monitoringSupport::RemoveOAs()
{
  if (oa_array)
  {
    for (int i=0; i < no_of_attributes; i++)
        {
          if (oa_array[i])
                delete oa_array[i];
        }
        delete oa_array;

        oa_array = NULL;
  }
  return OK;
}
	
// --------------------------------------------------------------
// Stops the monitoringSupport instance
// --------------------------------------------------------------
int monitoringSupport::ShutDown()
{
  if (wkup_status)
  {
	  wkup_status = 0;
	  cancelWakeUps(NULLCP);    // Disable wake ups
	  //MSscanner::cancelWakeUps(NULLCP);    // Disable wake ups 
  }
  status = 0;

  RemoveOAs();				// Release Memory 

  if (oaval) 
  {
	delete oaval;
    oaval = NULL;
  }

  if (stat_p) 
  {
	delete stat_p;
	stat_p = NULL;
  }
#ifdef DEBUG
   advise(NULL,"monitoringSupport::ShutDown : just executed");
#endif
  return OK;
}

// --------------------------------------------------------------
// Used by KS::wakeup method. 
// --------------------------------------------------------------
void monitoringSupport::update()
{
 AdminStateVal admState = *(AdminStateVal*)
    MSscanner::administrativeState()->get();
 OperStateVal operState = *(OperStateVal*)
    MSscanner::operationalState()->get();

 if (admState == as_unlocked && operState == os_enabled) 
 {
  int i;
  int alg;
  double dtmp ;

  if (status == 0)  				// Not correctly initialized yet
  {
     // StartUp is a virtual function used to specialize
     // the two derived MOCs in the initialization phase.
     if (StartUp() == NOTOK) 			// Not properly initialized
     {
	fprintf(stderr, "monitoringSupport::update initialisation of object failed \n");
	return;					// Failed to initialize...
     } // endif  StartUp
     status = 1;
  }	// endif status
  else if (CheckConsistency() == NOTOK)
  {
     void* oldStateVal = MSscanner::operationalState()->copy();
     MSscanner::operationalState() -> set(os_disabled);
     triggerStateChange(I_operationalState, MSscanner::_level, oldStateVal);

     advise (NULL, "Error: Changed observed Attributes in monitoringSupport::update()\n");

     return;
  }


#ifdef DEBUG
  char *str = (char *)((String *)MSscanner::_attrs[I_scannerId]->get());
  fprintf(stdout,"\nmonitoringSupport::update - Observing %d of object : %s\n",
							no_of_attributes,str);
#endif

// Update the attribute values
  for (i=0; i < no_of_attributes ; i++) // For each attribute
  {
    	// Update the value of each attribute and 
	// and store it in the oaval matrix
	oaval[i] = oa_array[i]->getval_attr();
  }

  // Now apply the OPERATION stored in operationType attribute 
  // Find derived result 
  alg = *(int *)((Integer *)_attrs[I_operationType])->get(); 
  switch( alg )
  {
    case ALG_FindSum:         		FindSum();          break;
    case ALG_FindDiv:         		FindDiv();          break;
    case ALG_FindMean:        		FindMean();         break;
    case ALG_FindSubstraction:        	FindSubstraction(); break;
    case ALG_FindProd:        		FindProd();         break;
    case ALG_FindProb:        		FindProb();         break;
    case ALG_FindSpecialSum:  		FindSpecialSum();   break;
	default:
	  fprintf( stderr, "Invalid algorithm %d in operationType\n",alg); 
  };

  // Error checking
  // result = max(result,0); 
  if (result < 0)
    result = 0;

  // Store it, in attribute derivedResult
  derivedResult()->setreal(result); 

  // Now, apply on the result of the operation of the algorithm
  // stored in algorithmType attribute
  stat_p->append_value(result);
  stat_p->print_sumstats();
  dtmp = stat_p->get_result();

  // Store the result in attribute estimateOfMean
  estimateOfMean()->setreal(dtmp); 

    if (derivedResult()->triggeredThreshold() == On) {
      _qosAlarmGauge   = derivedResult();
      _qosAlarmThldOid = _classInfo ->
                                getAttrOid(I_severityIndicatingGaugeThreshold);
      triggerEvent(I_qualityofServiceAlarm, _level);
    }

    if (estimateOfMean()->triggeredThreshold() == On) {
      _qosAlarmGauge   = estimateOfMean();
      _qosAlarmThldOid = _classInfo ->
                                getAttrOid(I_estimateOfMeanSeverityIndicatingGaugeThreshold);
      triggerEvent(I_qualityofServiceAlarm, _level);
    }


#ifdef DEBUG
  fprintf(stdout,"monitoringSupport::update - Derived result from operation : %lf\n", result);
  fprintf(stdout,"monitoringSupport::update - Derived result from sumstat   : %lf\n", dtmp);
#endif

  return;
 } // if unlocked (scan)
}

	
// --------------------------------------------------------------
// Initializes all the attributes of the new monitoringSupport 
// object instance to initial values
// --------------------------------------------------------------
int monitoringSupport::createRR (AVA*& err, void* pv, int)
{

  // correct OO style
  if (MSscanner::createRR(err, pv) == NOTOK)
    return NOTOK;

  static Bool bClassInitialised = False;

  // Initialise the variables for this instance
  _thldAlarmInfo                = NULL;

  status = 0;              // Set the status to inactive
  no_of_attributes = 0;    // Set number of attributes
                           // to zero
  oaval = NULL;            // Pointer to array of doubles
                           // containing the actual values
  stat_p = NULL;           // sumstats instance pointer,
                           // initialised to NULL

  oa_array = NULL;         // Initiailise NULL

#ifdef DEBUG
   fprintf(stdout, "monitoringSupport::createRR - Creating an MO with MO id : %s \n",
                   (char *)MSscanner::_attrs[I_scannerId]->get());
#endif

//Set types...

   derivedResult()->setType(OV_REAL);

   estimateOfMean()->setType(OV_REAL);

   severityIndicatingGaugeThreshold() ->setType(OV_REAL);

   estimateOfMeanSeverityIndicatingGaugeThreshold()->setType(OV_REAL);

   severityIndicatingTideMarkMax()->setType(TM_MAX);

   severityIndicatingTideMarkMin()->setType(TM_MIN);

//Associate and set initial values....

   severityIndicatingGaugeThreshold()
	->add(MSDERIVEDGAUGETHLDLOW, Off, MSDERIVEDGAUGETHLDHIGH, Off);

   derivedResult() ->associateThreshold(severityIndicatingGaugeThreshold());

   derivedResult()
	->associateTideMark(severityIndicatingTideMarkMax());

   derivedResult()
	->associateTideMark(severityIndicatingTideMarkMin());

   derivedResult()->setreal((double)0.0);

//Set types...

   estimateOfMeanSeverityIndicatingTideMarkMax()->setType(TM_MAX);

   estimateOfMeanSeverityIndicatingTideMarkMin()->setType(TM_MIN);


//Associate and set initial values....

   estimateOfMeanSeverityIndicatingGaugeThreshold()
	->add(MSDERIVEDGAUGETHLDLOW, Off, MSDERIVEDGAUGETHLDHIGH, Off);

  estimateOfMean()->associateThreshold(estimateOfMeanSeverityIndicatingGaugeThreshold());

  estimateOfMean()->associateTideMark(estimateOfMeanSeverityIndicatingTideMarkMax());

  estimateOfMean()->associateTideMark(estimateOfMeanSeverityIndicatingTideMarkMin());

   estimateOfMean()->setreal((double)0.0);

   if (!bClassInitialised)
   { 
    	bClassInitialised = True;
	_info.setDefault ( 
			   I_severityIndicatingGaugeThreshold,
	    		   new GaugeThresholdReal ( 
				 MSDERIVEDGAUGETHLDLOW, 
				 Off, 
				 MSDERIVEDGAUGETHLDHIGH, 
				 Off)
			 );
	_info.setDefault ( 
			   I_estimateOfMeanSeverityIndicatingGaugeThreshold,
	    		   new GaugeThresholdReal ( 
				 MSDERIVEDGAUGETHLDLOW, 
				 Off, 
				 MSDERIVEDGAUGETHLDHIGH, 
				 Off)
			 );
   }				   

#ifdef DEBUG
// Print a message that initialisation has finished
   fprintf( stdout, "Finished createRR\n");
#endif
   return OK;
} //monitoringSupport::createRR

int monitoringSupport::deleteRR (AVA*& err, void* info, Bool checkOnly, int)
{

#ifdef DEBUG
  fprintf(stdout, "monitoringSupport::deleteRR-Executing.....\n");
#endif

  ShutDown();

  return MSscanner::deleteRR(err, info, checkOnly);
} //monitoringSupport::deleteRR

// --------------------------------------------------------------
int monitoringSupport::set (CMISModifyOp mode, int attrId, int classLevel,
                      void* value, AVA*& err, Bool checkOnly, int)
{
    
  if (classLevel != _level) 
      return MSscanner::set(mode, attrId, classLevel, value, err, checkOnly);

  int pval;             // Passed value (if integer ...)
    int iPollInterval = 0;

#ifdef DEBUG
  fprintf(stdout, "monitoringSupport::set-Executing.....\n");
#endif

    AdminStateVal admState =  *(AdminStateVal*)
	MSscanner::administrativeState()->getval();

  switch (attrId)
  {
    case I_severityIndicatingGaugeThreshold:
    case I_estimateOfMeanSeverityIndicatingGaugeThreshold:
    case I_severityIndicatingTideMarkMax:
    case I_severityIndicatingTideMarkMin:
    case I_estimateOfMeanSeverityIndicatingTideMarkMax:
    case I_estimateOfMeanSeverityIndicatingTideMarkMin:
      if (admState == as_locked) {
        // The attributes :
        //
        // severityIndicatingGaugeThreshold
        // estimateOfMeanSeverityIndicatingGaugeThreshold
        // severityIndicatingTideMarkMax
        // severityIndicatingTideMarkMin
        // estimateOfMeanSeverityIndicatingTideMarkMax
        // estimateOfMeanSeverityIndicatingTideMarkMin
        //
        // in the probabilityEstimator managed object
        // class shall not be modified unless the value of the
        // administrative state is "locked"

        if (checkOnly)
          return OK;

        advise (NULL,
            "monitoringSupport::set() new collectionPeriod=%d \n",*(int *) value);
        }
      else {
        if (checkOnly)
          return NOTOK;

         err = new AVA(m_processingFailure, "adminStateUnlocked");
         return NOTOK;
      }
      break;


    case I_operationType:
      if (admState == as_locked) {

	if ( (*(int *)value) != 1 && (*(int *)value) != 2 && (*(int *)value) != 3 
		&& (*(int *)value) != 4 && (*(int *)value) != 5 && (*(int *)value) != 6
		&& (*(int *)value) != 7)
	   {
	   advise (NULL, 
	      "monitoringSupport::set(): invalid operationType value, set failed \n");

           if (checkOnly)
	      return NOTOK;

	   err = new AVA(m_invalidAttributeValue, "invalidValue");
	   return NOTOK;
	   }

	if (checkOnly)
           return OK;

	 }
      else {
	if (checkOnly)
          return NOTOK;

	 err = new AVA(m_processingFailure, "adminStateUnlocked");
	 return NOTOK;
	 }
      break;

    case I_algorithmType:
      if (admState == as_locked) {
         if ( (*(int *)value) != 0 && (*(int *)value) != 1 && (*(int *)value) != 2
                && (*(int *)value) != 3 )
          {
           advise (NULL,
              "monitoringSupport::set(): invalid algorithmType value, set failed \n");
           
           if (checkOnly)            
              return NOTOK;

           err = new AVA(m_invalidAttributeValue, "invalidValue");
           return NOTOK;
          }

         if (checkOnly)
           return OK;

         }
      else {
        if (checkOnly)
          return NOTOK;

         err = new AVA(m_processingFailure, "adminStateUnlocked");
         return NOTOK;
         }
      break;

    case I_collectionPeriod:
      if (admState == as_locked) {
        // The collection Period in the monitoringSupport managed object
        // class shall not be modified unless the value of the
        // administrative state is "locked"

        if (checkOnly)
          return OK;

        advise (NULL,
            "monitoringSupport::set() new collectionPeriod=%d \n",*(int *) value);
        }
      else {
        if (checkOnly)
          return NOTOK;

         err = new AVA(m_processingFailure, "adminStateUnlocked");
         return NOTOK;
      }
      break;

    default:
      break;

  } // Switch attrid

  return OK;
} //monitoringSupport::set

// -------------------------------------------------------------------------
// Functions related with the algorithms 
// applied to the observed attributes
// -------------------------------------------------------------------------

// -------------------------------------------------------------------------
// Simple sum of the observed attribute values
void monitoringSupport::FindSum()
{
   result = 0;
   for(int i=0; i < no_of_attributes; i++)
      result += oaval[i];
}

// -------------------------------------------------------------------------
// Mean of the observed attribute values
void monitoringSupport::FindMean( )
{
  result = 0;
  for(int i=0; i < no_of_attributes; i++ )
    result += oaval[i];
  result /= no_of_attributes;
}

// -------------------------------------------------------------------------
// Division of the observed attribute values
void monitoringSupport::FindProb()
{
  result = 0;
  if ( (oaval[1]+oaval[0]) == 0 )
    fprintf(stderr, "monitoringSupport::FindDiv - Division by zero\n");
  else
    result = oaval[0]/(oaval[1]+oaval[0]);
}

// -------------------------------------------------------------------------
// In this case the result is computed as follows:
//    result = value1 - (value2 + value3)
void monitoringSupport::FindSpecialSum()
{
  result = 0;
  result = oaval[0]-oaval[1]-oaval[2];
}

// -------------------------------------------------------------------------
// It computes simply the ratio value1/value2
void monitoringSupport::FindDiv()
{
    result = 0;
    if (oaval[1] == 0)
      fprintf(stderr, "monitoringSupport::FindDiv - Division by zero\n");
    else
      result = oaval[0]/oaval[1];
}

// -------------------------------------------------------------------------
// It computes simply the ratio value1*value2
void monitoringSupport::FindProd()
{
    result = 0;
    if (oaval[1] == 0)
      fprintf(stderr, "monitoringSupport::FindDiv - Division by zero\n");
    else
      result = oaval[0]*oaval[1];
}

// -------------------------------------------------------------------------
// It computes simply the ratio value1-value2
void monitoringSupport::FindSubstraction()
{
    result = 0;
    if (oaval[1] == 0)
      fprintf(stderr, "monitoringSupport::FindDiv - Division by zero\n");
    else
      result = oaval[0] - oaval[1];
}

int monitoringSupport::buildReport (int eventId, int classLevel,
                                void*& info, Bool& freeFlag)
{
    if (classLevel != _level) {
        return MSscanner::buildReport(eventId, classLevel, info, freeFlag);
        // for correct OO style...
    }

    freeFlag = False;   // we take control for efficiency

    switch (eventId) {

    case I_qualityofServiceAlarm:
      if((_thldAlarmInfo == NULL) &&
         (_thldAlarmInfo=_thldAlarmInfo_alloc())==NULL
      ) {
        return NOTOK;   // not alloc'd and can't alloc
      }

      _thldAlarmInfo->thresholdInfo->triggeredThreshold->mid_global =
                        _qosAlarmThldOid;       // NOT to be free'd

      _thldAlarmInfo->thresholdInfo->observedValue->offset = OV_REAL;
      _thldAlarmInfo->thresholdInfo->observedValue->un.real =
                        ((GaugeReal*) _qosAlarmGauge) -> getreal();

      info = _thldAlarmInfo;
    break;

    default:
      return NOTOK;

    } /* end switch */

    return OK;

} //monitoringSupport::buildReport

