/*
 * Copyright (c) 1992 University College London
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Department of Computer Science, University College London,
 * London, England. Neither the name of the University, nor its
 * registered trademark, UCL, may be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*/
/****************************************************************/
/*
	GDMOAttribute.h

	Class that represents a GDMO Attribute

	Author: J.Cowan		(j.cowan@.cs.ucl.ac.uk)

*/
/****************************************************************/

#ifndef	_SYMBOL
#include "Symbol.h"
#endif

#ifndef	_GENERIC_LIST_
#include "GenList.h"
#endif

#ifndef	_LIST_ITERATOR_
#include "ListIterator.h"
#endif

#ifndef	_GENERIC_STRING_
#include "GenStr.h"
#endif

/****************************************************************/

class	GDMOAttribute;

class	GDMOAttrList : public ListSymbol
{
protected:
	void	deleteElem (void*);
public:
	GDMOAttrList(Bool del, char *name)
	: ListSymbol(name, "GDMOAttribute", ObjectTypeDef, del) {}
	GDMOAttrList(char *name)
	: ListSymbol(name, "GDMOAttribute", ObjectTypeDef) {}
	~GDMOAttrList() { clear();}

	GDMOAttribute *FindAttr(char *attrName);
};

/****************************************************************/
/* Table for converting attribute syntaxes/types to code equivalent */
/****************************************************************/

class  ConvertEntry
{
	GString	_convertName;
	GString	_className;
	GString	_headerFName;
public:
	ConvertEntry(GString convertName, GString className,
			 GString headerFName)
	: _convertName(convertName), _className(className),
	  _headerFName(headerFName)
	{}
	char	*GetConvertName()	 {return (char *) _convertName;}
	char	*GetClassName()		{return (char *) _className;}
	char	*GetHeaderFName()	{return (char *) _headerFName;}
};

class	ConvertList : public List
{
protected:
	void	deleteElem (void*);
};

/****************************************************************/

class	ConvertTable
{
	ConvertList _convertList;
public:
	ConvertTable() {}	// stop compiler warning

	~ConvertTable()	{_convertList.clear();}

	ConvertEntry	*Lookup(char *);
	void		Initialise(char *fileName);
};

/****************************************************************/

typedef	enum	{Get=1, Replace=2, GetReplace=3, SetByCreate=4} GetReplaceType;

typedef	enum	{Add=1, Remove=2, AddRemove=3} AddReplaceType;

typedef	enum	{Equality=1, Ordering=2, SubStrings=4, SetComparison=8,
		 SetInterSection=16} MatchesType;

class GDMOAttribute : public MetaClass
{
	StringSymbol	_name, _syntax, _derived, _defaultValue,
			_initialValue, _permittedValue, _requiredValue,
			_codeName, _headerFileName;
	IntegerSymbol	_matchesFor, _replaceWithDefault, _getReplace,
			_addRemove;
	GDMOOid		_oid;
	static		ConvertTable _attrTable, _syntaxTable;
public:
	GDMOAttribute(char *name) :
			MetaClass("GDMOAttribute"),
			_name(name, "AttributeName"),
			_syntax("", "Syntax"),
			_derived("", "Derived"),
			_matchesFor(False, "MatchesFor"),
			_replaceWithDefault(False, "ReplaceWithDefault"),
			_defaultValue("", "DefaultValue"),
			_initialValue("", "InitialValue"),
			_permittedValue("", "PermittedValue"),
			_requiredValue("", "RequiredValue"),
			_getReplace(0, "GetReplace"),
			_addRemove(0, "AddRemove"),
			_oid("Oid"),
			_codeName("", "ClassName"),
			_headerFileName("", "headerFileName")
	{
		AddSymbol(&_name);
		AddSymbol(&_syntax);
		AddSymbol(&_derived);
		AddSymbol(&_matchesFor);
		AddSymbol(&_replaceWithDefault);
		AddSymbol(&_defaultValue);
		AddSymbol(&_initialValue);
		AddSymbol(&_permittedValue);
		AddSymbol(&_requiredValue);
		AddSymbol(&_getReplace);
		AddSymbol(&_addRemove);
		AddSymbol(&_oid);
		AddSymbol(&_codeName);
		AddSymbol(&_headerFileName);
	}

	~GDMOAttribute() {}

	void	AddSyntax(char *syntax);

	void	AddDerivation(char *derived, GDMOAttrList &);

	// return name of attribute

	char	*GetName()
	{
		return (char *) _name;
	}

	// return syntax of attribute

	char	*GetSyntax()
	{
		return (char *) _syntax;
	}

	// return derivation of attribute

	char	*GetDerivation()
	{
		return (char *) _derived;
	}

	GDMOOid	*GetOid() { return &_oid;}

	void	UpdatePkgAttrRef(GDMOAttribute *);
	void	UpdatePkgAttrRef(ConvertEntry *);

	static	void		AddSyntaxDatabase(char *syntaxDbase);
	static	void		AddAttributeDatabase(char *attrDbase);
	static	ConvertEntry	*LookupAttributeDatabase(char *attrName);
	static	ConvertEntry	*LookupSyntaxDatabase(char *syntaxName);
};

/****************************************************************/
/* Attribute Group */

class GDMOAttributeGroup : public MetaClass
{
	StringSymbol		_name;
	StringSymbolList	_elements;
	StringSymbolList	_extensions;
	IntegerSymbol		_fixed;
	GDMOOid			_oid;

	Bool	FindAttr(GDMOAttrList &, ListIterator &);

public:
	GDMOAttributeGroup(char *name) :
			MetaClass("GDMOAttributeGroup"),
			_name(name, "AttributeGroupName"),
			_elements("Elements"),
			_extensions("Extensions"),
			_fixed(False, "Fixed"),
			_oid("Oid")
	{
		AddSymbol(&_name);
		AddSymbol(&_elements);
		AddSymbol(&_extensions);
		AddSymbol(&_fixed);
		AddSymbol(&_oid);
	}

	Bool	UpdateRefs(GDMOAttrList &);

	char	*GetName()
	{
		return (char *) _name;
	}

	GDMOOid	*GetOid() { return &_oid;}
};

/****************************************************************/

class	GDMOAttrGroupList : public ListSymbol
{
protected:
	void	deleteElem (void*);
public:
	GDMOAttrGroupList(Bool del, char *name)
	: ListSymbol(name, "GDMOAttributeGroup", ObjectTypeDef, del) {}
	GDMOAttrGroupList(char *name)
	: ListSymbol(name, "GDMOAttributeGroup", ObjectTypeDef) {}
	~GDMOAttrGroupList() { clear();}

	GDMOAttributeGroup *FindAttrGroup(char *);
};

/****************************************************************/
// inline methods for GDMOAttrList

inline void  GDMOAttrList::deleteElem (void* val)
{  if (del) delete (GDMOAttribute *) val; }

/****************************************************************/
// inline methods for GDMOAttrGroupList

inline void  GDMOAttrGroupList::deleteElem (void* val)
{  if (del) delete (GDMOAttributeGroup *) val; }

/****************************************************************/
// inline methods for ConvertList

inline void  ConvertList::deleteElem (void* val)
{  if (del) delete (ConvertEntry *) val; }

/****************************************************************/





