/*
 * Copyright (c) 1992 University College London
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Department of Computer Science, University College London,
 * London, England. Neither the name of the University, nor its
 * registered trademark, UCL, may be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*/
/****************************************************************/
/*
	GDMOClas.cc

	Methods of a GDMO Managed Object Class definition

	Author: J.Cowan		(j.cowan@.cs.ucl.ac.uk)
*/
/****************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include "DebugMem.h"
#include "GDMOGen.h"
#include "GDMOOid.h"
#include "GDMOAttr.h"
#include "GDMOAct.h"
#include "GDMONot.h"
#include "GDMOPkg.h"
#include "GDMOClas.h"

/****************************************************************/
/* Add Name binding reference */

void GDMOClass::AddNameBinding(GDMONameBinding *nameBinding)
{
	_nameBindingList.append(nameBinding);

	// Subordinate Mo should be this!

	if (nameBinding->GetSubordinateMo() != this)
		printf("\nInternal error in GDMOClass::AddNameBinding");
}

/****************************************************************/
/* Add MO child reference */

void GDMOClass::AddChild(GDMOClass *moChild)
{
	_childList.append(moChild);
}

/****************************************************************/
// package references

void GDMOClass::AddPackageRef(char *packageName)
{
	GDMOPackageRef *packageRef = (NEW GDMOPackageRef(packageName));

	_packageRefList.append(packageRef);
}

/***********************************************************/
// parent references of class ManagedObject

void GDMOClass::AddParentRef(char *moClass)
{
	GDMOClassRef *MOref = (NEW GDMOClassRef(moClass));

	_parentRefList.append(MOref);

} /* AddParentRef */

/***********************************************************/
// Update references to classes and packages

Bool GDMOClass::UpdateRefs(GDMOList *moList, GDMOPackageList *packageList)
{

	if (UpdateMoRefs(&_parentRefList, moList) == False)
		return False;

	if (UpdateMoPackageRefs(&_packageRefList, packageList)
			== False)
		return False;

	if (UpdateChildRefs(moList) == False)
		return False;

	UpdateFromPackage();
	UpdateParentList();

	return True;

} /* UpdateRefs */

/***********************************************************/
// Update MO class refs

Bool GDMOClass::UpdateMoRefs(GDMORefList *refList, GDMOList *moList)
{
GDMOClassRef	*moRef;
ListIterator	refIterator(refList);

	while ((moRef=(GDMOClassRef *)refIterator.getNext()) !=
		(GDMOClassRef *) 0)
	{
	GDMOClass	*moClass;

		if ((moClass=moList->FindMo(moRef->GetName())) !=
				(GDMOClass *) 0)
			moRef->UpdateRef(moClass);
		else
		{
		GString	msg = GString("Cannot find MO ") +
			      GString(moRef->GetName());

			PrintError((char *)msg);
			return False;
		}
	}

	return True;

} /* UpdateMoRefs */

/***********************************************************/
// Update MO package refs

Bool GDMOClass::UpdateMoPackageRefs(GDMOPackageRefList *refList,
				    GDMOPackageList *packageList)
{
GDMOPackageRef	*packageRef;
ListIterator	refIterator(refList);

	while ((packageRef=(GDMOPackageRef *)refIterator.getNext()) !=
		(GDMOPackageRef *) 0)
	{
	GDMOPackage	  *package;

		if ((package=packageList->FindPackage(packageRef->GetName()))
				!= (GDMOPackage *) 0)
			packageRef->UpdateRef(package);
		else
		{
		GString	msg = GString("Cannot find package ") +
			      GString(packageRef->GetName());

			PrintError((char *)msg);
			return False;
		}
	}

	return True;

} /* UpdateMoRefs */

/****************************************************************/
// Update child lists in MOs

Bool GDMOClass::UpdateChildRefs(GDMOList *moList)
{
ListIterator	refIterator(&_parentRefList);
GDMOClassRef	*moRef;

	while ((moRef=(GDMOClassRef *)refIterator.getNext()) !=
		(GDMOClassRef *)0)
	{
	GDMOClass	*parentMo;

		if ((parentMo=moList->FindMo(moRef->GetName())) !=
				(GDMOClass *) 0)
			parentMo->AddChild(this);
		else
		{
		GString	msg = GString("Cannot find MO ") +
			      GString(moRef->GetName());

			PrintError((char *)msg);
			return False;
		}
	}
	return True;
}

/***********************************************************/
/* Update attributes, actions and notifications */

void GDMOClass::UpdateFromPackage()
{
GDMOPackageRef	*packageRef;
ListIterator	pkgIterator(&_packageRefList);

	while ((packageRef=(GDMOPackageRef *)pkgIterator.getNext()) !=
		(GDMOPackageRef *)0)
	{
	GDMOPackage	*package = packageRef->GetPackage();
	Symbol	*toSym = FindSymbol("BehaviourReferenceList");
	Symbol	*fromSym = package->FindSymbol("BehaviourReferenceList");

		fromSym->Copy(toSym);

		package->CopyPackage(&_attrList, &_actionList,
				     &_notificationList,
				     &_attrGroupList);
	}

} /* UpdateFromPackage */

/***********************************************************/
/* Update list of parents */

void GDMOClass::UpdateParentList()
{
GDMOClassRef	*moRef;
ListIterator	moIterator(&_parentRefList);

	while ((moRef=(GDMOClassRef *)moIterator.getNext()) !=
		(GDMOClassRef *) 0)
	{
	GDMOClass	*mo = moRef->GetMO();

		_parentList.append(mo);
	}

} /* UpdateParentList */

/***********************************************************/
/* Find	attribute */

GDMOAttribute *GDMOClass::FindAttr(char *attrName)
{
GDMOAttribute	*attr;

	if ((attr=_attrList.FindAttr(attrName)) != (GDMOAttribute *) 0)
		return attr;

	// Search parents for the attribute

	ListIterator	moIterator(&_parentRefList);
	GDMOClassRef	*moRef;

	while ((moRef=(GDMOClassRef *)moIterator.getNext()) !=
		(GDMOClassRef *) 0)
	{
	GDMOClass	*mo = moRef->GetMO();

		if ((attr=mo->FindAttr(attrName)) != (GDMOAttribute *) 0)
			return attr;
	}

	return (GDMOAttribute *) 0;

} /* FindAttr */

/***********************************************************/
// methods of class GDMOClassRef
/***********************************************************/

void GDMOClassRef::UpdateRef(GDMOClass *mo)
{
	_mo = mo;
}

/***********************************************************/
// methods of class GDMOList

GDMOClass *GDMOList::FindMo(char *moName)
{
GDMOClass	*moClass;
ListIterator	moIterator(this);

	while ((moClass=(GDMOClass *)moIterator.getNext()) !=
			(GDMOClass *) 0)
		if (!strcmp(moClass->GetName(), moName))
			break;

	return moClass;

} /* FindMo */

/***********************************************************/
/* Name Binding methods */
/***********************************************************/

Bool	GDMONameBinding::UpdateRefs(GDMOList *moList)
{
GDMOClass *mo;
GDMOAttribute *attr;
Bool	result = True;
GString	msg = GString("NameBinding: ") + _name;

	if ((mo=moList->FindMo((char *)_superiorName)) != (GDMOClass *)0)
		_superiorMo = mo;
	else
	{
		msg += GString(" missing superior: ") + _superiorName ;
		result = False;
	}

	if ((mo=moList->FindMo((char *)_subordinateName)) != (GDMOClass *) 0)
		_subordinateMo = mo;
	else
	{
		msg += GString(" missing subordinate: ") + _subordinateName;
		result = False;
	}

	if (_subordinateMo)
	{
		if ((attr=_subordinateMo->FindAttr((char *)_attributeName)) !=
				(GDMOAttribute *) 0)
			_attribute = attr;
		else
		{
			msg += GString(" missing attribute: ") +
			       _attributeName;
			result = False;
		}
	}

	if (!result)
		printf("\nError: %s", (char *) msg);
	else
		_subordinateMo->AddNameBinding(this);


	return result;
}

/***********************************************************/

void	GDMONameBinding::AddCreateModifier(int createModifier)
{
/*
	printf("\nCREATE MOD %d is %d will be %d",
		createModifier, _create.GetValue(),
		createModifier | _create.GetValue());
*/
	_create.Update(createModifier | _create.GetValue());

} /* AddCreateModifier */

/***********************************************************/

void	GDMONameBinding::AddDeleteModifier(int deleteModifier)
{

	_delete.Update(deleteModifier | _delete.GetValue());

} /* AddDeleteModifier */

/***********************************************************/
/* End of module */
/***********************************************************/




