/*
 * Copyright (c) 1992 University College London
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Department of Computer Science, University College London,
 * London, England. Neither the name of the University, nor its
 * registered trademark, UCL, may be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*/
/****************************************************************/
/*
        iimclex.l

        Lexical Analyser for IIMC parser statements

        Author: J.Cowan         (j.cowan@.cs.ucl.ac.uk)
*/
/****************************************************************/

/* %C       target language is C */

%{
typedef struct  Reserved
        {
        char    *_word;
        int     _token;
        };
static	char		*LexBuf;
static	int		LexIndex;
static  Reserved        *LookupReserved(char *);
static	int		LexInput(char *);
#ifdef	MAIN
static	void		InitIIMCLex(char *, int, IIMCBeginState);
#endif
static	int		BeginState;
#define DEBUGLEX(x)
/*
#define DEBUGLEX(x)     printf(x)
*/
#undef  YY_INPUT
#define YY_INPUT(buf, result, max_size) (result = LexInput(buf))
#undef  YY_DECL
#define YY_DECL int IIMC::Lex()

#define	BEGINPARSE	-1
%}

%x	TEXTSTATE VALUESTATE
%%

	/* Text that is eaten up */

"!!"                    { DEBUGLEX("\nBEGIN TEXTSTATE"); BEGIN TEXTSTATE;}
<TEXTSTATE>"!!"		{ DEBUGLEX("\nENDTEXTSTATE"); BEGIN 0; return TEXT;}
<TEXTSTATE>\n		{ DEBUGLEX(yytext); LineNumber++;}
<TEXTSTATE>.		{ DEBUGLEX(yytext);}

	/* Values that are just returned as strings */

<VALUESTATE>[^;]*	{
                        StringSymbol *sym;

				sym = (NEW StringSymbol(yytext, ""));	
				_symTable.Add(sym, False);
                        	yylval.text = sym->GetValue();
				DEBUGLEX("\nVALUE"); 
				DEBUGLEX(yytext); 
				BEGIN 0; return VALUE;
			}

[ \t]                   { ; /* remove white space */ }

        /* Numbers */

[0-9]*                  { 
                        StringSymbol *sym;

			sym = (NEW StringSymbol(yytext, ""));	
			_symTable.Add(sym, False);
                        yylval.text = sym->GetValue();
			
                        DEBUGLEX("\nNUMBER");
                        return NUMBER;
                        }

        /* Variables */

[a-zA-Z][-a-zA-Z0-9]*    {
                        Reserved        *reserved;

                        DEBUGLEX("\nYYText ");
                        DEBUGLEX(yytext);
                        if ((reserved=LookupReserved(yytext)) !=
                                        (Reserved *) 0)
			{
			int	token = reserved->_token;
				switch (token)
				{
				case BEGINPARSE:
					DEBUGLEX("BEGINSTATE");
					return BeginState;
				default:
					return token;
				}
			}
                        StringSymbol *sym;

			sym = (NEW StringSymbol(yytext, ""));	
			_symTable.Add(sym, False);
                        yylval.text = sym->GetValue();
                        DEBUGLEX("\nSTRING");
                        return STRING;
                        }

        /* Other lexemes */

";"                     { DEBUGLEX("\nSEMI"); return SEMI;}
"."                     { DEBUGLEX("\nDOT"); return DOT;}
","                     { DEBUGLEX("\nCOMMA"); return COMMA;}
"{"                     { DEBUGLEX("\nLBRACE"); return LBRACE;}
"}"                     { DEBUGLEX("\nRBRACE"); return RBRACE;}
[\n]                    { LineNumber++;}
<<EOF>>			{
				if (_restart)
				{
					_restart = False;
					DEBUGLEX("\nRESTART"); 
					YY_NEW_FILE; 
					BEGIN(0);
				}
				else
				{
					DEBUGLEX("\nTERMINATE"); 
					yyterminate();
				}
			}
%%

/****************************************************************/
/* Reserved words */

static  Reserved        reservedWords[] =
{
        {"BEGINPARSE",		BEGINPARSE},
        {"REFERENCE",		REFERENCE},
        {"DESCRIPTION",		DESCRIPTION},
        {"ENDPARSE",		ENDPARSE},
        {"INDEX",		INDEX},
        {"AUGMENTS",		AUGMENTS},
        {"UNITS",		UNITS},
        {"DEFVAL",		DEFVAL},
        {"DISPLAY-HINT",	DISPLAYHINT},
        {"DELETEATT",		DELETEATT},
        {"DELETEVALUE",		DELETEVALUE},
        {"IMPLIED",		IMPLIED},
        {"NULL",		NULLTOKEN}
};

static  Reserved *LookupReserved(char *text)
{
int     count;
int     lgth = sizeof(reservedWords) / sizeof(Reserved);
Reserved *entry;

        for (count=0; count<lgth; count++)
        {
                entry = &reservedWords[count];
#ifdef  __MSDOS__
                if (!stricmp(text, entry->_word))
#else
                if (!strcasecmp(text, entry->_word))
#endif
                {
		int	token = entry->_token;

			switch (token)
			{
			case DELETEVALUE: case DEFVAL:
				BEGIN VALUESTATE;

			default:
				break;
			}
                        DEBUGLEX("\n");
                        DEBUGLEX(entry->_word);
                        return entry;
                }
        }

        return (Reserved *) 0;

} /* LookupReserved */

/****************************************************************/
/* Input routine */

static	int	LexInput(char *buf)
{
char	ch = LexBuf[LexIndex] & 0x7f;

	//printf("\nCh is %c", ch);
	if (ch != '\0')
	{
		LexIndex++;
		buf[0] = ch;
		return 1;
	}
	// printf("\nEOF FILE");
	buf[0] = EOF;
	return YY_NULL;

} /* LexInput */

/****************************************************************/
/* Initialise Lex */

#ifdef	MAIN
static	void	InitIIMCLex(char *buf, int startLine, IIMCBeginState beginState)
#else
void	IIMC::InitLex(char *buf, int startLine, IIMCBeginState beginState)
#endif
{
	LexIndex = 0;
	LexBuf = buf;
	LineNumber = startLine;

	// printf("\nBuf: %s", LexBuf);
	switch (beginState)
	{
	case MOCBEGINSTATE:
		BeginState = BEGINMOCPARSE;
		break;

	case ATTRBEGINSTATE:
		BeginState = BEGINATTRIBUTEPARSE;
		break;

	case NBBEGINSTATE:
		BeginState = BEGINNAMEBINDINGPARSE;
		break;

	default:
		printf("\nUnknown start state in IIMCLex");
		exit(1);
	}

} /* InitLex */

/****************************************************************/
/* End of file */
/****************************************************************/




