/*
 * Copyright (c) 1992 University College London
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Department of Computer Science, University College London,
 * London, England. Neither the name of the University, nor its
 * registered trademark, UCL, may be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*/
/****************************************************************/
/*
	Symbol.h

	Symbol Table classes

	Author:	J.Cowan		(j.cowan@.cs.ucl.ac.uk)
*/
/****************************************************************/

#ifndef	_SYMBOL_

#define	_SYMBOL_

/****************************************************************/

#ifndef	_GENERIC_LIST_
#include "GenList.h"
#endif

#ifndef	_GENERIC_STRING_
#include "GenStr.h"
#endif

/****************************************************************/

typedef	enum {StringTypeDef, ListTypeDef, ObjectTypeDef,
	      ObjectRefTypeDef,	IntegerTypeDef, 
	      StringListTypeDef, IntegerListTypeDef,
	      ObjectRefListType} SymType;

/****************************************************************/
// Forward references

class		Symbol;
class		StringSymbol;
class		IntegerSymbol;
class		ListSymbol;
class		ObjectSymbol;
class		ObjectRefSymbol;
class		SymbolTable;

/****************************************************************/
/* Symbol table	Entry */
/****************************************************************/

class	Symbol
{
	char	*_name;
	void	FatalCast(char *msg)
	{
		printf("\n%s\n", msg);
		exit(1);
	}
public:
	Symbol(char *name) : _name(name) {}

	virtual	~Symbol() {}

	char		*GetName() { return _name; };

	virtual	char	*GetSymTypeName() = 0;

	virtual	SymType	GetSymType() = 0;

	// functions to	cast base cast to derived cast
	// To save effort not using pure functions

	virtual	StringSymbol	*GetStringSymbol()
	{
		FatalCast("Invalid String Cast");
		return (StringSymbol *)	0;
	}
	virtual	IntegerSymbol	*GetIntegerSymbol()
	{
		FatalCast("Invalid Integer Cast");
		return (IntegerSymbol *) 0;
	}
	virtual	ListSymbol	*GetListSymbol()
	{
		FatalCast("Invalid List	Cast");
		return (ListSymbol *) 0;
	}
	virtual	ObjectSymbol	*GetObjectSymbol()
	{
		FatalCast("Invalid Object Cast");
		return (ObjectSymbol *)	0;
	}
	virtual	ObjectRefSymbol	*GetObjectRefSymbol()
	{
		FatalCast("Invalid Object Reference Cast");
		return (ObjectRefSymbol	*) 0;
	}

	virtual	void	Copy(Symbol *) = 0;

}; /* Symbol */

/****************************************************************/

class	StringSymbol : public Symbol, public GString
{
public:
	StringSymbol(char *initStr, char *name)
	: GString(initStr), Symbol(name) {}

	~StringSymbol()	{}

	char	*GetSymTypeName() { return "String";}

	SymType	GetSymType() {return StringTypeDef;}

	char	*GetValue() { return print(); }

	void	Update(char *newVal) { StrReplace(newVal); }

	StringSymbol	*GetStringSymbol() { return this;}

	void	Copy(Symbol *sym)
	{
	StringSymbol	*ssym =	sym->GetStringSymbol();

		ssym->Update(GetValue());
	}

}; /* StringSymbol */

/****************************************************************/

class	IntegerSymbol :	public Symbol
{
	int	_intValue;
public:
	IntegerSymbol(int intValue, char *name)
	: _intValue(intValue), Symbol(name) {}

	char	*GetSymTypeName() { return "Integer";}

	SymType	GetSymType() { return IntegerTypeDef;}

	int	GetValue() { return _intValue; }

	void	Update(int newVal) { _intValue = newVal; }

	IntegerSymbol	*GetIntegerSymbol() { return this;}

	void	Copy(Symbol *sym)
	{
	IntegerSymbol	*isym =	sym->GetIntegerSymbol();

		isym->Update(GetValue());
	}

}; /* IntegerSymbol */

/****************************************************************/

class	ObjectSymbol : public Symbol
{
public:
	ObjectSymbol(char *name) : Symbol(name) {}

	virtual	~ObjectSymbol()	{}

	char	*GetSymTypeName() { return "Object"; }

	SymType	GetSymType()	{return	ObjectTypeDef;}

	virtual	SymbolTable	*GetSymbolTable() = 0;

	ObjectSymbol	*GetValue() { return this; }

	ObjectSymbol	*GetObjectSymbol() { return this;}

	void	Copy(Symbol *)
	{
		printf("\nSorry: Cannot	copy objects");
		exit(1);
	}

}; /* ObjectSymbol */

/****************************************************************/

class	ObjectRefSymbol	: public Symbol
{
	ObjectSymbol	*_object;
public:
	ObjectRefSymbol(ObjectSymbol *object, char *name) :
			_object(object), Symbol(name) {}

	ObjectRefSymbol(char *name) :
			_object((ObjectSymbol *) 0), Symbol(name) {}

	virtual	~ObjectRefSymbol() {}

	char	*GetSymTypeName() {return "ObjectRef";}

	SymType	GetSymType()	{return	ObjectRefTypeDef;}

	ObjectSymbol	*GetValue() { return _object; }

	void	Update(ObjectSymbol *newVal) { _object = newVal; }

	ObjectRefSymbol	*GetObjectRefSymbol() {	return this;}

	void	Copy(Symbol *sym)
	{
	ObjectRefSymbol	*osym =	sym->GetObjectRefSymbol();

		osym->Update(GetValue());
	}

}; /* ObjectRefSymbol */

/****************************************************************/

class	ListSymbol : public Symbol, public List
{
	GString	_memberTypeName;
	SymType	_memberType;

protected:
	virtual	void	deleteElem (void*);

public:
	ListSymbol(char	*name, char *memberTypeName, SymType memberType)
		: Symbol(name), _memberTypeName(memberTypeName),
		  _memberType(memberType) {}
	ListSymbol(char	*name, char *memberTypeName, SymType memberType,
		   Bool	del)
		: List(del), Symbol(name),
		  _memberTypeName(memberTypeName), _memberType(memberType) {}

	virtual	~ListSymbol() { clear(); }

	char	*GetSymTypeName() { return "List";}

	SymType	GetSymType() {return ListTypeDef;}

	char	*GetMemberTypeName() { return (char *) _memberTypeName;	}

	SymType	GetMemberType()	{ return _memberType; }

	ListSymbol	*GetListSymbol() { return this;}

	virtual	void	Copy(Symbol *)
	{
		printf("\nSorry: no list copy for %s",	GetName());
		exit(1);
	}

}; /* ListSymbol */

/****************************************************************/

class	StringSymbolList: public ListSymbol
{
protected:
	void	deleteElem (void*);
public:
	StringSymbolList(Bool del, char *name)
	: ListSymbol(name, "String", StringTypeDef, del) {}
	StringSymbolList(char *name)
	: ListSymbol(name, "String", StringTypeDef) {}

	~StringSymbolList() { clear();}

	void	Copy(Symbol *);
};

/****************************************************************/

class	IntegerSymbolList: public ListSymbol
{
protected:
	void	deleteElem (void*);
public:
	IntegerSymbolList(Bool del, char *name)
	: ListSymbol(name, "Integer", IntegerTypeDef, del) {}
	IntegerSymbolList(char *name)
	: ListSymbol(name, "Integer", IntegerTypeDef) {}

	~IntegerSymbolList() { clear();}

	void	Copy(Symbol *);
};

/****************************************************************/

class	ObjectRefSymbolList: public ListSymbol
{
protected:
	void	deleteElem (void*);
public:
	ObjectRefSymbolList(Bool del, char *name)
	: ListSymbol(name, "ObjectRef", ObjectRefTypeDef, del) {}
	ObjectRefSymbolList(char *name)
	: ListSymbol(name, "ObjectRef", ObjectRefTypeDef) {}

	~ObjectRefSymbolList() { clear();}

	void	Copy(Symbol *);
};

/****************************************************************/
/* Symbol Table	classes */
/****************************************************************/

class	SymbolList : public List
{
protected:
	void	deleteElem (void*);
public:
	SymbolList(Bool	delList) {del=delList;}

	~SymbolList() {	clear();}

}; /* SymbolList */

/****************************************************************/

class	SymbolTable
{
	SymbolList	_symList;
	void		DumpSymbolContents(Symbol *, int);
public:
	SymbolTable(Bool del) :	_symList(del) {}

	Bool	Add(Symbol	*symbol, Bool uniqueTest)
	{
	Bool	res = False;
	char	*name =	symbol->GetName();

		if (uniqueTest && Lookup(name) != (Symbol *) 0)

			printf("\nVariable %s already declared", name);
		else
		{
			_symList.append(symbol);
			res = True;
		}

		return res;
	}

	void	Clear()	{ _symList.clear(); }

	Symbol	*Lookup(char *name);

	void	DumpContents(int);

	void	DumpMembers(int);

}; /* SymbolTable */

/****************************************************************/
/* Class Dictionary classes */
/****************************************************************/
/* Class Member Entry */

class ClassMember
{
	GString		_name;
	Bool		_external;
	SymType		_type;
	GString		_listMemberName;
public:
	ClassMember(char *name) : _name(name)
	{}

	void	SetClassMemberExternal(Bool external) { _external = external; }

	void	SetClassMemberType(SymType type) { _type = type; }

	void	SetClassListMemberType(char *name) {_listMemberName = name; }

	char	*GetName()	{ return (char *) _name; }

	Bool	GetExternal()	{ return _external; }

	SymType	GetType()	{ return _type; }

	char	*GetListMember() { return (char *) _listMemberName; }
};

/****************************************************************/
/* Class Member List */

class	ClassMemberList : public List
{
protected:
	void	deleteElem (void*);
public:
	ClassMemberList() {}

	~ClassMemberList() { clear();}

}; /* ClassMemberList	*/

/****************************************************************/
/* Class Entry */

class ClassEntry
{
	GString		_name;
	ClassMemberList	_memberList;
public:
	ClassEntry(char	*name) : _name(name)
	{}

	char	*GetName() { return (char *) _name; }

	ClassMemberList	*GetMemberList() { return &_memberList; }

	ClassMember	*AddClassMember(char *name)
	{
	ClassMember	*classMember = (NEW ClassMember(name)); 

		_memberList.append(classMember);
		return (classMember);
	}

	ClassMember	*LookupClassMember(char *name);

	void		PrintClassEntry();

}; /* ClassEntry */

/****************************************************************/
/* Class List */

class	ClassList : public List
{
protected:
	void	deleteElem (void*);
public:
	ClassList() {}

	~ClassList() { clear();}

}; /* ClassList	*/

/****************************************************************/

class ClassDictionary
{
	static	ClassList	_classList;
	static	ClassEntry	*_classEntry;
	static	ClassMember	*_classMember;
private:
	static	void	ValidClassMember();
public:
	static	ClassEntry	*LookupClassEntry(char *name);

	static	void	DeleteClassDictionary()
	{
		_classList.clear();
	}
	static	void	PrintClassDictionary();

	static	void	AddClass(char *name);

	static	void	AddClassMember(char *name);

	static	void	SetClassMemberExternal(Bool external);

	static	void	SetClassMemberType(SymType type);

	static	void	SetClassListMemberType(char *name);
	
}; /* ClassDictionary */

/****************************************************************/
/* Meta	class */
/****************************************************************/

typedef	enum	{Assign=0, Or=1, And=2}	UpdateMode;

class MetaClass	: public ObjectSymbol
{
	ClassEntry		*_class;
	SymbolTable		_symTable;
	ObjectRefSymbol		_thisObject;
	static	SymbolTable	_globalObjectList;
private:
	void			CreateClassMembers();
	void			DeleteClassMembers();

public:
	MetaClass(char *name) :	ObjectSymbol(name),
				_symTable(False),
				_thisObject((ObjectSymbol *) this, "this")
	{
		/* Find class in class Dictionary */

		if (!(_class=ClassDictionary::LookupClassEntry(name)))
		{
			printf("\nUnknown Class %s", name);
			exit(1);
		}

		CreateClassMembers();

		/* Add this to symbol table */

		AddSymbol(&_thisObject);
	}

	~MetaClass()	{ DeleteClassMembers();}

	void	CheckClassMembers();

	void	AddSymbol(Symbol *symbol);

	static	void	AddGlobalSymbol(Symbol *symbol)
	{
		_globalObjectList.Add(symbol, True);
	}

	static	Symbol	*LookupGlobal(char *name)
	{
		return(_globalObjectList.Lookup(name));
	}

	void	UpdateSymbol(char *name, char *value);

	void	UpdateSymbol(char *name, int value, UpdateMode mode);

	void	UpdateSymbolList(char *name, Symbol *element);

	void	Copy(Symbol *toSym, Symbol *fromSym)
	{
		fromSym->Copy(toSym);
	}

	Symbol	*SymLookup(char	*name)
	{
		return(_symTable.Lookup(name));
	}

	Symbol	*FindSymbol(char *name)
	{
	Symbol	*sym = SymLookup(name);

		if (sym	== (Symbol	*) 0)
		{
			printf("\nCannot find symbol %s", name);
			exit(1);
		}
		return sym;
	}

	SymbolTable	*GetSymbolTable() {return &_symTable;}

	void	Copy(Symbol *)
	{
		printf("\nCannot copy a	MetaClass");
		exit(1);
	}

	void	DumpSymTableClasses()
	{
		ClassDictionary::PrintClassDictionary();
	}

	void	DumpSymTableContents()
	{
		_symTable.DumpContents(0);
	}

}; /* MetaClass	*/

/****************************************************************/
/* ListSymbol Inline functions */

inline void  ListSymbol::deleteElem (void* val)
{ if (del) delete (Symbol *) val;}

/****************************************************************/
/* StringSymbolList Inline functions */

inline void  StringSymbolList::deleteElem (void* val)
{ if (del) delete (StringSymbol *) val;}

/****************************************************************/
/* IntegerSymbolList Inline functions */

inline void  IntegerSymbolList::deleteElem (void* val)
{ if (del) delete (IntegerSymbol *) val;}

/****************************************************************/
/* IntegerSymbolList Inline functions */

inline void  ObjectRefSymbolList::deleteElem (void* val)
{ if (del) delete (ObjectRefSymbol *) val;}

/****************************************************************/
/* SymbolList Inline functions */

inline void  SymbolList::deleteElem (void* val)
{ if (del) delete (Symbol *) val;}

/****************************************************************/
/* ClassMemberList Inline functions */

inline void  ClassMemberList::deleteElem (void* val)
{  if (del) delete (ClassMember *) val; }

/****************************************************************/
/* ClassList Inline functions */

inline void  ClassList::deleteElem (void* val)
{  if (del) delete (ClassEntry *) val; }

/****************************************************************/

#endif

/****************************************************************/
/* End of module */
/****************************************************************/



