//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// Coordinator.cc - coordinator object realisation

// By George Pavlou, February 1991


#include <stdlib.h>
#include <signal.h>
#include <unistd.h>

#ifdef	__GNUG__
#ifndef	__HPUX__
#define SIGHANDLER	SignalHandler
#else
#define SIGHANDLER	void
#endif
#elif	__cplusplus
#ifndef AIXV3
#define SIGHANDLER	SIG_PF
#else	/* AIXV3 */
#define SIGHANDLER	void
#endif
#endif

#include <memory.h>
#include <errno.h>
#include <sys/types.h>
#ifndef	ICR1
#include <isode/acsap.h>
#else
#include <isode/ul/acsap.h>
#endif
#include "Coordinator.h"


int KsWakeUp::cmpmode;
Coordinator* Coordinator::_instance;

static const int sigMask = 0x7fffffff;		// all signals


Coordinator::Coordinator ()
{
    FD_ZERO(&_mask);
    _nfds = 0;
    for (register int i = 0; i < MAX_ENDPOINT; i++)
	_ks[i] = NULL;
    _nReportObjs = 0;
    _ksShutdown.setDeleteMode(False);
    _instance = this;

    // set temporary alarm handler
    _alarm = True;
#ifdef POSIX_SIGNALS
    struct sigaction sig_act;
#ifndef SOLARIS24
    sig_act.sa_handler = serviceAlarmTmp;
#else
    sig_act.sa_handler = (void(*)(int)) serviceAlarmTmp;
#endif
    sigemptyset(&(sig_act.sa_mask));
    sig_act.sa_flags = 0;
    sigaction(SIGALRM,&sig_act,NULL);
#else
#ifdef	__GNUG__
    signal(SIGALRM, serviceAlarmTmp);
#else
    signal(SIGALRM, (SIGHANDLER) serviceAlarmTmp);
#endif
#endif /* POSIX_SIGNALS */
}


int  Coordinator::scheduleWakeUps (KS* ks, char* token, long period,
					Bool ifReportObjs, Bool onceOnly)
{
    if (period <= 0) {
	fprintf(stderr, "Coordinator: invalid wake-up period %d", period);
	return NOTOK;
    }

    // insert this wake up in the list
    KsWakeUp *wakeUp = new KsWakeUp(ks, token, period, time((long*)0) + period,
					ifReportObjs, onceOnly);
    _wakeUps.insert(wakeUp);

    // arrange now for the next alarm (if needed)
    KsWakeUp *nextWakeUp = (KsWakeUp*) _wakeUps.first();

    if (nextWakeUp != wakeUp)	// the new one went furher down the list
	return OK;

    if (_alarm)
	alarm((unsigned) period);
    else
	nextWakeUpIntvlChange(period);

    return OK;
}


int  Coordinator::cancelWakeUps (KS* ks, char* token)
{
    // find and remove it
    KsWakeUp wakeUp(ks, token), *cancelWakeUp, *nextWakeUp;

    if ((cancelWakeUp = _wakeUps.remove(&wakeUp)) == NULL) {
	fprintf(stderr, "Coordinator: cannot find alarm to cancel\n");
	return NOTOK;
    }

    // arrange now for the next alarm (if needed)
    if (nextWakeUp = (KsWakeUp*) _wakeUps.first()) {
	if (cancelWakeUp->_wakeUpTime < nextWakeUp->_wakeUpTime) {
	    long nextAlarmIntvl = nextWakeUp->_wakeUpTime - time((long*)0);
	    if (_alarm) {
		alarm((unsigned) nextAlarmIntvl);
	    }
	    else
		nextWakeUpIntvlChange(nextAlarmIntvl);
	}
    }
    else	// the wakeUp removed was the only one in the list
	if (_alarm)
	    alarm(0);			// nothing scheduled, so cancel
	else
	    nextWakeUpIntvlChange(0);

    delete cancelWakeUp;
    return OK;
}


int  Coordinator::serviceWakeUp ()
{
    // check first if there is a wake up to service
    KsWakeUp *wakeUp, *nextWakeUp;
    if (! (wakeUp = (KsWakeUp*) _wakeUps.first())) {
	fprintf(stderr, "Coordinator.serviceWakeUp: alarm but empty \
wake up list!\n");
	return NOTOK;
    }

    // wake up the relevant knowledge source
    if (wakeUp->_ifReportObjs == False || _nReportObjs > 0)
	(wakeUp->_ks) -> wakeUp(wakeUp->_token);


    // arrange for the next wake up for this KS
    // (if a periodic one was requested)

    long now = time((long*)0);

    if (wakeUp == (KsWakeUp*) _wakeUps.first()) {	// not cancelled by KS
	wakeUp = (KsWakeUp*) _wakeUps.get();

        if (!wakeUp->_onceOnly) {
	    wakeUp->_wakeUpTime += wakeUp->_period;
	    if (wakeUp->_wakeUpTime < now)		// for when debugging
	        wakeUp->_wakeUpTime = now + wakeUp->_period;
	    _wakeUps.insert(wakeUp);
        }
        else
	    delete wakeUp;
    }

    // schedule the next wake up (if any)
    if ((nextWakeUp = (KsWakeUp*) _wakeUps.first()) == NULL) {
	// no more wake-ups for the time being...
	if (_alarm)
	    alarm(0);
	else
	    nextWakeUpIntvlChange(0);
	return OK;
    }

    long nextAlarmIntvl = nextWakeUp->_wakeUpTime - now;
    if (nextAlarmIntvl > 0) {
	if (_alarm)
	    alarm((unsigned) nextAlarmIntvl);
	else
	    nextWakeUpIntvlChange(nextAlarmIntvl);
    }
    else
	serviceWakeUp();	// wake-ups scheduled at roughly the same time

    return OK;
}


int  Coordinator::registerCommEndpoint (KS* ks, int fd)
{
    if (fd >= MAX_ENDPOINT) {
	fprintf(stderr, "Coordinator.registerCommEndpoint: fd too big - \
max %d\n", MAX_ENDPOINT);
	return NOTOK;
    }

    if (fd >= _nfds)
	_nfds = fd + 1;

    FD_SET(fd, &_mask);
    _ks[fd] = ks;

    commEndpointMaskChange (fd, MASK_ADD);

    return OK;
}


int  Coordinator::deregisterCommEndpoint (KS* ks, int fd)
{
    if (_ks[fd] != ks)
	return NOTOK;

    if (_nfds == fd + 1)
	_nfds = compactFDs(fd);

    FD_CLR(fd, &_mask);
    _ks[fd] = NULL;

    commEndpointMaskChange (fd, MASK_REMOVE);

    return OK;
}


int  Coordinator::compactFDs (int lastfd)
{
    for (int fd = lastfd - 1; fd >= 0; fd--)
	if (FD_ISSET (fd, &_mask))
	    break;
    return fd + 1;
}


static fd_set ifds;

int  Coordinator::readCommEndpoints_aux (fd_set* ifdsp)
{
    for (register int fd = 0; fd < _nfds; fd++)
        if (FD_ISSET(fd, ifdsp) && _ks[fd] != NULL) {
#ifdef POSIX_SIGNALS
            sigset_t sig_set, old_sig_set;
            sigfillset(&(sig_set)); /* block all signals */
            sigprocmask (SIG_BLOCK,&sig_set,&old_sig_set);
#else
            int oldmask = sigblock(sigMask);	// block signals
#endif /* POSIX_SIGNALS */
            _ks[fd]->readCommEndpoint(fd);      // KS input
#ifdef POSIX_SIGNALS
            sigprocmask (SIG_SETMASK, &old_sig_set, NULL);
#else
            sigsetmask(oldmask);	// restore signals
#endif /* POSIX_SIGNALS */
         }
 
    return OK;
}


int  Coordinator::readCommEndpoints ()
{
    ifds = _mask;
    switch (xselect(_nfds, &ifds, NULLFD, NULLFD, NOTOK)) {
    case NOTOK:
	return NOTOK;

    case OK:
	if (errno == EINTR)
	    errno = 0;
	return OK;

    default:	// data on fd
	break;
    }

    return readCommEndpoints_aux(&ifds);
}


void  Coordinator::shutdown (Bool doExit)
{
    // tell the knowledge sources to shutdown communication end-points
    for (register int fd = 0; fd < _nfds; fd++)
	if (FD_ISSET(fd, &_mask))
	    _ks[fd] -> shutdown(fd);

    // notify the knowledge sources that have explicitly requested so
    register KS* ks;
    while (ks = (KS*) _ksShutdown.get())
	ks -> shutdown(-1);

    // exit if requested
    if (doExit)
	exit(OK);
}


KsWakeUpList*  Coordinator::passWakeUpControl ()
{
    alarm((unsigned) 0);	// cancel any pending alarm
    return &_wakeUps;		// we are passing control elsewhere...
}


// temporary alarm signal handler in case an alarm happens before
// listening starts (the default handler causes the program to exit...)

void  Coordinator::serviceAlarmTmp (int)
{
#ifdef POSIX_SIGNALS
    struct sigaction sig_act;
#endif /* POSIX_SIGNALS */

    // do nothing but restore handler
    if (_instance->_alarm) {
#ifdef POSIX_SIGNALS
#ifndef SOLARIS24
    sig_act.sa_handler = serviceAlarmTmp;
#else
    sig_act.sa_handler = (void(*)(int)) serviceAlarmTmp;
#endif
        sigemptyset(&(sig_act.sa_mask));
        sig_act.sa_flags = 0;
        sigaction(SIGALRM,&sig_act,NULL);
#else
#ifdef	__GNUG__
        signal(SIGALRM, serviceAlarmTmp);
#else
        signal(SIGALRM, (SIGHANDLER) serviceAlarmTmp);
#endif
#endif /* POSIX_SIGNALS */
    }
}


void  Coordinator::serviceAlarm (int)
{
#ifdef POSIX_SIGNALS
    sigset_t sig_set, old_sig_set;
    struct sigaction sig_act;
    sigfillset(&(sig_set)); /* block all signals */
    sigprocmask (SIG_BLOCK,&sig_set,&old_sig_set);
#else
    int oldmask = sigblock(sigMask);	// block signals
#endif /* POSIX_SIGNALS */

    _instance->serviceWakeUp();			// service the KS wake-up

    // reset the signal handler
#ifdef POSIX_SIGNALS
#ifndef SOLARIS24
    sig_act.sa_handler = serviceAlarm;
#else
    sig_act.sa_handler = (void(*)(int)) serviceAlarm;
#endif
    sigemptyset(&(sig_act.sa_mask));
    sig_act.sa_flags = 0;
    sigaction(SIGALRM,&sig_act,NULL);
#else
#ifdef	__GNUG__
    signal(SIGALRM, serviceAlarm);
#else
    signal(SIGALRM, (SIGHANDLER) serviceAlarm);
#endif
#endif /* POSIX_SIGNALS */

#ifdef POSIX_SIGNALS
    sigprocmask (SIG_SETMASK, &old_sig_set, NULL);
#else
    sigsetmask(oldmask);	// restore signals
#endif /* POSIX_SIGNALS */
}


void  Coordinator::terminate (int)
{
#ifdef POSIX_SIGNALS
    sigset_t sig_set, old_sig_set;
    sigfillset(&(sig_set)); /* block all signals */
    sigprocmask (SIG_BLOCK,&sig_set,&old_sig_set);
#else
    int oldmask = sigblock(sigMask);	// block signals
#endif /* POSIX_SIGNALS */

    _instance->shutdown(True);

#ifdef POSIX_SIGNALS
    sigprocmask (SIG_SETMASK, &old_sig_set, NULL);
#else
    sigsetmask(oldmask);	// restore signals
#endif /* POSIX_SIGNALS */
    // restoring is redundant as shutdown(True) should exit
}


void  Coordinator::setAlarmSignalHandler ()
{
#ifdef POSIX_SIGNALS
    struct sigaction sig_act;
#endif /* POSIX_SIGNALS */

    _instance -> _alarm = True;			// OSIMIS handles alarm signals
#ifdef POSIX_SIGNALS
#ifndef SOLARIS24
    sig_act.sa_handler = serviceAlarm;
#else
    sig_act.sa_handler = (void(*)(int)) serviceAlarm;
#endif
    sigemptyset(&(sig_act.sa_mask));
    sig_act.sa_flags = 0;
    sigaction(SIGALRM,&sig_act,NULL);
#else
#ifdef	__GNUG__
    signal(SIGALRM, serviceAlarm);
#else
    signal(SIGALRM, (SIGHANDLER) serviceAlarm);
#endif
#endif /* POSIX_SIGNALS */

}


void  Coordinator::setTerminateSignalHandler ()
{
#ifdef POSIX_SIGNALS
    struct sigaction sig_act;
#ifndef SOLARIS24
    sig_act.sa_handler = terminate;
#else
    sig_act.sa_handler = (void(*)(int)) terminate;
#endif
    sigemptyset(&(sig_act.sa_mask));
    sig_act.sa_flags = 0;
    sigaction(SIGINT,&sig_act,NULL);
    sigaction(SIGQUIT,&sig_act,NULL);
    sigaction(SIGTERM,&sig_act,NULL);
#else
#ifdef	__GNUG__
    signal(SIGINT,  terminate);
    signal(SIGQUIT, terminate);
    signal(SIGTERM, terminate);
#else
    signal(SIGINT,  (SIGHANDLER) terminate);
    signal(SIGQUIT, (SIGHANDLER) terminate);
    signal(SIGTERM, (SIGHANDLER) terminate);
#endif
#endif /* POSIX_SIGNALS */
}


void  Coordinator::listen ()
{
    // set signal handlers

    setAlarmSignalHandler();
    setTerminateSignalHandler();

    // listen

    for (;;)
	readCommEndpoints();
}
