//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// Coordinator.h - coordinator object declaration

// By George Pavlou, February 1991


/*
 * The basic OSIMIS coordinator that coordinates activity in management
 * applications by acting as a central point for all external communication
 * endpoints and scheduled alarms. There should always be one only instance
 * of this or any derived classes in an application realised as one UNIX
 * process. Its services may be accessed through the Knowledge Source class.
 * It uses the UNIX select(2) facility, enhanced in ISODE as xselect to be
 * portable across all UNIX platforms. A first-come-first-served policy
 * is exercised with respect to both external communications and scheduled
 * real-time alarms (wake-ups).
 *
 * This class is designed in such a way to allow the integration with other
 * coordinating mechanisms through suitable derived classes (see
 * the X and Isode coordinator classes)
 */


#ifndef _COORDINATOR_
#define _COORDINATOR_


#ifdef	AIXV3
#include <stdio.h>
#endif

#ifdef  __GNUG__
#ifndef _string_h
#include <string.h>
#endif
#elif   __cplusplus
#ifndef __string_h
#include <string.h>
#endif
#endif

#ifndef _MANIFEST_
#ifndef	ICR1
#include <isode/manifest.h>
#else
#include <isode/compat/manifest.h>
#endif
#endif

#ifndef _ucbext_h
#include "ucbext.h"
#endif

#ifndef _GEN_LIST_
#include "GenList.h"
#endif

#ifndef _GENERIC_KS_
#include "GenericKS.h"
#endif


struct KsWakeUp
{
    KS*   _ks;
    char* _token;
    long  _period;
    long  _wakeUpTime;
    Bool  _ifReportObjs;
    Bool  _onceOnly;

#define TIMECMP		1
#define TOKENCMP	2

    static int timecmp (KsWakeUp*, KsWakeUp*);
    static int tokencmp (KsWakeUp*, KsWakeUp*);
    static int cmpmode;

    KsWakeUp (KS*, char*, long, long, Bool, Bool);
    KsWakeUp (KS*, char*);
    KsWakeUp ();
    ~KsWakeUp ();
};


class KsWakeUpList : public List
{
protected:
    int		compare (void*, void*);

public:
    void	insert (KsWakeUp*);
    KsWakeUp*	remove (KsWakeUp*);
};


class Coordinator
{
#   define MAX_ENDPOINT	128
#   define MASK_ADD	1
#   define MASK_REMOVE	2

    KsWakeUpList _wakeUps;
    List	 _ksShutdown;
    int		 _nReportObjs;
    Bool	 _alarm;
    static Coordinator* _instance;

    int		 compactFDs (int);

protected:
    fd_set	 _mask;
    int		 _nfds;
    KS* 	 _ks[MAX_ENDPOINT];

    // enables to change from xselect to other package-specific mechanism

    int		 readCommEndpoints_aux (fd_set*);
    virtual int	 readCommEndpoints ();

    // interface to other coordinating mechanisms

    void	setAlarmMode (Bool);	// take or not control of alarms
    int		serviceWakeUp ();	// alarm handler (in serviceAlarm)
    void	shutdown (Bool);	// terminate handler (in terminate)

    virtual int	commEndpointMaskChange (int, int);
    virtual int nextWakeUpIntvlChange (long);

public:
    // interface to objects that may relate to polling for notifications
    // (e.g. event discriminators, logs etc.)

    int         registerReportObj ();
    int         deregisterReportObj ();
    int		getNReportObjs ();


    // interface to knowledge sources

    int		scheduleWakeUps (KS*, char*, long, Bool, Bool);
    int		cancelWakeUps (KS*, char*);
    int		registerCommEndpoint (KS*, int);
    int		deregisterCommEndpoint (KS*, int);
    void	notifyShutdown (KS*);


    // interface to other coordinating mechanisms

    KsWakeUpList* passWakeUpControl ();
    KS**	getKSArray ();

    static void	serviceAlarm (int);
    static void	serviceAlarmTmp (int);
    static void	terminate (int);
    static void	setAlarmSignalHandler ();
    static void	setTerminateSignalHandler ();


    // central loop

    void	listen ();


    // instance access

    static Coordinator* getInstance ();


    // constructor

    Coordinator ();
};




// KsWakeUp inline methods

inline int KsWakeUp::timecmp (KsWakeUp* w1, KsWakeUp* w2)
{ return w1->_wakeUpTime - w2->_wakeUpTime; }

inline int KsWakeUp::tokencmp (KsWakeUp* w1, KsWakeUp* w2)
{ return w1->_ks == w2->_ks && strcmp(w1->_token, w2->_token) == 0 ?
  OK : NOTOK; }

inline KsWakeUp::KsWakeUp ()
{ _ks = NULLKS; _period = _wakeUpTime = 0; _token = NULLCP;
  _ifReportObjs = _onceOnly = False; }

inline KsWakeUp::KsWakeUp (KS* ks, char* token, long period,
		long wakeUpTime, Bool ifReportObjs, Bool onceOnly)
{ _ks = ks; _token = strdup(token); _period = period; _wakeUpTime = wakeUpTime;
  _ifReportObjs = ifReportObjs; _onceOnly = onceOnly; }

inline KsWakeUp::KsWakeUp (KS* ks, char* token)
{ _ks = ks; _period = _wakeUpTime = 0; _token = strdup(token);
  _ifReportObjs = _onceOnly = False; }

inline KsWakeUp::~KsWakeUp ()
{ if (_token) free(_token); }


// KsWakeUpList inline methods

inline int KsWakeUpList::compare (void* w1, void* w2)
{ return KsWakeUp::cmpmode == TIMECMP ?
  KsWakeUp::timecmp((KsWakeUp*) w1, (KsWakeUp*) w2) :
  (KsWakeUp::cmpmode == TOKENCMP ?
  KsWakeUp::tokencmp((KsWakeUp*) w1, (KsWakeUp*) w2) : NOTOK); }

inline void  KsWakeUpList::insert (KsWakeUp* wakeup)
{ KsWakeUp::cmpmode = TIMECMP; List::insert(wakeup); }

inline KsWakeUp*  KsWakeUpList::remove (KsWakeUp* wakeup)
{ KsWakeUp::cmpmode = TOKENCMP; return (KsWakeUp*) List::remove(wakeup); }


// Coordinator inline methods

inline int  Coordinator::nextWakeUpIntvlChange (long)
{ return OK; }

inline int  Coordinator::commEndpointMaskChange (int, int)
{ return OK; }

inline int  Coordinator::registerReportObj ()
{ return ++_nReportObjs; }

inline int  Coordinator::deregisterReportObj ()
{ if (_nReportObjs == 0) fprintf(stderr,
  "Coordinator::deregisterReportObj: no report objects registered!\n");
  return _nReportObjs == 0 ? NOTOK : --_nReportObjs; }

inline int  Coordinator::getNReportObjs ()
{ return _nReportObjs; }

inline void  Coordinator::notifyShutdown (KS* ks)
{ _ksShutdown.append(ks); }

inline void  Coordinator::setAlarmMode (Bool alarmMode)
{ _alarm = alarmMode; }

inline KS**  Coordinator::getKSArray ()
{ return _ks; }

inline Coordinator*  Coordinator::getInstance ()
{ return _instance; }


#endif
