//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
/******************************************************************/
/*
	File:	cmisx500.cc

	Notes:
		contains methods for directory programming class

	Author:
		James Cowan (J.Cowan@uk.ac.ucl.cs)
*/
/******************************************************************/

#include "cmis.h"
#include "cmisx500.h"

extern int  empty_asn1_type;	// from the msap library (attrv.c)

/******************************************************************/
/* Public methods */
/******************************************************************/

Boolean	X500_attribute::encode(char *oid_name, char *value,
			       OID *oid, PE *pe)
{
oid_table_attr	*oid_entry;

	// Free memory if calling object more than once

	free_memory();

	if (oid_entry=name2attr(oid_name))
	{
		*oid = oid_entry->oa_ot.ot_oid;
		if (_syntax=find_syntax(oid_entry))
		{
			if ((_address=parse(_syntax, value)) || empty_asn1_type)
			{
			PE	ret_pe;

				if (ret_pe=encode_pe(_syntax, _address))
				{
					_pe = ret_pe;
					*pe = ret_pe;

					return(True);
				}
			}
		}
	}
	return(False);

} /* encode */

/******************************************************************/

Boolean	X500_attribute::decode(OID oid, PE value, 
			       GString &str_oid, GString &str_value)
{
oid_table_attr	*oid_entry;

	if (oid_entry=oid2attr(oid))
	{
		// Free memory if calling object more than once

		free_address();

		_oid_name = oid_entry->oa_ot.ot_name;
		str_oid = _oid_name;
		if (_syntax=find_syntax(oid_entry))
		{
			if (_address=decode_pe(_syntax, value))
			{
				if (print(_syntax, _address, _value))
				{
					str_value = _value;
#ifdef	DEBUG_CLIENT
		CLIENT_TRACE("\nDecode %s Syntax %s Address %d Value %s",
			     (char *) _oid_name, get_syntax(),
			     (char *) _address, (char *) _value);
#endif
					return(True);
				}
			}
#ifdef	DEBUG_CLIENT
			else
				CLIENT_TRACE("\nCannot Decode %s Syntax %s",
			     		     (char *) _oid_name, get_syntax());
#endif
		}
	}
	else
		_oid_name = "Unknown oid";

	_value = "";
	str_value = _value;

	return(False);

}

/******************************************************************/

Boolean	X500_attribute::is_encodeable(char *oid_name)
{
oid_table_attr	*oid_entry;

	if (oid_entry=name2attr(oid_name))
	{
	sntx_table *syn;

		if (syn=find_syntax(oid_entry))
		{
			if (((ENCODE_FN) syn->s_encode != (ENCODE_FN) NULLIFP)
			   && ((PARSE_FN)syn->s_parse != (PARSE_FN)NULLIFP)) 

				return(True);
		}
	}

	return(False);

} /* is_encodeable */

/******************************************************************/
/* Private methods */
/******************************************************************/

sntx_table *X500_attribute::find_syntax(oid_table_attr *oid_entry)
{
sntx_table *ret_syn = get_syntax_table(oid_entry->oa_syntax);

#ifdef	DEBUG_CLIENT
	if (!ret_syn)
		CLIENT_TRACE("\nNo syntax");
#endif

	return(ret_syn);

} /* find_syntax */

/******************************************************************/

char *X500_attribute::find_syntax(char *oid_name)
{
oid_table_attr	*oid_entry;

	if (oid_entry=name2attr(oid_name))
	{
	sntx_table *syn;

		if (syn=find_syntax(oid_entry))

			return(syn->s_sntx);
	}

	return(0);

} /* find_syntax */

/******************************************************************/

void *	X500_attribute::parse(sntx_table *syn, char *value)
{
PARSE_FN	parse_fn = (PARSE_FN) syn->s_parse;

	if (parse_fn != (PARSE_FN) NULLIFP)
	{
	void *		ret_address;

		ret_address = parse_fn(value);
#ifdef	DEBUG_CLIENT
		if (!ret_address && !empty_asn1_type)
			CLIENT_TRACE("\nParse failure");
#endif
		return(ret_address);
	}
#ifdef	DEBUG_CLIENT
	else
		CLIENT_TRACE("\nCannot parse %s", get_syntax());
#endif

	return (0);

} /* parse */

/******************************************************************/

PE	X500_attribute::encode_pe(sntx_table *syn, void * address)
{
ENCODE_FN	encode_fn;
PE		ret_pe;

	encode_fn = (ENCODE_FN) syn->s_encode;
	if (encode_fn) 
	{
		ret_pe = encode_fn(address);
#ifdef	DEBUG_CLIENT
		if (!ret_pe)
			CLIENT_TRACE("\nEncode failed %s", get_syntax());
#endif

		return(ret_pe);
	}
#ifdef	DEBUG_CLIENT
	else
		CLIENT_TRACE("\nCannot encode %s", get_syntax());
#endif

	return (0);

} /* encode_pe */

/******************************************************************/

void *	X500_attribute::decode_pe(sntx_table *syn, PE value)
{
void *		address;
DECODE_FN	decode_fn = (DECODE_FN) syn->s_decode;

	if (decode_fn != (DECODE_FN) NULLIFP)
	{
		address = decode_fn(value);
#ifdef	DEBUG_CLIENT
		if (!address)
			CLIENT_TRACE("\nDecode failed %s", get_syntax());
#endif
		return(address);
	}
#ifdef	DEBUG_CLIENT
	else
		CLIENT_TRACE("\nCannot decode %s", get_syntax());
#endif

	return(0);

} /* decode_pe */

/******************************************************************/

Boolean	X500_attribute::print(sntx_table *syn, void * address,
			      GString &value)
{
PRINT_FN	print_fn = (PRINT_FN) syn->s_print;

	if (print_fn)
	{
	PS	ps;

		if ((ps = ps_alloc((IFP)str_open)) == NULLPS)
			return(False);
		print_fn(ps, address, READOUT);

		if (ps->ps_ptr)
		{
			*ps->ps_ptr = '\0';
			value = ps->ps_base;
		}
		else
		{
#ifdef	DEBUG_CLIENT
			CLIENT_TRACE("\nCannot print %s", get_syntax());
#endif
			value = "";
		}
		ps_free(ps);
		return(True);
	}

#ifdef	DEBUG_CLIENT
	CLIENT_TRACE("\nCannot print %s", get_syntax());
#endif
	return (False);

} /* print */

/******************************************************************/

void	X500_attribute::free_address(sntx_table *syn, void * address) 
{
FREE_FN	free_fn = (FREE_FN) syn->s_free;

	if (free_fn != (FREE_FN)NULLIFP)
	{
		free_fn(address);
		return;
	}

#ifdef	DEBUG_CLIENT
	CLIENT_TRACE("\nDo not know how to free %s", syn->s_sntx);
#endif

} /* free_address */

/******************************************************************/

void	X500_attribute::initialise()
{
	_address = 0; _pe = 0; _syntax = 0;

} /* initialise */

/******************************************************************/

void	X500_attribute::free_pe()
{
	if (_pe)
	{
		pe_free(_pe);
		_pe = 0;
	}

} /* free_pe */

/******************************************************************/

void	X500_attribute::free_address()
{
	if (_syntax && _address)
	{
		free_address(_syntax, _address);
		_address = 0;
	}

} /* free_address */

/******************************************************************/

void	X500_attribute::free_memory()
{
	free_pe(); free_address();

} /* free_memory */

/******************************************************************/
// Not used; routine that accesses the directory tables directly
/*
oid_table_attr *X500_attribute::find_oid_entry(char *oid_name)
{
int	count;
extern	int attrNumEntries;
extern	oid_table_attr	attrOIDTable[];
	
	for (count=0; count<attrNumEntries; count++)
	{
	oid_table_attr	*oid_table_ptr = &attrOIDTable[count];

		if (!strcmp(oid_name, oid_table_ptr->oa_ot.ot_name))
		{
		printf("\nCount %d %s %s %s %d",
			count, attrOIDTable[count].oa_ot.ot_name,
			attrOIDTable[count].oa_ot.ot_stroid,
			sprintoid(attrOIDTable[count].oa_ot.ot_oid),
			sprintoid(attrOIDTable[count].oa_ot.ot_oid),
			attrOIDTable[count].oa_syntax);

			return(oid_table_ptr);
		}
	}

	printf("\nOid not found ");
	return(0);

} 
*/
/******************************************************************/
/* End of Program */
/******************************************************************/


