//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
/******************************************************************/
/*

	File:	cmisx500.h

	Author:
		James Cowan (J.Cowan@uk.ac.ucl.cs)
*/
/******************************************************************/
/* Notes on class x500_attribute

This class encapsulates the directory programminng routines 
described in the ISODE manual.

The public methods are :


Boolean		encode(char *, char *, OID *, PE *);

	encodes a string representation of an oid and its value and
	returns pointers to an OID and a PE. This routine should be
	overloaded to allow the encoding of C structures


Boolean		decode(OID, PE, GString &, GString &);
Boolean		decode(OID, PE);

	decodes an OID and a PE. The function is overloaded so that
	it will either return the string representations of the OID
	and the PE or not; you can retrieve them later.


Boolean	is_encodeable(char *)

	Returns whether an object identifier can be encoded

char	*get_syntax();

	returns the syntax name of what has been encoded or decoded
	eg "Integer".


void *	get_address()

	returns the address of the C structure which has just been
	decoded or encoded


char	*get_oid_name()

	returns the string representation of the OID that has
	just been encoded or decoded


char	*get_value()

	returns the string representation of the PE that has
	just been encoded or decoded

get_syntax, get_address, get_oid_name and get_value all return 0
if there is nothing valid to return - eg if the encode or decode
routines failed.


Memory management

Heap memory is handled in the following way:

The C struct address (read by get_address) is allocated during encoding
and decoding and is automatically freed by the class. When the encoding
is overloaded to allow the encoding of C structures, the class will not
allocate or deallocate the address.

The string representations of OID and PE (read by get_oid_name and
get_value) are allocated and deallocated by the class.

The OID variables used by the class are not allocated by the class -
either the encoder provides it or the decoder returns a pointer to the
OID in the ISODE syntax tables.

The PE variable is allocated and freed when encoding, but NOT when it
is being decoded - the calling application does that.
*/
/******************************************************************/
/* Type declarations etc */

extern "C"
{
	sntx_table 	*get_syntax_table (short);
}

typedef	void *	(* PARSE_FN)(char *);
typedef	PE	(* ENCODE_FN)(void *);
typedef	void	(* PRINT_FN)(PS, void *, int);
typedef	void *	(* DECODE_FN)(PE);
typedef	void	(* FREE_FN)(void *);

/******************************************************************/

class	X500_attribute
{
	void *		_address;	 // these variables need to be freed
	PE		_pe;		// ""

	GString		_oid_name, _value;
	sntx_table	*_syntax;

	oid_table_attr	*find_oid_entry(char *);
	char		*find_syntax(char *);
	sntx_table	*find_syntax(oid_table_attr *);
	void *		parse(sntx_table *, char *);
	Boolean		print(sntx_table *, void * , GString &);
	PE		encode_pe(sntx_table *, void *); 
	void *		decode_pe(sntx_table *, PE);
	void		free_address(sntx_table *, void *); 
	void		initialise();
	void		free_pe();
	void		free_address();
	void		free_memory();

public:
	X500_attribute()
	{
	initialise();
	}

	~X500_attribute()
	{
		free_memory();
	}

	Boolean		encode(char *, char *, OID *, PE *);
	Boolean		decode(OID, PE, GString &, GString &);
	Boolean		decode(OID oid, PE pe)
	{
		return(decode(oid, pe, _oid_name, _value));
	}

	Boolean	is_encodeable(char *);

	char		*get_syntax()
	{
		if (_syntax)
			return(_syntax->s_sntx);
		else
			return(0);
	}

	char		*get_syntax(char *oid_name)
	{
		return(find_syntax(oid_name));
	}

	void *		get_address()
	{
		return(_address);
	}
	char		*get_oid_name()
	{
		return((char *) _oid_name);
	}
	char		*get_value()
	{
		return((char *) _value);
	}
};

/******************************************************************/
/* End of module */
/******************************************************************/


