//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// EvSink.cc - a primitive event sink that receives and prints
//	       event reports from a remote MIB

// By George Pavlou, February 1991


#ifdef  __GNUG__
#ifndef	__HPUX__
#define SIGHANDLER	SignalHandler
#else   /* __HPUX__ */
#define SIGHANDLER	void
#endif
#elif   __cplusplus
#ifndef AIXV3
#define SIGHANDLER	SIG_PF
#else	/* AIXV3 */
#define SIGHANDLER	void
#endif
#endif

#include <stdlib.h>
#include <signal.h>
#include "msap.h"
#include "Error.h"
#include "SmiSntx.h"
#include "Syntaxes.h"
#include <memory.h>
#if defined(SECURITY)
#include "mauth.h"
#include "mintegrity.h"
#endif

#if defined(DEBUG)
#include "msap-pp.h"
FILE *pduDump = (FILE *) 0;
#endif

#define	ALL	0
#define FILTER	1

int msd,
    request_id = 0,
    request_invoke = 0,
    invoke = 0;
MName  efdiscr;
MSAPIndication mis;

static int   do_create (int, CMISFilter*, MSAPIndication*);
static int   do_delete (int, MSAPIndication*);
static int   do_wait (int, MSAPIndication*);
static void  terminate (int);

void
usage()
{
    fprintf(stderr, "usage:\n");
    fprintf(stderr, "evsink { <agent> <host> | -S <agent> } \
\t-f [<filter> | <eventype> ...]\n");
#if defined(SECURITY)
    fprintf(stderr, "\t[-P <peer DN>] [-D <DSA name/addr>] [-I]\n\n");
#endif
}

main (int argc, char** argv)
{
    char* application = NULLCP;
    char* host = NULLCP;
    char buf[1024], **arg, *optFlags = "Sf:";
    CMISFilter* filter = NULLMFILTER;
    int errFlag, argIndex;

#if defined(SECURITY)
    int IFlag = 0;
    char *peer = (char *) 0,
         *dsa = (char *) 0;

    optFlags = "Sf:P:D:I";
#endif

    // parse command line arguments

    if (argc < 3) {
        usage();
        exit (0);
    }


    buf[0] = '\0';
    for(arg = argv, argIndex = 0; argIndex < argc && *arg != (char *) 0;
        ++arg, ++argIndex) {

        /*
        ** Can't use getopt(3) as it thinks the options all come first
        ** and any arg that does not have a '-' in front of it will
        ** cause it to stop parsing args. <agent> and <host> do not
        ** have '-' in front of them. Below is a mimic of getopt(3).
        */
        int opt = (*arg)[0] == '-' ? (*arg)[1] : 0;
        char *optArg;

        if (opt != 0) {
            /* Known option? */
            char *cp = strchr(optFlags, opt);
            if (cp == (char *) 0)
                opt = -1;

            /* Arguments? */
            if (*(++cp) == ':') {
                optArg = *(++arg);
                if (optArg == (char *) 0)
                    opt = -2;
            }
            else optArg = (char *) 0;
        }

        switch (opt) {

        case -2: /* Missing argument */
            fprintf(stderr, "Option -%c needs an argument\n", opt);
            errFlag = 1;
            break;

        case -1: /* Unknown option */
            fprintf(stderr, "Unknown option: -%c\n", opt);
            errFlag = 1;
            break;

        case 0:
            break;

        case ('S'): /* Single command line application name */
            if (!application)
                application = optArg;
            else {
                fprintf(stderr, "Use -S once only\n");
                errFlag = 1;
            }
            break;

        case ('f'): /* Filter value for Create */
        {
            char *bp = buf;
            int manyNames = 0;

            if (bp[0] != '\0') { /* already given -f */
                fprintf(stderr, "use -f once only\n");
                errFlag = 1;
                break;
            }

            for( ;
                argIndex < argc && *arg != (char *) 0 && (*arg)[0] != '\0';
                ++arg, ++argIndex) {

                if ((*arg)[0] == '-') { /* Next option */
                    --arg;
                    --argIndex;
                    break;
                }

                if ((*arg)[0] == '(') { /* filter expression */
                    memcpy(buf, (*arg), strlen(*arg));
                    break;
                }

                if (bp == buf) { /* start of list */

                    /* more than one name? */
                    if (*(arg + 1) != (char *) 0 &&
                        (*(arg + 1))[0] != '\0' &&
                        (*(arg + 1))[0] != '-') {
                        sprintf(bp, "(");
                        bp = buf + strlen(buf);

                        manyNames = 1;
                    }
                }

                /* notification name */
                sprintf(bp, "(eventType=%s)", *arg);
                bp = buf + strlen(buf);

                if (manyNames) {

                    /* end of list? */
                    if (*(arg + 1) == (char *) 0 ||
                        (*(arg + 1))[0] == '\0' ||
                        (*(arg + 1))[0] == '-') {
                        sprintf(bp, ")");
                        break;
                    }

                    /* OR filter */
                    else {
                        sprintf(bp, "|");
                        bp = buf + strlen(buf);
                    }
                }
            }
        }
            break;

#if defined(SECURITY)
        case ('P'): /* peer DN */
            if (peer == (char *) 0)
                peer = optArg;
            else {
                fprintf(stderr, "Use -P once only\n");
                errFlag = 1;
            }
            break;

        case ('D'): /* DSA name or address */
            if (dsa == (char *) 0)
                dsa = optArg;
            else {
                fprintf(stderr, "Use -D once only\n");
                errFlag = 1;
            }
            break;

        case ('I'): /* Use integrity checks */
            if (!IFlag)
                IFlag = 1;
            else {
                fprintf(stderr, "Use -I once only\n");
                errFlag = 1;
            }
            break;
#endif
        } /* switch */
    }

    // load syntaxes

    if (initialiseSyntaxes(NULLCP, evrepSyntaxes) == NOTOK)
        error(NULLCP, "evsink: cannot initialise syntaxes");

#if defined(SECURITY)
    initialiseOsisec(argc, argv);
    initialiseMd5KeySpace();
    initialiseIdSpace();
#endif

    // convert the supplied arguments (if any) to a CMIS filter

    if (buf[0] != '\0' && (filter = str2mfilter(buf)) == NULLMFILTER)
	error(NULLCP, "evsink: invalid filter! %s", buf);

    // now try to connect to the remote agent

    MInitOption *miop[2]; /* Connect options */

#ifdef SECURITY
    MInitOption mio;
    AuthAssocIntegrityInfo *aaii;
    aaii = AuthAssocIntegrityInfo_make();

    if (peer != (char *) 0)
        aaii->peer = strdup(peer);
    if (dsa != (char *) 0)
        aaii->dsa = strdup(dsa);
    if (IFlag)
        memcpy((void *) aaii->sessionKey, (void *) makeMd5Key(), 8);

    mio.optionId = MIO_AuthAssocIntegrityInfo;
    mio.optionInfo = (void *) aaii;

    miop[0] = &mio;
    miop[1] = (MInitOption *) 0;

#else

   miop[0] = miop[1] = (MInitOption *) 0;

#endif

    host = application ? NULLCP : argv[2];
    application = application ? application : argv[1];

    if ((msd = m_initialise(application, host, miop)) < 0)
        error((char *) 0, "cannot establish association to %s at %s",
              argv[1], argv[2]);

#if defined(SECURITY)
    if (aaii != (AuthAssocIntegrityInfo *) 0)
        AuthAssocIntegrityInfo_free(aaii);
#endif

    // set signal handlers

#ifdef	__GNUG__
    signal(SIGINT,  terminate);
    signal(SIGQUIT, terminate);
    signal(SIGTERM, terminate);
#else
    signal(SIGINT,  (SIGHANDLER) terminate);
    signal(SIGQUIT, (SIGHANDLER) terminate);
    signal(SIGTERM, (SIGHANDLER) terminate);
#endif

    // create the discriminator

    do_create(msd, filter, &mis);

    // wait to receive event reports

#if defined(DEBUG2)
    if ((pduDump = fopen("evsink.rx.log", "w")) != (FILE *) 0)
        ppSetOutputFile(pduDump);
#endif

    for (;;)
	do_wait(msd, &mis);
}


static int  do_create (int msd, CMISFilter* filter, MSAPIndication* mi)
{
    struct MIDentifier obj_class;
    CMISParam	attrs[2];
    int		nattrs = 1,	// default destination address only
		status;

    // create a EFDiscriminator MO to receive event reports
    obj_class.mid_type = MID_GLOBAL;
    obj_class.mid_global = name2oid("eventForwardingDiscriminator");


    // initialise the destination address attribute
    DestAddressVal addr;
    addr.da_type = DA_MSD;	// this management association descriptor
    addr.da_next = NULL;

    attrs[0].mp_id.mid_type = MID_GLOBAL;
    attrs[0].mp_id.mid_global = name2oid("destination");
    if ((attrs[0].mp_val = DestAddress_encode(&addr)) == NULLPE) {
	advise(NULLCP, "cannot build destination address attribute");
	terminate(NOTOK);
    }

    // initialise the discriminator construct attribute (if any)

    if (filter) {
	attrs[1].mp_id.mid_type = MID_GLOBAL;
	attrs[1].mp_id.mid_global = name2oid("discriminatorConstruct");
	if ((attrs[1].mp_val = CMISFilter_encode(filter)) == NULLPE) {
	    advise(NULLCP, "CMISFilter_encode failed");
	    terminate(NOTOK);
	}

	nattrs = 2;	// we supply initial values for two attributes

	// free allocated memory
	mfilter_free(filter);
    }

    // now create
#if defined(SECURITY)
    request_id = makeInitiatorId(msd, request_id);
#else
    request_id = 1;
#endif
    status = M_Create(msd, request_id, &obj_class, NULLMN, 0, NULLMN,
                      NULLMACCESS, nattrs, attrs, mi);

    request_invoke = mi->mi_invoke;

    // free allocated memory
    mid_free(&obj_class);
    for (int i = 0; i < nattrs; i++)
	mparm_free(&attrs[i]);

    return status;
}


static int  do_delete (int msd, MSAPIndication* mi)
{
    // delete the EFDiscriminator we created
#if defined(SECURITY)
    request_id = makeInitiatorId(msd, request_id);
#else
    ++request_id;
#endif
    int status = M_Delete(msd, request_id, NULLMID, &efdiscr,
                          NULLMSCOPE, NULLMFILTER, NULLMACCESS, NULLMSYNC, mi);

    mn_free(&efdiscr);
    return status;
}


static int  do_wait (int msd, MSAPIndication* mi)
{
    char * inst;
    PE pdu;

    M_WaitReqAux(msd, NOTOK, mi, &pdu);

#if defined(DEBUG)
    if (pduDump != (FILE *) 0)
        ppMSAPIndication(mi);
#endif

#if defined(SECURITY)
    if (mi->mi_invoke != 0) {

        MD5Value md5;

        memset((void *) &md5, 0, sizeof(MD5Value));

        if (pdu == (PE) 0) {
            if (mi->mi_type == MI_ERROR)
                pdu = int2prim(request_id);

            else
                printf("Unexpected MSAPIndication type: %d\n", mi->mi_type);
        }

        if (mi->mi_type == MI_EVENT_REP)
            invoke = makeResponderId(msd, invoke);

        if (makeMd5Value(msd, pdu, &md5) != 0)
            printf("Could not make integrity check for PDU\n");

        register
        int chk1 = ENCODE_ID(request_id, md5.md5_words[0]),
            chk2 = ENCODE_ID(invoke, md5.md5_words[0]);

        register
        int i = mi->mi_invoke == chk1 ? request_id :
            (mi->mi_invoke == chk2 ? invoke : 0);

        printf("check %s - invokeId=%#08x id=%#08x chk=%#08x\n",
               i == 0 ? "failed" : "ok",
               (u_int) mi->mi_invoke, (u_int) i, md5.md5_words[0]);

        if (i == 0)
            printf("chk1=%#08x chk1=%#08x invoke=%#08x request_id=%#08x\n",
                   (u_int) chk1, (u_int) chk2, (u_int) invoke,
                   (u_int) request_id);
    }
#endif

    if (pdu != (PE) 0)
        pe_free(pdu);

    switch (mi->mi_type) {
    case MI_CREATE_RES:
    {
	struct CMISCreateRes * c = &mi->mi_createres.result;
	printf("Create Result - id %#08x, ", (u_int) mi->mi_createres.id);

	if (c->cr_class.mid_global)
            printf("class %s, ", oid2name(c->cr_class.mid_global, OIDPART));

	inst = dn2str(c->cr_inst.mn_dn);
	printf("instance {%s}", inst);
	free(inst);
	efdiscr = c->cr_inst;		// struct copy
					// will be used to delete the discr.
	c->cr_inst.mn_dn = NULLDN;	// not to be freed by mi_free()
	if (c->cr_time)
            printf(", time %s\n\n\n", c->cr_time);
	else
	    printf("\n\n\n");
    }
	break;

    case MI_DELETE_RES:
    {
	struct CMISDeleteRes* d = &mi -> mi_deleteres.result;
	printf("Delete Result - id %#08x, ", (u_int) mi -> mi_deleteres.id);
	if (d->dr_class.mid_global)
            printf("class %s, ", oid2name(d->dr_class.mid_global, OIDPART));

	inst = dn2str(d->dr_inst.mn_dn);
	printf("instance {%s}", inst);
	free(inst);

	if (d->dr_time)
            printf(", time %s\n", d->dr_time);
	else
	    printf("\n\n");
    }
	break;

    case MI_EVENT_REP:
    {
	struct CMISEventReportArg* e = &mi -> mi_eventrepinv.args;
	char *reportid, *reportval;

	if (e->ea_info)
	    ava2str(e->ea_type.mid_global, e->ea_info, &reportid, &reportval);
	else
	    reportid = strdup(oid2name(e->ea_type.mid_global, OIDPART));
	printf("EventReport - %s\n", reportid);

	if (e->ea_class.mid_global)
            printf("class: %s\n", oid2name(e->ea_class.mid_global, OIDPART));
        if (e->ea_inst.mn_dn) {
	    char * inst;
	    inst = dn2str(e->ea_inst.mn_dn);
	    printf("inst: {%s}\n", inst);
	    free(inst);
	}
	if (e->ea_time)
            printf("time: %.24s\n", e->ea_time);

	if (reportval)
	    fprintf(stdout, "report: %s\n", reportval);
	fprintf(stdout, "\n\n");
	free(reportid); if (reportval) free(reportval);
    }
	break;

    case MI_PREJECT:
    case MI_ABORT:
	fprintf(stderr, "evsink: received %s, exits\n",
		mi->mi_type == MI_PREJECT ?
				"provider reject" : "association abort");
	exit(NOTOK);

    default:
	printf("Unexpected mindication type: %d\n", mi->mi_type);
	mi_free(mi);
	return NOTOK;
    }
    
    mi_free(mi);
    return mi->mi_type == MI_DELETE_RES ? DONE : OK;
}


static void  terminate (int signal)
{
    if (efdiscr.mn_dn) {
	do_delete (msd, &mis);
	while (do_wait (msd, &mis) != DONE);
    }
    m_terminate(msd);

#if defined(DEBUG2)
    if (pduDump != (FILE *) 0)
        fclose(pduDump);
#endif

    exit(signal);
}
