//
// Copyright (c) 1995 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// TclRMIB.cc

// By Thurain Tin / Rong Shi, January 1995


#include <stdlib.h>
#include <string.h>

#ifdef  __GNUG__
#include <std.h>
#endif

#include "TclRMIB.h"
#include "AnyType.h"
#include "Error.h"
#include "Syntaxes.h"
#include "TkCoord.h"

static int			agentId, mgrId;
static TclRMIBAgentList		agentList;
static TclRMIBManagerList	mgrList;
static ListIterator		agentIterator(&agentList);
static ListIterator		mgrIterator(&mgrList);
static TclKS*			tclks;
static TkCoordinator*		coord;

#define MAXATTRS		100

int parseFilterArg (int, char**, char**, int*);
int parseAvaArg (int, char**, char**, char**, int*);


TclRMIBAgent* FindTclRMIBAgent (Tcl_Interp* interp, char* cp)
{
    int id;
    if ((id = atoi(cp)) == 0) {
	sprintf(interp->result, "invalid agent id %s", cp);
	return NULLTCLRMIBAGENT;
    }

    register TclRMIBAgent* agent;
    agentIterator.reset();
    while (agent = (TclRMIBAgent*) agentIterator.getNext()) {
	if (agent->id == id) {
	    return agent;
	}
    }

    sprintf(interp->result, "no agent found with id %s", cp);
    return NULLTCLRMIBAGENT;
}


TclRMIBManager* FindTclRMIBManager (Tcl_Interp* interp, char* cp)
{
    int id;
    if ((id = atoi(cp)) == 0) {
	sprintf(interp->result, "invalid manager id %s", cp);
	return NULLTCLRMIBMANAGER;
    }

    register TclRMIBManager* mgr;
    mgrIterator.reset();
    while (mgr = (TclRMIBManager*) mgrIterator.getNext()) {
	if (mgr->id == id) {
	    return mgr;
	}
    }

    sprintf(interp->result, "no manager found with id %s", cp);
    return NULLTCLRMIBMANAGER;
}


void AssembleMO (Tcl_Interp* interp, CMISObject* obj)
{
    Tcl_AppendResult(interp, " {", obj->getErrorStr(),
		" {", obj->getClass() ? obj->getClass() : "",
		"} {", obj->getInstance() ? obj->getInstance() : "",
		"} {", obj->getTime() ? obj->getTime() : "",
		"}", (char *) NULL);

    CMISErrors err = obj->getError();
    if (err == m_noError || err == m_getListError || err == m_setListError) {
	switch (obj->getOperation()) {
	case M_GET:
	case M_SET:
	case M_CREATE:
	{
	    AVAArray* avaArray = obj->getAttrs();
	    if (avaArray) {
		for (register int i = 0, n = avaArray->getCount(); i < n; i++) {
		    AVA* attr = (AVA*) avaArray->get(i);
		    if (attr) {
			Attr* val = attr->getError()
					? NULLATTR : attr->getValue();
			char* cp = NULLCP;
			Tcl_AppendResult(interp, " {",
				val ? attr->getName() : attr->getErrorStr(),
				val ? " {" : " ",
				val ? cp = val->print() : attr->getName(),
				val ? "}}" : "}", (char *) NULL);
			if (cp) free(cp);
		    }
		}
	    }
	}
	    break;

	case M_ACTION:
	{
	    AVA* action = obj->getActionRes();
	    if (action) {
		Attr* val = action->getValue();
		char* cp = NULLCP;
		Tcl_AppendResult(interp, " {", action->getName(), " {",
			val ? cp = val->print() : "", "}}", (char *) NULL);
		if (cp) free(cp);
	    }
	}
	    break;

	case M_DELETE:
	    break;
	}
    }

    if (err == m_noSuchAttribute || err == m_missingAttributeValue ||
	    err == m_invalidAttributeValue || err == m_processingFailure) {
	AVA* errInfo = obj->getErrorInfo();
	if (errInfo) {
	    Tcl_AppendResult(interp, " {", errInfo->getName(), (char*) NULL);
	    if (err == m_invalidAttributeValue || err == m_processingFailure) {
		Attr* val = errInfo->getValue();
		char* cp = NULLCP;
		Tcl_AppendResult(interp, " {", val ? cp = val->print() : "",
				"}", (char *) NULL);
		if (cp) free(cp);
	    }
	    Tcl_AppendResult(interp, "}", (char *) NULL);
	}
    }

    Tcl_AppendResult(interp, "}", (char*) NULL);
}


void AssembleMOList (Tcl_Interp* interp, char* id, char* agent, char* host,
	CMISObjectList* objList, CMISObject* obj = NULLCMISOBJECT)
{
    Tcl_AppendElement(interp, id);

    char replyId[16];
    sprintf(replyId, "%d", objList ? objList->getInvoke() : obj->getInvoke());
    Tcl_AppendElement(interp, replyId);

    Tcl_AppendElement(interp, agent);
    Tcl_AppendElement(interp, host);

    int operType = objList ? objList->getOperation() : obj->getOperation();
    switch (operType) {
    case M_GET:
	Tcl_AppendElement(interp, "M_GET");
	break;
    case M_SET:
	Tcl_AppendElement(interp, "M_SET");
	break;
    case M_ACTION:
	Tcl_AppendElement(interp, "M_ACTION");
	break;
    case M_CREATE:
	Tcl_AppendElement(interp, "M_CREATE");
	break;
    case M_DELETE:
	Tcl_AppendElement(interp, "M_DELETE");
	break;
    }

    Tcl_AppendElement(interp,
		objList ? objList->getErrorStr() : obj->getErrorStr());

    if (objList) {
	ListIterator iterator;
	iterator.setList(objList);
	CMISObject* obj;
	while (obj = (CMISObject*) iterator.getNext())
	    AssembleMO(interp, obj);
    }
    else
    if (obj)
	AssembleMO(interp, obj);
}


int TclRMIBManager::eventNotification (char* objClass, char* objInst,
	char* eventType, char* eventTime, Attr* report, RMIBAgent* agent)
{
    char* cp;
    Tcl_VarEval(this->interp, this->eproc, " ", this->Id,
		" ", agent->getAgentName(), " ", agent->getHostName(),
		" ", objClass, " ", objInst, " ", eventType, " ",
		eventTime, " {", cp = report->print(), "}", (char *) NULL);

    free(cp);
    return OK;
}


int TclRMIBManager::generalResult (CMISObjectList* objList,
	RMIBAgent* agent)
{
    sprintf(interp->result, " ");
    Tcl_AppendResult(this->interp, " {", (char*) NULL);

    AssembleMOList(this->interp, this->Id, agent->getAgentName(),
		agent->getHostName(), objList);

    Tcl_AppendResult(this->interp, "}", (char*) NULL);

    char* cp = interp->result;
    interp->result = (char *) NULL;
    Tcl_VarEval(this->interp, this->rproc, cp, (char *) NULL);

    delete objList;
    free(cp);
    return OK;
}


int TclRMIBManager::singleGetResult (CMISObject* obj, RMIBAgent* agent)
{
    sprintf(interp->result, " ");
    Tcl_AppendResult(this->interp, " {", (char*) NULL);

    AssembleMOList(this->interp, this->Id, agent->getAgentName(),
		agent->getHostName(), NULLCMISOBJECTLIST, obj);

    Tcl_AppendResult(this->interp, "}", (char*) NULL);

    char* cp = interp->result;
    interp->result = (char *) NULL;
    Tcl_VarEval(this->interp, this->rproc, cp, (char *) NULL);

    delete obj;
    free(cp);
    return OK;
}


int TclRMIBManager::createResult (CMISObject* obj, RMIBAgent* agent)
{
    sprintf(interp->result, " ");
    Tcl_AppendResult(this->interp, " {", (char*) NULL);

    AssembleMOList(this->interp, this->Id, agent->getAgentName(),
		agent->getHostName(), NULLCMISOBJECTLIST, obj);

    Tcl_AppendResult(this->interp, "}", (char*) NULL);

    char* cp = interp->result;
    interp->result = (char *) NULL;
    Tcl_VarEval(this->interp, this->rproc, cp, (char *) NULL);

    delete obj;
    free(cp);
    return OK;
}


int TclRMIBManager::noTimeAssertion (RMIBAgent* agent)
{
    if (!this->nproc)
	return NOTOK;

    Tcl_VarEval(this->interp, this->nproc,
		" ", this->Id, " ", agent->getAgentName(),
		" ", agent->getHostName(), (char*) NULL);
    return OK;
}


int TclRMIBManager::brokenAssociation (RMIBAgent* agent)
{
    Tcl_VarEval(this->interp, this->bproc,
		" ", this->Id, " ", agent->getAgentName(),
		" ", agent->getHostName(), (char*) NULL);
    return OK;
}


DEFINE_CALLBACK (RmibAgentCmd)
{
    char* appl = NULLCP;
    char* host = NULLCP;
    char* time = NULLCP;
    char* entity = NULLCP;

    if (argc > 2) {
	int state = 0;
	for (int i = 1; i < argc; i++)
	    switch (state) {
	    case 0:
		if (*argv[i] == '-' && strlen(argv[i]) == 2)
		    switch (*(argv[i]+1)) {
		    case 'a':
			state = 1;
			break;
		    case 'h':
			state = 2;
			break;
		    case 't':
			state = 3;
			break;
		    case 'e':
			state = 4;
			break;
		    default:
			goto usage;
		    }
		else
		    goto usage;
		break;

	    case 1:
		appl = argv[i];
		state = 0;
		break;

	    case 2:
		host = argv[i];
		state = 0;
		break;

	    case 3:
		time = argv[i];
		state = 0;
		break;

	    case 4:
		entity = argv[i];
		state = 0;
		break;
	    }
    }

    TclRMIBAgent* agent;
    agentList.append(agent = new TclRMIBAgent(++agentId, appl, host));

    if (time)
	agent->setTimeOut(atoi(time));

    if (entity)
	agent->setTitle(entity);

    sprintf(interp->result, "%d", agentId);
    CALLBACK_OK

usage:
    sprintf(interp->result, "usage: rmib_agent ?-a appl? ?-h host?\n\
\t?-t timeout? ?-e title?");
    CALLBACK_ERROR
}


DEFINE_CALLBACK (RmibManagerCmd)
{
    char* eproc = NULLCP;
    char* rproc = NULLCP;
    char* bproc = NULLCP;
    char* nproc = NULLCP;
    int i, state = 0;

    if (argc < 3)
	goto usage;

    for (i = 1; i < argc; i++)
	switch (state) {
	case 0:
	    if (*argv[i] == '-' && strlen(argv[i]) == 2)
		switch (*(argv[i]+1)) {
		case 'e':
		    state = 1;
		    break;
		case 'r':
		    state = 2;
		    break;
		case 'b':
		    state = 3;
		    break;
		case 'n':
		    state = 4;
		    break;
		default:
		    goto usage;
		}
	    else
		goto usage;
	    break;

	case 1:
	    eproc = argv[i];
	    state = 0;
	    break;

	case 2:
	    rproc = argv[i];
	    state = 0;
	    break;

	case 3:
	    bproc = argv[i];
	    state = 0;
	    break;

	case 4:
	    nproc = argv[i];
	    state = 0;
	    break;
	}

    mgrList.append(new TclRMIBManager(++mgrId, interp,
			eproc, rproc, bproc, nproc));
    sprintf(interp->result, "%d", mgrId);
    CALLBACK_OK

usage:
    sprintf(interp->result, "usage: rmib_manager ?-e eproc? ?-r rproc?\n\
\t?-b bproc? ?-n nproc?");
    CALLBACK_ERROR
}


DEFINE_CALLBACK (RmibDestroyCmd)
{
    int i, state = 0;

    if (argc < 3)
	goto usage;

    for (i = 1; i < argc; i++)
	switch (state) {
	case 0:
	    if (*argv[i] == '-' && strlen(argv[i]) == 2)
		switch (*(argv[i]+1)) {
		case 'a':
		    state = 1;
		    break;
		case 'm':
		    state = 2;
		    break;
		default:
		    goto usage;
		}
	    else
		goto usage;
	    break;

	case 1:
	{
	    TclRMIBAgent* agent;
	    if (!(agent = FindTclRMIBAgent(interp, argv[i])))
		CALLBACK_ERROR

	    if (agent->getMsd() != UNCONNECTED)
		agent->disconnect();

	    TclRMIBAgent* tclAgent;
	    if (tclAgent = (TclRMIBAgent*) agentList.remove(agent)) {
		sprintf(interp->result, "%s", argv[i]);
		delete tclAgent;
		CALLBACK_OK
	    }
	    sprintf(interp->result, "cannot destroy agent with id %s", argv[i]);
	    CALLBACK_ERROR
	}
	    state = 0;
	    break;

	case 2:
	{
	    TclRMIBManager* mgr;
	    if (!(mgr = FindTclRMIBManager(interp, argv[i])))
		CALLBACK_ERROR

	    TclRMIBManager* tclMgr;
	    if (tclMgr = (TclRMIBManager*) mgrList.remove(mgr)) {
		sprintf(interp->result, "%s", argv[i]);
		delete tclMgr;
		CALLBACK_OK
	    }
	    sprintf(interp->result, "cannot destroy mgr with id %s", argv[i]);
	    CALLBACK_ERROR
	}
	    state = 0;
	    break;
	}

usage:
    sprintf(interp->result, "usage: rmib_destroy ?-a|m objId?");
    CALLBACK_ERROR
}


DEFINE_CALLBACK (RmibAgentInfoCmd)
{
    Bool assocInfo = False;
    if (argc > 1) {
	if (*argv[1] == '-' && strlen(argv[1]) == 2 && *(argv[1]+1) == 'a')
	     assocInfo = True;
	else
	    goto usage;
    }

    register TclRMIBAgent* agent;
    agentIterator.reset();
    Tcl_AppendResult(interp, "{ ", (char *) NULL);

    while (agent = (TclRMIBAgent*) agentIterator.getNext()) {
	char buf[16];
	if (assocInfo && agent->getMsd() == UNCONNECTED)
	    continue;
	sprintf(buf, "%d ", agent->id);
	Tcl_AppendResult(interp, buf, (char *) NULL);
    }

    Tcl_AppendResult(interp, "}", (char *) NULL);
    CALLBACK_OK

usage:
    sprintf(interp->result, "usage: rmib_agent_info ?-a?");
    CALLBACK_ERROR
}


DEFINE_CALLBACK (RmibConnectCmd)
{
    if (argc < 2)
	goto usage;

    TclRMIBAgent* agent;
    if (!(agent = FindTclRMIBAgent(interp, argv[1])))
	CALLBACK_ERROR

    if (agent->getMsd() != UNCONNECTED) {
	sprintf(interp->result, "agent is already in association!");
	CALLBACK_ERROR
    }

    if (argc > 2) {
	int state = 0;
	for (int i = 2; i < argc; i++)
	    switch (state) {
	    case 0:
		if (*argv[i] == '-' && strlen(argv[i]) == 2)
		    switch (*(argv[i]+1)) {
		    case 'a':
			state = 1;
			break;
		    case 'h':
			state = 2;
			break;
		    case 't':
			state = 3;
			break;
		    default:
			goto usage;
		    }
		else
		    goto usage;
		break;

	    case 1:
		agent->setAgentName(argv[i]);
		state = 0;
		break;

	    case 2:
		agent->setHostName(argv[i]);
		state = 0;
		break;

	    case 3:
		agent->setTimeOut(atoi(argv[i]));
		state = 0;
		break;
	    }
    }

    if (!agent->getAgentName()
#ifndef USE_X500				/* if no LT hooks */
		|| !agent->getHostName()
#endif
        ) {
	sprintf(interp->result, "agent %s or host %s name missing!",
			agent->getAgentName(), agent->getHostName());
	CALLBACK_ERROR
    }

    if (agent->connect() == NOTOK) {
	sprintf(interp->result, "association establishment failed!");
	CALLBACK_ERROR
    }

    sprintf(interp->result, "%s", argv[1]);
    CALLBACK_OK

usage:
    sprintf(interp->result, "usage: m_connect agentId ?-a appl? ?-h host?\n\
\t?-t timeout?");
    CALLBACK_ERROR
}


DEFINE_CALLBACK (RmibDisconnectCmd)
{
    if (argc != 2)
	goto usage;

    TclRMIBAgent* agent;
    if (!(agent = FindTclRMIBAgent(interp, argv[1])))
	CALLBACK_ERROR

    if (agent->getMsd() == UNCONNECTED) {
	sprintf(interp->result, "agent is not in association!");
	CALLBACK_ERROR
    }

    agent->disconnect();
    sprintf(interp->result, "%s", argv[1]);
    CALLBACK_OK

usage:
    sprintf(interp->result, "usage: m_disconnect agentId");
    CALLBACK_ERROR
}


DEFINE_CALLBACK (RmibGetCmd)
{
    char* objClass = NULLCP;
    char* objInst = NULLCP;
    char* scope = NULLCP;
    char* sync = NULLCP;
    char* filter = NULLCP;
    char* attrs[MAXATTRS];
    char* mgrId = NULLCP;
    Bool individual = False;
    int i, j, k, l = 0, n, nattrs = 0, state = 0, result;
    TclRMIBManager* mgr = NULLTCLRMIBMANAGER;
    CMISScope* cmisScope = NULLMSCOPE;
    CMISSync cmisSync = s_bestEffort;
    CMISObjectList* objList = NULLCMISOBJECTLIST;

    if (argc < 2)
	goto usage;

    TclRMIBAgent* agent;
    if (!(agent = FindTclRMIBAgent(interp, argv[1])))
	CALLBACK_ERROR

    if (agent->getMsd() == UNCONNECTED) {
	sprintf(interp->result, "agent is not in association!");
	CALLBACK_ERROR
    }

    if (argc > 2) {
	for (i = 2; i < argc; i++)
	    switch (state) {
	    case 0:
		if (*argv[i] == '-' && strlen(argv[i]) == 2)
		    switch (*(argv[i]+1)) {
		    case 'c':
			state = 1;
			break;
		    case 'i':
			state = 2;
			break;
		    case 's':
			state = 3;
			break;
		    case 'f':
			state = 4;
			break;
		    case 'a':
			state = 5;
			break;
		    case 'm':
			state = 6;
			break;
		    case 'o':
			individual = True;
			break;
		    default:
			goto usage;
		    }
		else
		    goto usage;
		break;

	    case 1:
		objClass = argv[i];
		state = 0;
		break;

	    case 2:
		objInst = argv[i];
		state = 0;
		break;

	    case 3:
		scope = argv[i];
		if (i < argc-1 && *argv[i+1] != '-')
		    sync = argv[++i];
		state = 0;
		break;

	    case 4:
		if (parseFilterArg(argc-i, &argv[i], &filter, &j) == NOTOK)
		    goto usage;
		i += j-1;
		state = 0;
		break;

	    case 5:
	    {
		Bool more = False;
		for (j = i; ; j++) {
		    if (*argv[j] == '-' && strlen(argv[j]) == 2) {
			more = True;
			break;
		    }
		    else
		    if (j == argc-1)
			break;
		}

		n = more ? j-i : j-i+1;
		nattrs += n;

		if (nattrs > MAXATTRS) {
		    sprintf(interp->result, "too many # attrs");
		    if (filter)
			free(filter);
		    CALLBACK_ERROR
		}

		for (k = i; k < i+n; k++, l++)
		    attrs[l] = argv[k];
		attrs[l] = NULLCP;

		i = more ? j-1 : j;
	    }
		state = 0;
		break;

	    case 6:
		mgrId = argv[i];
		state = 0;
		break;
	    }
    }

    if (mgrId && !(mgr = FindTclRMIBManager(interp, mgrId))) {
	if (filter)
	    free(filter);
	CALLBACK_ERROR
    }

    if (mgr && !mgr->rproc) {
	sprintf(interp->result, "manager has no appropriate callback proc!");
	if (filter)
	    free(filter);
	CALLBACK_ERROR
    }

    if (scope && (cmisScope = str2scope(scope)) == NULLMSCOPE) {
	sprintf(interp->result, "invalid scope %s", scope);
	if (filter)
	    free(filter);
	CALLBACK_ERROR
    }

    if (sync && (cmisSync = str2sync(sync)) == s_invalid) {
	sprintf(interp->result, "invalid sync %s", sync);
	if (filter)
	    free(filter);
	if (cmisScope)
	    free((char*) cmisScope);
	CALLBACK_ERROR
    }

    result = agent->Get(!objClass && !objInst ? "system" : objClass, objInst,
			cmisScope ? cmisScope->sc_type : Sc_BaseObject,
			cmisScope ? cmisScope->sc_level : 0, filter,
			NULLMACCESS, cmisSync, nattrs ? attrs : (char**) NULL,
			objList, mgr, mgr ? individual : False);

    if (filter)
	free(filter);

    if (cmisScope)
	free((char*) cmisScope);

    if ((!mgr && !objList) || (mgr && result == NOTOK)) {
	sprintf(interp->result, "Get operation failed!");
	CALLBACK_ERROR
    }

    if (!mgr) {
	AssembleMOList(interp, argv[1], agent->getAgentName(),
			agent->getHostName(), objList);
	delete objList;
    }
    else {
	sprintf(interp->result, "%d", result);
    }

    CALLBACK_OK

usage:
    sprintf(interp->result, "usage: m_get agentId ?-c class? ?-i instance?\n\
\t?-s scope ?sync?? ?-f filter? ?-a attr ...?\n\t?-m mgrId ?-o??");
    if (filter)
	free(filter);
    CALLBACK_ERROR
}


DEFINE_CALLBACK (RmibCancelGetCmd)
{
    if (argc < 3)
	goto usage;

    TclRMIBAgent* agent;
    if (!(agent = FindTclRMIBAgent(interp, argv[1])))
	CALLBACK_ERROR

    if (agent->getMsd() == UNCONNECTED) {
	sprintf(interp->result, "agent is not in association!");
	CALLBACK_ERROR
    }

    agent->CancelGet(atoi(argv[2]));
    sprintf(interp->result, "%s", argv[2]);
    CALLBACK_OK

usage:
    sprintf(interp->result, "usage: m_cancel_get agentId invokeId");
    CALLBACK_ERROR
}


DEFINE_CALLBACK (RmibSetCmd)
{
    char* objClass = NULLCP;
    char* objInst = NULLCP;
    char* scope = NULLCP;
    char* sync = NULLCP;
    char* filter = NULLCP;
    AVAArray* avaArray = NULLAVAARRAY;
    char* mgrId = NULLCP;
    int i, j, nattrs = 0, state = 0, result;
    TclRMIBManager* mgr = NULLTCLRMIBMANAGER;
    CMISScope* cmisScope = NULLMSCOPE;
    CMISSync cmisSync = s_bestEffort;
    CMISObjectList* objList = NULLCMISOBJECTLIST;

    if (argc < 2)
	goto usage;

    TclRMIBAgent* agent;
    if (!(agent = FindTclRMIBAgent(interp, argv[1])))
	CALLBACK_ERROR

    if (agent->getMsd() == UNCONNECTED) {
	sprintf(interp->result, "agent is not in association!");
	CALLBACK_ERROR
    }

    if (argc > 2) {
	for (i = 2; i < argc; i++)
	    switch (state) {
	    case 0:
		if (*argv[i] == '-' && strlen(argv[i]) == 2)
		    switch (*(argv[i]+1)) {
		    case 'c':
			state = 1;
			break;
		    case 'i':
			state = 2;
			break;
		    case 's':
			state = 3;
			break;
		    case 'f':
			state = 4;
			break;
		    case 'w':
			state = 5;
			break;
		    case 'a':
			state = 6;
			break;
		    case 'r':
			state = 7;
			break;
		    case 'd':
			state = 8;
			break;
		    case 'm':
			state = 9;
			break;
		    default:
			goto usage;
		    }
		else
		    goto usage;
		break;

	    case 1:
		objClass = argv[i];
		state = 0;
		break;

	    case 2:
		objInst = argv[i];
		state = 0;
		break;

	    case 3:
		scope = argv[i];
		if (i < argc-1 && *argv[i+1] != '-')
		    sync = argv[++i];
		state = 0;
		break;

	    case 4:
		if (parseFilterArg(argc-i, &argv[i], &filter, &j) == NOTOK)
		    goto usage;
		i += j-1;
		state = 0;
		break;

	    case 5:
	    case 6:
	    case 7:
	    case 8:
	    {
		char *attrType, *attrValue;
		if (parseAvaArg(argc-i, &argv[i],
				&attrType, &attrValue, &j) == NOTOK)
		    goto usage;

		CMISModifyOp modifyOp;
		switch (state) {
		case 5:
		    modifyOp = m_replace;
		    break;
		case 6:
		    modifyOp = m_addValue;
		    break;
		case 7:
		    modifyOp = m_removeValue;
		    break;
		case 8:
		    modifyOp = m_setToDefault;
		    attrValue = NULLCP;
		    break;
		}

		AVA* ava = new AVA(attrType, attrValue, modifyOp);

		if (ava->createError()) {
		    sprintf(interp->result, "invalid attribute, type %s value %s",
				attrType, attrValue);
		    delete ava;

		    if (attrType)
			free(attrType);

		    if (attrValue)
			free(attrValue);

		    if (avaArray) {
			avaArray->clear();
			delete avaArray;
		    }

		    if (filter)
			free(filter);
		    CALLBACK_ERROR
		}

		if (attrType)
		    free(attrType);

		if (attrValue)
		    free(attrValue);

		if (nattrs == 0)
		    avaArray = new AVAArray(10);

		avaArray->add(ava);
		nattrs++;
		i += j;
	    }
		state = 0;
		break;

	    case 9:
		mgrId = argv[i];
		state = 0;
		break;
	    }
    }

    if (!avaArray) {
	sprintf(interp->result, "no attributes given to modify!");
	if (filter)
	    free(filter);
	CALLBACK_ERROR
    }

    if (mgrId && !(mgr = FindTclRMIBManager(interp, mgrId))) {
	if (filter)
	    free(filter);
	avaArray->clear();
	delete avaArray;
	CALLBACK_ERROR
    }

    if (mgr && !mgr->rproc) {
	sprintf(interp->result, "manager has no appropriate callback proc!");
	if (filter)
	    free(filter);
	avaArray->clear();
	delete avaArray;
	CALLBACK_ERROR
    }

    if (scope && (cmisScope = str2scope(scope)) == NULLMSCOPE) {
	sprintf(interp->result, "invalid scope %s", scope);
	if (filter)
	    free(filter);
	avaArray->clear();
	delete avaArray;
	CALLBACK_ERROR
    }

    if (sync && (cmisSync = str2sync(sync)) == s_invalid) {
	sprintf(interp->result, "invalid sync %s", sync);
	if (filter)
	    free(filter);
	avaArray->clear();
	delete avaArray;
	if (cmisScope)
	    free((char*) cmisScope);
	CALLBACK_ERROR
    }

    result = agent->Set(objClass, objInst,
			cmisScope ? cmisScope->sc_type : Sc_BaseObject,
			cmisScope ? cmisScope->sc_level : 0, filter,
			NULLMACCESS, cmisSync, avaArray, objList, mgr);

    if (filter)
	free(filter);

    avaArray->clear();
    delete avaArray;

    if (cmisScope)
	free((char*) cmisScope);

    if ((!mgr && !objList) || (mgr && result == NOTOK)) {
	sprintf(interp->result, "Set operation failed!");
	CALLBACK_ERROR
    }

    if (!mgr) {
	AssembleMOList(interp, argv[1], agent->getAgentName(),
			agent->getHostName(), objList);
	delete objList;
    }
    else {
	sprintf(interp->result, "%d", result);
    }

    CALLBACK_OK

usage:
    sprintf(interp->result, "usage: m_set agentId ?-c class? ?-i instance?\n\
\t?-s scope ?sync?? ?-f filter?\n\t?-w|a|r|d attrType?=attrValue?? ...\n\
\t?-m mgrId?");
    if (filter)
	free(filter);
    if (avaArray) {
	avaArray->clear();
	delete avaArray;
    }
    CALLBACK_ERROR
}


DEFINE_CALLBACK (RmibActionCmd)
{
    char* objClass = NULLCP;
    char* objInst = NULLCP;
    char* scope = NULLCP;
    char* sync = NULLCP;
    char* filter = NULLCP;
    AVA* action = NULLAVA;
    char* mgrId = NULLCP;
    int i, j, state = 0, result;
    TclRMIBManager* mgr = NULLTCLRMIBMANAGER;
    CMISScope* cmisScope = NULLMSCOPE;
    CMISSync cmisSync = s_bestEffort;
    CMISObjectList* objList = NULLCMISOBJECTLIST;

    if (argc < 2)
	goto usage;

    TclRMIBAgent* agent;
    if (!(agent = FindTclRMIBAgent(interp, argv[1])))
	CALLBACK_ERROR

    if (agent->getMsd() == UNCONNECTED) {
	sprintf(interp->result, "agent is not in association!");
	CALLBACK_ERROR
    }

    if (argc > 2) {
	for (i = 2; i < argc; i++)
	    switch (state) {
	    case 0:
		if (*argv[i] == '-' && strlen(argv[i]) == 2)
		    switch (*(argv[i]+1)) {
		    case 'c':
			state = 1;
			break;
		    case 'i':
			state = 2;
			break;
		    case 's':
			state = 3;
			break;
		    case 'f':
			state = 4;
			break;
		    case 'a':
			state = 5;
			break;
		    case 'm':
			state = 6;
			break;
		    default:
			goto usage;
		    }
		else
		    goto usage;
		break;

	    case 1:
		objClass = argv[i];
		state = 0;
		break;

	    case 2:
		objInst = argv[i];
		state = 0;
		break;

	    case 3:
		scope = argv[i];
		if (i < argc-1 && *argv[i+1] != '-')
		    sync = argv[++i];
		state = 0;
		break;

	    case 4:
		if (parseFilterArg(argc-i, &argv[i], &filter, &j) == NOTOK)
		    goto usage;
		i += j-1;
		state = 0;
		break;

	    case 5:
	    {
		Bool more = False;
		for (j = i; j < argc; j++) {
		    if (*argv[j] == '-' && strlen(argv[j]) == 2) {
			more = True;
			break;
		    }
		}

		char* actionType  = NULLCP;
		char* actionValue = NULLCP;
		if (parseAvaArg(argc-i, &argv[i],
				&actionType, &actionValue, &j) == NOTOK)
		    goto usage;

		char actionTypeInfo[64];
		sprintf(actionTypeInfo, "%s%s", actionType, "Info");

		action = new AVA(actionTypeInfo, actionValue);
		action->adjustActionOid();

		if (action->createError()) {
		    sprintf(interp->result, "invalid action, type %s value %s",
				actionType, actionValue);
		    delete action;

		    if (actionType)
			free(actionType);

		    if (actionValue)
			free(actionValue);

		    if (filter)
			free(filter);
		    CALLBACK_ERROR
		}

		if (actionType)
		    free(actionType);

		if (actionValue)
		    free(actionValue);

		i = more ? i+j-1 : i+j;
	    }
		state = 0;
		break;

	    case 6:
		mgrId = argv[i];
		state = 0;
		break;
	    }
    }

    if (!action) {
	sprintf(interp->result, "no action specified!");
	if (filter)
	    free(filter);
	CALLBACK_ERROR
    }

    if (mgrId && !(mgr = FindTclRMIBManager(interp, mgrId))) {
	if (filter)
	    free(filter);
	delete action;
	CALLBACK_ERROR
    }

    if (mgr && !mgr->rproc) {
	sprintf(interp->result, "manager has no appropriate callback proc!");
	if (filter)
	    free(filter);
	delete action;
	CALLBACK_ERROR
    }

    if (scope && (cmisScope = str2scope(scope)) == NULLMSCOPE) {
	sprintf(interp->result, "invalid scope %s", scope);
	if (filter)
	    free(filter);
	delete action;
	CALLBACK_ERROR
    }

    if (sync && (cmisSync = str2sync(sync)) == s_invalid) {
	sprintf(interp->result, "invalid sync %s", sync);
	if (filter)
	    free(filter);
	delete action;
	if (cmisScope)
	    free((char*) cmisScope);
	CALLBACK_ERROR
    }

    result = agent->Action(objClass, objInst,
			cmisScope ? cmisScope->sc_type : Sc_BaseObject,
			cmisScope ? cmisScope->sc_level : 0, filter,
			NULLMACCESS, cmisSync, action, objList, mgr);

    if (filter)
	free(filter);

    delete action;

    if (cmisScope)
	free((char*) cmisScope);

    if ((!mgr && !objList) || (mgr && result == NOTOK)) {
	sprintf(interp->result, "Action operation failed!");
	CALLBACK_ERROR
    }

    if (!mgr) {
	AssembleMOList(interp, argv[1], agent->getAgentName(),
			agent->getHostName(), objList);
	delete objList;
    }
    else {
	sprintf(interp->result, "%d", result);
    }

    CALLBACK_OK

usage:
    sprintf(interp->result, "usage: m_action agentId ?-c class? ?-i instance?\n\
\t?-s scope ?sync?? ?-f filter?\n\t?-a actionType?=actionValue?? ?-m mgrId?");
    if (filter)
	free(filter);
    if (action)
	delete action;
    CALLBACK_ERROR
}


DEFINE_CALLBACK (RmibCreateCmd)
{
    char* objClass = NULLCP;
    char* objInst = NULLCP;
    int instType = CA_OBJECT_INST;
    char* refInst = NULLCP;
    AVAArray* avaArray = NULLAVAARRAY;
    char* mgrId = NULLCP;
    int i, j, nattrs = 0, state = 0, result;
    TclRMIBManager* mgr = NULLTCLRMIBMANAGER;
    CMISObject* obj = NULLCMISOBJECT;

    if (argc < 2)
	goto usage;

    TclRMIBAgent* agent;
    if (!(agent = FindTclRMIBAgent(interp, argv[1])))
	CALLBACK_ERROR

    if (agent->getMsd() == UNCONNECTED) {
	sprintf(interp->result, "agent is not in association!");
	CALLBACK_ERROR
    }

    if (argc > 2) {
	for (i = 2; i < argc; i++)
	    switch (state) {
	    case 0:
		if (*argv[i] == '-' && strlen(argv[i]) == 2)
		    switch (*(argv[i]+1)) {
		    case 'c':
			state = 1;
			break;
		    case 'i':
			state = 2;
			break;
		    case 's':
			state = 3;
			break;
		    case 'r':
			state = 4;
			break;
		    case 'a':
			state = 5;
			break;
		    case 'm':
			state = 6;
			break;
		    default:
			goto usage;
		    }
		else
		    goto usage;
		break;

	    case 1:
		objClass = argv[i];
		state = 0;
		break;

	    case 2:
	    case 3:
		objInst = argv[i];
		if (state == 3)
		    instType = CA_SUPERIOR_INST;
		state = 0;
		break;

	    case 4:
		refInst = argv[i];
		state = 0;
		break;

	    case 5:
	    {
		char *attrType = NULLCP, *attrValue = NULLCP;
		if (parseAvaArg(argc-i, &argv[i], &attrType, &attrValue, &j)
			== NOTOK || !attrValue || *attrValue == '\0')
		    goto usage;

		AVA* ava = new AVA(attrType, attrValue);

		if (ava->createError()) {
		    sprintf(interp->result, "invalid attribute, type %s value %s",
				attrType, attrValue);
		    delete ava;
		    if (avaArray) {
			avaArray->clear();
			delete avaArray;
		    }

		    if (attrType)
			free(attrType);

		    if (attrValue)
			free(attrValue);

		    CALLBACK_ERROR
		}

		if (attrType)
		    free(attrType);

		if (attrValue)
		    free(attrValue);

		if (nattrs == 0)
		    avaArray = new AVAArray(10);

		avaArray->add(ava);
		nattrs++;
		i += j;
	    }
		state = 0;
		break;

	    case 6:
		mgrId = argv[i];
		state = 0;
		break;
	    }
    }

    if (mgrId && !(mgr = FindTclRMIBManager(interp, mgrId))) {
	if (avaArray) {
	    avaArray->clear();
	    delete avaArray;
	}
	CALLBACK_ERROR
    }

    if (mgr && !mgr->rproc) {
	sprintf(interp->result, "manager has no appropriate callback proc!");
	if (avaArray) {
	    avaArray->clear();
	    delete avaArray;
	}
	CALLBACK_ERROR
    }

    result = agent->Create(objClass, objInst, instType,
			refInst, NULLMACCESS, avaArray, obj, mgr);

    if (avaArray) {
	avaArray->clear();
	delete avaArray;
    }

    if ((!mgr && !obj) || (mgr && result == NOTOK)) {
	sprintf(interp->result, "Create operation failed!");
	CALLBACK_ERROR
    }

    if (!mgr) {
	AssembleMOList(interp, argv[1], agent->getAgentName(),
			agent->getHostName(), NULLCMISOBJECTLIST, obj);
	delete obj;
    }
    else {
	sprintf(interp->result, "%d", result);
    }

    CALLBACK_OK

usage:
    sprintf(interp->result, "usage: m_create agentId ?-c class?\n\
\t?-i instance | -s superiorInstance?\n\
\t?-r referenceInstance? ?-a attrType=attrValue? ...\n\
\t?-m mgrId?");
    if (avaArray) {
	avaArray->clear();
	delete avaArray;
    }
    CALLBACK_ERROR
}


DEFINE_CALLBACK (RmibDeleteCmd)
{
    char* objClass = NULLCP;
    char* objInst = NULLCP;
    char* scope = NULLCP;
    char* sync = NULLCP;
    char* filter = NULLCP;
    char* mgrId = NULLCP;
    int i, j, state = 0, result;
    TclRMIBManager* mgr = NULLTCLRMIBMANAGER;
    CMISScope* cmisScope = NULLMSCOPE;
    CMISSync cmisSync = s_bestEffort;
    CMISObjectList* objList = NULLCMISOBJECTLIST;

    if (argc < 2)
	goto usage;

    TclRMIBAgent* agent;
    if (!(agent = FindTclRMIBAgent(interp, argv[1])))
	CALLBACK_ERROR

    if (agent->getMsd() == UNCONNECTED) {
	sprintf(interp->result, "agent is not in association!");
	CALLBACK_ERROR
    }

    if (argc > 2) {
	for (i = 2; i < argc; i++)
	    switch (state) {
	    case 0:
		if (*argv[i] == '-' && strlen(argv[i]) == 2)
		    switch (*(argv[i]+1)) {
		    case 'c':
			state = 1;
			break;
		    case 'i':
			state = 2;
			break;
		    case 's':
			state = 3;
			break;
		    case 'f':
			state = 4;
			break;
		    case 'm':
			state = 5;
			break;
		    default:
			goto usage;
		    }
		else
		    goto usage;
		break;

	    case 1:
		objClass = argv[i];
		state = 0;
		break;

	    case 2:
		objInst = argv[i];
		state = 0;
		break;

	    case 3:
		scope = argv[i];
		if (i < argc-1 && *argv[i+1] != '-')
		    sync = argv[++i];
		state = 0;
		break;

	    case 4:
		if (parseFilterArg(argc-i, &argv[i], &filter, &j) == NOTOK)
		    goto usage;
		i += j-1;
		state = 0;
		break;

	    case 5:
		mgrId = argv[i];
		state = 0;
		break;
	    }
    }

    if (!objInst) {
	sprintf(interp->result, "no object instance specified!");
	if (filter)
	    free(filter);
	CALLBACK_ERROR
    }

    if (mgrId && !(mgr = FindTclRMIBManager(interp, mgrId))) {
	if (filter)
	    free(filter);
	CALLBACK_ERROR
    }

    if (mgr && !mgr->rproc) {
	sprintf(interp->result, "manager has no appropriate callback proc!");
	if (filter)
	    free(filter);
	CALLBACK_ERROR
    }

    if (scope && (cmisScope = str2scope(scope)) == NULLMSCOPE) {
	sprintf(interp->result, "invalid scope %s", scope);
	if (filter)
	    free(filter);
	CALLBACK_ERROR
    }

    if (sync && (cmisSync = str2sync(sync)) == s_invalid) {
	sprintf(interp->result, "invalid sync %s", sync);
	if (filter)
	    free(filter);
	if (cmisScope)
	    free((char*) cmisScope);
	CALLBACK_ERROR
    }

    result = agent->Delete(objClass, objInst,
			cmisScope ? cmisScope->sc_type : Sc_BaseObject,
			cmisScope ? cmisScope->sc_level : 0, filter,
			NULLMACCESS, cmisSync, objList, mgr);

    if (filter)
	free(filter);

    if (cmisScope)
	free((char*) cmisScope);

    if ((!mgr && !objList) || (mgr && result == NOTOK)) {
	sprintf(interp->result, "Delete operation failed!");
	CALLBACK_ERROR
    }

    if (!mgr) {
	AssembleMOList(interp, argv[1], agent->getAgentName(),
			agent->getHostName(), objList);
	delete objList;
    }
    else {
	sprintf(interp->result, "%d", result);
    }

    CALLBACK_OK

usage:
    sprintf(interp->result, "usage: m_delete agentId ?-c class? ?-i instance?\n\
\t?-s scope ?sync?? ?-f filter? ?-m mgrId?");
    if (filter)
	free(filter);
    CALLBACK_ERROR
}


DEFINE_CALLBACK (RmibNotifyCmd)
{
    char* objClass = NULLCP;
    char* objInst = NULLCP;
    char* eventType = NULLCP;
    char* timeExpr = NULLCP;
    char* filter = NULLCP;
    int i, j, state = 0;
    Bool notifyBroken = False, allEvent = False, cancel = False;

    if (argc < 3)
	goto usage;

    TclRMIBAgent* agent;
    if (!(agent = FindTclRMIBAgent(interp, argv[1])))
	CALLBACK_ERROR

    TclRMIBManager* mgr;
    if (!(mgr = FindTclRMIBManager(interp, argv[2])))
	CALLBACK_ERROR

    if (argc > 3) {
	for (i = 3; i < argc; i++)
	    switch (state) {
	    case 0:
		if (*argv[i] == '-' && strlen(argv[i]) == 2)
		    switch (*(argv[i]+1)) {
		    case 'c':
			state = 1;
			break;
		    case 'i':
			state = 2;
			break;
		    case 'e':
			state = 3;
			break;
		    case 't':
			state = 4;
			break;
		    case 'f':
			state = 5;
			break;
		    case 'a':
			allEvent = True;
			break;
		    case 'b':
			notifyBroken = True;
			break;
		    case 's':
			cancel = True;
			break;
		    default:
			goto usage;
		    }
		else
		    goto usage;
		break;

	    case 1:
		objClass = argv[i];
		state = 0;
		break;

	    case 2:
		objInst = argv[i];
		state = 0;
		break;

	    case 3:
		eventType = argv[i];
		state = 0;
		break;

	    case 4:
		if (parseFilterArg(argc-i, &argv[i], &timeExpr, &j) == NOTOK)
		    goto usage;
		i += j-1;
		state = 0;
		break;

	    case 5:
		if (parseFilterArg(argc-i, &argv[i], &filter, &j) == NOTOK)
		    goto usage;
		i += j-1;
		state = 0;
		break;
	    }
    }

    if ((notifyBroken && !mgr->bproc) || ((filter || allEvent ||
	    objClass || objInst || eventType || timeExpr) && !mgr->eproc)) {
	sprintf(interp->result, "manager has no appropriate callback proc!");
	CALLBACK_ERROR
    }

    if (notifyBroken)
	agent->notifyBrokenAssociation(mgr);

    if (filter || allEvent) {
	if (!cancel)
	    agent->receiveEvent(filter, mgr);
	else
	    agent->stopReceiveEvent(filter, mgr);
    }
    else
    if (objClass || objInst || eventType || timeExpr) {
	if (!cancel)
	    agent->receiveEvent(objClass, objInst, eventType, timeExpr, mgr);
	else
	    agent->stopReceiveEvent(objClass, objInst, eventType, timeExpr, mgr);
    }

    if (timeExpr)
	free(timeExpr);

    if (filter)
	free(filter);

    CALLBACK_OK

usage:
    sprintf(interp->result, "usage: m_notify agentId mgrId\n\
\t??-c managedObjectClass? ?-i managedObjectInstance? ?-e eventType?\n\
\t?-t eventTimeExpr? | ?-f filter? | ?-a?? ?-b? ?-s?");
    if (timeExpr)
	free(timeExpr);
    if (filter)
	free(filter);
    CALLBACK_ERROR
}


DEFINE_CALLBACK (RmibScheduleWakeUpsCmd)
{
    int i;
    if (argc < 3)
	goto usage;

    for (i = 2; i < argc; i += 2)
	tclks->schedule(atoi(argv[i]), argv[i-1]);

    if (!tclks->interp)
	tclks->interp = interp;

    CALLBACK_OK

usage:
    sprintf(interp->result, "usage: schedule_wakeups wproc1 intvl1 ..");
    CALLBACK_ERROR
}


DEFINE_CALLBACK (RmibCancelWakeUpsCmd)
{
    if (argc < 2)
	tclks->cancelAll();

    for (int i = 1; i < argc; i++)
	tclks->cancel(argv[i]);

    CALLBACK_OK

usage:
    sprintf(interp->result, "usage: cancel_wakeups wproc ..");
    CALLBACK_ERROR
}


int TclRMIB_Init (Tcl_Interp* interp)
{
    Tcl_CreateCommand(interp, "rmib_agent", RmibAgentCmd,
		(ClientData) 0, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "rmib_manager", RmibManagerCmd,
		(ClientData) 0, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "rmib_destroy", RmibDestroyCmd,
		(ClientData) 0, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "rmib_agent_info", RmibAgentInfoCmd,
		(ClientData) 0, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "m_connect", RmibConnectCmd,
		(ClientData) 0, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "m_disconnect", RmibDisconnectCmd,
		(ClientData) 0, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "m_get", RmibGetCmd,
		(ClientData) 0, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "m_cancel_get", RmibCancelGetCmd,
		(ClientData) 0, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "m_set", RmibSetCmd,
		(ClientData) 0, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "m_action", RmibActionCmd,
		(ClientData) 0, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "m_create", RmibCreateCmd,
		(ClientData) 0, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "m_delete", RmibDeleteCmd,
		(ClientData) 0, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "m_notify", RmibNotifyCmd,
		(ClientData) 0, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "schedule_wakeups", RmibScheduleWakeUpsCmd,
		(ClientData) 0, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "cancel_wakeups", RmibCancelWakeUpsCmd,
		(ClientData) 0, (Tcl_CmdDeleteProc *) NULL);

    tclks = new TclKS();
    coord = new TkCoordinator();

    // load in the application's syntaxes here
    if (initialiseSyntaxes(NULLCP, allSyntaxes) != OK)
	return TCL_ERROR;

    return TCL_OK;
}
