//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// TimeManager.cc - time manager realisation

// By George Pavlou, February 1994


#include <stdlib.h>
#include "msap.h"
#include "Error.h"
#include "RMIBAgent.h"
#include "SmiSntx.h"
#include "TimeManager.h"

#if	defined(SOLARIS) || defined(__linux__)
extern "C" {
int nint(double);
}
#endif


TimeManager* TimeManager::_instance;


TimeManager::TimeManager (int timeMode)
{
    _simTimeIntvl = DEFAULT_INTVL;
    _simulatedTime = 0.0;
    _timeMode = timeMode;
    _simulatedTimeOID = NULLOID;
    _rmibAgent = NULL;

    _instance = this;
}


int  TimeManager::initialise (RMIBAgent* rmibAgent)
{
    // request the attributeValueChange event from the systemSim class

    if (rmibAgent -> receiveEvent("systemSim",
			NULLCP, "attributeValueChange", NULLCP, this) == NOTOK)
	return NOTOK;
    _rmibAgent = rmibAgent;
    _simulatedTimeOID = name2oid("simulatedTime");

    return OK;
}


int  TimeManager::eventNotification (char*, char*, char* eventype, char*,
						Attr* report, RMIBAgent*)
{
    static int count = 0;
    static double prevTime = 0;
    KsWakeUp* wakeUp;

    if (strcasecmp(eventype, "attributeValueChange") != 0)
	return NOTOK;

    type_SMI_AttrValChangeInfo* attrValChange =
			(type_SMI_AttrValChangeInfo*) report -> get();


    // check in case the systemSim object emits the attributeValueChange
    // notification for other attributes in addition to simulatedTime

    if (!oid_equal(((MID) attrValChange->attrValChangeDefinition
		->attrValChangeDefinition_member->attributeId)->mid_global,
		_simulatedTimeOID))
	return OK;

    double* newTime = real_dec(attrValChange->attrValChangeDefinition
			->attrValChangeDefinition_member->newAttrValue);
    _simulatedTime = *newTime;
    free((char*) newTime);

    switch (count) {
    case 0:	// very first time: we operate in simulated time
	// we still let the coordinator handle alarms (_timeMode = TM_REAL)
	// as we don't know yet the reporting interval to round up intervals
	prevTime = _simulatedTime;
	++count;
	break;

    case 1:	// second time: we know the reporting interval now
    {
        KsWakeUpList* coordWakeUps = Coordinator::getInstance()
							->passWakeUpControl();
        _simTimeIntvl = _simulatedTime - prevTime;
	while (wakeUp = (KsWakeUp*) coordWakeUps->get()) {
	    // round up wake-up interval
	    if (nint(wakeUp->_period) % nint(_simTimeIntvl) != 0)
		wakeUp->_period =
			nint(wakeUp->_period) / nint(_simTimeIntvl) +
			nint(_simTimeIntvl);
	    // put all alarms their period away from now
	    wakeUp->_wakeUpTime = nint(_simulatedTime) + wakeUp->_period;
	    _wakeUps.append(wakeUp);
	}
    }

	_timeMode = TM_SIMULATED;	// we take care of wake-ups now
	advise(NULLCP, "TimeManager: we operate in simulated time, \
interval is %.1f\n", _simTimeIntvl);

	++count;
	break;

    default:	// finally we are in business...
	if ((wakeUp = (KsWakeUp*) _wakeUps.first()) &&
				wakeUp->_wakeUpTime == nint(_simulatedTime))
	    serviceWakeUp();
	break;
    }

    return OK;
}


void  TimeManager::updateTime (double simulatedTime)
{
    _simulatedTime = simulatedTime;
    KsWakeUp* wakeUp = (KsWakeUp*) _wakeUps.first();
    if (wakeUp && wakeUp->_wakeUpTime == nint(_simulatedTime))
	serviceWakeUp();
}


int  TimeManager::serviceWakeUp ()
{
    KsWakeUp *wakeUp = (KsWakeUp*) _wakeUps.get(), *nextWakeUp;
    int nReportObjs = Coordinator::getInstance()->getNReportObjs();

    // wake up the relevant knowledge source
    if (wakeUp->_ifReportObjs == False || nReportObjs > 0)
	(wakeUp->_ks) -> wakeUp(wakeUp->_token);


    // arrange for the next wake up for this KS
    // (if a periodic one was requested)
    if (!wakeUp->_onceOnly) {
	wakeUp->_wakeUpTime += wakeUp->_period;
	_wakeUps.insert(wakeUp);
    }
    else
	delete wakeUp;

    if ((nextWakeUp = (KsWakeUp*) _wakeUps.first()) == NULL)
        // no more wake-ups for the time being...
        return OK;

    long nextAlarmIntvl = nextWakeUp->_wakeUpTime - nint(_simulatedTime);
    if (nextAlarmIntvl <= 0)
	serviceWakeUp();	// wake-ups scheduled at roughly the same time

    return OK;
}


int  TimeManager::scheduleWakeUps (KS* ks, char* token, long intvl,
					Bool ifReporting, Bool onceOnly)
{
    if (intvl <= 0 || intvl % nint(_simTimeIntvl) != 0) {
	advise(NULLCP, "TimeManager: scheduleWakeUps: interval should be a \
positive integer multiple of %d!",
			_simTimeIntvl);
	return NOTOK;
    }

    _wakeUps.insert(new KsWakeUp(ks, token ? token : "",
		intvl, nint(_simulatedTime) + intvl, ifReporting, onceOnly));
    return OK;
}


int  TimeManager::cancelWakeUps (KS* ks, char* token)
{
    KsWakeUp wakeUp(ks, token ? token : ""), *cancelWakeUp;

    if ((cancelWakeUp = _wakeUps.remove(&wakeUp)) == NULL) {
	fprintf(stderr, "TimeManager: cannot find wake-up to cancel\n");
	return NOTOK;
    }

    delete cancelWakeUp;
    return OK;
}

