/*
 * Copyright (c) 1994 University College London
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Department of Computer Science, University College London.
 * The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

/*
** Autnetication dialogue exchange mechanism
*/

/*
** Saleem N. Bhatti
** March 1994
*/

/*
** To use the authentication, the follwoing call should be made after the
** initialiseSyntaxes() call:
**
**    void initialiseOsisec(int argc, char **argv);
**
** The information required for authenticated association is passed to one of
** the following function calls in AuthAssocIntegrityInfo structure:
**
**    int makeCredentials(AuthAssocIntegrityInfo *info, PE *pe);
**    int getCredentials(PE pe, AuthAssocIntegrityInfo **info);
**
**    int makeAcseUserInfo(AuthAssocIntegrityInfo *info, External **external);
**    int getAcseUserInfo(External *external, AuthAssocIntegrityInfo **info);
**
** makeAcseUserInfo() and getAcseUserInfo() make use of makeCredentials() and
** getCredentials() respectively, wrapping up the 'magic PE' produced in the
** EXTERNAL syntax. Alternatively, an AuthAssocIntegrityInfo structure can be
** passed to as a MInitOption to:
**
**    int m_initialise(char *agent, char *host, MInitOption **opt)
**
** as part of the NULL terminated opt array. m_initialise() will call
** makeAcseUserInfo().
**
** As a utility, the follwoing function calls are available for reading the
** OSISEC ~/.authrc file.
**
**    FILE *openAuthrc();
**    int rewindFile(FILE *fp);
**    int readAuthrc(FILE *fp, char *name, char *value);
**    void closeAuthrc(FILE *fp);
*/

#include <stdio.h>
#if defined(__GNUG__)
#include <std.h>
#else
#include <stdlib.h>
#endif
#include <sys/types.h>
#include <unistd.h>
#include <string.h>

#if defined(__STDC__) || defined(__cplusplus)
#include <stdarg.h>
#else
#include <varargs.h>
#endif

#include "mauth.h"
#include "CredSntx.h"

#if defined(SECURITY)

#include "quipu/bind.h"         /* for ds_bind_arg */
#include "quipu/DAS-types.h"    /* {en,de}code_DAS_DirectoryBindArgument() */
#include "osisec/serv-access.h" /* in OSISEC/h/ for macros */
#include "osisec/osisec.h"      /* All the stuff I don't know about ... ? */

/* Handy #defines */
extern int sys_nerr;
extern char *sys_errlist[];
extern int errno;

#define ERROR(_s) \
    fprintf(stderr, "%s: errno %d - %s\n", _s, errno, errno < sys_nerr ? sys_errlist[errno] : "???")

#define NEW(_s) (_s *) calloc(1, sizeof(_s))

FILE *
openAuthrc()
{
    register
    char *home = getenv("HOME");
    register
    FILE *fp;
    char authrc[256];

    if (home == (char *) 0) {
        fprintf(stderr, "openAuthrc() no home directory found\n");
        return (FILE *) 0;
    }

    sprintf(authrc, "%s/.authrc", home);

    if ((fp = fopen(authrc, "r")) == (FILE *) 0) {
        ERROR("openAuthrc(): open()");
        return (FILE *) 0;
    }

    return fp;
}


int
#if defined(__STDC__) || defined(__cplusplus)
rewindFile(FILE *fp)
#else
rewindFile(fp)
FILE *fp;
#endif
{
    if (lseek(fileno(fp), (off_t) 0, SEEK_SET) == (off_t) -1) {
        ERROR("rewindFile()");
        return -1;
    }
}

static
char *S_authrcNames[] = { /* From OSISEC manual, Section 4.3 */
    "authority",
    "username",
    "password",
    "secret_key",
    "osisec", /* Many instances */
    "dsap", /* Many instances */
    "isode", /* Many instances */
    "salt",
    "cit",
    "certificate",
    "crosscerts", /* Many instances */
    "usercert", /* Many instances */
    "userinfo", /* Many instances */
    "cainfo", /* Many instances */
    "dsacrosscerts", /* Many instances */
    "key_directory",
    "public_directory",
    "private_directory",
    "issued_directory",
    "sievefile",
    "dsa", /* This is not part of OSISEC; it was put in by me! */
    (char *) 0
};

int
#if defined(__STDC__) || defined(__cplusplus)
readAuthrc(FILE *fp,
           char *name,
           char *value)
#else
readAuthrc(fp, name, value)
FILE *fp;
char *name;
char *value;
#endif
{
    register
    char **np, *cp;
    char line[2048]; /* I hope this is enough! */

    value[0] = '\0';

    /* Known name? */
    for(np = S_authrcNames; *np != (char *) 0; ++np) {
        if (strcmp(name, *np) == 0)
            break;
    }
    if (*np == (char *) 0)
        return -1;

    /* Get value */
    while(fgets(line, sizeof(line), fp) == line) {

        char n[64];

        if (line[0] == '#')
            continue; /* comment */

        if ((cp = strchr(line, '\n')) != (char *) 0)
            *cp = '\0';

        /*
        ** Check the variable name first just in case the user has not
        ** passed a big enough buffer, which could result in writing beyond
        ** the array bounds.
        */
        if (sscanf(line, " %63[^: ] :", n) != 1)
            continue;

        if (strcmp(n, name) != 0)
            continue;

        cp = line + (strlen(n));

        return (sscanf(cp, " : %[^\n]", value) == 1 ? 0 : -1);
    }

    value[0] = '\0';
    return -1;
}

void
#if defined(__STDC__) || defined(__cplusplus)
closeAuthrc(FILE *fp)
#else
closeAuthrc(fp)
FILE *fp;
#endif
{ fclose(fp); }

AuthAssocIntegrityInfo *
AuthAssocIntegrityInfo_make()
{ return NEW(AuthAssocIntegrityInfo); }

void
#if defined(__STDC__) || defined(__cplusplus)
AuthAssocIntegrityInfo_free(AuthAssocIntegrityInfo *v)
#else
AuthAssocIntegrityInfo_free(v)
AuthAssocIntegrityInfo *v;
#endif
{
    if (v == (AuthAssocIntegrityInfo *) 0)
        return;

    if (v->name != (char *) 0)
        free(v->name);
    if (v->peer != (char *) 0)
        free(v->peer);
    if (v->ca != (char *) 0)
        free(v->ca);
    if (v->dsa != (char *) 0)
        free(v->dsa);

    free(v);
}

/* Utils ... probably should be in msap */
#define MAKE_EXTERNAL_ASN1(_pe, _info) /* (PE _pe, External *_info) */ \
{ if (_pe != (PE) 0) \
      _info = external_build(EXTERN_ASN1_TYPE, (char *) 0, 0, _pe, (struct qbuf *) 0); \
  else _info = (External *) 0; }

#define GET_EXTERNAL_ASN1(_info, _pe) /* (External *_info, PE _pe) */ \
{ if (_info != (External *) 0 && \
      _info->encoding != (struct choice_UNIV_0 *) 0 && \
      _info->encoding->offset == EXTERN_ASN1_TYPE) \
      _pe = _info->encoding->un.single__ASN1__type; \
  else _pe = (PE) 0; }

static
char *S_AUTH_errorMessages[] = {
    "No ACSE user info",
    "Authentication failed",
    "Could not perform Public Key encryption/decryption",
    "Error accessing X.500 directory service",
    "Could not read users private info from ~/.authrc",
    "General error",
    "No error"
};

static char **S_AUTH_msg = &S_AUTH_errorMessages[6];

char *
#if defined(__STDC__) || defined(__cplusplus)
authErrorMessage(int e)
#else
authErrorMessage(e)
int e;
#endif
{
    static
    char number[20];

    if (e > 0 || e < -6) {
        sprintf(number, "%d", e);
        return number;
    }
    else return S_AUTH_msg[e];
}

int
#if defined(__STDC__) || defined(__cplusplus)
makeCredentials(AuthAssocIntegrityInfo *info,
                 PE *pe)
#else
makeCredentials(info, pe)
AuthAssocIntegrityInfo *info;
PE *pe;
#endif
{
    register
    char *name = (char *) 0, *peer = (char *) 0, *ca = (char *) 0,
         *dsa = (char *) 0, *bp = (char *) 0;

    struct GenericPublicKey *publicKey = (struct GenericPublicKey *) 0;
    struct GenericParameters *params = (struct GenericParameters *) 0;

    char *keyValue = (char *) 0;
    int keyLength = 0;
    struct signature *sig;
    struct alg_id *alg;

    char buffer[1024];
    DN dn = (DN) 0, peerDn = (DN) 0;
    PE token = (PE) 0;
    struct type_Cred_SessionCredentials credentials;

    *pe = (PE) 0;

    /*
    ** Get the user name; default: read from ~/.authrc
    */
    if (info->name == (char *) 0 || info->ca == (char *) 0) {

        register
        FILE *fp;

        if ((fp = openAuthrc()) == (FILE *) 0) {
#if defined(DEBUG)
            fprintf(stderr, "makeCredentials(): openAuthrc()\n");
#endif
            return AUTH_error;
        }

        if (info->name == (char *) 0) {
            if (readAuthrc(fp, "username", buffer) != 0) {
#if defined(DEBUG)
                fprintf(stderr, "makeCredentials(): readAuthrc(username)\n");
#endif
                closeAuthrc(fp);
                return AUTH_privateInfo;
            }
            name = strdup(buffer);
        }
        else name = strdup(info->name);

        if (info->ca == (char *) 0) { /* This is not used for now! */
            if (readAuthrc(fp, "authority", buffer) != 0) {
#if defined(DEBUG)
                fprintf(stderr, "makeCredentials(): readAuthrc(authority)\n");
#endif
                free(name);
                closeAuthrc(fp);
                return AUTH_privateInfo;
            }
            ca = strdup(buffer);
        }
        else ca = strdup(info->ca);

        if (info->dsa == (char *) 0) {
            if (readAuthrc(fp, "dsa", buffer) != 0) {
#if defined(DEBUG)
                fprintf(stderr, "makeCredentials(): readAuthrc(dsa)\n");
#endif
                free(ca);
                free(name);
                closeAuthrc(fp);
                return AUTH_privateInfo;
            }
            dsa = strdup(buffer);
        }
        else dsa = strdup(info->dsa);

        closeAuthrc(fp);
    }

    /*
    ** Make the encrypted session key
    */
    /* Name of peer */
    peer = strdup(info->peer == (char *) 0 ? name : info->peer);
    if ((peerDn = str2dn(peer)) == (DN) 0) {
#if defined(DEBUG)
        fprintf(stderr, "makeCredentials(): str2dn(%s)\n", peer);
#endif
        free(peer);
        free(dsa);
        free(ca);
        free(name);
        return AUTH_error;
    }
#if defined(DEBUG)
    fprintf(stderr, "name : %s\n", name);
    fprintf(stderr, "peer : %s\n", peer);
    fprintf(stderr, "ca   : %s\n", ca);
    fprintf(stderr, "dsa  : %s\n", dsa);
{
    register
    int z;

    fprintf(stderr, "key  : ");
    for(z = 0; z < 8; ++z)
        fprintf(stderr, "%02x", (u_int) ((u_char) info->sessionKey[z]));
    fprintf(stderr, "\n\n");
}
#endif
    free(peer);

    /* Get the Public Key info */
    if (getPublicKey(peerDn, dsa, &publicKey, &params) != OK) {
#if defined(DEBUG)
        fprintf(stderr, "makeCredentials(): getPublicKey()\n");
#endif
        dn_free(peerDn);
        free(dsa);
        free(ca);
        free(name);
        return AUTH_x500Error;
    }
    free(dsa);

    /*
    ** We want to send the sessionKey protected from inspection by
    ** anyone. To do this we must encrypt it with the peer's publicKey
    ** so that it can be decrypted only by the peer. As the sessionKey is an
    ** arbitrary 64bit value, someone could forge a nonsense value. To
    ** prevent this, the whole token is eventually signed by the sender.
    */
    /*
    ** Purify reports an abr of 1 in str2bitstr() and so we copy the key
    ** to a buffer with one more byte than we actually need.
    **
    ** Does not seem to make much difference, so take it out for now.
    */
#if 0
    bp = (char *) malloc(9);
    memcpy(bp, info->sessionKey, 8);
    bp[8] = (char) 0;
#endif
    bp = info->sessionKey;
    /* Encrypt */
    if (STRINGENCRYPT(bp, 8, &keyValue, &keyLength, publicKey, params) != OK) {
#if defined(DEBUG)
        fprintf(stderr, "makeCredentials(): STRINGENCRYPT()\n");
#endif
        /* free(bp); */
        dn_free(peerDn);
        free(ca);
        free(name);
        return AUTH_cryptError;
    }
    /* free(bp); */

    if ((dn = str2dn(name)) == (DN) 0) {
#if defined(DEBUG)
        fprintf(stderr, "makeCredentials(): str2dn(%s)\n", name);
#endif
        free(keyValue);
        dn_free(peerDn);
        free(ca);
        free(name);
        return AUTH_error;
    }
    free(name);

    /*
    ** Make the ACSE token to be used for authenticating the user
    */
    token = MKACSETOKEN(dn);
    if (token == (PE) 0) {
#if defined(DEBUG)
        fprintf(stderr, "makeCredentials(): MKACSETOKEN()\n");
#endif
        free(keyValue);
        dn_free(peerDn);
        free(ca);
        return AUTH_cryptError;
    }
    dn_free(dn);
    free(ca);

    /*
    ** Make the magic PE
    */
    /* Session token */
    credentials.token = NEW(struct type_Cred_SessionToken);

    /* Session key */
    credentials.token->sessionKey = NEW(struct type_Cred_SessionKey);
    credentials.token->sessionKey->keyId = oid_cpy(str2oid(SESSIONKEY_OID));
    credentials.token->sessionKey->key = oct2prim(keyValue, keyLength);
    free(keyValue);

    /* Recipient name */
    credentials.token->recipient = (struct type_IF_DistinguishedName *) peerDn;

    /* Signature algorithm identifier */
    alg = NEW(struct alg_id);
    alg->algorithm = oid_cpy(str2oid(MD2WITHRSA_OID));
    alg->asn = (PE) 0;
    alg->p_type = ALG_PARM_ABSENT;
    credentials.token->algorithm = (struct type_AF_AlgorithmIdentifier *) alg;

    /* Signature */
    sig = SIGNED((char *) credentials.token, _ZSessionTokenCred, &_ZCred_mod);
    if (sig == (struct signature *) 0) {
        SessionToken_free(credentials.token);
        free(keyValue);
        dn_free(peerDn);
        free(ca);
        return AUTH_cryptError;        
    }
    credentials.signature = strb2bitstr(sig->encrypted, sig->n_bits,
                                        PE_CLASS_UNIV, PE_PRIM_BITS);

    /* Credentials */
    credentials.credentials = NEW(struct type_Cred_MsapCredentials);
    credentials.credentials->credentialId = oid_cpy(str2oid(CREDENTIAL_OID));
    credentials.credentials->credentials = token;

    *pe = SessionCredentials_encode(&credentials);

    /* Free it all up */
    sig_free(sig);
    MsapCredentials_free(credentials.credentials);
    SessionToken_free(credentials.token);
    pe_free(credentials.signature);

#if defined(DEBUG)
    if (*pe == (PE) 0)
        fprintf(stderr, "makeCredentials(): SessionCredentials_encode()\n");
#endif
    return (*pe == (PE) 0 ? AUTH_error : 0);
}


int
#if defined(__STDC__) || defined(__cplusplus)
getCredentials(PE pe, AuthAssocIntegrityInfo **info)
#else
getCredentials(pe, info)
PE pe;
AuthAssocIntegrityInfo **info;
#endif
{
    register
    struct type_Cred_MsapCredentials *mc =
        (struct type_Cred_MsapCredentials *) 0;
    register
    struct type_Cred_SessionToken *st =
        (struct type_Cred_SessionToken *) 0;
    register
    struct type_Cred_SessionKey *sk = (struct type_Cred_SessionKey *) 0;
    register
    struct type_Cred_SessionCredentials *credentials =
        (struct type_Cred_SessionCredentials *) 0;
    DN dn = (DN) 0;
    struct ds_bind_arg *token = (struct ds_bind_arg *) 0;
    struct signature sig;
    struct alg_id *alg;

    char name[512], ca[512], dsa[512], *sessionKey = (char *) 0;
    int sessionKeyLength = 0;
    register
    char *peer = (char *) 0, *bp = (char *) 0;
    register
    int e;
    
    *info = (AuthAssocIntegrityInfo *) 0;

    credentials = SessionCredentials_decode(pe);
    if (credentials == (struct type_Cred_SessionCredentials *) 0) {
#if defined(DEBUG)
        fprintf(stderr, "getCredentials(): SessionCredentials_decode()\n");
#endif
        return AUTH_error;
    }

    /*
    ** Verify credentials -- ACSE token
    */
    mc = credentials->credentials;
    if (mc != (struct type_Cred_MsapCredentials *) 0) {

        register
        FILE *fp;
        register
        char *id;

        /*
        ** Get authority name and DSA name
        */
        if ((fp = openAuthrc()) == (FILE *) 0) {
#if defined(DEBUG)
            fprintf(stderr, "getCredentials(): openAuthrc()\n");
#endif
            SessionCredentials_free(credentials);
            return AUTH_error;
        }

        if (readAuthrc(fp, "username", name) != 0) {
#if defined(DEBUG)
            fprintf(stderr, "getCredentials(): readAuthrc(name)\n");
#endif
            closeAuthrc(fp);
            SessionCredentials_free(credentials);
            return AUTH_privateInfo;
        }

        if (readAuthrc(fp, "authority", ca) != 0) {
#if defined(DEBUG)
            fprintf(stderr, "getCredentials(): readAuthrc(authority)\n");
#endif
            closeAuthrc(fp);
            SessionCredentials_free(credentials);
            return AUTH_privateInfo;
        }

        if (readAuthrc(fp, "dsa", dsa) != 0) {
#if defined(DEBUG)
            fprintf(stderr, "makeCredentials(): readAuthrc(dsa)\n");
#endif
            closeAuthrc(fp);
            SessionCredentials_free(credentials);
            return AUTH_privateInfo;
        }

        closeAuthrc(fp);

        /*
        ** Check credentials
        */
        st = credentials->token;
        /* Check that we are who the peer expects us to be! */
        if (st->recipient != (struct type_IF_DistinguishedName *) 0) {
            register
            char *dnStr = dn2str(st->recipient);

            if (dnStr == (char *) 0 || strcmp(dnStr, name) != 0) {
#if defined(DEBUG)
                if (dnStr == (char *) 0)
                    fprintf(stderr, "getCredentials(): dn2str() recipient\n");
                else {
                    fprintf(stderr, "getCredentials(): I am\n    '%s'\n", name);
                    fprintf(stderr, "but peer expects\n    '%s'\n", dnStr);
                }
#endif
                if (dnStr != (char *) 0)
                    free(dnStr);
                SessionCredentials_free(credentials);
                return AUTH_error;
            }

            free(dnStr);
        }
        else { /* This should not happen */
#if defined(DEBUG)
            fprintf(stderr, "getCredentials(): peer did not send a DN\n");
#endif
            SessionCredentials_free(credentials);
            return AUTH_error;
        }

        /* Check type */
        id = sprintoid(mc->credentialId);
        if (strcmp(id, CREDENTIAL_OID) != 0) {
#if defined(DEBUG)
            fprintf(stderr, "getCredentials(): unknown OID %s\n", id);
#endif
            SessionCredentials_free(credentials);
            return AUTH_error;
        }

        /*
        ** CKACSETOKEN() just decodes the credentials and does CKBIND().
        ** As I need to decode the credentials anyway to get the peer
        ** DN, I may as well use CKBIND() and not CKACSETOKEN().
        */
        if (decode_DAS_DirectoryBindArgument(mc->credentials, 1, (int *) 0,
            (char **) 0, &token) != OK) {
#if defined(DEBUG)
            fprintf(stderr, "getCredentials(): decode_DAS_DirectoryBindArgument() credentials\n");
#endif
            SessionCredentials_free(credentials);
            return AUTH_error;
        }

        /* Authenticate peer */
        if ((dn = str2dn(ca)) == (DN) 0) {
#if defined(DEBUG)
            fprintf(stderr, "getCredentials(): str2dn(%s)\n", ca);
#endif
            bind_arg_free(token);
            free((void *) token);
            SessionCredentials_free(credentials);
            return AUTH_error;
        }

        switch(CKBIND(token, dn)) {
#if 0
        /*
        ** The OSISEC manual (Section 6.6.3, page 70) says that
        ** the return values from CKASCETOKEN()/CKBIND() should be:
        */
        case DSE_NOERROR:       e = 0;               break;
        case DSE_SERVICEERROR:  e = AUTH_x500Error;  break;
        case DSE_SECURITYERROR: e = AUTH_authFailed; break;
        default:                e = AUTH_error;      break; /* impossible */
#endif

        /*
        ** ... but looking at the code in OSISEC/osisec/common/dstok.c
        ** I find:
        */
        case OK: e = 0;              break;
        default: e = AUTH_authFailed; break;

        /* ... hmmm */
        }
        dn_free(dn);

        if (e != 0) {
#if defined(DEBUG)
            fprintf(stderr, "getCredentials(): CKBIND() credentials\n");
#endif
            bind_arg_free(token);
            free((void *) token);
            SessionCredentials_free(credentials);
            return e;
        }
    }

    /*
    ** Get sessionKey
    */
    sk = st->sessionKey;
    if (sk != (struct type_Cred_SessionKey *) 0) {

        register
        char *keyValue = (char *) 0, *id;
        int keyLength = 0;

        struct GenericPublicKey *publicKey = (struct GenericPublicKey *) 0;
        struct GenericParameters *params = (struct GenericParameters *) 0;

        /* Check type */
        id = sprintoid(sk->keyId);
        if (strcmp(id, SESSIONKEY_OID) != 0) {
#if defined(DEBUG)
            fprintf(stderr, "getCredentials(): unknown OID %s\n", id);
#endif
            bind_arg_free(token);
            free((void *) token);
            SessionCredentials_free(credentials);
            return AUTH_error;
        }

        /* Get value */
        keyValue = prim2str(sk->key, &keyLength);

        /*
        ** Purify reports an abr of 1 in str2bitstr() and so we copy the key
        ** to a buffer with one more byte than we actually need.
        **
        ** Does not seem to make much difference, so take it out for now.
        */
#if 0
        bp = (char *) malloc(keyLength + 1);
        memcpy(bp, keyValue, keyLength);
        bp[keyLength] = (char) 0;
#endif
        bp = keyValue;
        /* Decrypt */
        if (STRINGDECRYPT(bp, keyLength, &sessionKey, &sessionKeyLength,
                          SECRETKEY, SECRETPARAMETERS) != OK) {
#if defined(DEBUG)
            fprintf(stderr, "getCredentials(): STRINGDECRYPT()\n");
#endif
            /* free(bp); */
            free(keyValue);
            bind_arg_free(token);
            free((void *) token);
            SessionCredentials_free(credentials);
            return AUTH_cryptError;
        }
        /* free(bp); */
        free(keyValue);

        if (sessionKeyLength != 8) {
#if defined(DEBUG)
            fprintf(stderr, "getCredentials(): key length %d\n",
                    sessionKeyLength);
#endif
            free(sessionKey);
            bind_arg_free(token);
            free((void *) token);
            SessionCredentials_free(credentials);
            return AUTH_error;
        }

        /* Check the signature */
        if (st->algorithm != (struct type_AF_AlgorithmIdentifier *) 0) {
            /* Public Key info */
            if (getPublicKey(token->dba_dn, dsa, &publicKey, &params) != OK) {
#if defined(DEBUG)
                fprintf(stderr, "get(): getPublicKey()\n");
#endif
                free(sessionKey);
                bind_arg_free(token);
                free((void *) token);
                SessionCredentials_free(credentials);
                return AUTH_x500Error;
            }

            id = sprintoid(st->algorithm->algorithm);
            if (strcmp(id, MD2WITHRSA_OID) != 0) {
#if defined(DEBUG)
                fprintf(stderr, "getCredentials(): unknown OID %s\n", id);
#endif
                free(sessionKey);
                bind_arg_free(token);
                free((void *) token);
                SessionCredentials_free(credentials);
                return AUTH_error;
            }

            memset((void *) &sig, 0, sizeof(struct signature));
            alg = (struct alg_id *) st->algorithm;
            sig.alg.algorithm = alg->algorithm;
            sig.alg.asn = alg->asn;
            sig.alg.p_type = alg->p_type;
            sig.encrypted = bitstr2strb(credentials->signature, &sig.n_bits);
            switch(VERIFY((char *) credentials->token, _ZSessionTokenCred,
                       &_ZCred_mod, &sig, publicKey, params,
                       HASHPARAMETERS)) {

            case DSE_NOERROR:       e = 0;               break;
            case DSE_SERVICEERROR:  e = AUTH_x500Error;  break;
            case DSE_SECURITYERROR: e = AUTH_authFailed; break;
            default:                e = AUTH_error;      break;
            }

            if (e != 0) {
#if defined(DEBUG)
                fprintf(stderr, "getCredentials(): VERIFY()\n");
#endif
                free(sig.encrypted);
                free(sessionKey);
                bind_arg_free(token);
                free((void *) token);
                SessionCredentials_free(credentials);
                return e;
            }

            free(sig.encrypted);
        }
    }

    /*
    ** Pass back the info
    */
    peer = dn2str(token->dba_dn);
    if (peer == (char *) 0) {
#if defined(DEBUG)
        fprintf(stderr, "getCredentials(): dn2str() peer\n");
#endif
        free(sessionKey);
        bind_arg_free(token);
        free((void *) token);
        SessionCredentials_free(credentials);
        return AUTH_error;
    }

    *info = AuthAssocIntegrityInfo_make();
    (*info)->peer = peer;
    (*info)->ca = strdup(ca);
    (*info)->dsa = strdup(dsa);
    memcpy((*info)->sessionKey, sessionKey, 8);

#if defined(DEBUG)
    fprintf(stderr, "peer : %s\n", (*info)->peer);
    fprintf(stderr, "ca   : %s\n", (*info)->ca);
    fprintf(stderr, "dsa  : %s\n", (*info)->dsa);
{
    register
    int z;

    fprintf(stderr, "key  : ");
    for(z = 0; z < 8; ++z)
        fprintf(stderr, "%02x", (u_int) ((u_char) (*info)->sessionKey[z]));
    fprintf(stderr, "\n\n");
}
#endif

    free(sessionKey);
    bind_arg_free(token);
    free((void *) token);
    SessionCredentials_free(credentials);

    return 0;
}

int
#if defined(__STDC__) || defined(__cplusplus)
makeAcseUserInfo(AuthAssocIntegrityInfo *info,
                 External **external)
#else
makeAcseUserInfo(info, external)
AuthAssocIntegrityInfo *info;
External **external;
#endif
{
    PE pe;
    register
    int e = AUTH_error;

    *external = (External *) 0;

    if (info == (AuthAssocIntegrityInfo *) 0)
        return AUTH_error;

    if ((e = makeCredentials(info, &pe)) != 0) {
#if defined(DEBUG)
        advise((char *) 0, "makeAcseUserInfo(): makeCredentials()");
#endif
        return e;
    }
    else
        MAKE_EXTERNAL_ASN1(pe, *external);

    return 0;
}

int
#if defined(__STDC__) || defined(__cplusplus)
getAcseUserInfo(External *external,
                AuthAssocIntegrityInfo **info)
#else
getAcseUserInfo(external, info)
External *external;
AuthAssocIntegrityInfo **info;
#endif
{
    PE credentials = (PE) 0;
    register
    int e = AUTH_error;

    *info = (AuthAssocIntegrityInfo *) 0;

    if (external == (External *) 0)
        return AUTH_noUserInfo;

    GET_EXTERNAL_ASN1(external, credentials);

    if (credentials != (PE) 0)
        e = getCredentials(credentials, info);

    if (e != 0)
        advise((char *) 0, "getAcseUserInfo(): getCredentials()");

    return e;
}


/* OSISEC stuff */

/*
** I don't understand most of the following ... just pirated it from Luca
*/
static LLog _pgmc_log = {
    "/tmp/osisec.log",
    NULLCP,
    NULLCP,
    LLOG_NONE,
    LLOG_FATAL,
    -1,
    LLOGNIL,
    NOTOK /* LLOGCLS | LLOGCRT | LLOGZER | LLOGTTY, NOTOK */
};

LLog *pgm_log = &_pgmc_log;
char *pgmname;

extern struct SecurityServices *dsap_security;
extern LLog *log_dsap;
extern LLog *psap_log;

void
#if defined(__STDC__) || defined(__cplusplus)
adios(const char *what, ...)
#else
adios(va_alist)
va_dcl
#endif
{
#if !(defined(__STDC__) || defined(__cplusplus))
    register
    char *what;
#endif
    va_list args;

#if defined(__STDC__) || defined(__cplusplus)
    va_start(args, what);
#else
    va_start(args);
    what = va_arg(args, char *);
#endif
    _ll_log(pgm_log, LLOG_FATAL, args);
    va_end(args);

    _exit(1);
}


static void
#if defined(__STDC__) || defined(__cplusplus)
dsapinit(int *argc,
         char **vec)
#else
dsapinit (argc, vec)
int *argc;
char **vec;
#endif
{
    int       argp, narg;
    register char   *ap;
    char  **argptr,
    *args[10];

    argp = 0;
    narg = 1;
    args[argp++] = pgmname;
    for (argptr = vec, argptr++; ap = *argptr; argptr++, narg++) {
        if (*ap == '-')
            switch (*++ap) {
            case 'T':
                if ((ap = *++argptr) == NULL || *ap == '-')
                    break;
                args[argp++] = "-T";
                args[argp++] = ap;
                shuffle_up ((*argc)--, vec, narg);
                shuffle_up ((*argc)--, vec, narg);
                break;

            case 'i':
                if ((ap = *++argptr) == NULL || *ap == '-')
                    break;
                args[argp++] = "-c";
                args[argp++] = ap;
                shuffle_up ((*argc)--, vec, narg);
                shuffle_up ((*argc)--, vec, narg);
                break;

            case 't':
                if ((ap = *++argptr) == NULL || *ap == '-')
                    break;
                args[argp++] = "-t";
                args[argp++] = ap;
                shuffle_up ((*argc)--, vec, narg);
                shuffle_up ((*argc)--, vec, narg);
                break;

            default:
                continue;
            }

        break;
    }
    args[argp] = NULLCP;

    dsap_init (&argp, (argptr = args, &argptr));
}

static void
#if defined(__STDC__) || defined(__cplusplus)
osisecinit(int *argc,
           char **vec)
#else
osisecinit (argc, vec)
int *argc;
char **vec;
#endif
{
    int       argp, narg;
    register char   *ap;
    char  **argptr,
    *args[10];

    narg = 1;
    argp = 0;
    args[argp++] = pgmname;
    for (argptr = vec, argptr++; ap = *argptr ; argptr++, narg++) {
        if (*ap == '-')
            switch (*++ap) {
            case 'A':
                if ((ap = *++argptr) == NULL || *ap == '-')
                    break;
                args[argp++] = "-A";
                args[argp++] = ap;
                shuffle_up ((*argc)--, vec, narg);
                shuffle_up ((*argc)--, vec, narg);
                break;

            default:
                continue;
            }

        break;
    }
    args[argp] = NULLCP;

    osisec_init (&argp, (argptr = args, &argptr));
}

extern char *isodetcpath; /* This is in the ISODE library */

void
#if defined(__STDC__) || defined(__cplusplus)
initialiseOsisec(int argc,
                 char **argv)
#else
initialiseOsisec(argc, argv)
int argc;
char **argv;
#endif
{
    static
    int done = 0;

    char *args[2];
    int nargs;

    if (done)
        return;
    done = 1;
{
    /*
    ** Pints owed to Kevin for spotting this one ...
    **
    ** There are two versions of struct ds_bind_arg; there's one in
    ** OSISEC/h/quipu/bind.h (which is the one I want) and one other
    ** in both ISODE/h/quipu/bind.h and
    ** OSIMIS/include/isode/quipu/bind.h.
    */
    struct ds_bind_arg test;

    /*
    ** So ...
    **
    ** The next line should cause a compile error if I am picking up
    ** the wrong one.
    */
    test.dba_authdn = test.dba_toauthdn = (DN) 0;
}
    fprintf(stderr, "initialising OSISEC ...\n");

#if 0
    /* Register syntaxes: quipu and certificate */
    /* This is in OSIMIS/msap/SmiSntx.c         */
    quipu_syntaxes();
    certificate_syntax();
#endif

    if (pgmname = rindex(argv[0], '/'))
        pgmname++;
    if (pgmname == NULL || pgmname[0] == '\0')
        pgmname = argv[0];

    /* Handle logging: use LLOG_ALL to log all events */
    ll_hdinit (pgm_log, pgmname);       /* Initialise logging */
    log_dsap->ll_stat |= LLOGNIL;
    log_dsap->ll_events |= LLOG_ALL;
    psap_log->ll_stat |= LLOGNIL;
    psap_log->ll_events |= LLOG_ALL;
    log_osisec->ll_stat |= LLOGNIL;
    log_osisec->ll_events |= LLOG_ALL;

    /*
    ** False args for dsapinit() and osisecinit() to prevent any
    ** possible argv flags clashes. his does mean, howver that you
    ** cannot pass any flags to dsap or osisec.
    */
    args[0] = pgmname;
    args[1] = (char *) 0;
    nargs = 1;

    dsapinit(&nargs, args);    /* Initialise dsap */
    osisecinit(&nargs, args);  /* Initialise osisec */

    dsap_security = use_serv_X509(); /* X.509 security */
    /*
    ** DO NOT USE
    **
    **    use_sig_md2withrsa();
    */
    use_sig_rSAMD2withRSAENCRYPTION();
    fprintf(stderr, "OSISEC encryption : Using MD2 with RSA for signatures\n");

    osisec_tailor(OSISECUSERTAILORFILE, pgm_log); /* Set-up logging */

    fprintf(stderr, "OSISEC etcdir     : %s\n", isodetcpath);
    fprintf(stderr, "OSISEC user tailor: %s\n", OSISECUSERTAILORFILE);
    fprintf(stderr, "OSISEC initialised\n");
}

#else /* SECURITY */

/*
** Dummy stubs
*/

int
#if defined(__STDC__) || defined(__cplusplus)
makeCredentials(AuthAssocIntegrityInfo *info, PE *pe)
#else
makeCredentials(info, pe)
AuthAssocIntegrityInfo *info;
PE *pe;
#endif
{
    *pe = (PE) 0;
    fprintf(stderr, "makeCredentials(): security options not enabled\n");
    return 0;
}

int
#if defined(__STDC__) || defined(__cplusplus)
getCredentials(PE pe, AuthAssocIntegrityInfo **info)
#else
getCredentials(pe, info)
PE pe;
AuthAssocIntegrityInfo **info;
#endif
{
    *info = (AuthAssocIntegrityInfo *) 0;
    fprintf(stderr, "getCredentials(): security options not enabled\n");
    return 0;
}

FILE *
openAuthrc()
{
    fprintf(stderr, "openAuthrc(): security options not enabled\n");
    return (FILE *) 0;
}

int
#if defined(__STDC__) || defined(__cplusplus)
makeAcseUserInfo(AuthAssocIntegrityInfo *info,
                 External **external)
#else
makeAcseUserInfo(info, external)
AuthAssocIntegrityInfo *info;
External **external;
#endif
{
    *external = (External *) 0;
    fprintf(stderr, "makeAcseUserInfo(): security options not enabled\n");
    return 0;
}

int
#if defined(__STDC__) || defined(__cplusplus)
getAcseUserInfo( External *external,
                AuthAssocIntegrityInfo **info)
#else
getAcseUserInfo(external, info)
External *external;
AuthAssocIntegrityInfo **info;
#endif
{
    *info = (AuthAssocIntegrityInfo *) 0;
    fprintf(stderr, "getAcseUserInfo(): security options not enabled\n");
    return 0;
}

int
#if defined(__STDC__) || defined(__cplusplus)
rewindFile(FILE *fp)
#else
rewindFile(fp)
FILE *fp;
#endif
{
    fprintf(stderr, "rewindFile(): security options not enabled\n");
    return 0;
}

int
#if defined(__STDC__) || defined(__cplusplus)
readAuthrc(FILE *fp, char *name, char *value)
#else
readAuthrc(fp, name, value)
FILE *fp;
char *name;
char *value;
#endif
{
    value[0] = '\0';
    fprintf(stderr, "readAuthrc(): security options not enabled\n");
    return 0;
}


void
#if defined(__STDC__) || defined(__cplusplus)
closeAuthrc(FILE *fp)
#else
closeAuhtrc(fp)
FILE *fp;
#endif
{
    fprintf(stderr, "closeAuthrc(): security options not enabled\n");
}

void
#if defined(__STDC__) || defined(__cplusplus)
initialiseOsisec(int argc, char **argv)
#else
initialiseOsisec(argc, argv)
int argc;
char **argv;
#endif
{
    fprintf(stderr, "initialiseOsisec(): security options not enabled\n");
}


#endif /* SECURITY */
