/*
 * Copyright (c) 1994 University College London
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Department of Computer Science, University College London.
 * The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

/*
** A simple plug-in interface to MSAP for MD5 based integrity checks
*/

/*
** Saleem N. Bhatti
** February 1994
*/

/*
** The 8 byte MD5 session keys are stored in a static array. The index to the
** array is the value of the file descriptor of the association on which that
** session key is being used. The keys are managed by a set of simple function
** calls, which are fairly self explanatory:
**
**    void initialiseMd5KeySpace();
**    char *getMd5Key(const int fd);
**    int setMd5Key(const int fd, const char *key);
**    int clearMd5Key(const int fd);
**    char *makeMd5Key();
**
** initialiseMd5KeySpace() should be called once to zero the key space for
** use. setMd5Key() will clear any existing keys for fd before setting the new
** value given. setMd5Key() will make a copy of the space pointed to be key.
** setMd5Key() returns 0 on success or -1 if passed a bad fd.  getMd5Key()
** returns a valid pointer if a key exists for that fd or NULL if no key
** exists or a bad value of fd was passed. makeMd5Key() returns a random key
** value. The pointers returned by getMd5Key() and makeMd5Key() should not be
** free()d. clearMd5Key() clears the key for fd, but it's use is not necessary
** as setMd5Key() clears any previous key for fd; clearMd5Key() can be used to
** assist in memory debugging.
*/

/*
** The ROS invokeID value is formed from a XOR the id passed in the MSAP call
** with the first 32bits of the 128bit MD5 value evaluated fro the PDU. The id
** value forms part of a sequence that is generated by a function based on a
** pseudo random number generator -- random(3). This is seeded by the using
** some bits from the MD5 key for that association. As the manager and agent
** in general communicate asynchronously, they can not easily use one shared
** sequence. The agent uses a sequence that is generated by using the first
** 32bits of the MD5 key and the manager seed is the last 32bits of the MD5
** key. In fact the two sequences are selected as to wether you are the
** initiator of the association or the responder. The function calls used are:
**
**    void initialiseIdSpace();
**    int getResponderIdSeed(const int fd);
**    int getInitiatorIdSeed(const int fd);
**    int makeResponderId(const int fd, const int prev);
**    int makeInitiatorId(const int fd, const int prev);
**    void clearId(const int fd);
**
** initialiseIdSpace() should be called once to zero the Id space for use.
** getResponderIdSeed() and getInitiatorIdSeed() return the seed values for
** the responder sequence and initiator sequnce respectively, howvere it is
** not ncessary to use these functions directly: calls to makeResponderId() or
** makeInitiatorId() with prev = 0 will result in the sequnce being
** initilaised. Further, it can be detected when there is not state present
** for the sequence when it is started and so again the sequnce is
** automatically initilaised. makeResponderId() and makeInitiatorId() clear
** any previous state for fd and restart the sequence if prev = 0. clearId()
** clears the Id state for fd, but it's use is not necessary as
** makeResponderId() and makeInitiatorId() clear any previous state for fd;
** clearId() can be used to assist in memory debugging.
**
** To advance a sequence but return to a previous point, P, in the sequence,
** the state of the sequence can be saved and restored by using the interface
** below:
**
**    void saveResponderIdState(const int fd, char **state);
**    void restoreResponderIdState(const int fd, char *state);
**    void saveInitiatorIdState(const int fd, char **state);
**    void restoreInitiatorIdState(const int fd, char *state);
**
** The state returned from a saveResponderIdState() and saveInitiatorIdState()
** is malloc()ed space but should not be free()d as restoreResponderIdState()
** and restoreInitiatorIdState() will use it and it will be freed through the
** normal use of makeResponderId() or makeInitiatorId().
*/

/*
** Once the key has been set-up the, the MD5 hash value for a PE can be
** evaluated by using:
**
**    int makeMd5Value(const int fd, PE pdu, MD5Value *check);
**
** The macro ENCODE_ID() XORs its argument and DECODE_ID() takes the XORed
** value and one of the values that was XORed and returns the other value.
*/

#include <stdio.h>
#if defined(__GNUG__)
#include <std.h>
#else
#include <stdlib.h>
#endif
#include <string.h>
#include <memory.h>

#include "mintegrity.h"
#include "md5.h"

extern int pe2ssdu _a_((PE, char **, int *)); /* in ISODE/psap/pe2ssdu.c */

/*
** Just in case cc complains ...
*/
#if !defined(__STDC__) || !defined(__cplusplus)
char *initstate();
char *setstate();
#endif

/*
** Store keys here ... aarrgh.
*/
static
char *S_md5Keys[FD_SETSIZE];

/*
** Clear the key storage space.
*/
void
initialiseMd5KeySpace()
{
    static
    int flag = 0;

    if (flag) return;
    memset((void *) S_md5Keys, 0, sizeof(char *) * FD_SETSIZE);
    flag = 1;
}


/*
** Get a stored key. Index is the file descriptor for the association.
**
** Return  valid pointer : if there is a key for that fd
**         (char *) 0    : otherwise
*/
char *
#if defined(__STDC__) || defined(__cplusplus)
getMd5Key(const int fd)
#else
getMd5Key(fd)
int fd;
#endif
{
    if (fd < 0) /* actually < 2 */
        return (char *) 0;

    return S_md5Keys[fd];
}

/*
** Store a key. Index is the file descriptor for the association.
** key is copied.
**
** Return  0 : success
**        -1 : Bad fd value
*/
int
#if defined(__STDC__) || defined(__cplusplus)
setMd5Key(const int fd,
          const char *key)
#else
setMd5Key(fd, key)
int fd;
char *key;
#endif
{
    if (fd < 0) /* should be < 2 */
        return -1;

    if (S_md5Keys[fd] == (char *) 0) /* use existing space */
        S_md5Keys[fd] = (char *) malloc(8);

    memcpy((void *) S_md5Keys[fd], (void *) key, 8);

    return 0;
}

/*
** Clear the key for this fd.
*/
void
#if defined(__STDC__) || defined(__cplusplus)
clearMd5Key(const int fd)
#else
clearMd5Key(fd)
int fd;
#endif
{
    if (fd < 0) /* should be < 2 */
        return;

    if (S_md5Keys[fd] != (char *) 0) {
        free((void *) S_md5Keys[fd]);
        S_md5Keys[fd] = (char *) 0;
    }
}

static
long S_keyState[8] = {
    0x495abfe6, 0xa7f5ae73, 0x73f2ef57, 0x65f803d7,
    0xb7360c92, 0xc4e1aae4, 0x4799950f, 0xfe531847
};

char *
makeMd5Key()
{
    static
    char key[8], state[32];
    static
    unsigned int seed = 0;
    register
    int i;
    register
    char *prevState;

    /* Set up a random(ish) seed for the random(ish) number generator */
    /* 42 is the answer ... we can't be wrong! */
    if (seed == 0) {
        seed = (getpid() * 42) ^ time((time_t *) 0);
        memcpy((void *) state, (void *) S_keyState, 32);
        prevState = initstate(seed, state, 32);
    }
    else prevState = setstate(state);

    for(i = 0; i < 8; ++i)
        key[i] = (char) (random() & 0xff);

    setstate(prevState);

    return key;
}


#if defined(__STDC__) || defined(__cplusplus)
const
#endif
static char *S_types[] = {
    "MI_UNKNOWN",
    "MI_UREJECT",
    "MI_PREJECT",
    "MI_ABORT",
    "MI_TERM",
    "MI_ERROR",
    "MI_EVENT_REP",
    "MI_EVENT_REPC",
    "MI_EVENT_REP_RES",
    "MI_LINKED_REPLY",
    "MI_GET",
    "MI_GET_RES",
    "MI_SET",
    "MI_SETC",
    "MI_SET_RES",
    "MI_ACTION",
    "MI_ACTIONC",
    "MI_ACTION_RES",
    "MI_CREATE",
    "MI_CREATE_RES",
    "MI_DELETE",
    "MI_DELETE_RES",
    "MI_CANCEL_GET",
    "MI_EMPTY_RES",
    (char *) 0
};


/*
** Store state for random number generator ... double aargh!
*/
static
char *S_responderStates[FD_SETSIZE];
static
char *S_initiatorStates[FD_SETSIZE];

#define STATE_SIZE 32

static
long S_initState[8] = { /* Values from the Sun manual page for random(3)! */
0x9a319039, 0x32d9c024, 0x9b663182, 0x5da1f342,
0x7449e56b, 0xbeb1dbb0, 0xab5c5918, 0x946554fd
};

/*
** Clear the state storage space.
*/
void
initialiseIdSpace()
{
    static
    int flag = 0;

    if (flag) return;
    memset((void *) S_responderStates, 0, sizeof(char *) * FD_SETSIZE);
    memset((void *) S_initiatorStates, 0, sizeof(char *) * FD_SETSIZE);
    flag = 1;
}

/*
** Get the Responder seed for this assoc. The FIRST 32bits of the sessionKey.
*/

int
#if defined(__STDC__) || defined(__cplusplus)
getResponderIdSeed(const int fd)
#else
getResponderIdSeed(fd)
int fd;
#endif
{
    register
    char *key;
    int i;

    if (fd < 0 || fd >= FD_SETSIZE) /* < 2 */
        return 0;

    if ((key = getMd5Key(fd)) == (char *) 0)
        return 0;

    memcpy((void *) &i, (void *) key, 4);

    return i;
}

/*
** Get the Initiator seed for this assoc. The LAST 32bits of the sessionKey.
*/

int
#if defined(__STDC__) || defined(__cplusplus)
getInitiatorIdSeed(const int fd)
#else
getInitiatorIdSeed(fd)
int fd;
#endif
{
    register
    char *key;
    int i;

    if (fd < 0 || fd >= FD_SETSIZE) /* < 2 */
        return 0;

    if ((key = getMd5Key(fd)) == (char *) 0)
        return 0;

    memcpy((void *) &i, (void *) &(key[4]), 4);

    return i;
}

char G_nullKey[8] = { '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0' };

#define NULL_KEY(fd) \
    (S_md5Keys[fd] != (char *) 0 && \
     memcmp((void *) S_md5Keys[fd], (void *) G_nullKey, 8) == 0)

int
#if defined(__STDC__) || defined(__cplusplus)
makeResponderId(const int fd, const int prev)
#else
makeResponderId(fd, prev)
int fd;
int prev;
#endif
{
    register
    int i, *ip, seed, nullFlag = 0;
    register
    char *state = (char *) 0, *prevState = (char *) 0;

    if (fd < 0 || fd >= FD_SETSIZE)
        return 0;

    if ((seed = getResponderIdSeed(fd)) == 0 || /* fd not in use - no key */
        NULL_KEY(fd))                           /* NULL key for this fd */
        nullFlag = 1;

    state = S_responderStates[fd];

    if (prev == 0 || state == (char *) 0) { /* This is the first value */

        if (state != (char *) 0)
            free(state);
        state = (char *) malloc(STATE_SIZE);

        if (state == (char *) 0) {
            fprintf(stderr, "makeResponderId(): malloc() failed\n");
            return 0;
        }

        if (nullFlag)
            memset((void *) state, 0, STATE_SIZE);
        else
            memcpy((void *) state, (void *) S_initState, STATE_SIZE);

        S_responderStates[fd] = state;
        if (!nullFlag)
            prevState = initstate(seed, state, STATE_SIZE);
    }
    else if (!nullFlag)
        prevState = setstate(state);

    if (nullFlag) {
        ip = (int *) state;
        i = ++(*ip);
    }
    else i = random();

    if (!nullFlag)
        setstate(prevState);

    return i;
}

int
#if defined(__STDC__) || defined(__cplusplus)
makeInitiatorId(const int fd, const int prev)
#else
makeInitiatorId(fd, prev)
int fd;
int prev;
#endif
{
    register
    int i, *ip, seed, nullFlag = 0;
    register
    char *state = (char *) 0, *prevState = (char *) 0;

    if (fd < 0 || fd >= FD_SETSIZE)
        return 0;

    if ((seed = getResponderIdSeed(fd)) == 0 || /* fd not in use - no key */
        NULL_KEY(fd))                           /* NULL key for this fd */
        nullFlag = 1;

    state = S_initiatorStates[fd];

    if (prev == 0 || state == (char *) 0) { /* This is the first value */

        if (state != (char *) 0)
            free(state);
        state = (char *) malloc(STATE_SIZE);

        if (state == (char *) 0) {
            fprintf(stderr, "makeInitiatorId(): malloc() failed\n");
            return 0;
        }

        if (nullFlag)
            memset((void *) state, 0, STATE_SIZE);
        else
            memcpy((void *) state, (void *) S_initState, STATE_SIZE);

        S_initiatorStates[fd] = state;
        if (!nullFlag)
            prevState = initstate(seed, state, STATE_SIZE);
    }
    else if (!nullFlag)
        prevState = setstate(state);

    if (nullFlag) {
        ip = (int *) state;
        i = ++(*ip);
    }
    else i = random();

    if (!nullFlag)
        setstate(prevState);

    return i;
}

void
#if defined(__STDC__) || defined(__cplusplus)
clearId(const int fd)
#else
clearId(fd)
int fd;
#endif
{
    if (fd < 0 || fd >= FD_SETSIZE)
        return;

    if (S_responderStates[fd] != (char *) 0) {
        free(S_responderStates[fd]);
        S_responderStates[fd] = (char *) 0;
    }
    if (S_initiatorStates[fd] != (char *) 0) {
        free(S_initiatorStates[fd]);
        S_initiatorStates[fd] = (char *) 0;
    }
}

void
#if defined(__STDC__) || defined(__cplusplus)
saveResponderIdState(const int fd, char **state)
#else
saveResponderIdState(fd, state)
int fd;
char **state;
#endif
{
    *state = (char *) 0;

    if (fd < 0 || fd >= FD_SETSIZE || state == (char **) 0)
        return;

    if (S_responderStates[fd] == (char *) 0)
        return;

    *state = (char *) malloc(STATE_SIZE);
    if (*state == (char *) 0) {
        fprintf(stderr, "saveResponderIdState(): malloc() failed\n");
        return;
    }
    memcpy((void *) *state, (void *) S_responderStates[fd], STATE_SIZE);
}

void
#if defined(__STDC__) || defined(__cplusplus)
restoreResponderIdState(const int fd, char *state)
#else
restoreResponderIdState(fd, state)
int fd;
char *state;
#endif
{
    if (fd < 0 || fd >= FD_SETSIZE || state == (char *) 0)
        return;

    if (S_responderStates[fd] != (char *) 0)
        free(S_responderStates[fd]);

    S_responderStates[fd] = state;
}

void
#if defined(__STDC__) || defined(__cplusplus)
saveInitiatorIdState(const int fd, char **state)
#else
saveInitiatorIdState(fd, state)
int fd;
char **state;
#endif
{
    *state = (char *) 0;

    if (fd < 0 || fd >= FD_SETSIZE || state == (char **) 0)
        return;

    if (S_initiatorStates[fd] == (char *) 0)
        return;

    *state = (char *) malloc(STATE_SIZE);
    if (*state == (char *) 0) {
        fprintf(stderr, "saveInitiatorIdState(): malloc() failed\n");
        return;
    }
    memcpy((void *) *state, (void *) S_initiatorStates[fd], STATE_SIZE);
}

void
#if defined(__STDC__) || defined(__cplusplus)
restoreInitiatorIdState(const int fd, char *state)
#else
restoreInitiatorIdState(fd, state)
int fd;
char *state;
#endif
{
    if (fd < 0 || fd >= FD_SETSIZE || state == (char *) 0)
        return;

    if (S_responderStates[fd] == (char *) 0)
        free(S_responderStates[fd]);

    S_initiatorStates[fd] = state;
}

/*
** Calculate the MD5 integrity check given a PE and the fd of the assoc.
**
** The MD5 algorithm is executed over:
**
**   <BER of PDU><8 byte sessionKey>
**
** The first 128bit MD5 value is returned.
**
** RETURN  0 : success
**        -1 : failure (*check = 0)
*/
int
#if defined(__STDC__) || defined(__cplusplus)
makeMd5Value(const int fd,
             PE pdu,
             MD5Value *check)
#else
makeMd5Value(fd, pdu, check)
int fd;
PE pdu;
MD5Value *check;
#endif
{
    char *buffer = (char *) 0,
         *sessionKey = getMd5Key(fd);
    int length1 = 0, length2 = 0;

    MD5Context context;

    if (check == (MD5Value *) 0) {
#if defined(DEBUG)
        fprintf(stderr, "makeMd5Value(): NULL check\n");
#endif
        return -1;
    }

    memset((void *) check, 0, sizeof(MD5Value));

    /* If no session key is passed then simply make a MD5 check */
#if defined(DEBUG)
    if (sessionKey == (char *) 0)
        fprintf(stderr, "makeMd5Value(): no sessionKey for fd %d\n", fd);
#endif

    /* If the key is all zeros then skip the check */
    if ((sessionKey != (char *) 0) &&
        (memcmp((void *) sessionKey, (void *) G_nullKey, 8) == 0))
        return 0;

    if (pdu == (PE) 0) {
#if defined(DEBUG)
        fprintf(stderr, "makeMd5Value(): NULL pdu\n");
#endif
        return -1;
    }

    /* PDU -> BER */
    length1 = ps_get_abs(pdu); /* the absolute length */
    if (pe2ssdu(pdu, &buffer, &length2) != OK) {
        fprintf(stderr, "makeMd5Value(): pe2ssdu failed\n");
        free(buffer);
        return -1;
    }

    /* Sanity check ... this shouldn't happen */
    if (length1 != length2) {
#if defined(DEBUG)
        fprintf(stderr, "makeMd5Value(): l1:%d l2:%d\n", length1, length2);
#endif
        free(buffer);
        return -1;
    }

    buffer = realloc(buffer, length1 + 8); /* how do you detect a fail here? */
    /* Copy session key into buffer */
    if (sessionKey != (char *) 0) {
        memcpy((void *) &(buffer[length1]), (void *) sessionKey, 8);
        length1 += 8;
    }

    /*
    ** Get the MD5 value
    **
    ** The "RSA Data Security, Inc. MD5 Message-Digest Algorithm" is
    ** used. The code is "derived from the RSA Data Security, Inc. MD5
    ** Message-Digest Algorithm". (Actually, it is lifted straight
    ** from RFC 1321 with some ANSI-fication -- see md5.h md5.c and
    ** global.h)
    */
    md5Init(&context);
    md5Update(&context, (unsigned char *) buffer, length1);
    md5Final(check->md5_bytes, &context);

    free(buffer);

    return 0;
}

/*
** IGNORE EVERYTHING BELOW THIS LINE!
*/

#if 0
/*
** These functions are for me only -- DO NOT USE -- they are not tested
*/
int makeMsapPdu _a_((const int fd, const int type, void *parm, PE *pdu, const int errorFlag));
int makeMsapPduCheck _a_((const int fd, MSAPIndication *mi, int *msapId, int *check));
int getMsapInvokeId _a_((const MSAPIndication *mi)); /* ROS invokeId of mi */
int setMsapInvokeId _a_((const int id, MSAPIndication *mi));
/*
**
*/
#endif


#if 0

#if defined(__cplusplus)
extern "C" {
#endif
/*
** All these should be in msap.c
*/
int build_CMIP_EventReportArgument _a_((PE *, int, int, char *, void *));
int build_CMIP_EventReportResult _a_((PE *, int, int, char *, void *));
int build_CMIP_LinkedReplyArgument _a_((PE *, int, int, char *, void *));
int build_CMIP_GetArgument _a_((PE *, int, int, char *, void *));
int build_CMIP_GetResult _a_((PE *, int, int, char *, void *));
int build_CMIP_SetArgument _a_((PE *, int, int, char *, void *));
int build_CMIP_SetResult _a_((PE *, int, int, char *, void *));
int build_CMIP_ActionArgument _a_((PE *, int, int, char *, void *));
int build_CMIP_ActionResult _a_((PE *, int, int, char *, void *));
int build_CMIP_CreateArgument _a_((PE *, int, int, char *, void *));
int build_CMIP_CreateResult _a_((PE *, int, int, char *, void *));
int build_CMIP_DeleteArgument _a_((PE *, int, int, char *, void *));
int build_CMIP_DeleteResult _a_((PE *, int, int, char *, void *));
int build_CMIP_GetListError _a_((PE *, int, int, char *, void *));
int build_CMIP_SetListError _a_((PE *, int, int, char *, void *));
int build_CMIP_ActionError _a_((PE *, int, int, char *, void *));
int build_CMIP_DeleteError _a_((PE *, int, int, char *, void *));

/* .. and these are defined below ... */
int dummyCMIPBuildFunction _a_((PE *, int, int, char *, void *));
int build_CMIP_EmptyResult _a_((PE *, int, int, char *, void *));
#if defined(__cplusplus)
}
#endif

/*
** A Dummy build function ...
*/
int
#if defined(__STDC__) || defined(__cplusplus)
dummyCMIPBuildFunction(PE *pe,
                       int i1,
                       int i2,
                       char *str,
                       void *parm)
#else
dummyCMIPBuildFunction(pe, i1, i2, str, parm)
PE *pe;
int i1;
int i2;
char *str;
void *parm;
#endif
{
#if defined(DEBUG)
    fprintf(stderr, "dummyCMIPBuildFunction(): check for invalid PDU type?\n");
#endif
    *pe = (PE) 0;
    return NOTOK;
}

int
#if defined(__STDC__) || defined(__cplusplus)
build_CMIP_EmptyResult(PE *pe,
                       int i1,
                       int i2,
                       char *str,
                       void *parm)
#else
build_CMIP_EmptyResult(pe, i1, i2, str, parm)
PE *pe;
int i1;
int i2;
char *str;
void *parm;
#endif
{
    *pe = int2prim(*((int *) parm));
    return OK;
}

/*
** A lookup table of encode functions for PDU types. The index to the table
** is the #defined MI_* constants in msap.h. The functions are the ones in
** msap.c generated by pepy.
*/

typedef int (*CMIPBuildFunction) _a_((PE *, int, int, char *, void *));

#if defined(__STDC__) || defined(__cplusplus)
const
#endif
static
CMIPBuildFunction S_cmipBuildFunctions[] = {
    dummyCMIPBuildFunction, /* Unknown */
    dummyCMIPBuildFunction, /* UReject */
    dummyCMIPBuildFunction, /* PReject */
    dummyCMIPBuildFunction, /* Abort */
    dummyCMIPBuildFunction, /* Term */
    dummyCMIPBuildFunction, /* Error */
    build_CMIP_EventReportArgument,
    build_CMIP_EventReportArgument,
    build_CMIP_EventReportResult,
    build_CMIP_LinkedReplyArgument,
    build_CMIP_GetArgument,
    build_CMIP_GetResult,
    build_CMIP_SetArgument,
    build_CMIP_SetArgument,
    build_CMIP_SetResult,
    build_CMIP_ActionArgument,
    build_CMIP_ActionArgument,
    build_CMIP_ActionResult,
    build_CMIP_CreateArgument,
    build_CMIP_CreateResult,
    build_CMIP_DeleteArgument,
    build_CMIP_DeleteResult,
    build_CMIP_EmptyResult, /* CancelGet */
    build_CMIP_EmptyResult,
    (CMIPBuildFunction) 0
};

#if defined(__STDC__) || defined(__cplusplus)
const
#endif
static
CMIPBuildFunction S_cmipBuildErrorFunctions[] = {
    dummyCMIPBuildFunction, /* Unknown */
    dummyCMIPBuildFunction, /* UReject */
    dummyCMIPBuildFunction, /* PReject */
    dummyCMIPBuildFunction, /* Abort */
    dummyCMIPBuildFunction, /* Term */
    dummyCMIPBuildFunction, /* Error */
    dummyCMIPBuildFunction, /* EventReport */
    dummyCMIPBuildFunction, /* EventReportConf */
    dummyCMIPBuildFunction, /* EventReportRes */
    dummyCMIPBuildFunction, /* LinkedReply */
    dummyCMIPBuildFunction, /* Get */
    build_CMIP_GetListError,
    dummyCMIPBuildFunction, /* Set */
    build_CMIP_SetListError,
    build_CMIP_SetListError,
    dummyCMIPBuildFunction, /* Action */
    build_CMIP_ActionError,
    build_CMIP_ActionError,
    dummyCMIPBuildFunction, /* Create */
    dummyCMIPBuildFunction, /* CreateRes */
    dummyCMIPBuildFunction, /* Delete */
    build_CMIP_DeleteError,
    dummyCMIPBuildFunction, /* CancelGet */
    dummyCMIPBuildFunction, /* EmptyRes */
    (CMIPBuildFunction) 0
};

#endif

#if 0
/*
** Create PDUs with integrity checks
**
** Return  OK    : success
**         NOTOK : failure
*/
int
#if defined(__STDC__) || defined(__cplusplus)
makeMsapPdu(const int fd,
            const int type,
            void *parm,
            PE *pdu,
            const int errorFlag)
#else
makeMsapPdu(fd, type, parm, pdu, errorFlag)
int fd;
int type;
void *parm;
PE *pdu;
int errorFlag;
#endif
{
    CMIPBuildFunction build = errorFlag ? S_cmipBuildErrorFunctions[type]:
                                          S_cmipBuildFunctions[type];

#if defined(DEBUG)
    if (build == dummyCMIPBuildFunction)
        fprintf(stderr, "makeMsapPdu(): %d %s %s\n",
                type, S_types[type], errorFlag ? "errorFlag" : "");
#endif

    return (*build)(pdu, 1, 0, (char *) 0, parm);
}

int
#if defined(__STDC__) || defined(__cplusplus)
makeMsapPduCheck(const int fd,
                 MSAPIndication *mi,
                 int *msapId,
                 int *check)
#else
makeMsapPduCheck(fd, mi, msapId, check)
int fd;
MSAPIndication *mi;
int *msapId;
int *check;
#endif
{
    register
    void *parm = (void *) 0;
    register
    int errorFlag = 0, id = 0;

    PE pdu;
    MD5Value md5;

    switch (mi->mi_type) {
    case MI_UNKNOWN:
    case MI_UREJECT:
    case MI_PREJECT:
    case MI_ABORT:
    case MI_TERM:
    case MI_ERROR:
        break;

    case MI_EVENT_REP:
    case MI_EVENT_REPC:
        parm = (void *) &mi->mi_eventrepinv.args;
        id = mi->mi_eventrepinv.id;
        break;

    case MI_EVENT_REP_RES:
        parm = (void *) &mi->mi_eventrepres.result;
        id = mi->mi_eventrepres.id;
        break;

    case MI_LINKED_REPLY:
        id = mi->mi_linkedreplyinv.id;

        switch (mi->mi_linkedreplyinv.args.lr_type) {

        case LR_UNKNOWN:
            /* Nothing */
            break;

        case LR_GET_LISTERR:
            errorFlag = 1;
        case LR_GET_RES:
            parm = (void *) &mi->mi_linkedreplyinv.args.lr_getres;
            break;

        case LR_SET_LISTERR:
            errorFlag = 1;
        case LR_SET_RES:
            parm = (void *) &mi->mi_linkedreplyinv.args.lr_setres;
            break;

        case LR_ACTION_RES:
            parm = (void *) &mi->mi_linkedreplyinv.args.lr_actionres;
            break;

        case LR_PROC_FAILURE:
            /* Nothing */
            break;

        case LR_DELETE_RES:
            parm = (void *) &mi->mi_linkedreplyinv.args.lr_deleteres;
            break;

        case LR_ACTION_ERR:
            parm = (void *) &mi->mi_linkedreplyinv.args.lr_actionerr;
            errorFlag = 1;
            break;

        case LR_DELETE_ERR:
            parm = (void *) &mi->mi_linkedreplyinv.args.lr_deleteerr;
            errorFlag = 1;
            break;
        }
        break;

    case MI_GET:
        parm = (void *) &mi->mi_getinv.args;
        id = mi->mi_getinv.id;
        break;

    case MI_GET_RES:
        parm = (void *) &mi->mi_getres.result;
        id = mi->mi_getres.id;
        break;

    case MI_SET:
    case MI_SETC:
        parm = (void *) &mi->mi_setinv.args;
        id = mi->mi_setinv.id;
        break;

    case MI_SET_RES:
        parm = (void *) &mi->mi_setres.result;
        id = mi->mi_setres.id;
        break;

    case MI_ACTION:
    case MI_ACTIONC:
        parm = (void *) &mi->mi_actioninv.args;
        id = mi->mi_actioninv.id;
        break;

    case MI_ACTION_RES:
        parm = (void *) &mi->mi_actionres.result;
        id = mi->mi_actionres.id;
        break;

    case MI_CREATE:
        parm = (void *) &mi->mi_createinv.args;
        id = mi->mi_createinv.id;
        break;

    case MI_CREATE_RES:
        parm = (void *) &mi->mi_createres.result;
        id = mi->mi_createres.id;
        break;

    case MI_DELETE:
        parm = (void *) &mi->mi_deleteinv.args;
        id = mi->mi_deleteinv.id;
        break;

    case MI_DELETE_RES:
        parm = (void *) &mi->mi_deleteres.result;
        id = mi->mi_deleteres.id;
        break;

    case MI_CANCEL_GET:
        parm = (void *) &mi->mi_cancelgetinv.args.cga_getinvoke;
        id = mi->mi_cancelgetinv.id;
        break;

    case MI_EMPTY_RES:
        parm = (void *) &mi->mi_emptyres.id;
        id = mi->mi_emptyres.id;
        break;
    }

    if (msapId != (int *) 0)
        *msapId = id;

    /* Nothing to check */
    if (id == 0) {
#if defined(DEBUG)
        fprintf(stderr, "makeMsapPduCheck(): no check fd=%d %s\n",
                fd, S_types[mi->mi_type]);
#endif
        return OK;
    }

    if (check != (int *) 0) {

        if (makeMsapPdu(fd, mi->mi_type, parm, &pdu, errorFlag) != OK) {
#if defined(DEBUG)
            fprintf(stderr, "makeMsapPduCheck(): build failed fd=%d %s\n",
                fd, S_types[mi->mi_type]);
#endif
            return NOTOK;
        }

        if (makeMd5Value(fd, pdu, &md5) != OK) {
#if defined(DEBUG)
            fprintf(stderr, "makeMsapPduCheck(): MD5 failed fd=%d %s\n",
                    fd, S_types[mi->mi_type]);
#endif
            return NOTOK;
        }
        pe_free(pdu);

        *check = (int) md5.md5_words[0];
    }

    return OK;
}


int
#if defined(__STDC__) || defined(__cplusplus)
getMsapInvokeId(const MSAPIndication *mi)
#else
getMsapInvokeId(mi)
MSAPIndication *mi;
#endif
{
    register
    int id = 0;

    switch (mi->mi_type) {
    case MI_UNKNOWN:
    case MI_UREJECT:
    case MI_PREJECT:
    case MI_ABORT:
    case MI_TERM:
    case MI_ERROR:
        break;

    case MI_EVENT_REP:
    case MI_EVENT_REPC:
        id = mi->mi_eventrepinv.id;
        break;

    case MI_EVENT_REP_RES:
        id = mi->mi_eventrepres.id;
        break;

    case MI_LINKED_REPLY:
        id = mi->mi_linkedreplyinv.id;
        /* id = mi->mi_linkedreplyinv.args.lr_linkid; */
        break;

    case MI_GET:
        id = mi->mi_getinv.id;
        break;

    case MI_GET_RES:
        id = mi->mi_getres.id;
        break;

    case MI_SET:
    case MI_SETC:
        id = mi->mi_setinv.id;
        break;

    case MI_SET_RES:
        id = mi->mi_setres.id;
        break;

    case MI_ACTION:
    case MI_ACTIONC:
        id = mi->mi_actioninv.id;
        break;

    case MI_ACTION_RES:
        id = mi->mi_actionres.id;
        break;

    case MI_CREATE:
        id = mi->mi_createinv.id;
        break;

    case MI_CREATE_RES:
        id = mi->mi_createres.id;
        break;

    case MI_DELETE:
        id = mi->mi_deleteinv.id;
        break;

    case MI_DELETE_RES:
        id = mi->mi_deleteres.id;
        break;

    case MI_CANCEL_GET:
        id = mi->mi_cancelgetinv.id;
        break;

    case MI_EMPTY_RES:
        id = mi->mi_emptyres.id;
        break;
    }

    return id;
}

int
#if defined(__STDC__) || defined(__cplusplus)
setMsapInvokeId(const int id, MSAPIndication *mi)
#else
setMsapInvokeId(id, mi)
int id;
MSAPIndication *mi;
#endif
{
    if (mi == (MSAPIndication *) 0)
        return 0;

    switch (mi->mi_type) {
    case MI_UNKNOWN:
    case MI_UREJECT:
    case MI_PREJECT:
    case MI_ABORT:
    case MI_TERM:
    case MI_ERROR:
        break;

    case MI_EVENT_REP:
    case MI_EVENT_REPC:
        mi->mi_eventrepinv.id = id;
        break;

    case MI_EVENT_REP_RES:
        mi->mi_eventrepres.id = id;
        break;

    case MI_LINKED_REPLY:
        mi->mi_linkedreplyinv.id = id;
        /* mi->mi_linkedreplyinv.args.lr_linkid = id; */
        break;

    case MI_GET:
        mi->mi_getinv.id = id;
        break;

    case MI_GET_RES:
        mi->mi_getres.id = id;
        break;

    case MI_SET:
    case MI_SETC:
        mi->mi_setinv.id = id;
        break;

    case MI_SET_RES:
        mi->mi_setres.id = id;
        break;

    case MI_ACTION:
    case MI_ACTIONC:
        mi->mi_actioninv.id = id;
        break;

    case MI_ACTION_RES:
        mi->mi_actionres.id = id;
        break;

    case MI_CREATE:
        mi->mi_createinv.id = id;
        break;

    case MI_CREATE_RES:
        mi->mi_createres.id = id;
        break;

    case MI_DELETE:
        mi->mi_deleteinv.id = id;
        break;

    case MI_DELETE_RES:
        mi->mi_deleteres.id = id;
        break;

    case MI_CANCEL_GET:
        mi->mi_cancelgetinv.id = id;
        break;

    case MI_EMPTY_RES:
        mi->mi_emptyres.id = id;
        break;
    }

    return id;
}
#endif
