/*
 * Copyright (c) 1993 University College London
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Department of Computer Science, University College London.
 * The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*/

/************************************************************************* 
 * midas_xmp.h 
 * 
 * VERSION 2.0 August 2nd 1993 
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~ 
 * 
 * This file contains definitions for the MIDAS XMP convergence 
 * interface. 
 * 
 * Where the MIDAS interface corresponds exactly to the XMP one 
 * the proper XMP constants and types have been used. 
 * 
 * The XOM object model has NOT been used. Instead C data structures 
 * have been specified for use whenever an XOM object is specified  
 * in "real" XMP. Wherever possible, existing structures from ISODE 
 * and MSAP have been used for this purpose. 
 * 
 * Through out this file reference is made to: 
 * 
 * [1]	X/OPen preliminaey Specification, System Management Protocols 
 *      API (XMP), X/Open Company Ltd., July 1992 
 * 
 *************************************************************************/ 
 
#ifndef _MIDAS_XMP_ 
#define _MIDAS_XMP_ 

#if defined(__STDC__) || defined(__cplusplus)
#define _a_(_args_) _args_
#else
#define _a_(_args_) ()
#endif
 
/************************************************************************* 
 * Basic OM Data types 
 * ~~~~~~~~~~~~~~~~~~~ 
 *  
 * These are defined as "system dependent" in the standard. The following  
 * seem to be rational choices for our Unix systems 
 *************************************************************************/ 
 
#include <sys/types.h> 
 
typedef int  		OM_sint; 
typedef short    	OM_sint16; 
typedef long 		OM_sint32; 
typedef u_int    	OM_uint; 
typedef u_short  	OM_uint16; 
typedef u_long   	OM_uint32; 
 
 
/************************************************************************* 
 * Additional OM types 
 * ~~~~~~~~~~~~~~~~~~~ 
 *  
 * These are the types specified in the XOM and XMP documents which will be  
 * used unaltered. 
 *************************************************************************/ 
 
typedef OM_uint16    	OM_type;    /*Values as on p.65 of [1]*/ 
typedef OM_uint16    	OM_syntax;  /*Values as on p.64 of [1]*/ 
typedef OM_uint32    	OM_boolean; 
typedef OM_sint32    	OM_enumeration; 
typedef OM_sint32    	OM_integer; 
 
 
/************************************************************************* 
 * Types used from ISODE  
 * ~~~~~~~~~~~~~~~~~~~~~ 
 *  
 * Those used are noted below for information only. 
 *************************************************************************/ 
 
/* 
 * From psap.h 
 */ 
 
#ifndef _PSAP_ 
#include <isode/psap.h> 
#endif 
 
/* 
 * struct OIDentifier   
 * OID       	- a pointer to an OIDentifier structure 
 * NULLOID      - a null pointer to same 
 */ 
 
/* 
 * From isodaddrs.h 
 */ 
 
#ifndef _ISOADDRS_ 
#include <isode/isoaddrs.h> 
#endif 
 
/* 
 * struct PSAPaddr       
 * NULLPA       - a null pointer to a struct PSAPaddr 
 * 
 * AEI      	- a pointer to an AEInfo structure 
 * NULLAEI      - a null pointer to same 
 */ 
 
/************************************************************************* 
 * Types used from MSAP  
 * ~~~~~~~~~~~~~~~~~~~~~ 
 *  
 * Those used are noted below for information only. 
 *************************************************************************/ 
 
/* 
 * From mparm.h 
 */ 
 
#ifndef _MPARM_ 
#include "mparm.h" 
#endif 
 
/* 
 * External   	- a structure (struct type_UNIV_EXTERNAL) to represent 
 *                an ASN.1 "external" type. 
 * NULLEXTERN   - a NULL pointer to same 
 * 
 * MID       	- a pointer to a MIDentifier structure which identifies a 
 *                MO class 
 * 
 * MN           - a pointer to a MName structure which identifies an 
 *                MO instance 
 * 
 *           	- the following map 1-1 onto the similarly named XOM 
 *                structures passed in the "argument" field of XMP calls 
 * struct CMISXXXXarg 
 * struct CMISXXXXres 
 * struct CMISLinkedReplyArg   
/*  
 
/* 
 * From msap.h 
 */ 
 
#ifndef _MSAP_ 
#include "msap.h" 
#endif 
 
/* 
 * MSAPIndication    - structure holding the result of a management  
 * 		       operation 
 */ 
 
 
/************************************************************************* 
 * Types defined specially for the MIDAS XMP interface  
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ 
 *  
 * These structures are mainly used to replace the "Objects" used as  
 * parameters in "real" XMP calls. The information content is mainly the 
 * same but existing ISODE and MSAP structures are used wherever possible. 
 * 
 * Real XMP objects consist of arrays of "attributes" many of which may be 
 * included or ommitted as desired. In the structures below, attributes  
 * are represented either by simple types or by pointers to complex types.  
 * The ommission of an attribute is indicated in the case of a simple type 
 * by the assignment of the dummy value MIDAS_NOT_PRESENT and in the case 
 * of a complex type by the use of a NULL pointer.  
 *************************************************************************/ 
 
#define MIDAS_NOT_PRESENT 0  /*used to indicate absence of a simple type*/ 
 
/* 
 * MIDAS_OID_LIST 
 * ~~~~~~~~~~~~~~ 
 * 
 * Not strictly necessary. In principle we need this to manage a 
 * list of OIDs corresponding to presentation contexts. (However, 
 * this functionality is not available from MSAP.) 
 * 
 */ 
 
typedef struct MIDAS_OID_list { 
     struct MIDAS_OID_list   		* MIDAS_OID_next; 
     struct OIDentifier  		MIDAS_OID_this; 
} MIDAS_OID_LIST; 
#define NULLOIDLIST ((MIDAS_OID_LIST*)(0)) 
 
 
/*  
 * Values returned from functions and output parameters. 
 * 
 * Essentially the same information is yielded by both MSAP and XMP calls. 
 * However, MSAP "returns" a simple OK/NOTOK status and places all other 
 * result information (including error information) in an MSAPIndication  
 * structure. XMP "returns" MP_SUCCESS or (in the case of an 
 * error) an XOM object which contains error information - this is nearly  
 * equivalent to an instance of an MSAPIndication structure but is 
 * slightly richer as it can also report errors specific to XMP. 
 * 
 * The structure MIDAS_RESULT is defined for use as an output 
 * parameter to MIDAS/XMP calls. A pointer to a pointer to a structure 
 * of this type is included as a parameter to most calls. The called  
 * function will set this pointer either to the address of a 
 * dynamically allocated structure of type MIDAS_RESULT or to the value 
 * NULLMIDASRESULT. 
 * 
 * Broadly speaking NULLMIDASRESULT is used: 
 * 1.	For synchronous calls which fail 
 * 2.	For asynchronous calls 
 * 
 * A further complication occurs in the case of linked replies. In the 
 * synchronous case XMP requires that the call does not return until  
 * all parts of the linked reply have been received (or an error is 
 * reported). This means that the MIDAS/XMP interface must have a 
 * means of representing sequences of partial results. MIDAS_RESULT 
 * includes a link pointer to cover this requirement. 
 * 
 * NOTE THAT SINCE THE MIDAS_RESULT STRUCTURE IS DYNAMICALLY ALLOCATED 
 * THE CALLING CODE MUST FREE THE STORAGE. 
 * 
 * The only other output parameter is an invoke ID of type OM_sint32. 
 * A pointer to one of these is included in most calls and will be 
 * filled in by the called function. 
 * 
 * The structure MIDAS_STATUS is defined to hold the information returned 
 * by a call. All calls return a pointer to a statically allocated 
 * instance of this structure (in the case of errors) or to one of the 
 * constants MIDAS_SUCCESS, MIDAS_NO_WORKSPACE, MIDAS_INVALID SESSION or  
 * MIDAS_INSUFFICIENT_RESOURCES. 
 * 
 * The following structures make use of as much of MSAP as possible 
 * whilst preserving the spirit of XMP 
 * 
 */ 
 
/* 
 * 
 * MIDAS_RESULT 
 * ~~~~~~~~~~~~ 
 * 
 * This is used to convey the results of a synchronous request and the 
 * argument of a response invocation. 
 * 
 * A link to another instance of the structure covers the 
 * linked_reply case. This pointer should be set to NULLMIDASRESULT 
 * if there is not a linked reply. 
 * 
 * When used to convey a linked reply, the following restrictions  
 * apply to the use of the MSAPIndication structures: 
 * 
 * 1.	The value of mi_type must be MI_LINKED_REPLY 
 * 
 * 2.	The field selected from the union must be mi_linkedreplyinv  
 * 
 * If the operation succeeds, the value of the mi_errtype field will 
 * be m_no_error. Otherwise it will be one of the other values defined 
 * for the CMISErrors type.  
 * 
 * The constant NULLMIDASRESULT is used to indicate the absence 
 * of a response parameter when a call returns. 
 * 
 */ 
 
typedef struct MIDAS_result { 
     struct MIDAS_result 	* MIDAS_result_next; 
     MSAPIndication   		MIDAS_result_info; 
} MIDAS_RESULT; 
#define NULLMIDASRESULT ((MIDAS_RESULT *)(0)) 
  
/* 
 * MIDAS_STATUS 
 * ~~~~~~~~~~~~ 
 * 
 * Status and related structures - a pointer to an instance of 
 * the MIDAS_status structure is returned by most XMP calls. 
 * 
 * If the call succeeds then MIDAS_SUCCESS is returned. 
 * 
 * If the call fails then either: 
 * 
 *   1.  MIDAS_NO_WORKSPACE, MIDAS_INVALID SESSION or  
 *       MIDAS_INSUFFICIENT_RESOURCES is returned 
 * 
 *   2.  A pointer to a valid MIDAS_STATUS structure is 
 *       returned in which case the MIDAS_problem and 
 *       MIDAS_parameter fields (may) contain additional 
 *       information. 
 * 
 * The MIDAS_STATUS structure can report errors of 4 types: 
 *   1.	 Communications error (pp 156 & 165 of [1]) 
 * 
 *   2.  Library error (pp 157 & 165 of [1]) 
 * 
 *   3.  Service error (corresponds to the values of CMISErrors) 
 * 
 *   4.	 System error (Unix errno) 
 * 
 */ 
 
typedef struct MIDAS_problem { 
	enum problem_type { 
		midas_communication_error, 
		midas_library_error, 
		midas_service_error, 
		midas_system_error 
	} midas_problem_type; 
 
	union problem_value { 
		CMISErrors 	cmis_error_val;	    /*problem_type is  
						      midas_service_error*/ 
		int		error_val;	    /*other problem_types*/ 
	} midas_problem_value; 
} MIDAS_PROBLEM; 
 
	 
typedef struct MIDAS_status { 
 	MIDAS_PROBLEM          		MIDAS_problem; 
 	MSAPIndication     		* MIDAS_parameter; /*additional info 
							     as appropriate*/ 
} MIDAS_STATUS, *MIDAS_STATUS_P;   
 
#define NULL_MIDAS_STATUS    		((MIDAS_STATUS*)0)   
#define MIDAS_SUCCESS        		((MIDAS_STATUS*)0) 
#define MIDAS_NO_WORKSPACE   		((MIDAS_STATUS*)1) 
#define MIDAS_INVALID_SESSION    	((MIDAS_STATUS*)2) 
#define MIDAS_INSUFFICIENT_RESOURCES   	((MIDAS_STATUS*)3) 
 
/* 
 * 
 * MIDAS_FEATURE 
 * ~~~~~~~~~~~~~ 
 * 
 * Feature structure - equivalent to MP_feature on p.166 of [1] 
 * 
 * Note that the response field gets filled in by the called party to 
 * indicate acceptance or otherwise of the proposed feature 
 * 
 * A linked list of these structures are supplied as an argument to 
 * the mp_negotiate call (but will be ignored in the first 
 * implementation). 
 * 
 */ 
 
typedef struct MIDAS_feature { 
     struct OIDentifier  		feature; 
     OM_sint             		request; 
     OM_boolean          		response; 
     struct MIDAS_feature    		* MIDAS_feature_next; 
} MIDAS_FEATURE;  
#define NULLMIDASFEATURE     ((MIDAS_FEATURE *)(0)) 
#define MP_ACTIVATE		0 
#define MP_DEACTIVATE		1 
#define MP_QUERY_STATE		2 
#define MP_QUERY_SUPPORTEED	3 
 
/* 
 * MIDAS_SESSION 
 * ~~~~~~~~~~~~~ 
 * 
 * Session class (p.109) 
 * 
 * Notes: 
 *   1.  responder address and responder title should be set to  
 *       NULLPA and NULLAEI repectively in the first release since 
 *       the association with the responder will be made outside 
 *       the interface and the relevant file descriptor will be passsed 
 *       in the context object. 
 * 
 *   2.  presentation_context_list and SMASE_user_info are place- 
 *       holders for future releases. They will be ignored in  
 *       release 1.0. 
 * 
 *   3.  The file_descriptor will be ignored. Should be set to  
 *       MP_NO_VALID_FILE_DESCRIPTOR 
 * 
 */ 
 
 
typedef struct MIDAS_session { 
    struct PSAPaddr     		* requestor_address; 
    AEI             			requestor_title; 
    OM_sint32           		role; 
#define  MP_T_MANAGING           	1 
#define  MP_T_MONITORING         	2 
#define  MP_T_PERFORMING         	4 
#define  MP_T_REPORTING          	8 
    OM_sint32           		file_descriptor; 
#define  MP_NO_VALID_FILE_DESCRIPTOR 	-1 
    struct PSAPaddr     		* responder_address; 
    AEI                 		responder_title; 
    External            		* access_control; 
    OID             			authentication_mechanism; 
    char                		* authentication_value; 
    OM_sint32           		CMIS_functional_units; 
#define  MP_T_FU_CANCEL_GET          		1 
#define  MP_T_FU_FILTER              		2 
#define  MP_T_FU_MULTIPLE_OBJECT_SELECTION      4 
#define  MP_T_FU_EXTENDED_SERVICE    		8 
#define  MP_T_FU_MULTIPLE_REPLY      		16 
    OID             			application_context; 
    MIDAS_OID_LIST      		* presentation_context_list; 
    External            		* user_info; 
    caddr_t         			* SMASE_user_data; 
    struct MIDAS_session		* MIDAS_session_next; 
} MIDAS_SESSION; 
 
/* 
 * MIDAS_CONTEXT 
 * ~~~~~~~~~~~~~ 
 * 
 * Context class (p.98) 
 *  
 * Notes: 
 *   1.  extensions is a place-holder for future releases. It 
 *       should be set to NULL in release 1.0. 
 * 
 *   2.  connection_id must be a valid Unix file-descriptor 
 *    
 *   3.  responder address and responder title should be set to  
 *       NULLPA and NULLAEI repectively in the first release since 
 *       the association with the responder will be made outside 
 *       the interface and the relevant file descriptor will be passsed 
 *       in the context object in the connection_id field.  
 */ 
 
typedef struct MIDAS_context { 
    caddr_t       			* extensions; 
    External            		* access_control; 
    OM_sint32           		connection_id; 
    OM_enumeration      		mode; 
#define  MP_T_CONFIRMED          	1 
#define  MP_T_NON_CONFIRMED      	2 
    OM_enumeration      		priority; 
#define  MP_T_HIGH               	1 
#define  MP_T_LOW                	2 
#define  MP_T_MEDIUM             	3 
    struct PSAPaddr     		* responder_address; 
    AEI             			responder_title; 
    OM_boolean          		asynchronous; 
    OM_sint32           		size_limit; 
    OM_sint32           		time_limit; 
} MIDAS_CONTEXT; 
 
/* 
 * MIDAS_WORKSPACE 
 * ~~~~~~~~~~~~~~~ 
 * 
 * A pointer of this type is returned by a call to mp_initialise(). 
 * It is not significant in the first release which supports a single 
 * workspace only. 
 * 
 */ 
 
typedef caddr_t MIDAS_WORKSPACE; 
 
/* 
 * MIDAS_WAITING_SESSIONS 
 * ~~~~~~~~~~~~~~~~~~~~~~ 
 * 
 * A pointer to one of these is used as a call/return parameter to mp_wait(). 
 * 
 * N.B. Since we do the association control outside XMP the session does not 
 * really "know" about associations. Our mp_wait() waits on ALL fd's except  
 * stdin stdout and stderr which is unlikely to be what you really want. 
 * 
 */ 
 
typedef struct midas_waiting_sessions 	{ 
			MIDAS_SESSION 		*bound_session; 
			OM_boolean			activated; 
			struct midas_waiting_sessions 	
			*next;
} MIDAS_WAITING_SESSIONS; 
 
 
/************************************************************************* 
 * Function prototypes 
 * 
 * These are included mainly for information so that implementors can 
 * see how the structures defined above are used in the various function  
 * calls. There is no requirement that an ANSI C compiler should be 
 * used. 
 * 
 * Unless otherwise stated, the function arguments are in the same order  
 * and have the same meanings as in [1]  
 * 
 ************************************************************************/ 
 
#if defined(__cplusplus) 
extern "C" { 
#endif 
 
 
MIDAS_STATUS_P mp_abandon_M _a_(( 
     MIDAS_SESSION *,  
     OM_sint32       			/*invoke_id*/ 
)); 
  
MIDAS_STATUS_P mp_action_req_M _a_(( 
     MIDAS_SESSION *, 
     MIDAS_CONTEXT *, 
     CMISActionArg *,    		/*argument*/ 
     MIDAS_RESULT **,	   		/*result_return*/ 
     /*Set to NULLMIDASRESULT for async and non-confirmed modes*/ 
     OM_sint32 *     			/*invoke_id_return*/ 
)); 
 
  
MIDAS_STATUS_P mp_action_rsp_M _a_(( 
     MIDAS_SESSION *, 
     MIDAS_CONTEXT *, 
     MIDAS_RESULT *,    		/*response*/ 
     OM_sint32 *     			/*invoke_id_return*/ 
)); 
 
 
MIDAS_STATUS_P mp_bind_M _a_(( 
     MIDAS_SESSION *,     
     MIDAS_WORKSPACE, 
     MIDAS_SESSION ** 			/*bound_session_return*/ 
)); 
 
 
MIDAS_STATUS_P mp_cancel_get_req_M _a_(( 
     MIDAS_SESSION *, 
     MIDAS_CONTEXT *, 
     CMISCancelGetArg *,    		/*argument*/ 
     OM_sint32 *     			/*invoke_id_return*/ 
)); 
 
  
MIDAS_STATUS_P mp_cancel_get_rsp_M _a_(( 
     MIDAS_SESSION *, 
     MIDAS_CONTEXT *, 
     CMISErrors *,    			/*response*/ 
      /*Set to NULLCMISERRORS to indicate success*/ 
#define NULLCMISERRORS ((CMISErrors *)(0)) 
     OM_sint32 *  	   		/*invoke_id_return*/ 
)); 
 
 
MIDAS_STATUS_P mp_create_req_M _a_(( 
     MIDAS_SESSION *, 
     MIDAS_CONTEXT *, 
     CMISCreateArg *,    		/*argument*/ 
     MIDAS_RESULT **, 	  		/*result_return*/ 
     /*Set to NULLMIDASRESULT for async mode*/ 
     OM_sint32 *     			/*invoke_id_return*/ 
)); 
 
  
MIDAS_STATUS_P mp_create_rsp_M _a_(( 
     MIDAS_SESSION *, 
     MIDAS_CONTEXT *, 
     MIDAS_RESULT *,    		/*response*/ 
     OM_sint32 *     			/*invoke_id_return*/ 
)); 
 
 
MIDAS_STATUS_P mp_delete_req_M _a_(( 
     MIDAS_SESSION *, 
     MIDAS_CONTEXT *, 
     CMISDeleteArg *,    		/*argument*/ 
     MIDAS_RESULT **,   		/*result_return*/ 
     /*Set to NULLMIDASRESULT for async mode*/ 
     OM_sint32 *     			/*invoke_id_return*/ 
)); 
 
  
MIDAS_STATUS_P mp_delete_rsp_M _a_(( 
     MIDAS_SESSION *, 
     MIDAS_CONTEXT *, 
     MIDAS_RESULT *,    		/*response*/ 
     OM_sint32 *     			/*invoke_id_return*/ 
)); 
 
MIDAS_STATUS_P mp_event_report_req_M _a_(( 
     MIDAS_SESSION *, 
     MIDAS_CONTEXT *, 
     CMISEventReportArg *,    		/*argument*/ 
     MIDAS_RESULT **,   		/*result_return*/ 
     /*Set to NULLMIDASRESULT for async and non-confirmed modes*/ 
     OM_sint32 *     			/*invoke_id_return*/ 
)); 
 
  
MIDAS_STATUS_P mp_event_report_rsp_M _a_(( 
     MIDAS_SESSION *, 
     MIDAS_CONTEXT *, 
     MIDAS_RESULT *,    		/*response*/ 
     OM_sint32 *     			/*invoke_id_return*/ 
)); 
 
MIDAS_STATUS_P mp_get_req_M _a_(( 
     MIDAS_SESSION *, 
     MIDAS_CONTEXT *, 
     CMISGetArg *,    			/*argument*/ 
     MIDAS_RESULT **,   		/*result_return*/ 
     /*Set to NULLMIDASRESULT for async mode*/ 
     OM_sint32 *     			/*invoke_id_return*/ 
)); 
 
  
MIDAS_STATUS_P mp_get_rsp_M _a_(( 
     MIDAS_SESSION *, 
     MIDAS_CONTEXT *, 
     MIDAS_RESULT *,    		/*response*/ 
     OM_sint32 *     			/*invoke_id_return*/ 
)); 
 
MIDAS_WORKSPACE mp_initialisze_M _a_(( 
     void 
)); 
 
MIDAS_STATUS_P mp_negotiate_M _a_(( 
     MIDAS_FEATURE *, 			/*feature_list*/ 
     MIDAS_WORKSPACE 
)); 
 
MIDAS_STATUS_P mp_receive_M _a_(( 
     MIDAS_SESSION *, 
     MIDAS_CONTEXT *, 
     OM_sint *,      			/*primitive_return*/ 
     OM_sint *,      			/*completion_flag_return*/ 
     MIDAS_STATUS**, 		/*operation_notification_status_return*/ 
     MIDAS_RESULT **,   		/*result_or_argument_return*/ 
     OM_sint32 *     			/*invoke_id_return*/ 
)); 
 
 
MIDAS_STATUS_P mp_set_req_M _a_(( 
     MIDAS_SESSION *, 
     MIDAS_CONTEXT *, 
     CMISSetArg *,    			/*argument*/ 
     MIDAS_RESULT **,   		/*result_return*/ 
     /*Set to NULLMIDASRESULT for async or unconfirmed modes*/ 
     OM_sint32 *     			/*invoke_id_return*/ 
)); 
 
  
MIDAS_STATUS_P mp_set_rsp_M _a_(( 
     MIDAS_SESSION *, 
     MIDAS_CONTEXT *, 
     MIDAS_RESULT *,    		/*response*/ 
     OM_sint32 *     			/*invoke_id_return*/ 
)); 
 
MIDAS_STATUS_P mp_shutdown_M _a_(( 
     MIDAS_WORKSPACE 
)); 
 
MIDAS_STATUS_P mp_unbind_M _a_(( 
     MIDAS_SESSION * 
)); 
 
MIDAS_STATUS_P mp_wait_M _a_(( 
     MIDAS_WAITING_SESSIONS *,  /*bound_session_list*/ 
     MIDAS_WORKSPACE, 
     OM_uint32       			/*timeout*/ 
)); 
      
#if defined(__cplusplus) 
} 
#endif 
 
#undef _a_(_args_)
              
#endif 
