/*
 * Copyright (c) 1993 University College London
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Department of Computer Science, University College London.
 * The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*/

#include "xmp.h"
#include "midas_xmp.h"
#include "msap.h"
#include "mp_utilities.h"

#if defined(SECURITY)
#include "mintegrity.h"
#include "minvoke.h"
#include "mp_integrity.h"
#endif

#if !defined(_a_)
#if defined(__STDC__) || defined(__cplusplus)
#define _a_(_args_) _args_
#else
#define _a_(_args_) ()
#endif
#endif

static int receive_error_return _a_((MIDAS_STATUS*, MIDAS_RESULT*, enum problem_type, int, MIDAS_STATUS**));
static OM_sint maptype2xmp _a_((int));
static OM_sint mapltype2xmp _a_((int));
extern int get_association_descriptor _a_((MIDAS_SESSION *, MIDAS_CONTEXT *));

/*
 * mp_receive_M - get the argument of a management operation/notification
 * or retrieve the (partial) result of an asynchronously executed
 * management operation/notification
 *
 * N.B. On return operation_notification_status_return may have been set to
 * point to a statically allocated structure
 *
 * N.B. On return result_or_argument_return may have been set to point to a
 * dynamically allocated structure. It is the caller's responsibility to free
 * this.
 *
 *
 * N.B. Responder_address in context is not filled in correctly.
 *
 */

MIDAS_STATUS *
mp_receive_M(session, context, primitive_return, completion_flag_return,
           operation_notification_status_return, 
	   result_or_argument_return, invoke_id_return)
    MIDAS_SESSION   *session;
    MIDAS_CONTEXT   *context;
    OM_sint         *primitive_return;
    OM_sint         *completion_flag_return;
    MIDAS_STATUS    **operation_notification_status_return;
    MIDAS_RESULT    **result_or_argument_return;
    OM_sint32       *invoke_id_return;
{

    static MIDAS_STATUS status; /*Will be returned if an error occurs*/
    static MIDAS_RESULT mrp;
    register MSAPIndication *mip;
    int assoc_descriptor;
    PE pdu;

    status.MIDAS_parameter = (MSAPIndication *)NULL;
    
    mrp.MIDAS_result_next = (MIDAS_RESULT *)NULL;
    mip = &(mrp.MIDAS_result_info);

    /*Get the association descriptor*/
    assoc_descriptor = get_association_descriptor(session, context);

    if(assoc_descriptor == NOTOK){
        status.MIDAS_problem.midas_problem_type = midas_library_error; 
        status.MIDAS_problem.midas_problem_value.error_val
                                  = MP_E_INVALID_CONNECTION_ID;

        fprintf(stderr, "mp_receive: invalid association descriptor\n");
        return (&status);

    }

    if (mp_wait_pdu(assoc_descriptor, mip, &status, 0) == NOTOK) 
	return(&status);

    /*be optimistic*/

    *operation_notification_status_return = MIDAS_SUCCESS;
    *result_or_argument_return = &mrp; 

    switch (mip->mi_type) {
        case MI_GET:
        case MI_SET:
        case MI_SETC:
        case MI_ACTION:
        case MI_ACTIONC:
        case MI_CREATE:
        case MI_DELETE:
        case MI_CANCEL_GET:
        case MI_EVENT_REP:
        case MI_EVENT_REPC:
            /*succesful returns*/
            *primitive_return = maptype2xmp(mip->mi_type);
            *completion_flag_return = MP_INCOMING;
            *invoke_id_return = mip->mi_invoke;
            return(MIDAS_SUCCESS);

        case MI_GET_RES:
        case MI_SET_RES:
        case MI_ACTION_RES:
        case MI_CREATE_RES:
        case MI_DELETE_RES:
        case MI_EMPTY_RES:
        case MI_EVENT_REP_RES:
            /*succesful returns*/
            *primitive_return = maptype2xmp(mip->mi_type);
            *completion_flag_return = MP_COMPLETED;
            *invoke_id_return = mip->mi_invoke;
            return(MIDAS_SUCCESS);

        case MI_LINKED_REPLY:
            /*look inside the wrapper*/
		    switch (mip->mi_linkedreplyinv.args.lr_type){
			    case LR_GET_RES:
			    case LR_SET_RES:
                case LR_ACTION_RES:
                case LR_DELETE_RES:
                    /*succesful returns*/
				    *primitive_return =
		    		mapltype2xmp(mip->mi_linkedreplyinv.args.lr_type);
                    *completion_flag_return = MP_PARTIAL;
                    *invoke_id_return = mip->mi_emptyres.id;
				    return(MIDAS_SUCCESS);

                case LR_ACTION_ERR:
                case LR_GET_LISTERR:
                case LR_SET_LISTERR:
                case LR_DELETE_ERR:
                    /*remote has refused to do the operation. Set up a
                                        status object to be returned*/

                    receive_error_return(&status, &mrp, midas_service_error,
                       (int)(mip->mi_linkedreplyinv.args.lr_actionerr.ae_error),
                       operation_notification_status_return);

                    return(MIDAS_SUCCESS);  /*N.B. the operation has failed
                                              but this call has succeeded*/


                case LR_PROC_FAILURE:
                    /* should not really happen*/
                    receive_error_return(&status, &mrp, midas_service_error,
                        (int)(m_processingFailure),
                        operation_notification_status_return);

                    return(MIDAS_SUCCESS);  /*N.B. the operation has failed
                                              but this call has succeeded*/

                default:
                    /*Unexpected type*/
                    receive_error_return(&status, &mrp, midas_service_error,
                        MP_E_UNRECOGNIZED_OPERATION,
                        operation_notification_status_return);

                    return(MIDAS_SUCCESS);  /*N.B. the operation has failed
                                              but this call has succeeded*/

            }

            case MI_PREJECT:
            case MI_UREJECT:
            case MI_ERROR:
                /*Problem in supporting comms layers affecting this
                  operation only*/

                receive_error_return(&status, &mrp, midas_communication_error,
                                     MP_E_COMMUNICATIONS_PROBLEM,    /*p 166*/
                                     operation_notification_status_return);

                return(MIDAS_SUCCESS);  /*N.B. the operation has failed
                                          but this call has succeeded*/

            case MI_ABORT:
            case MI_TERM:
                /*Supporting comms layers have collapsed completely*/
                receive_error_return(&status, &mrp, midas_communication_error,
                                     MP_E_BROKEN_SESSION,            /*p 166*/
                                     operation_notification_status_return);

                return(MIDAS_SUCCESS);  /*N.B. the operation has failed
                                          but this call has succeeded*/

            case MI_UNKNOWN:
            default:
                /*Unexpected midindication type*/
                receive_error_return(&status, &mrp, midas_service_error,
                                     MP_E_UNRECOGNIZED_OPERATION,
                                      operation_notification_status_return);

                return(MIDAS_SUCCESS);  /*N.B. the operation has failed
                                          but this call has succeeded*/
	    }
}



/*
 * set up return values for a failed recieve operation
 *
 */

static int
receive_error_return(statp, mrp, problem_type, error_val, onsrpp)
    MIDAS_STATUS *statp;
    MIDAS_RESULT *mrp;
    enum problem_type problem_type;
    int error_val;            /*should be union problem_value*/
    MIDAS_STATUS **onsrpp;{


    error_status(statp, &mrp->MIDAS_result_info, problem_type, error_val);

    *onsrpp = statp;

    return((int)MIDAS_SUCCESS);  /*N.B. the operation has failed but this call
                                                          has succeeded*/
}


/*
 * Map msap primitives to xmp ones
 *
 */

static OM_sint
maptype2xmp(msapprim)
    int msapprim;{

    switch(msapprim){
        case MI_EVENT_REP:
            return(MP_EVENT_REPORT_IND);
        case MI_EVENT_REPC:
            return(MP_EVENT_REPORT_CNF);
        case MI_GET:
            return(MP_GET_IND);
        case MI_GET_RES:
            return(MP_GET_CNF);
        case MI_SET:
        case MI_SETC:
            return(MP_SET_IND);
        case MI_SET_RES:
            return(MP_SET_CNF);
        case MI_ACTION:
        case MI_ACTIONC:
            return(MP_ACTION_IND);
        case MI_ACTION_RES:
            return(MP_ACTION_CNF);
        case MI_CREATE:
            return(MP_CREATE_IND);
        case MI_CREATE_RES:
            return(MP_CREATE_CNF);
        case MI_DELETE:
            return(MP_DELETE_IND);
        case MI_DELETE_RES:
            return(MP_DELETE_CNF);
        case MI_CANCEL_GET:
            return(MP_CANCEL_GET_IND);
        default:
            return(0);
    }
}

/*
 * Map msap primitives to xmp ones
 *
 */

static OM_sint
mapltype2xmp(msapprim)
    int msapprim;{

    switch(msapprim){
        case LR_GET_RES:
            return(MP_GET_CNF);
        case LR_SET_RES:
            return(MP_SET_CNF);
        case LR_ACTION_RES:
            return(MP_ACTION_CNF);
        case LR_DELETE_RES:
            return(MP_DELETE_CNF);
        default:
            return(0);
   }
}
